/*
 * libmm-player
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, YeJin Cho <cho.yejin@samsung.com>,
 * Seungbae Shin <seungbae.shin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __MM_PLAYER_STREAMING_H__
#define	__MM_PLAYER_STREAMING_H__

#include <glib.h>
#include <gst/gst.h>
#include <string.h>
#include <math.h>
#include <dlog.h>
#include "mm_player.h"

#define MAX_FILE_BUFFER_NAME_LEN 256
#define MAX_BUFFER_PERCENT 100.0

#define DEFAULT_BUFFER_HIGH_WATERMARK 0.99
#define DEFAULT_BUFFER_LOW_WATERMARK  0.1 /* max of low watermark */

/* low watermark min criteria to avoid buffer underrun */
#define LOW_WATERMARK_MIN_BYTE_LEVEL 0.05
#define LOW_WATERMARK_MIN_TIME_VALUE 500

/* low and high watermark criteria */
#define MIN_BUFFER_WATERMARK 0.0
#define MAX_BUFFER_WATERMARK 1.0

/* buffering time criteria (ms) */
#define DEFAULT_BUFFERING_TIME   (3 * 1000)
#define MIN_BUFFERING_TIME       (0)
#define MAX_BUFFERING_TIME       (10 * 1000)

#define DEFAULT_PREBUFFERING_TIME      (3 * 1000)
#define DEFAULT_REBUFFERING_TIME       (10 * 1000)
#define DEFAULT_ADAPTIVE_REBUFFER_TIME (5 * 1000)
#define DEFAULT_LIVE_REBUFFER_TIME     (3 * 1000)

/* buffer size */
#define DEFAULT_BUFFER_SIZE_BYTES      (5 * 1024 * 1024)  /* 5 MBytes */
#define DEFAULT_RING_BUFFER_SIZE_BYTES (20 * 1024 * 1024) /* 20 MBytes, for queue2 */
#define MIN_BUFFER_SIZE_BYTES          (1 * 1024 * 1024)  /* 1 MBytes */
#define MAX_BUFFER_SIZE_BYTES          (32 * 1024 * 1024) /* 32 MBytes */
#define MAX_BUFFER_SIZE_TIME           (15 * 1000)        /* 15 sec, for mq */


#define GET_BYTE_FROM_BIT(bit) (bit / 8)
#define GET_BIT_FROM_BYTE(byte) (byte * 8)
#define CALC_WATERMARK(a, b) ((gdouble)(a) / (gdouble)(b))
#define GET_WATERMARK(a, b, c, d) \
	do { \
		if (((a) > 0) && ((b) > 0)) { \
			d = CALC_WATERMARK(a, b);	\
		} else { \
			LOGW("set default watermark value %f", c); \
			d = c; \
		} \
	} while (0)

#define GET_VALID_VALUE(a, min, max) \
	do { \
		if (min >= max) {\
			LOGW("invalid max value, take min"); \
			a = min; \
		} else { \
			a = MAX(a, min); \
			a = MIN(a, max); \
		} \
	} while (0)

#define PLAYER_BUFFER_CAST(handle)	((streaming_buffer_t *)(handle))
#define PLAYER_STREAM_CAST(sr)		((mmplayer_streaming_t *)(sr))

#define GET_CURRENT_BUFFERING_BYTE(handle)	(PLAYER_BUFFER_CAST(handle)->buffering_bytes)
#define GET_CURRENT_BUFFERING_TIME(handle)	(PLAYER_BUFFER_CAST(handle)->buffering_time)

#define IS_MUXED_BUFFERING_MODE(sr)		(PLAYER_STREAM_CAST(sr)->streaming_buffer_type == BUFFER_TYPE_MUXED) ? (TRUE) : (FALSE)
#define IS_DEMUXED_BUFFERING_MODE(sr)	(PLAYER_STREAM_CAST(sr)->streaming_buffer_type == BUFFER_TYPE_DEMUXED) ? (TRUE) : (FALSE)

#define GET_MAX_BUFFER_SIZE_BYTES(type) ((type == BUFFER_TYPE_MUXED) ? (DEFAULT_RING_BUFFER_SIZE_BYTES) : (MAX_BUFFER_SIZE_BYTES))
#define GET_NEW_BUFFERING_BYTE(size, max) ((size) < (max)) ? (size) : (max)

typedef enum {
	BUFFER_TYPE_DEFAULT,
	BUFFER_TYPE_MUXED = BUFFER_TYPE_DEFAULT,	/* queue2 */
	BUFFER_TYPE_DEMUXED,		/* multi Q in decodebin */
	BUFFER_TYPE_MAX,
} buffer_type_e;

typedef enum {
	MUXED_BUFFER_TYPE_MEM_QUEUE, /* push mode in queue2 */
	MUXED_BUFFER_TYPE_MEM_RING_BUFFER, /* pull mode in queue2 */
	MUXED_BUFFER_TYPE_MAX,
} muxed_buffer_type_e;

typedef enum {
	MM_PLAYER_BUFFERING_DEFAULT = 0x00,
	MM_PLAYER_BUFFERING_IN_PROGRESS = 0x01,
	MM_PLAYER_BUFFERING_ABORT = 0x02,
	MM_PLAYER_BUFFERING_COMPLETE = 0x04,
} mmplayer_buffering_state_e;

typedef struct {
	mmplayer_buffering_mode_e mode;
	gboolean is_pre_buffering;
	gint prebuffer_time; /* ms */
	gint rebuffer_time;  /* ms */
} streaming_requirement_t;

typedef struct {
	GstElement *buffer;			/* buffering element of playback pipeline */

	guint buffering_bytes;		/* max buffer size of byte */
	gint buffering_time;		/* max buffer size of time */
	gdouble buffer_high_watermark;
	gdouble buffer_low_watermark;

	gboolean is_live;
} streaming_buffer_t;

typedef struct {
	gboolean buffering_monitor;
	gint64 prev_pos;
	gint buffering_time;	// DEFAULT_BUFFERING_TIME
} streaming_default_t;

typedef struct {
	buffer_type_e	streaming_buffer_type;
	streaming_buffer_t buffer_handle[BUFFER_TYPE_MAX];	/* front buffer : queue2 */

	streaming_requirement_t buffering_req;
	streaming_default_t default_val;

	mmplayer_buffering_state_e buffering_state;
	gboolean	is_adaptive_streaming;

	gint		buffering_percent;
	guint		buffer_max_bitrate;
	guint		buffer_avg_bitrate;
	gboolean	need_update;
	gboolean	need_sync;
	gint		ring_buffer_size;
} mmplayer_streaming_t;


mmplayer_streaming_t *_mm_player_streaming_create(void);

void _mm_player_streaming_initialize(mmplayer_streaming_t *streaming_player, gboolean buffer_init);

void _mm_player_streaming_destroy(mmplayer_streaming_t *streaming_player);

void _mm_player_streaming_set_queue2(mmplayer_streaming_t *streamer, GstElement *buffer,
					gboolean use_buffering, muxed_buffer_type_e type, guint64 content_size);

void _mm_player_streaming_set_multiqueue(mmplayer_streaming_t *streamer, GstElement *buffer);

void _mm_player_streaming_sync_property(mmplayer_streaming_t *streamer, GstElement *decodebin);

void _mm_player_streaming_buffering(mmplayer_streaming_t *streamer, GstMessage *buffering_msg,
					guint64 content_size, gint64 position, gint64 duration);

void _mm_player_streaming_set_content_bitrate(mmplayer_streaming_t *streaming_player, guint max_bitrate, guint avg_bitrate);

#endif
