/*
 * libmm-player
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, YeJin Cho <cho.yejin@samsung.com>,
 * Seungbae Shin <seungbae.shin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __MM_PLAYER_UTILS_H__
#define __MM_PLAYER_UTILS_H__
#include <limits.h>
#include <glib.h>
#include <gst/gst.h>
#include <dlog.h>
#include <mm_player_ini.h>
#include <mm_types.h>
#include <mm_error.h>
#include <mm_message.h>
#include "mm_player_priv.h"

#ifdef __cplusplus
	extern "C" {
#endif

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "MM_PLAYER"

/* general */
#ifndef ARRAY_SIZE
#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))
#endif

#define MMPLAYER_MIN_INT (INT_MIN)
#define MMPLAYER_MAX_INT (INT_MAX)

#define MMPLAYER_FREEIF(x) \
	do {	\
		if (x) { \
			g_free(x); \
			x = NULL;	\
		} \
	} while (0)

#define MMPLAYER_GET_ATTRS(x_player) ((mmplayer_t *)x_player)->attrs

/* command */
#define MMPLAYER_CMD_LOCK(x_player)                          g_mutex_lock(&((mmplayer_t *)x_player)->cmd_lock)
#define MMPLAYER_CMD_TRYLOCK(x_player)                       g_mutex_trylock(&((mmplayer_t *)x_player)->cmd_lock)
#define MMPLAYER_CMD_UNLOCK(x_player)                        g_mutex_unlock(&((mmplayer_t *)x_player)->cmd_lock)

/* playback */
#define MMPLAYER_PLAYBACK_LOCK(x_player)                     g_mutex_lock(&((mmplayer_t *)x_player)->playback_lock)
#define MMPLAYER_PLAYBACK_UNLOCK(x_player)                   g_mutex_unlock(&((mmplayer_t *)x_player)->playback_lock)

/* capture thread */
#define MMPLAYER_CAPTURE_THREAD_LOCK(x_player)               g_mutex_lock(&((mmplayer_t *)x_player)->capture_thread_mutex)
#define MMPLAYER_CAPTURE_THREAD_UNLOCK(x_player)             g_mutex_unlock(&((mmplayer_t *)x_player)->capture_thread_mutex)
#define MMPLAYER_CAPTURE_THREAD_WAIT(x_player)               g_cond_wait(&((mmplayer_t *)x_player)->capture_thread_cond, &((mmplayer_t *)x_player)->capture_thread_mutex)
#define MMPLAYER_CAPTURE_THREAD_SIGNAL(x_player)             g_cond_signal(&((mmplayer_t *)x_player)->capture_thread_cond);

/* gapless play thread */
#define MMPLAYER_GAPLESS_PLAY_THREAD_LOCK(x_player)             g_mutex_lock(&((mmplayer_t *)x_player)->gapless_play_thread_mutex)
#define MMPLAYER_GAPLESS_PLAY_THREAD_UNLOCK(x_player)           g_mutex_unlock(&((mmplayer_t *)x_player)->gapless_play_thread_mutex)
#define MMPLAYER_GAPLESS_PLAY_THREAD_WAIT(x_player)             g_cond_wait(&((mmplayer_t *)x_player)->gapless_play_thread_cond, &((mmplayer_t *)x_player)->gapless_play_thread_mutex)
#define MMPLAYER_GAPLESS_PLAY_THREAD_SIGNAL(x_player)           g_cond_signal(&((mmplayer_t *)x_player)->gapless_play_thread_cond);

/* gst bus msg thread */
#define MMPLAYER_BUS_MSG_THREAD_LOCK(x_player)                   g_mutex_lock(&((mmplayer_t *)x_player)->bus_msg_thread_mutex)
#define MMPLAYER_BUS_MSG_THREAD_UNLOCK(x_player)                 g_mutex_unlock(&((mmplayer_t *)x_player)->bus_msg_thread_mutex)
#define MMPLAYER_BUS_MSG_THREAD_WAIT(x_player)                   g_cond_wait(&((mmplayer_t *)x_player)->bus_msg_thread_cond, &((mmplayer_t *)x_player)->bus_msg_thread_mutex)
#define MMPLAYER_BUS_MSG_THREAD_WAIT_UNTIL(x_player, end_time)   g_cond_wait_until(&((mmplayer_t *)x_player)->bus_msg_thread_cond, &((mmplayer_t *)x_player)->bus_msg_thread_mutex, end_time)
#define MMPLAYER_BUS_MSG_THREAD_SIGNAL(x_player)                 g_cond_signal(&((mmplayer_t *)x_player)->bus_msg_thread_cond);

/* handling fakesink */
#define MMPLAYER_FSINK_LOCK(x_player)                        g_mutex_lock(&((mmplayer_t *)x_player)->fsink_lock)
#define MMPLAYER_FSINK_UNLOCK(x_player)                      g_mutex_unlock(&((mmplayer_t *)x_player)->fsink_lock)

/* handling update tag */
#define MMPLAYER_UPDATE_TAG_LOCK(x_player)                   g_mutex_lock(&((mmplayer_t *)x_player)->update_tag_lock)
#define MMPLAYER_UPDATE_TAG_UNLOCK(x_player)                 g_mutex_unlock(&((mmplayer_t *)x_player)->update_tag_lock)

/* video stream bo */
#define MMPLAYER_VIDEO_BO_LOCK(x_player)                     g_mutex_lock(&((mmplayer_t *)x_player)->video_bo_mutex)
#define MMPLAYER_VIDEO_BO_UNLOCK(x_player)                   g_mutex_unlock(&((mmplayer_t *)x_player)->video_bo_mutex)
#define MMPLAYER_VIDEO_BO_WAIT(x_player)                     g_cond_wait(&((mmplayer_t *)x_player)->video_bo_cond, &((mmplayer_t *)x_player)->video_bo_mutex)
#define MMPLAYER_VIDEO_BO_WAIT_UNTIL(x_player, end_time)     g_cond_wait_until(&((mmplayer_t *)x_player)->video_bo_cond, &((mmplayer_t *)x_player)->video_bo_mutex, end_time)
#define MMPLAYER_VIDEO_BO_SIGNAL(x_player)                   g_cond_signal(&((mmplayer_t *)x_player)->video_bo_cond);

/* subtitle info */
#define MMPLAYER_SUBTITLE_INFO_LOCK(x_player)                     g_mutex_lock(&((mmplayer_t *)x_player)->subtitle_info_mutex)
#define MMPLAYER_SUBTITLE_INFO_UNLOCK(x_player)                   g_mutex_unlock(&((mmplayer_t *)x_player)->subtitle_info_mutex)
#define MMPLAYER_SUBTITLE_INFO_WAIT(x_player)                     g_cond_wait(&((mmplayer_t *)x_player)->subtitle_info_cond, &((mmplayer_t *)x_player)->subtitle_info_mutex)
#define MMPLAYER_SUBTITLE_INFO_WAIT_UNTIL(x_player, end_time)     g_cond_wait_until(&((mmplayer_t *)x_player)->subtitle_info_cond, &((mmplayer_t *)x_player)->subtitle_info_mutex, end_time)
#define MMPLAYER_SUBTITLE_INFO_SIGNAL(x_player)                   g_cond_signal(&((mmplayer_t *)x_player)->subtitle_info_cond);

#if 0
#define MMPLAYER_FENTER();					LOGD("<ENTER>");
#define MMPLAYER_FLEAVE();					LOGD("<LEAVE>");
#else
#define MMPLAYER_FENTER();
#define MMPLAYER_FLEAVE();
#endif

#define MMPLAYER_RETURN_IF_FAIL(expr) \
	do { \
		if (!(expr)) { \
			LOGW("failed [%s]", #expr); \
			return; \
		} \
	} while (0)

#define MMPLAYER_RETURN_VAL_IF_FAIL(expr, var) \
	do { \
		if (!(expr)) { \
			LOGW("failed [%s]", #expr); \
			return (var); \
		} \
	} while (0)

/* debug caps string */
#define MMPLAYER_LOG_GST_CAPS_TYPE(x_caps) \
	do { \
		gchar *caps_type = NULL; \
		if (x_caps) { \
			caps_type = gst_caps_to_string(x_caps); \
			LOGD("caps: %s", caps_type); \
			MMPLAYER_FREEIF(caps_type); \
		} else {\
			LOGW("caps is null"); \
		} \
	} while (0)

/* message posting */
#define MMPLAYER_POST_MSG(x_player, x_msgtype, x_msg_param) \
	do { \
		LOGD("[handle: %p] posting %s to application", x_player, #x_msgtype); \
		_mmplayer_post_message(x_player, x_msgtype, x_msg_param); \
	} while (0)

/* setting player state */
#define MMPLAYER_SET_STATE(x_player, x_state) \
	do { \
		LOGD("[handle: %p] update state machine to %d", x_player, x_state); \
		_mmplayer_set_state(x_player, x_state); \
	} while (0)

#define MMPLAYER_CHECK_STATE(x_player, x_command) \
	do {	\
		LOGD("[handle: %p] checking player state before doing %s", x_player, #x_command); \
		switch (_mmplayer_check_state(x_player, x_command)) { \
		case MM_ERROR_PLAYER_INVALID_STATE: \
			return MM_ERROR_PLAYER_INVALID_STATE; \
			break; \
			/* NOTE : for robustness of player. we won't treat it as an error */ \
		case MM_ERROR_PLAYER_NO_OP: \
			return MM_ERROR_NONE; \
			break; \
		case MM_ERROR_PLAYER_DOING_SEEK: \
			return MM_ERROR_PLAYER_DOING_SEEK; \
		default: \
			break; \
		}	\
	} while (0)

/* setting element state */
#define MMPLAYER_ELEMENT_SET_STATE(x_element, x_state) \
	do {	\
		LOGD("setting state [%s:%d] to [%s]", #x_state, x_state, GST_ELEMENT_NAME(x_element)); \
		if (GST_STATE_CHANGE_FAILURE == gst_element_set_state(x_element, x_state)) { \
			LOGE("failed to set state %s to %s", #x_state, GST_ELEMENT_NAME(x_element)); \
			goto STATE_CHANGE_FAILED; \
		}	\
	} while (0)

#define MMPLAYER_CHECK_NULL(x_var) \
	do { \
		if (!x_var) { \
			LOGE("[%s] is NULL", #x_var); \
			goto ERROR; \
		} \
	} while (0)

/* generating dot */
#define MMPLAYER_GENERATE_DOT_IF_ENABLED(x_player, x_name) \
	do { \
		if (x_player->ini.generate_dot) { \
			GST_DEBUG_BIN_TO_DOT_FILE(GST_BIN(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst), \
			GST_DEBUG_GRAPH_SHOW_ALL, x_name); \
		} \
	} while (0)

/* signal manipulation */
#define MMPLAYER_SIGNAL_CONNECT(x_player, x_object, x_type, x_signal, x_callback, x_arg) \
	do { \
		mmplayer_signal_item_t *item = NULL; \
		item = (mmplayer_signal_item_t *)g_malloc(sizeof(mmplayer_signal_item_t)); \
		if (!item) { \
			LOGE("cannot connect signal [%s]", x_signal); \
		} else { \
			item->obj = G_OBJECT(x_object); \
			item->sig = g_signal_connect(G_OBJECT(x_object), x_signal, \
						x_callback, x_arg); \
			if ((x_type >= MM_PLAYER_SIGNAL_TYPE_AUTOPLUG) && (x_type < MM_PLAYER_SIGNAL_TYPE_MAX)) \
				x_player->signals[x_type] = g_list_append(x_player->signals[x_type], item); \
			else \
				LOGE("wrong signal type [%d]", x_type); \
		} \
	} while (0)

/* create element */
#define MMPLAYER_CREATE_ELEMENT(x_bin, x_id, x_factory, x_name, x_bucket, x_player) \
	do {\
		x_bin[x_id].id = x_id;\
		x_bin[x_id].gst = gst_element_factory_make(x_factory, x_name);\
		if (!x_bin[x_id].gst) {\
			LOGE("failed to create %s", x_factory);\
			goto ERROR;\
		} else {\
			if (x_player->ini.set_dump_element_flag)\
				__mmplayer_add_dump_buffer_probe(x_player, x_bin[x_id].gst);\
		} \
		x_bucket = g_list_append(x_bucket, &x_bin[x_id]);\
	} while (0);

/* release element resource */
#define MMPLAYER_RELEASE_ELEMENT(x_player, x_bin, x_id) \
	do { \
		if (x_bin[x_id].gst) { \
			gst_element_set_state(x_bin[x_id].gst, GST_STATE_NULL); \
			gst_bin_remove(GST_BIN(x_player->pipeline->mainbin[MMPLAYER_M_PIPE].gst), x_bin[x_id].gst); \
			x_bin[x_id].gst = NULL; \
			LOGD("release done [element %d]", x_id); \
		} \
	} while (0)

/* state */
#define MMPLAYER_PREV_STATE(x_player) ((mmplayer_t *)x_player)->prev_state
#define MMPLAYER_CURRENT_STATE(x_player) ((mmplayer_t *)x_player)->state
#define MMPLAYER_PENDING_STATE(x_player) ((mmplayer_t *)x_player)->pending_state
#define MMPLAYER_TARGET_STATE(x_player) ((mmplayer_t *)x_player)->target_state
#define MMPLAYER_STATE_GET_NAME(state) _mmplayer_get_state_name(state)

#define MMPLAYER_PRINT_STATE(x_player) \
	do { \
		LOGD("-- prev %s, current %s, pending %s, target %s --", \
			MMPLAYER_STATE_GET_NAME(MMPLAYER_PREV_STATE(x_player)), \
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(x_player)), \
			MMPLAYER_STATE_GET_NAME(MMPLAYER_PENDING_STATE(x_player)), \
			MMPLAYER_STATE_GET_NAME(MMPLAYER_TARGET_STATE(x_player))); \
	} while (0)

#define MMPLAYER_STATE_CHANGE_TIMEOUT(x_player) ((mmplayer_t *)x_player)->state_change_timeout

/* streaming */
#define MMPLAYER_IS_STREAMING(x_player) _mmplayer_is_streaming(x_player)
#define MMPLAYER_IS_RTSP_STREAMING(x_player) _mmplayer_is_rtsp_streaming(x_player)
#define MMPLAYER_IS_HTTP_STREAMING(x_player) _mmplayer_is_http_streaming(x_player)
#define MMPLAYER_IS_HTTP_LIVE_STREAMING(x_player) _mmplayer_is_http_live_streaming(x_player)
#define MMPLAYER_IS_LIVE_STREAMING(x_player) _mmplayer_is_live_streaming(x_player)
#define MMPLAYER_IS_DASH_STREAMING(x_player) _mmplayer_is_dash_streaming(x_player)
#define MMPLAYER_IS_SMOOTH_STREAMING(x_player) _mmplayer_is_smooth_streaming(x_player)
#define MMPLAYER_IS_MS_BUFF_SRC(x_player) _mmplayer_is_ms_buff_src(x_player)

#define MMPLAYER_URL_HAS_DASH_SUFFIX(x_player) _mmplayer_has_suffix(x_player, ".mpd")
#define MMPLAYER_URL_HAS_HLS_SUFFIX(x_player) _mmplayer_has_suffix(x_player, ".m3u8")

#define MMPLAYER_STREAM_TYPE_GET_NAME(type) _mmplayer_get_stream_type_name(type)

/*===========================================================================================
|																							|
|  GLOBAL FUNCTION PROTOTYPES																|
|																							|
========================================================================================== */

const gchar *_mmplayer_get_state_name(int state);
const gchar *_mmplayer_get_stream_type_name(int type);

gboolean _mmplayer_is_streaming(mmplayer_t *player);
gboolean _mmplayer_is_rtsp_streaming(mmplayer_t *player);
gboolean _mmplayer_is_live_streaming(mmplayer_t *player);
gboolean _mmplayer_is_http_streaming(mmplayer_t *player);
gboolean _mmplayer_is_http_live_streaming(mmplayer_t *player);
gboolean _mmplayer_is_dash_streaming(mmplayer_t *player);
gboolean _mmplayer_is_smooth_streaming(mmplayer_t *player);
gboolean _mmplayer_is_ms_buff_src(mmplayer_t *player);
gboolean _mmplayer_has_suffix(mmplayer_t *player, const gchar *suffix);
gboolean _mmplayer_is_videosink_ready(mmplayer_t *player, int surface_type);

gboolean _mmplayer_post_message(mmplayer_t *player, enum MMMessageType msgtype, MMMessageParamType *param);
gboolean _mmplayer_dump_pipeline_state(mmplayer_t *player);

bool _mmplayer_is_sdp_file(const char *path);
int _mmplayer_exist_file_path(const char *file_path);
char **_mmplayer_get_cookie_list(const char *cookies);
const char *_mmplayer_get_charset(const char *file_path);
int _mmplayer_get_pixtype(unsigned int fourcc);
bool _mmplayer_get_storage_info(const char *path, mmplayer_storage_info_t *storage_info);

media_format_mimetype_e _mmplayer_convert_audio_pcm_str_to_media_format_mime(const gchar *audio_pcm_str);

#ifdef __cplusplus
	}
#endif

#endif /* __MM_PLAYER_UTILS_H__ */
