/*
 * libmm-player
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, YeJin Cho <cho.yejin@samsung.com>,
 * Seungbae Shin <seungbae.shin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <gst/gst.h>
#include <string.h>

#include <dlog.h>
#include <mm_types.h>
#include <mm_message.h>

#include "mm_player.h"
#include "mm_player_priv.h"
#include "mm_player_attrs.h"
#include "mm_player_utils.h"
#include "mm_player_ini.h"
#include "mm_player_capture.h"
#include "mm_player_tracks.h"
#include "mm_player_es.h"
#include "mm_player_360.h"
#include "mm_player_gst.h"

int mm_player_create(MMHandleType *player)
{
	int result = MM_ERROR_PLAYER_INTERNAL;
	mmplayer_t *new_player = NULL;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* alloc player structure */
	new_player = g_try_new0(mmplayer_t, 1);
	if (!new_player) {
		LOGE("Cannot allocate memory for player\n");
		result = MM_ERROR_PLAYER_RESOURCE_LIMIT;
		goto ERROR;
	}

	/* create player cmd lock */
	g_mutex_init(&new_player->cmd_lock);

	/* create player playback lock */
	g_mutex_init(&new_player->playback_lock);

	/* load ini files */
	if (MM_ERROR_NONE != mm_player_ini_load(&new_player->ini)) {
		LOGE("can't load ini");
		goto ERROR;
	}

	if (MM_ERROR_NONE != mm_player_audio_effect_ini_load(&new_player->ini)) {
		LOGE("can't load audio ini");
		goto ERROR;
	}

	/* create player */
	result = _mmplayer_create_player((MMHandleType)new_player);
	if (result != MM_ERROR_NONE) {
		LOGE("failed to create player");
		if (result != MM_ERROR_PLAYER_RESOURCE_LIMIT)
			result = MM_ERROR_PLAYER_INTERNAL;
		goto ERROR;
	}

	*player = (MMHandleType)new_player;

	return MM_ERROR_NONE;

ERROR:

	if (new_player) {
		_mmplayer_destroy((MMHandleType)new_player);
		g_mutex_clear(&new_player->cmd_lock);
		g_mutex_clear(&new_player->playback_lock);

		MMPLAYER_FREEIF(new_player);
	}

	*player = (MMHandleType)0;
	return result; /* MM_ERROR_PLAYER_INTERNAL or MM_ERROR_PLAYER_RESOURCE_LIMIT */
}

int mm_player_destroy(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* destroy the gst bus msg thread if it is remained.
	   this funct have to be called before getting cmd lock. */
	_mmplayer_bus_msg_thread_destroy(player);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_destroy(player);

	MMPLAYER_CMD_UNLOCK(player);

	g_mutex_clear(&((mmplayer_t *)player)->cmd_lock);
	g_mutex_clear(&((mmplayer_t *)player)->playback_lock);

	memset((mmplayer_t *)player, 0x00, sizeof(mmplayer_t));

	/* free player */
	g_free((void *)player);

	return result;
}

int mm_player_realize(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_realize(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_abort_pause(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* destroy the gst bus msg thread not to be blocked in pause(without cmd lock). */
	_mmplayer_bus_msg_thread_destroy(player);

	/* abort the pause operation for preparing(without cmd lock). */
	result = _mmplayer_abort_pause(player);

	return result;
}

int mm_player_unrealize(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_unrealize(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_message_callback(MMHandleType player, MMMessageCallback callback, void *user_param)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_message_callback(player, callback, user_param);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_audio_decoded_callback(MMHandleType player, mmplayer_audio_extract_opt_e opt, mm_player_audio_decoded_callback callback, void *user_param)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_audio_decoded_cb(player, opt, callback, user_param);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_video_decoded_callback(MMHandleType player, mm_player_video_decoded_callback callback, void *user_param)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_video_decoded_cb(player, callback, user_param);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_do_video_capture(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_do_video_capture(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_volume(MMHandleType player, float volume)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_volume(player, volume);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_volume(MMHandleType player, float *volume)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(volume, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_volume(player, volume);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_mute(MMHandleType player, bool mute)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_mute(player, mute);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_mute(MMHandleType player, bool *mute)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(mute, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_mute(player, mute);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_state(MMHandleType player, mmplayer_state_e *state)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(state, MM_ERROR_COMMON_INVALID_ARGUMENT);

	*state = MM_PLAYER_STATE_NULL;

	result = _mmplayer_get_state(player, (int *)state);

	return result;
}

/* NOTE : It does not support some use cases, eg using colorspace converter */
int mm_player_change_videosink(MMHandleType player, MMDisplaySurfaceType display_surface_type, int wl_surface_id)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_change_videosink(player, display_surface_type, wl_surface_id);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_start(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_start(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int  mm_player_stop(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_stop(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_pause(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_pause(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_resume(MMHandleType player)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_resume(player);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_play_speed(MMHandleType player, float rate, bool streaming)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_playspeed(player, rate, streaming);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_position(MMHandleType player, int64_t pos)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_position(player, pos);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_position(MMHandleType player, int64_t *pos)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(pos, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_gst_get_position((mmplayer_t *)player, pos);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_duration(MMHandleType player, int64_t *dur)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(dur, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_duration(player, dur);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}


int mm_player_get_buffer_position(MMHandleType player, int *start_pos, int *end_pos)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(start_pos && end_pos, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_buffer_position(player, start_pos, end_pos);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_external_subtitle_path(MMHandleType player, const char *path)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_external_subtitle_path(player, path);

	MMPLAYER_CMD_UNLOCK(player);
	return result;
}

int mm_player_adjust_subtitle_position(MMHandleType player, int pos)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_adjust_subtitle_postion(player, pos);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_subtitle_silent(MMHandleType player, int silent)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_subtitle_silent(player, silent);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_subtitle_silent(MMHandleType player, int *silent)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_subtitle_silent(player, silent);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_attribute(MMHandleType player,  char **err_attr_name, const char *first_attribute_name, ...)
{
	int result = MM_ERROR_NONE;
	va_list var_args;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(first_attribute_name, MM_ERROR_COMMON_INVALID_ARGUMENT);

	va_start(var_args, first_attribute_name);
	result = _mmplayer_set_attribute(player, err_attr_name, first_attribute_name, var_args);
	va_end(var_args);

	return result;
}

int mm_player_get_attribute(MMHandleType player,  char **err_attr_name, const char *first_attribute_name, ...)
{
	int result = MM_ERROR_NONE;
	va_list var_args;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(first_attribute_name, MM_ERROR_COMMON_INVALID_ARGUMENT);

	va_start(var_args, first_attribute_name);
	result = _mmplayer_get_attribute(player, err_attr_name, first_attribute_name, var_args);
	va_end(var_args);

	return result;
}

int mm_player_get_attribute_info(MMHandleType player,  const char *attribute_name, mmplayer_attrs_info_t *info)
{
	int result = MM_ERROR_NONE;


	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(attribute_name, MM_ERROR_COMMON_INVALID_ARGUMENT);
	MMPLAYER_RETURN_VAL_IF_FAIL(info, MM_ERROR_COMMON_INVALID_ARGUMENT);

	result = _mmplayer_get_attributes_info((MMHandleType)player, attribute_name, info);

	return result;
}

int mm_player_get_track_count(MMHandleType player, mmplayer_track_type_e type, int *count)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(count, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_track_count(player, type, count);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_select_track(MMHandleType player, mmplayer_track_type_e type, int index)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_select_track(player, type, index);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_current_track(MMHandleType player, mmplayer_track_type_e type, int *index)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(index, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_current_track(player, type, index);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_track_language_code(MMHandleType player,  mmplayer_track_type_e type, int index, char **code)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_track_language_code(player, type, index, code);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_uri(MMHandleType player, const char *uri)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_uri(player, uri);

	MMPLAYER_CMD_UNLOCK(player);

	return result;

}

int mm_player_set_next_uri(MMHandleType player, const char *uri)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_next_uri(player, uri, FALSE);

	MMPLAYER_CMD_UNLOCK(player);

	return result;

}

int mm_player_get_next_uri(MMHandleType player, char **uri)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_next_uri(player, uri);

	MMPLAYER_CMD_UNLOCK(player);

	return result;

}

int mm_player_has_closed_caption(MMHandleType player, bool *exist)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(exist, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_has_closed_caption(player, exist);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

void mm_player_video_stream_internal_buffer_unref(void *buffer)
{
	_mm_player_video_stream_internal_buffer_unref(buffer);
}

int mm_player_submit_packet(MMHandleType player, media_packet_h packet)
{

	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* no lock here, otherwise callback for the "need-data" signal of appsrc will be blocking */
	//MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_submit_packet(player, packet);

	//MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_video_info(MMHandleType player, media_format_h format)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_video_info(player, format);

	MMPLAYER_CMD_UNLOCK(player);

	return result;

}

int mm_player_set_audio_info(MMHandleType player, media_format_h format)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_audio_info(player, format);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_subtitle_info(MMHandleType player, mmplayer_subtitle_stream_info_t *subtitle_stream_info)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_subtitle_info(player, subtitle_stream_info);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_media_stream_buffer_max_size(MMHandleType player, mmplayer_stream_type_e type, unsigned long long max_size)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_media_stream_max_size(player, type, max_size);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_media_stream_buffer_max_size(MMHandleType player, mmplayer_stream_type_e type, unsigned long long *max_size)
{
	int result = MM_ERROR_NONE;
	guint64 _max_size = 0;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(max_size, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_media_stream_max_size(player, type, &_max_size);
	*max_size = _max_size;

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_media_stream_buffer_min_percent(MMHandleType player, mmplayer_stream_type_e type, unsigned min_percent)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_media_stream_min_percent(player, type, min_percent);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_media_stream_buffer_min_percent(MMHandleType player, mmplayer_stream_type_e type, unsigned int *min_percent)
{
	int result = MM_ERROR_NONE;

	LOGD("\n");

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(min_percent, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_media_stream_min_percent(player, type, min_percent);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_audio_stream_changed_callback(MMHandleType player, mm_player_stream_changed_callback callback, void *user_param)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_audiostream_changed_cb(player, callback, user_param);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_timeout(MMHandleType player, int *timeout)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(timeout, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_timeout(player, timeout);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_media_stream_dynamic_resolution(MMHandleType player, bool drc)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_media_stream_dynamic_resolution(player, drc);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_release_video_stream_bo(MMHandleType player, void *bo)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(bo, MM_ERROR_COMMON_INVALID_ARGUMENT);

	//MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_video_stream_release_bo(player, bo);

	//MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_sound_stream_info(MMHandleType player, char *stream_type, int stream_index)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL((stream_type && (stream_index >= 0)), MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = mm_player_set_attribute(player, NULL, "sound_stream_type", stream_type, strlen(stream_type), "sound_stream_index", stream_index, NULL);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_manage_external_storage_state(MMHandleType player, int id, int state)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_manage_external_storage_state(player, id, state);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_adaptive_variant_info(MMHandleType player, int *num, char **var_info)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(num && var_info, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_adaptive_variant_info(player, num, var_info);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_max_adaptive_variant_limit(MMHandleType player, int bandwidth, int width, int height)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_max_adaptive_variant_limit(player, bandwidth, width, height);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_max_adaptive_variant_limit(MMHandleType player, int *bandwidth, int *width, int *height)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(bandwidth && width && height, MM_ERROR_COMMON_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_max_adaptive_variant_limit(player, bandwidth, width, height);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_streaming_buffering_time(MMHandleType player, int *prebuffer_ms, int *rebuffer_ms)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(prebuffer_ms && rebuffer_ms, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_streaming_buffering_time(player, prebuffer_ms, rebuffer_ms);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_audio_only(MMHandleType player, bool audio_only)
{
	int result = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_audio_only(player, audio_only);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_audio_only(MMHandleType player, bool *audio_only)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(audio_only, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_audio_only(player, audio_only);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_is_content_spherical(MMHandleType player, bool *is_spherical)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(is_spherical, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_is_content_spherical(player, is_spherical);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_set_enabled(MMHandleType player, bool enabled)
{
	int result = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_set_enabled(player, enabled);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_is_enabled(MMHandleType player, bool *enabled)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(enabled, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_is_enabled(player, enabled);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_set_direction_of_view(MMHandleType player, float yaw, float pitch)
{
	int result = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_set_direction_of_view(player, yaw, pitch);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_get_direction_of_view(MMHandleType player, float *yaw, float *pitch)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(yaw && pitch, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_get_direction_of_view(player, yaw, pitch);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_set_zoom(MMHandleType player, float level)
{
	int result = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_set_zoom(player, level);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_get_zoom(MMHandleType player, float *level)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(level, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_get_zoom(player, level);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_set_field_of_view(MMHandleType player, int horizontal_degrees, int vertical_degrees)
{
	int result = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_set_field_of_view(player, horizontal_degrees, vertical_degrees);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_360_get_field_of_view(MMHandleType player, int *horizontal_degrees, int *vertical_degrees)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(horizontal_degrees && vertical_degrees, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_360_get_field_of_view(player, horizontal_degrees, vertical_degrees);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_codec_type(MMHandleType player, mmplayer_stream_type_e stream_type, mmplayer_codec_type_e codec_type)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_codec_type(player, stream_type, codec_type);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_replaygain_enabled(MMHandleType player, bool enabled)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_replaygain_enabled(player, enabled);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_is_replaygain_enabled(MMHandleType player, bool *enabled)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(enabled, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_is_replaygain_enabled(player, enabled);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_video_roi_area(MMHandleType player, double scale_x, double scale_y, double scale_width, double scale_height)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_video_roi_area(player, scale_x, scale_y, scale_width, scale_height);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_get_video_roi_area(MMHandleType player, double *scale_x, double *scale_y, double *scale_width, double *scale_height)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_get_video_roi_area(player, scale_x, scale_y, scale_width, scale_height);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_set_client_pid(MMHandleType player, int pid)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_set_client_pid(player, pid);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_audio_offload_is_activated(MMHandleType player, bool *activated)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(activated, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);

	result = _mmplayer_audio_offload_is_activated(player, activated);

	MMPLAYER_CMD_UNLOCK(player);

	return result;
}

int mm_player_is_audio_control_available(MMHandleType player, mmplayer_audio_control_opt_e opt, bool *available)
{
	int result = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(available, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_CMD_LOCK(player);
	result = _mmplayer_is_audio_control_available(player, opt, available);
	MMPLAYER_CMD_UNLOCK(player);

	return result;
}