/*
 * libmm-player
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <mm_attrs.h>
#include <mm_error.h>
#include "mm_player_priv.h"
#include "mm_player_360.h"
#include "mm_player_utils.h"
#include <gst/gst.h>
#include <gst/gstutils.h>
#include <glib.h>

static bool
__mmplayer_check_video_360_used(mmplayer_t *player)
{
	/* check video 360 plugin is created */
	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
		player->pipeline &&
		player->pipeline->videobin &&
		player->pipeline->videobin[MMPLAYER_V_BIN].gst &&
		player->pipeline->videobin[MMPLAYER_V_360].gst,
		false);

	return true;
}

static bool
__mmplayer_check_audio_sink(mmplayer_t *player)
{
	/* check audio sink is created */
	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
		player->pipeline &&
		player->pipeline->videobin &&
		player->pipeline->videobin[MMPLAYER_A_BIN].gst &&
		player->pipeline->videobin[MMPLAYER_A_SINK].gst,
		false);

	return true;
}

int
_mmplayer_360_is_content_spherical(MMHandleType hplayer, bool *is_spherical)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(is_spherical, MM_ERROR_INVALID_ARGUMENT);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	if (player->state < MM_PLAYER_STATE_READY) {
		*is_spherical = false;
		return MM_ERROR_PLAYER_INVALID_STATE;
	}

	*is_spherical = (bool)player->is_content_spherical;

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_set_enabled(MMHandleType hplayer, bool enabled)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	player->is_video360_enabled = (gboolean)enabled;

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		/* We will get here if player is pending ready or ready and above */
		LOGD("set enabled %d", player->is_video360_enabled);
		g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"passthrough", !player->is_video360_enabled, NULL);
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_is_enabled(MMHandleType hplayer, bool *enabled)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gboolean is_enabled;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(enabled, MM_ERROR_INVALID_ARGUMENT);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		g_object_get(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"passthrough", &is_enabled, NULL);
		*enabled = !((bool)is_enabled);
	} else {
		*enabled = (bool)player->is_video360_enabled;
	}

	LOGD("get enabled %d", *enabled);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_set_direction_of_view(MMHandleType hplayer, float yaw, float pitch)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if (yaw > (float)(M_PI) || yaw < (float)(-M_PI) || pitch > (float)(M_PI_2) || pitch < (float)(-M_PI_2)) {
		LOGE("invalid argument %f, %f", yaw, pitch);
		return MM_ERROR_INVALID_ARGUMENT;
	}

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	player->video360_yaw_radians = yaw;
	player->video360_pitch_radians = pitch;

	if (player->is_content_spherical) {
		/* We will get here if player is pending ready or ready and above */
		if (__mmplayer_check_video_360_used(player)) {
			LOGD("set yaw %f, pitch %f for video", yaw, pitch);
			g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
					"pose-yaw", (int)(yaw * 180.0f / M_PI),
					"pose-pitch", (int)(pitch * 180.0f / M_PI), NULL);
		}
		if (player->is_openal_plugin_used && __mmplayer_check_audio_sink(player)) {
			g_object_set(G_OBJECT(player->pipeline->audiobin[MMPLAYER_A_SINK].gst),
					"source-orientation-y", (int)(yaw * 180.0f / M_PI),
					"source-orientation-x", (int)(pitch * 180.0f / M_PI), NULL);
		}
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_get_direction_of_view(MMHandleType hplayer, float *yaw, float *pitch)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int yaw_degrees = 0;
	int pitch_degrees = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(yaw && pitch, MM_ERROR_INVALID_ARGUMENT);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		g_object_get(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"pose-yaw", &yaw_degrees, "pose-pitch", &pitch_degrees, NULL);
		*yaw = M_PI * yaw_degrees / 180.0f;
		*pitch = M_PI * pitch_degrees / 180.0f;

		LOGD("get yaw %f, pitch %f", *yaw, *pitch);
		return MM_ERROR_NONE;
	}

	*yaw = player->video360_yaw_radians;
	*pitch = player->video360_pitch_radians;

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_set_zoom(MMHandleType hplayer, float level)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if (level < 1.0f || level > VIDEO360_MAX_ZOOM) {
		LOGE("invalid argument %f", level);
		return MM_ERROR_INVALID_ARGUMENT;
	}

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	player->video360_zoom = level;

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		/* We will get here if player is pending ready or ready and above */

		LOGD("set level %f", level);
		g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"zoom", 1.0f / level, NULL);
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_get_zoom(MMHandleType hplayer, float *level)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	float current_zoom = 0.0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(level, MM_ERROR_INVALID_ARGUMENT);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		g_object_get(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"zoom", &current_zoom, NULL);
		*level = 1.0f / current_zoom;

		LOGD("get level %f", *level);
		return MM_ERROR_NONE;
	}

	*level = player->video360_zoom;

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_set_field_of_view(MMHandleType hplayer, int horizontal_degrees, int vertical_degrees)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if (horizontal_degrees < 1 || horizontal_degrees > 360 || vertical_degrees < 1 || vertical_degrees > 180) {
		LOGE("invalid argument %d, %d", horizontal_degrees, vertical_degrees);
		return MM_ERROR_INVALID_ARGUMENT;
	}

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	player->video360_horizontal_fov = horizontal_degrees;
	player->video360_vertical_fov = vertical_degrees;

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		/* We will get here if player is pending ready or ready and above */
		LOGD("set h-fov %d, v-fov %d", horizontal_degrees, vertical_degrees);
		g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"horizontal-fov", horizontal_degrees, "vertical-fov", vertical_degrees, NULL);
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_360_get_field_of_view(MMHandleType hplayer, int *horizontal_degrees, int *vertical_degrees)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(horizontal_degrees && vertical_degrees, MM_ERROR_INVALID_ARGUMENT);

	LOGD("state %s, spherical info %d",
			MMPLAYER_STATE_GET_NAME(MMPLAYER_CURRENT_STATE(player)), player->is_content_spherical);

	if (player->is_content_spherical && __mmplayer_check_video_360_used(player)) {
		g_object_get(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_360].gst),
				"horizontal-fov", horizontal_degrees, "vertical-fov", vertical_degrees, NULL);

		LOGD("get h-fov %d, v-fov %d", *horizontal_degrees, *vertical_degrees);
		return MM_ERROR_NONE;
	}

	*horizontal_degrees = player->video360_horizontal_fov;
	*vertical_degrees = player->video360_vertical_fov;

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}
