/*
 * libmm-player
 *
 * Copyright(c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, YeJin Cho <cho.yejin@samsung.com>,
 * Seungbae Shin <seungbae.shin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*===========================================================================================
|																							|
|  INCLUDE FILES																			|
|																							|
========================================================================================== */
#include <dlog.h>
#include "mm_player_utils.h"
#include "mm_player_capture.h"
#include "mm_player_priv.h"

#include <mm_util_image.h>
#include <mm_util_imgp.h>
#include <gst/video/video-info.h>

//#define	CAPTURE_OUTPUT_DUMP	1
/*---------------------------------------------------------------------------
|    LOCAL VARIABLE DEFINITIONS for internal								|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    LOCAL FUNCTION PROTOTYPES:												|
---------------------------------------------------------------------------*/
static GstPadProbeReturn __mmplayer_video_capture_probe(GstPad *pad, GstPadProbeInfo *info, gpointer u_data);
static int  __mmplayer_get_video_frame_from_buffer(mmplayer_t *player, GstPad *pad, GstBuffer *buffer);
static gpointer __mmplayer_capture_thread(gpointer data);
static void __csc_tiled_to_linear_crop(unsigned char *yuv420_dest, unsigned char *nv12t_src, int yuv420_width, int yuv420_height, int left, int top, int right, int buttom);
static int __tile_4x2_read(int x_size, int y_size, int x_pos, int y_pos);
static int __mm_player_convert_colorspace(mmplayer_t *player, unsigned char *src_data, size_t src_size, mm_util_color_format_e src_fmt, unsigned int src_w, unsigned int src_h, mm_util_color_format_e dst_fmt);
static int __mm_player_convert_NV12_tiled(mmplayer_t *player);
static int __mm_player_convert_NV12(mmplayer_t *player);
static int __mm_player_convert_I420(mmplayer_t *player);
static int __mm_player_convert_BGRx(mmplayer_t *player);
#ifdef CAPTURE_OUTPUT_DUMP
static void capture_output_dump(mmplayer_t *player);
#endif

/*===========================================================================================
|																							|
|  FUNCTION DEFINITIONS																		|
|																							|
========================================================================================== */
int
_mmplayer_initialize_video_capture(mmplayer_t *player)
{
	int ret = MM_ERROR_NONE;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	/* create capture mutex */
	g_mutex_init(&player->capture_thread_mutex);

	/* create capture cond */
	g_cond_init(&player->capture_thread_cond);

	player->capture_thread_exit = FALSE;

	/* create video capture thread */
	player->capture_thread =
			g_thread_try_new("capture_thread", __mmplayer_capture_thread, (gpointer)player, NULL);

	if (!player->capture_thread) {
		ret = MM_ERROR_PLAYER_RESOURCE_LIMIT;
		goto ERROR;
	}

	return ret;

ERROR:
	/* capture thread */
	g_mutex_clear(&player->capture_thread_mutex);
	g_cond_clear(&player->capture_thread_cond);
	return ret;
}

int
_mmplayer_release_video_capture(mmplayer_t *player)
{
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	/* release capture thread */
	MMPLAYER_CAPTURE_THREAD_LOCK(player);
	player->capture_thread_exit = TRUE;
	MMPLAYER_CAPTURE_THREAD_SIGNAL(player);
	MMPLAYER_CAPTURE_THREAD_UNLOCK(player);

	LOGD("waitting for capture thread exit");
	g_thread_join(player->capture_thread);
	g_mutex_clear(&player->capture_thread_mutex);
	g_cond_clear(&player->capture_thread_cond);
	LOGD("capture thread released");

	return MM_ERROR_NONE;
}

int
_mmplayer_do_video_capture(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;
	GstPad *pad = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* capturing or not */
	if (player->video_capture_cb_probe_id || player->capture.data
			|| player->captured.data[0] || player->captured.data[1]) {
		LOGW("capturing... we can't do any more");
		return MM_ERROR_PLAYER_INVALID_STATE;
	}
	gint surface_type = 0;
	mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &surface_type);

	/* check if video pipeline is linked or not */
	if (!player->pipeline->videobin) {
		LOGW("not ready to capture");
		return MM_ERROR_PLAYER_INVALID_STATE;
	}

	pad = gst_element_get_static_pad(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "sink");

	if (player->state != MM_PLAYER_STATE_PLAYING) {
		if (player->state == MM_PLAYER_STATE_PAUSED) {
			// get last buffer from video sink
			GstSample *sample = NULL;

			gst_element_get_state(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, NULL, NULL, 5 * GST_SECOND); //5 seconds
			g_object_get(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "last-sample", &sample, NULL);

			if (sample) {
				GstBuffer *buf = NULL;
				buf = gst_sample_get_buffer(sample);
				if (buf) {
					if (__mmplayer_get_video_frame_from_buffer(player, pad, buf) != MM_ERROR_NONE)
						ret = MM_ERROR_PLAYER_INTERNAL;
				} else {
					LOGW("failed to get video frame");
				}
				gst_sample_unref(sample);
			}
			gst_object_unref(GST_OBJECT(pad));
			pad = NULL;
			return ret;
		} else {
			LOGW("invalid state(%d) to capture", player->state);
			gst_object_unref(GST_OBJECT(pad));
			pad = NULL;
			return MM_ERROR_PLAYER_INVALID_STATE;
		}
	}

	/* register probe */
	player->video_capture_cb_probe_id = gst_pad_add_probe(pad, GST_PAD_PROBE_TYPE_BUFFER,
		__mmplayer_video_capture_probe, player, NULL);

	gst_object_unref(GST_OBJECT(pad));
	pad = NULL;

	MMPLAYER_FLEAVE();

	return ret;
}

int
__mmplayer_handle_orientation(mmplayer_t *player, int orientation, int format)
{
	unsigned char *src_buffer = NULL;
	int ret = MM_ERROR_NONE;
	mm_util_image_h src_image = NULL;
	mm_util_image_h dst_image = NULL;
	size_t dst_size = 0;
	mm_util_img_rotate_type rot_enum = MM_UTIL_ROTATE_NUM;

	player->capture.orientation = orientation;

	if (orientation == 0) {
		LOGE("no need handle orientation : %d", orientation);
		player->capture.width = player->captured.width[0];
		player->capture.height = player->captured.height[0];
		return MM_ERROR_NONE;
	}

	src_buffer = (unsigned char *)player->capture.data;

	/* convert orientation degree into enum here */
	switch (orientation) {
	case 90:
		rot_enum = MM_UTIL_ROTATE_90;
		break;
	case 180:
		rot_enum = MM_UTIL_ROTATE_180;
		break;
	case 270:
		rot_enum = MM_UTIL_ROTATE_270;
		break;
	default:
		LOGE("wrong rotate value");
		break;
	}

	LOGD("source buffer for rotation = %p and rotation = %d", src_buffer, rot_enum);

	ret = mm_image_create_image(player->captured.width[0], player->captured.height[0], format, src_buffer, (size_t)player->capture.size, &src_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to create image");
		return ret;
	}

	ret = mm_util_rotate_image(src_image, rot_enum, &dst_image);
	mm_image_destroy_image(src_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to rotate image");
		return ret;
	}

	mm_image_debug_image(dst_image, NULL);

	ret = mm_image_get_image(dst_image, &player->capture.width, &player->capture.height, NULL, &player->capture.data, &dst_size);
	mm_image_destroy_image(dst_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to get image");
		return ret;
	}

	free(src_buffer);

	player->capture.size = (int)dst_size;
	player->capture.orientation = orientation;

	player->captured.width[0] = player->captured.stride_width[0] = player->capture.width;
	player->captured.height[0] = player->captured.stride_height[0] = player->capture.height;

	return ret;
}

static gpointer
__mmplayer_capture_thread(gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	MMMessageParamType msg = {0, };
	int orientation = 0;
	int display_angle = 0;
	int ret = 0;
	int i;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, NULL);

	MMPLAYER_CAPTURE_THREAD_LOCK(player);
	while (!player->capture_thread_exit) {
		LOGD("capture thread started. waiting for signal");
		MMPLAYER_CAPTURE_THREAD_WAIT(player);

		if (player->capture_thread_exit) {
			LOGD("exiting capture thread");
			goto EXIT;
		}
		LOGD("capture thread is received signal");

		/* NOTE: Don't use MMPLAYER_CMD_LOCK() here.
		 * Because deadlock can be happened if other player api is used in message callback.
		 */
		if (player->video_cs == MM_PLAYER_COLORSPACE_NV12_TILED) {
			/* Colorspace conversion : NV12T-> NV12-> RGB888 */
			ret = __mm_player_convert_NV12_tiled(player);
			if (ret != MM_ERROR_NONE) {
				LOGE("failed to covert NV12T");
				msg.code = ret;
				goto ERROR;
			}
		} else if (player->video_cs == MM_PLAYER_COLORSPACE_NV12) {
			ret = __mm_player_convert_NV12(player);
			if (ret != MM_ERROR_NONE) {
				LOGE("failed to covert NV12");
				msg.code = ret;
				goto ERROR;
			}
		} else if (player->video_cs == MM_PLAYER_COLORSPACE_I420) {
			ret = __mm_player_convert_I420(player);
			if (ret != MM_ERROR_NONE) {
				LOGE("failed to covert I420");
				msg.code = ret;
				goto ERROR;
			}
		} else if (player->video_cs == MM_PLAYER_COLORSPACE_BGRx) {
			ret = __mm_player_convert_BGRx(player);
			if (ret != MM_ERROR_NONE) {
				LOGE("failed to covert BGRx");
				msg.code = ret;
				goto ERROR;
			}
		} else {
			LOGE("invalid format");
			msg.code = MM_ERROR_PLAYER_INTERNAL;
			goto ERROR;
		}

		ret = _mmplayer_get_video_angle((MMHandleType)player, &display_angle, &orientation);
		if (ret != MM_ERROR_NONE) {
			LOGE("failed to get rotation angle");
			goto ERROR;
		}

		LOGD("orientation value = %d user_angle = %d", orientation, display_angle);

		ret = __mmplayer_handle_orientation(player, orientation, MM_UTIL_COLOR_RGB24);
		if (ret != MM_ERROR_NONE) {
			LOGE("failed to convert nv12 linear");
			goto ERROR;
		}

		player->capture.fmt = MM_PLAYER_COLORSPACE_RGB888;
		msg.data = &player->capture;
		msg.size = player->capture.size;

		if (player->cmd >= MMPLAYER_COMMAND_START) {
			MMPLAYER_POST_MSG(player, MM_MESSAGE_VIDEO_CAPTURED, &msg);
			LOGD("returned from capture message callback");
		}

		continue;
ERROR:
		for (i = 0; i < player->captured.plane_num; i++)
			MMPLAYER_FREEIF(player->captured.data[i]);

		msg.union_type = MM_MSG_UNION_CODE;

		MMPLAYER_POST_MSG(player, MM_MESSAGE_VIDEO_NOT_CAPTURED, &msg);
	}

EXIT:
	MMPLAYER_CAPTURE_THREAD_UNLOCK(player);
	return NULL;
}

/**
  * The output is fixed as RGB888
  */
static int
__mmplayer_get_video_frame_from_buffer(mmplayer_t *player, GstPad *pad, GstBuffer *buffer)
{
	gint i = 0;
	guint plane_size = 0;
	GstCaps *caps = NULL;
	GstVideoFrame vframe;
	GstVideoInfo vinfo;
	guint8 *pixels = NULL;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(buffer, MM_ERROR_INVALID_ARGUMENT);

	/* get fourcc */
	caps = gst_pad_get_current_caps(pad);

	MMPLAYER_RETURN_VAL_IF_FAIL(caps, MM_ERROR_INVALID_ARGUMENT);
	MMPLAYER_LOG_GST_CAPS_TYPE(caps);

	/* init capture image buffer */
	memset(&player->capture, 0x00, sizeof(mmplayer_video_capture_t));

	gst_video_info_from_caps(&vinfo, caps);

	gst_caps_unref(caps);

	LOGI("captured format is %s", gst_video_format_to_string(GST_VIDEO_INFO_FORMAT(&vinfo)));

	player->captured.width[0] = GST_VIDEO_INFO_WIDTH(&vinfo);
	player->captured.height[0] = GST_VIDEO_INFO_HEIGHT(&vinfo);

	switch (GST_VIDEO_INFO_FORMAT(&vinfo)) {
	case GST_VIDEO_FORMAT_NV12:
	case GST_VIDEO_FORMAT_SN12:
		player->video_cs = MM_PLAYER_COLORSPACE_NV12;
		player->captured.width[1] = player->captured.width[0];
		player->captured.height[1] = player->captured.height[0] >> 1;
		break;
	case GST_VIDEO_FORMAT_I420:
	case GST_VIDEO_FORMAT_S420:
		player->video_cs = MM_PLAYER_COLORSPACE_I420;
		player->captured.width[1] = player->captured.width[2] = player->captured.width[0] >> 1;
		player->captured.height[1] = player->captured.height[2] = player->captured.height[0] >> 1;
		break;
	case GST_VIDEO_FORMAT_ST12:
		player->video_cs = MM_PLAYER_COLORSPACE_NV12_TILED;
		player->captured.width[1] = player->captured.width[0];
		player->captured.height[1] = player->captured.height[0] >> 1;
		break;
	case GST_VIDEO_FORMAT_BGRx:
		player->video_cs = MM_PLAYER_COLORSPACE_BGRx;
		break;
	default:
		player->video_cs = MM_PLAYER_COLORSPACE_MAX;
		ret = MM_ERROR_PLAYER_INTERNAL;
		break;
	}

	if (ret != MM_ERROR_NONE) {
		LOGE("unknown format to capture");
		return ret;
	}

	if (gst_video_frame_map(&vframe, &vinfo, buffer, GST_MAP_READ)) {
		for (i = 0; i < GST_VIDEO_FRAME_N_PLANES(&vframe); i++) {
			player->captured.stride_width[i] = GST_VIDEO_FRAME_PLANE_STRIDE(&vframe, i);
			player->captured.stride_height[i] = player->captured.height[i];
			plane_size = player->captured.stride_width[i] * player->captured.stride_height[i];
			if (!plane_size) {
				LOGE("invalid plane size");
				gst_video_frame_unmap(&vframe);
				return MM_ERROR_PLAYER_INTERNAL;
			}
			pixels = GST_VIDEO_FRAME_PLANE_DATA(&vframe, i);
			player->captured.data[i] = g_malloc(plane_size);
			memcpy(player->captured.data[i], pixels, plane_size);
		}
		player->captured.plane_num = GST_VIDEO_FRAME_N_PLANES(&vframe);
		gst_video_frame_unmap(&vframe);
	}
	/* do convert colorspace */
	MMPLAYER_CAPTURE_THREAD_SIGNAL(player);

	return ret;
}

static GstPadProbeReturn
__mmplayer_video_capture_probe(GstPad *pad, GstPadProbeInfo *info, gpointer u_data)
{
	mmplayer_t *player = (mmplayer_t *)u_data;
	GstBuffer *buffer = NULL;
	int ret = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(info->data, GST_PAD_PROBE_REMOVE);
	MMPLAYER_FENTER();

	buffer = gst_pad_probe_info_get_buffer(info);
	ret = __mmplayer_get_video_frame_from_buffer(player, pad, buffer);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to get video frame");
		return GST_PAD_PROBE_REMOVE;
	}

	/* remove probe to be called at one time */
	if (player->video_capture_cb_probe_id) {
		gst_pad_remove_probe(pad, player->video_capture_cb_probe_id);
		player->video_capture_cb_probe_id = 0;
	}

	MMPLAYER_FLEAVE();

	return GST_PAD_PROBE_OK;
}

static int
__mm_player_convert_colorspace(mmplayer_t *player, unsigned char *src_data, size_t src_size, mm_util_color_format_e src_fmt, unsigned int src_w, unsigned int src_h, mm_util_color_format_e dst_fmt)
{
	int ret = MM_ERROR_NONE;
	mm_util_image_h src_image = NULL;
	mm_util_image_h dst_image = NULL;
	size_t size = 0;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_INTERNAL);
	SECURE_LOGD("src size info. width: %d, height: %d", src_w, src_h);

	ret = mm_image_create_image(src_w, src_h, src_fmt, src_data, src_size, &src_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to create image for capture, %d", ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	ret = mm_util_convert_colorspace(src_image, dst_fmt, &dst_image);
	mm_image_destroy_image(src_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to convert for capture, %d", ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}
	mm_image_debug_image(dst_image, NULL);

	ret = mm_image_get_image(dst_image, NULL, NULL, NULL, &player->capture.data, &size);
	mm_image_destroy_image(dst_image);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to get image for capture, %d", ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	player->capture.size = (int)size;

	return MM_ERROR_NONE;
}

/*
 * Get tiled address of position(x,y)
 *
 * @param x_size
 *   width of tiled[in]
 *
 * @param y_size
 *   height of tiled[in]
 *
 * @param x_pos
 *   x position of tield[in]
 *
 * @param src_size
 *   y position of tield[in]
 *
 * @return
 *   address of tiled data
 */
static int
__tile_4x2_read(int x_size, int y_size, int x_pos, int y_pos)
{
	int pixel_x_m1, pixel_y_m1;
	int roundup_x;
	int linear_addr0, linear_addr1, bank_addr ;
	int x_addr;
	int trans_addr;

	pixel_x_m1 = x_size -1;
	pixel_y_m1 = y_size -1;

	roundup_x = ((pixel_x_m1 >> 7) + 1);

	x_addr = x_pos >> 2;

	if ((y_size <= y_pos+32) && (y_pos < y_size) &&
		(((pixel_y_m1 >> 5) & 0x1) == 0) && (((y_pos >> 5) & 0x1) == 0)) {
		linear_addr0 = (((y_pos & 0x1f) <<4) | (x_addr & 0xf));
		linear_addr1 = (((y_pos >> 6) & 0xff) * roundup_x + ((x_addr >> 6) & 0x3f));

		if (((x_addr >> 5) & 0x1) == ((y_pos >> 5) & 0x1))
			bank_addr = ((x_addr >> 4) & 0x1);
		else
			bank_addr = 0x2 | ((x_addr >> 4) & 0x1);
	} else {
		linear_addr0 = (((y_pos & 0x1f) << 4) | (x_addr & 0xf));
		linear_addr1 = (((y_pos >> 6) & 0xff) * roundup_x + ((x_addr >> 5) & 0x7f));

		if (((x_addr >> 5) & 0x1) == ((y_pos >> 5) & 0x1))
			bank_addr = ((x_addr >> 4) & 0x1);
		else
			bank_addr = 0x2 | ((x_addr >> 4) & 0x1);
	}

	linear_addr0 = linear_addr0 << 2;
	trans_addr = (linear_addr1 <<13) | (bank_addr << 11) | linear_addr0;

	return trans_addr;
}

/*
 * Converts tiled data to linear
 * Crops left, top, right, buttom
 * 1. Y of NV12T to Y of YUV420P
 * 2. Y of NV12T to Y of YUV420S
 * 3. UV of NV12T to UV of YUV420S
 *
 * @param yuv420_dest
 *   Y or UV plane address of YUV420[out]
 *
 * @param nv12t_src
 *   Y or UV plane address of NV12T[in]
 *
 * @param yuv420_width
 *   Width of YUV420[in]
 *
 * @param yuv420_height
 *   Y: Height of YUV420, UV: Height/2 of YUV420[in]
 *
 * @param left
 *   Crop size of left
 *
 * @param top
 *   Crop size of top
 *
 * @param right
 *   Crop size of right
 *
 * @param buttom
 *   Crop size of buttom
 */
static void
__csc_tiled_to_linear_crop(unsigned char *yuv420_dest, unsigned char *nv12t_src, int yuv420_width, int yuv420_height,
							int left, int top, int right, int buttom)
{
	int i, j;
	int tiled_offset = 0, tiled_offset1 = 0;
	int linear_offset = 0;
	int temp1 = 0, temp2 = 0, temp3 = 0, temp4 = 0;

	temp3 = yuv420_width-right;
	temp1 = temp3-left;
	/* real width is greater than or equal 256 */
	if (temp1 >= 256) {
		for (i = top; i < yuv420_height-buttom; i = i+1) {
			j = left;
			temp3 = (j >> 8) << 8;
			temp3 = temp3 >> 6;
			temp4 = i >> 5;
			if (temp4 & 0x1) {
				/* odd fomula: 2 + x + (x >> 2) << 2 + x_block_num * (y - 1) */
				tiled_offset = temp4 - 1;
				temp1 = ((yuv420_width + 127) >> 7) << 7;
				tiled_offset = tiled_offset * (temp1 >> 6);
				tiled_offset = tiled_offset + temp3;
				tiled_offset = tiled_offset + 2;
				temp1 = (temp3 >> 2) << 2;
				tiled_offset = tiled_offset + temp1;
				tiled_offset = tiled_offset << 11;
				tiled_offset1 = tiled_offset + 2048 * 2;
				temp4 = 8;
			} else {
				temp2 = ((yuv420_height+31) >> 5) << 5;
				if ((i + 32) < temp2) {
					/* even1 fomula: x + ((x + 2) >> 2) << 2 + x_block_num * y */
					temp1 = temp3 + 2;
					temp1 = (temp1 >> 2) << 2;
					tiled_offset = temp3 + temp1;
					temp1 = ((yuv420_width + 127) >> 7) << 7;
					tiled_offset = tiled_offset + temp4 * (temp1 >> 6);
					tiled_offset = tiled_offset << 11;
					tiled_offset1 = tiled_offset + 2048 * 6;
					temp4 = 8;
				} else {
					/* even2 fomula: x + x_block_num * y */
					temp1 = ((yuv420_width + 127) >> 7) << 7;
					tiled_offset = temp4 * (temp1 >> 6);
					tiled_offset = tiled_offset + temp3;
					tiled_offset = tiled_offset << 11;
					tiled_offset1 = tiled_offset + 2048 * 2;
					temp4 = 4;
				}
			}

			temp1 = i & 0x1F;
			tiled_offset = tiled_offset + 64 * temp1;
			tiled_offset1 = tiled_offset1 + 64 * temp1;
			temp2 = yuv420_width - left - right;
			linear_offset = temp2 * (i - top);
			temp3 = ((j + 256) >> 8) << 8;
			temp3 = temp3 - j;
			temp1 = left & 0x3F;
			if (temp3 > 192) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset + temp1, 64 - temp1);
				memcpy(yuv420_dest + linear_offset + 64 - temp1, nv12t_src + tiled_offset + 2048, 64);
				memcpy(yuv420_dest + linear_offset + 128 - temp1, nv12t_src + tiled_offset1, 64);
				memcpy(yuv420_dest + linear_offset + 192 - temp1, nv12t_src + tiled_offset1 + 2048, 64);
				linear_offset = linear_offset + 256 - temp1;
			} else if (temp3 > 128) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset + 2048 + temp1, 64 - temp1);
				memcpy(yuv420_dest + linear_offset + 64 - temp1, nv12t_src + tiled_offset1, 64);
				memcpy(yuv420_dest + linear_offset + 128 - temp1, nv12t_src + tiled_offset1 + 2048, 64);
				linear_offset = linear_offset + 192 - temp1;
			} else if (temp3 > 64) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset1 + temp1, 64 - temp1);
				memcpy(yuv420_dest + linear_offset + 64 - temp1, nv12t_src + tiled_offset1 + 2048, 64);
				linear_offset = linear_offset + 128 - temp1;
			} else if (temp3 > 0) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset1 + 2048 + temp1, 64 - temp1);
				linear_offset = linear_offset + 64 - temp1;
			}

			tiled_offset = tiled_offset + temp4 * 2048;
			j = (left >> 8) << 8;
			j = j + 256;
			temp2 = yuv420_width - right - 256;
			for (; j <= temp2; j = j + 256) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 64);
				tiled_offset1 = tiled_offset1 + temp4 * 2048;
				memcpy(yuv420_dest + linear_offset + 64, nv12t_src + tiled_offset + 2048, 64);
				memcpy(yuv420_dest + linear_offset + 128, nv12t_src + tiled_offset1, 64);
				tiled_offset = tiled_offset + temp4 * 2048;
				memcpy(yuv420_dest + linear_offset + 192, nv12t_src + tiled_offset1 + 2048, 64);
				linear_offset = linear_offset + 256;
			}

			tiled_offset1 = tiled_offset1 + temp4 * 2048;
			temp2 = yuv420_width - right - j;
			if (temp2 > 192) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 64);
				memcpy(yuv420_dest + linear_offset + 64, nv12t_src + tiled_offset + 2048, 64);
				memcpy(yuv420_dest + linear_offset + 128, nv12t_src + tiled_offset1, 64);
				memcpy(yuv420_dest + linear_offset + 192, nv12t_src + tiled_offset1 + 2048, temp2 - 192);
			} else if (temp2 > 128) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 64);
				memcpy(yuv420_dest + linear_offset + 64, nv12t_src + tiled_offset + 2048, 64);
				memcpy(yuv420_dest + linear_offset + 128, nv12t_src + tiled_offset1, temp2 - 128);
			} else if (temp2 > 64) {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 64);
				memcpy(yuv420_dest + linear_offset + 64, nv12t_src + tiled_offset + 2048, temp2 - 64);
			} else {
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, temp2);
			}
		}
	} else if (temp1 >= 64) {
		for (i = top; i < (yuv420_height - buttom); i++) {
			j = left;
			tiled_offset = __tile_4x2_read(yuv420_width, yuv420_height, j, i);
			temp2 = ((j + 64) >> 6) << 6;
			temp2 = temp2 - j;
			linear_offset = temp1 * (i - top);
			temp4 = j & 0x3;
			tiled_offset = tiled_offset + temp4;
			memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, temp2);
			linear_offset = linear_offset + temp2;
			j = j + temp2;
			if ((j + 64) <= temp3) {
				tiled_offset = __tile_4x2_read(yuv420_width, yuv420_height, j, i);
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 64);
				linear_offset = linear_offset + 64;
				j += 64;
			}
			if ((j + 64) <= temp3) {
				tiled_offset = __tile_4x2_read(yuv420_width, yuv420_height, j, i);
				memcpy(yuv420_dest+linear_offset, nv12t_src + tiled_offset, 64);
				linear_offset = linear_offset + 64;
				j += 64;
			}
			if (j < temp3) {
				tiled_offset = __tile_4x2_read(yuv420_width, yuv420_height, j, i);
				temp2 = temp3 - j;
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, temp2);
			}
		}
	} else {
		for (i = top; i < (yuv420_height - buttom); i++) {
			linear_offset = temp1 * (i - top);
			for (j = left; j < (yuv420_width - right); j += 2) {
				tiled_offset = __tile_4x2_read(yuv420_width, yuv420_height, j, i);
				temp4 = j & 0x3;
				tiled_offset = tiled_offset + temp4;
				memcpy(yuv420_dest + linear_offset, nv12t_src + tiled_offset, 2);
				linear_offset = linear_offset + 2;
			}
		}
	}
}

#ifdef CAPTURE_OUTPUT_DUMP /* for capture output dump */
static void
capture_output_dump(mmplayer_t *player)
{
	unsigned char *temp = NULL;
	char file[100] = { 0, };
	FILE *fp = NULL;
	int ret = 0;
	int i = 0, j = 0;

	LOGE("capture output dump start. size = %d", player->capture.size);
	sprintf(file, "/tmp/dec_output_dump_%dx%d.yuv", player->captured.width[0], player->captured.height[0]);
	fp = fopen(file, "ab");

	for (i = 0; i < player->captured.plane_num; i++) {
		temp = player->captured.data[i];

		for (j = 0; j < player->captured.height[i]; j++) {
			ret = fwrite(temp, player->captured.width[i], 1, fp);
			temp += player->captured.stride_width[i];
		}
	}
	LOGE("capture yuv dumped!! ret = %d", ret);
	fclose(fp);

	temp = (unsigned char *)player->capture.data;
	sprintf(file, "/tmp/dec_output_dump_%dx%d.rgb", player->captured.width[0], player->captured.height[0]);
	fp = fopen(file, "ab");
	ret = fwrite(temp, player->capture.size, 1, fp);
	fclose(fp);
	LOGE("capture rgb dumped!! ret = %d", ret);
}
#endif

static int
__mm_player_convert_NV12_tiled(mmplayer_t *player)
{
	/* Colorspace conversion : NV12T-> NV12-> RGB888 */
	int i;
	int ret = MM_ERROR_NONE;
	unsigned char *src_buffer = NULL;
	unsigned char *linear_y_plane = NULL;
	unsigned char *linear_uv_plane = NULL;
	guint linear_y_plane_size;
	guint linear_uv_plane_size;
	guint width = player->captured.width[0];
	guint height = player->captured.height[0];

	linear_y_plane_size = width * height;
	linear_uv_plane_size = linear_y_plane_size / 2;

	if (!linear_y_plane_size || !linear_uv_plane_size) {
		LOGE("invalid plane size");
		return MM_ERROR_PLAYER_INTERNAL;
	}
	linear_y_plane = (unsigned char *)g_malloc(linear_y_plane_size);
	linear_uv_plane = (unsigned char *)g_malloc(linear_uv_plane_size);

	/* NV12 tiled to linear */
	__csc_tiled_to_linear_crop(linear_y_plane,
		player->captured.data[0], width, height, 0, 0, 0, 0);
	__csc_tiled_to_linear_crop(linear_uv_plane,
		player->captured.data[1], width, height / 2, 0, 0, 0, 0);

	src_buffer = (unsigned char *)g_malloc(linear_y_plane_size + linear_uv_plane_size);

	memcpy(src_buffer, linear_y_plane, linear_y_plane_size);
	memcpy(src_buffer + linear_y_plane_size, linear_uv_plane, linear_uv_plane_size);

	/* NV12 linear to RGB888 */
	ret = __mm_player_convert_colorspace(player, src_buffer, (size_t)(linear_y_plane_size + linear_uv_plane_size),
		MM_UTIL_COLOR_NV12, width, height, MM_UTIL_COLOR_RGB24);

	MMPLAYER_FREEIF(src_buffer);
	MMPLAYER_FREEIF(linear_y_plane);
	MMPLAYER_FREEIF(linear_uv_plane);

	for (i = 0; i < player->captured.plane_num; i++)
		MMPLAYER_FREEIF(player->captured.data[i]);

	return ret;
}

static int
__mm_player_convert_NV12(mmplayer_t *player)
{
	unsigned char *src_buffer = NULL;
	unsigned char *p_buf = NULL;
	unsigned char *temp = NULL;
	int planes[MAX_BUFFER_PLANE] = {0, };
	int ret = MM_ERROR_NONE;
	int i, j;

	/* using original width otherwises, app can't know aligned to resize */
	planes[0] =  player->captured.stride_width[0] * player->captured.stride_height[0];
	planes[1] = player->captured.stride_width[1] * player->captured.stride_height[1];
	guint src_buffer_size = planes[0] + planes[1];

	if (!src_buffer_size) {
		LOGE("invalid data size");
		return MM_ERROR_PLAYER_INTERNAL;
	}
	src_buffer = (unsigned char *)g_malloc(src_buffer_size);

	p_buf = src_buffer;

	temp = player->captured.data[0];

	/* set Y plane */
	for (i = 0; i < player->captured.height[0]; i++) {
		memcpy(p_buf, temp, player->captured.width[0]);
		p_buf += player->captured.width[0];
		temp += player->captured.stride_width[0];
	}

	temp = player->captured.data[1];

	/* set UV plane*/
	for (j = 0; j < player->captured.height[1]; j++) {
		memcpy(p_buf, temp, player->captured.width[1]);
		p_buf += player->captured.width[1];
		temp += player->captured.stride_width[1];
	}

	/* NV12 -> RGB888 */
	ret = __mm_player_convert_colorspace(player, (unsigned char *)src_buffer, (size_t)src_buffer_size,
			MM_UTIL_COLOR_NV12, player->captured.width[0],
			player->captured.height[0], MM_UTIL_COLOR_RGB24);

#ifdef CAPTURE_OUTPUT_DUMP
	capture_output_dump(player);
#endif
	MMPLAYER_FREEIF(src_buffer);

	for (i = 0; i < player->captured.plane_num; i++)
		MMPLAYER_FREEIF(player->captured.data[i]);

	return ret;
}

static int
__mm_player_convert_I420(mmplayer_t *player)
{
	unsigned char *src_buffer = NULL;
	unsigned char *p_buf = NULL;
	unsigned char *temp = NULL;
	int planes[MAX_BUFFER_PLANE] = {0, };
	int ret = MM_ERROR_NONE;
	int i;

	/* using original width otherwises, app can't know aligned to resize */
	planes[0] = player->captured.stride_width[0] * player->captured.stride_height[0];
	planes[1] = planes[2] = (player->captured.stride_width[0] >> 1)
								* (player->captured.stride_height[0] >> 1);

	if (!planes[0] || !planes[1] || !planes[2]) {
		LOGE("invalid plane size");
		return MM_ERROR_PLAYER_INTERNAL;
	}
	src_buffer = (unsigned char *)g_malloc(planes[0] + planes[1] + planes[2]);

	/* set Y plane */
	p_buf = src_buffer;

	temp = player->captured.data[0];

	for (i = 0; i < player->captured.height[0]; i++) {
		memcpy(p_buf, temp, player->captured.width[0]);
		temp += player->captured.stride_width[0];
		p_buf += player->captured.width[0];
	}

	/* set U plane */
	temp = player->captured.data[1];

	for (i = 0; i < player->captured.height[1]; i++) {
		memcpy(p_buf, temp, player->captured.width[1]);
		temp += player->captured.stride_width[1];
		p_buf += player->captured.width[1];
	}

	/* set V plane */
	temp = player->captured.data[2];

	for (i = 0; i < player->captured.height[2]; i++) {
		memcpy(p_buf, temp, player->captured.width[2]);
		temp += player->captured.stride_width[2];
		p_buf += player->captured.width[2];
	}

	/* I420 -> RGB888 */
	ret = __mm_player_convert_colorspace(player, (unsigned char *)src_buffer, (size_t)(planes[0] + planes[1] + planes[2]),
			MM_UTIL_COLOR_I420, player->captured.width[0], player->captured.height[0], MM_UTIL_COLOR_RGB24);

#ifdef CAPTURE_OUTPUT_DUMP
	capture_output_dump(player);
#endif
	MMPLAYER_FREEIF(src_buffer);

	for (i = 0; i < player->captured.plane_num; i++)
		MMPLAYER_FREEIF(player->captured.data[i]);

	return ret;
}

static int
__mm_player_convert_BGRx(mmplayer_t *player)
{
	int i;
	guint size;
	int ret = MM_ERROR_NONE;
	unsigned char *src_buffer = NULL;

	size = (player->captured.stride_width[0] * player->captured.stride_height[0]);

	if (!size) {
		LOGE("invalid size");
		return MM_ERROR_PLAYER_INTERNAL;
	}
	src_buffer = (unsigned char *)g_malloc(size);
	memcpy(src_buffer, player->captured.data[0], size);

	/* BGRx -> RGB888 */
	ret = __mm_player_convert_colorspace(player, (unsigned char *)src_buffer, (size_t)size, MM_UTIL_COLOR_BGRX,
			player->captured.width[0], player->captured.height[0], MM_UTIL_COLOR_RGB24);

#ifdef CAPTURE_OUTPUT_DUMP
	capture_output_dump(player);
#endif
	MMPLAYER_FREEIF(src_buffer);

	for (i = 0; i < player->captured.plane_num; i++)
		MMPLAYER_FREEIF(player->captured.data[i]);

	return ret;
}
