/*
 * libmm-player
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, naveen cherukuri <naveen.ch@samsung.com>,
 * YeJin Cho <cho.yejin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <dlog.h>
#include <mm_error.h>
#include "mm_player_utils.h"
#include "mm_player_tracks.h"
#include "mm_player_attrs.h"

#define INVALID_TRACK_INDEX -1
#define LANGUAGE_CODE_SIZE (3 + 1) /* Size of ISO-639-1, and considering the nul-terminator */

/*---------------------------------------------------------------------------------------
|    LOCAL FUNCTION PROTOTYPES:								                      |
---------------------------------------------------------------------------------------*/
static int __mmplayer_track_get_language(mmplayer_t *player, mmplayer_track_type_e type, gint stream_index, gchar **code);


/*=======================================================================================
|  FUNCTION DEFINITIONS									                     |
=======================================================================================*/
int
_mmplayer_get_track_count(MMHandleType hplayer,  mmplayer_track_type_e type, int *count)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	MMHandleType attrs = 0;

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(count, MM_ERROR_INVALID_ARGUMENT);
	MMPLAYER_RETURN_VAL_IF_FAIL((MMPLAYER_CURRENT_STATE(player) == MM_PLAYER_STATE_PAUSED)
		 || (MMPLAYER_CURRENT_STATE(player) == MM_PLAYER_STATE_PLAYING),
		MM_ERROR_PLAYER_INVALID_STATE);

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("cannot get content attribute");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	*count = 0;

	switch (type) {
	case MM_PLAYER_TRACK_TYPE_AUDIO:
		if (player->selector[MM_PLAYER_TRACK_TYPE_AUDIO].total_track_num > 0)
			*count = player->selector[MM_PLAYER_TRACK_TYPE_AUDIO].total_track_num;
		break;
	case MM_PLAYER_TRACK_TYPE_TEXT: /* internal or external */
		ret = mm_attrs_get_int_by_name(attrs, "content_text_track_num", count);
		break;
	default:
		ret = MM_ERROR_INVALID_ARGUMENT;
		break;
	}

	LOGD("type: %d, the num of track: %d", type, *count);

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_select_track(MMHandleType hplayer, mmplayer_track_type_e type, int index)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("track type: %d, index: %d", type, index);

	MMPLAYER_SUBTITLE_INFO_LOCK(player);

	if (type == MM_PLAYER_TRACK_TYPE_TEXT && player->subtitle_language_list) { /* external subtitle */
		GstElement *subparse = NULL;
		mmplayer_lang_info_t *temp = NULL;
		guint num_of_list = 0;

		if (!player->pipeline || !player->pipeline->textbin[MMPLAYER_T_FAKE_SINK].gst) {
			ret = MM_ERROR_PLAYER_NOT_INITIALIZED;
			goto EXIT;
		}

		num_of_list = g_list_length(player->subtitle_language_list);
		if (index < 0 || index >= num_of_list) {
			LOGE("req track index is wrong");
			ret = MM_ERROR_INVALID_ARGUMENT;
			goto EXIT;
		}

		temp = g_list_nth_data(player->subtitle_language_list, index);
		if (!temp) {
			LOGE("fail to get lang from list");
			ret = MM_ERROR_PLAYER_INTERNAL;
			goto EXIT;
		}
		subparse = player->pipeline->mainbin[MMPLAYER_M_SUBPARSE].gst;
		LOGD("setting to language %s", temp->language_code);
		g_object_set(G_OBJECT(subparse), "current-language", temp->language_key, NULL);

		_mmplayer_sync_subtitle_pipeline(player);

	} else { /* audio or internal subtitle */
		ret = _mmplayer_change_track_language(hplayer, type, index);
		if (ret != MM_ERROR_NONE) {
			LOGE("failed to change track");
			goto EXIT;
		}
	}

	if (type == MM_PLAYER_TRACK_TYPE_TEXT) {
		if (mm_player_set_attribute(hplayer, NULL,
				"current_text_track_index", index, NULL) != MM_ERROR_NONE) {
			LOGE("failed to set text track index");
			ret = MM_ERROR_PLAYER_INTERNAL;
			goto EXIT;
		}
	}

	LOGD("current index is updated");

EXIT:
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);
	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_current_track(MMHandleType hplayer, mmplayer_track_type_e type, int *index)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	MMHandleType attrs = 0;
	gint count = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(index, MM_ERROR_INVALID_ARGUMENT);

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_VAL_IF_FAIL(attrs, MM_ERROR_PLAYER_INTERNAL);

	MMPLAYER_SUBTITLE_INFO_LOCK(player);

	LOGD("track type: %d", type);

	*index = INVALID_TRACK_INDEX;

	switch (type) {
	case MM_PLAYER_TRACK_TYPE_AUDIO:
		if (player->selector[MM_PLAYER_TRACK_TYPE_AUDIO].total_track_num <= 0) {
			LOGW("there is no audio track");
			ret = MM_ERROR_PLAYER_NO_OP;
			goto EXIT;
		}

		*index = player->selector[MM_PLAYER_TRACK_TYPE_AUDIO].active_pad_index;
		break;
	case MM_PLAYER_TRACK_TYPE_TEXT: /* internal or external */
		mm_attrs_get_int_by_name(attrs, "content_text_track_num", &count);
		if (count <= 0) {
			LOGD("there is no text track");
			ret = MM_ERROR_PLAYER_NO_OP;
			goto EXIT;
		}

		mm_attrs_get_int_by_name(attrs, "current_text_track_index", index);
		break;
	default:
		LOGD("invalid input type");
		ret = MM_ERROR_INVALID_ARGUMENT;
		goto EXIT;
	}

	if (*index < 0) {
		LOGE("invalid track index");
		ret = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	LOGD("current track index: %d", *index);

EXIT:
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);
	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_track_language_code(MMHandleType hplayer, mmplayer_track_type_e type, int index, char **code)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gchar *lang_code = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(code, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_SUBTITLE_INFO_LOCK(player);

	if (type == MM_PLAYER_TRACK_TYPE_TEXT && player->subtitle_language_list) { /* external subtitle */
		mmplayer_lang_info_t *language_list = NULL;

		lang_code = (gchar *)g_malloc0(LANGUAGE_CODE_SIZE * sizeof(char));

		language_list = g_list_nth_data(player->subtitle_language_list, index);
		if (language_list == NULL) {
			LOGE("invalid index %d", index);
			ret = MM_ERROR_INVALID_ARGUMENT;
			goto EXIT;
		}
		g_strlcpy(lang_code, language_list->language_code, LANGUAGE_CODE_SIZE);
	} else { /* audio or internal subtitle */
		if (player->selector[type].total_track_num <= 0) {
			LOGW("language list is not available. [type:%d]", type);
			ret = MM_ERROR_PLAYER_NO_OP;
			goto EXIT;
		}

		ret = __mmplayer_track_get_language(player, type, index, &lang_code);
	}

EXIT:
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);

	*code = NULL;
	if (ret == MM_ERROR_NONE) {
		if (g_strcmp0(lang_code, "und"))
			*code = strndup(lang_code, 2);
		else
			*code = strndup("und", 3);
	}
	MMPLAYER_FREEIF(lang_code);

	MMPLAYER_FLEAVE();
	return ret;
}

void
_mmplayer_track_initialize(mmplayer_t *player)
{
	mmplayer_track_type_e type = MM_PLAYER_TRACK_TYPE_AUDIO;

	MMPLAYER_FENTER();

	for (; type < MM_PLAYER_TRACK_TYPE_MAX ; type++) {
		/* active_pad_index is initialized when player is created or destroyed.
		   and the value can be set by calling _mmplayer_change_track_language()
		   before pipeline is created.*/
		player->selector[type].total_track_num = 0;
		player->selector[type].channels = g_ptr_array_new();
	}
}

void
_mmplayer_track_destroy(mmplayer_t *player)
{
	mmplayer_track_type_e type = MM_PLAYER_TRACK_TYPE_AUDIO;
	MMPLAYER_FENTER();

	if (mm_player_set_attribute((MMHandleType)player, NULL, "content_text_track_num", 0,
			"current_text_track_index", INVALID_TRACK_INDEX, NULL) != MM_ERROR_NONE)
		LOGE("failed to reset track attr");

	for (; type < MM_PLAYER_TRACK_TYPE_MAX ; type++) {
		player->selector[type].active_pad_index = 0;
		player->selector[type].total_track_num = 0;

		if (player->selector[type].channels)
			g_ptr_array_free(player->selector[type].channels, TRUE);
		player->selector[type].channels = NULL;
	}
}

void
_mmplayer_track_update_selector_info(mmplayer_t *player, mmplayer_track_type_e type, GstPad *sinkpad)
{
	MMPLAYER_FENTER();

	player->selector[type].total_track_num++;
	g_ptr_array_add(player->selector[type].channels, sinkpad);

	LOGD("type: %d, track cnt: %d", type, player->selector[type].total_track_num);
}

void
_mmplayer_track_update_text_attr_info(mmplayer_t *player, GstMessage *msg)
{
	 const GValue *lang_list = NULL;
	 mmplayer_lang_info_t *temp = NULL;
	 gchar *current_lang = NULL;
	 gint  track_index = INVALID_TRACK_INDEX;
	 guint track_count = 0, index = 0;

	 MMPLAYER_FENTER();
	 MMPLAYER_RETURN_IF_FAIL(player && msg);

	if (!gst_message_get_structure(msg)) {
		LOGE("failed to get msg structure");
		return;
	}

	lang_list = gst_structure_get_value(gst_message_get_structure(msg), "lang_list");
	if (!lang_list) {
		LOGW("language list is null");
		return;
	}

	track_count = g_list_length((GList *)g_value_get_pointer(lang_list));
	if (track_count == 0) {
		LOGW("track count is zero");
		return;
	}

	MMPLAYER_SUBTITLE_INFO_LOCK(player);
	player->subtitle_language_list = (GList *)g_value_get_pointer(lang_list);
	mm_player_set_attribute((MMHandleType)player, NULL,
			"content_text_track_num", (gint)track_count, NULL);
	g_object_get(G_OBJECT(msg->src), "current-language", &current_lang, NULL);

	LOGI("track count: %d, current track lang: %s", track_count, current_lang);
	if (!current_lang)
		goto EXIT;

	for (index = 0 ; index < track_count ; index++) {
		temp = g_list_nth_data(player->subtitle_language_list, index);
		if (temp) {
			LOGI("[%d] lang_key: %s, lang_code: %s",
				index, temp->language_key, temp->language_code);

			if (!g_strcmp0(temp->language_key, current_lang)) {
				LOGD("current track index : %d", index);
				track_index = index;
				break;
			}
		}
	}

	if (track_index > INVALID_TRACK_INDEX)
		mm_player_set_attribute((MMHandleType)player, NULL,
				"current_text_track_index", track_index, NULL);
	else
		LOGE("failed to find current track index");

EXIT:
	MMPLAYER_SUBTITLE_INFO_SIGNAL(player);
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);

	MMPLAYER_FLEAVE();
}

static int
__mmplayer_track_get_language(mmplayer_t *player, mmplayer_track_type_e type, gint stream_index, gchar **code)
{
	GstTagList *tag_list = NULL;
	gchar *tag = NULL;
	GstPad *sinkpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL((code && (stream_index >= 0) &&
		(stream_index < player->selector[type].total_track_num)), MM_ERROR_INVALID_ARGUMENT);

	LOGD("type: %d, track count: %d, intput idx: %d", type, player->selector[type].total_track_num, stream_index);

	*code = (gchar *)g_malloc0(LANGUAGE_CODE_SIZE * sizeof(char));

	sinkpad = g_ptr_array_index(player->selector[type].channels, stream_index);

	g_object_get(sinkpad, "tags", &tag_list, NULL);
	if (tag_list)
		gst_tag_list_get_string(tag_list, GST_TAG_LANGUAGE_CODE, &tag);

	if (tag) {
		LOGD("language code: %s, len: %zu", tag, strlen(tag));
		g_strlcpy(*code, tag, LANGUAGE_CODE_SIZE);
		MMPLAYER_FREEIF(tag);
	} else {
		LOGD("no language code info - und");
		g_strlcpy(*code, "und", LANGUAGE_CODE_SIZE);
	}

	if (tag_list)
		gst_tag_list_unref(tag_list);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}
