/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "mm_error.h"
#include "mm_player_utils.h"
#include "gtest_mm_player_priv.h"

MMPlayerPriv::MMPlayerPriv()
{
	priv_player = NULL;
}

MMPlayerPriv::~MMPlayerPriv()
{
	if (priv_player) {
		_mmplayer_destroy((MMHandleType)priv_player);
		g_mutex_clear(&priv_player->cmd_lock);
		g_mutex_clear(&priv_player->playback_lock);

		g_free(priv_player);
		priv_player = NULL;
	}
}

int MMPlayerPriv::Create(void)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *new_player = NULL;

	/* alloc player structure */
	new_player = (mmplayer_t *)g_malloc0(sizeof(mmplayer_t));

	/* create player lock */
	g_mutex_init(&new_player->cmd_lock);

	/* create player lock */
	g_mutex_init(&new_player->playback_lock);

	/* load ini files */
	if (mm_player_ini_load(&new_player->ini) != MM_ERROR_NONE) {
		LOGE("can't load ini");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	if (mm_player_audio_effect_ini_load(&new_player->ini) != MM_ERROR_NONE) {
		LOGE("can't load audio ini");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	/* create player */
	ret = _mmplayer_create_player((MMHandleType)new_player);

	priv_player = new_player;

	return ret;
}

int MMPlayerPriv::Destroy(void)
{
	int ret = MM_ERROR_NONE;

	if (!priv_player)
		return MM_ERROR_PLAYER_NOT_INITIALIZED;

	__mmplayer_bus_msg_thread_destroy(priv_player);

	MMPLAYER_CMD_LOCK(priv_player);

	ret = _mmplayer_destroy(priv_player);

	MMPLAYER_CMD_UNLOCK(priv_player);

	g_mutex_clear(&((mmplayer_t *)priv_player)->cmd_lock);
	g_mutex_clear(&((mmplayer_t *)priv_player)->playback_lock);

	memset((mmplayer_t *)priv_player, 0x00, sizeof(mmplayer_t));

	/* free player */
	g_free((void *)priv_player);
	priv_player = NULL;

	return ret;
}

int MMPlayerPriv::SetUri(const gchar *uri)
{
	int ret = MM_ERROR_NONE;

	MMPLAYER_CMD_LOCK(priv_player);
	ret = _mmplayer_set_uri(priv_player, uri);
	MMPLAYER_CMD_UNLOCK(priv_player);

	return ret;
}

int MMPlayerPriv::Realize(void)
{
	int ret = MM_ERROR_NONE;

	MMPLAYER_CMD_LOCK(priv_player);
	ret = _mmplayer_realize(priv_player);
	MMPLAYER_CMD_UNLOCK(priv_player);

	return ret;
}

int MMPlayerPriv::Unrealize(void)
{
	int ret = MM_ERROR_NONE;

	MMPLAYER_CMD_LOCK(priv_player);
	ret = _mmplayer_unrealize(priv_player);
	MMPLAYER_CMD_UNLOCK(priv_player);

	return ret;
}

int MMPlayerPriv::GetCurrentState(mmplayer_state_e *state)
{
	int ret = MM_ERROR_NONE;

	*state = MM_PLAYER_STATE_NULL;
	ret = _mmplayer_get_state(priv_player, (int *)state);

	return ret;
}

int MMPlayerPriv::ParseProfile(const gchar *uri, int *uri_type)
{
	int ret = MM_ERROR_NONE;
	mmplayer_parse_profile_t profile;

	memset((void *)&profile, 0, sizeof(mmplayer_parse_profile_t));

	ret = _mmplayer_parse_profile(uri, NULL, &profile);
	if (uri_type)
		*uri_type = profile.uri_type;

	return ret;
}
