/*
 * libmm-player
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: JongHyuk Choi <jhchoi.choi@samsung.com>, YeJin Cho <cho.yejin@samsung.com>,
 * Seungbae Shin <seungbae.shin@samsung.com>, YoungHwan An <younghwan_.an@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*===========================================================================================
|																							|
|  INCLUDE FILES																			|
|																							|
========================================================================================== */
#include <glib.h>
#include <gst/gst.h>
#include <gst/video/videooverlay.h>
#include <gst/audio/gstaudiobasesink.h>
#include <unistd.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/time.h>
#include <stdlib.h>
#include <dlog.h>
#include <gio/gio.h>

#include <mm_error.h>
#include <mm_attrs.h>
#include <mm_sound.h>

#include "mm_player_priv.h"
#include "mm_player_ini.h"
#include "mm_player_capture.h"
#include "mm_player_utils.h"
#include "mm_player_tracks.h"
#include "mm_player_360.h"
#include "mm_player_gst.h"

#include <system_info.h>
#include <sound_manager.h>
#include <gst/allocators/gsttizenmemory.h>
#include <tbm_surface_internal.h>

/*===========================================================================================
|																							|
|  LOCAL DEFINITIONS AND DECLARATIONS FOR MODULE											|
|																							|
========================================================================================== */

/*---------------------------------------------------------------------------
|    GLOBAL CONSTANT DEFINITIONS:											|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    IMPORTED VARIABLE DECLARATIONS:										|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    IMPORTED FUNCTION DECLARATIONS:										|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    LOCAL #defines:														|
---------------------------------------------------------------------------*/
#define TRICK_PLAY_MUTE_THRESHOLD_MAX	2.0
#define TRICK_PLAY_MUTE_THRESHOLD_MIN	0.0

#define MM_VOLUME_FACTOR_DEFAULT		1.0
#define MM_VOLUME_FACTOR_MIN			0
#define MM_VOLUME_FACTOR_MAX			1.0

/* Don't need to sleep for sound fadeout
 * fadeout related fucntion will be deleted(Deprecated)
 */
#define MM_PLAYER_FADEOUT_TIME_DEFAULT	0

#define DEFAULT_PLAYBACK_RATE			1.0

#define PLAYER_DISPLAY_MODE_DST_ROI		5

#define ADAPTIVE_VARIANT_DEFAULT_VALUE -1 /* auto */

#define PLAYER_SPHERICAL_DEFAULT_YAW   0  /* sync from video360 plugin */
#define PLAYER_SPHERICAL_DEFAULT_PITCH 0
#define PLAYER_SPHERICAL_DEFAULT_H_FOV 120
#define PLAYER_SPHERICAL_DEFAULT_V_FOV 67

#define SPATIAL_AUDIO_CAPS             "audio/x-raw,format=S16LE,channels=4"
#define FEATURE_NAME_SPHERICAL_VIDEO   "http://tizen.org/feature/multimedia.player.spherical_video"

#define FAKE_SINK_MAX_LATENESS         G_GINT64_CONSTANT(20000000) /* set 20ms as waylandsink */

#define DEFAULT_PCM_OUT_FORMAT         "F32LE"
#define DEFAULT_PCM_OUT_SAMPLERATE     44100
#define DEFAULT_PCM_OUT_CHANNEL        2

/*---------------------------------------------------------------------------
|    LOCAL CONSTANT DEFINITIONS:											|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    LOCAL DATA TYPE DEFINITIONS:											|
---------------------------------------------------------------------------*/
/* NOTE : GstAutoplugSelectResult is defined in gstplay-enum.h but not exposed
 We are defining our own and will be removed when it actually exposed */
typedef enum {
	GST_AUTOPLUG_SELECT_TRY,
	GST_AUTOPLUG_SELECT_EXPOSE,
	GST_AUTOPLUG_SELECT_SKIP
} GstAutoplugSelectResult;

/*---------------------------------------------------------------------------
|    GLOBAL VARIABLE DEFINITIONS:											|
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    LOCAL VARIABLE DEFINITIONS:											|
---------------------------------------------------------------------------*/
static sound_stream_info_h stream_info;

/*---------------------------------------------------------------------------
|    LOCAL FUNCTION PROTOTYPES:												|
---------------------------------------------------------------------------*/
static int		__mmplayer_gst_create_pipeline(mmplayer_t *player);
static int		__mmplayer_gst_destroy_pipeline(mmplayer_t *player);
static int		__mmplayer_gst_create_text_pipeline(mmplayer_t *player);
static int		__mmplayer_gst_create_video_sink_bin(mmplayer_t *player, GstCaps *caps, MMDisplaySurfaceType surface_type);
static int		__mmplayer_gst_create_audio_sink_bin(mmplayer_t *player);
static int		__mmplayer_gst_create_text_sink_bin(mmplayer_t *player);

static void		__mmplayer_gst_decode_no_more_pads(GstElement *elem, gpointer data);
static void		__mmplayer_gst_create_sinkbin(GstElement *decodebin, GstPad *pad, gpointer data);
static void		__mmplayer_gst_decode_unknown_type(GstElement *elem,  GstPad *pad, GstCaps *caps, gpointer data);
static gboolean __mmplayer_gst_decode_autoplug_continue(GstElement *bin,  GstPad *pad, GstCaps *caps,  gpointer data);
static void __mmplayer_gst_decode_pad_removed(GstElement *elem,  GstPad *new_pad, gpointer data);
static void __mmplayer_gst_decode_drained(GstElement *bin, gpointer data);
static void	__mmplayer_pipeline_complete(GstElement *decodebin,  gpointer data);
static gboolean __mmplayer_is_midi_type(gchar *str_caps);
static gboolean __mmplayer_is_only_mp3_type(gchar *str_caps);
static void	__mmplayer_set_audio_attrs(mmplayer_t *player, GstCaps *caps);

static gboolean	__mmplayer_update_subtitle(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data);
static void		__mmplayer_release_misc(mmplayer_t *player);
static void		__mmplayer_release_misc_post(mmplayer_t *player);
static gboolean	__mmplayer_init_gstreamer(mmplayer_t *player);
static void __mmplayer_video_stream_decoded_preroll_cb(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data);
static void __mmplayer_video_stream_decoded_render_cb(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data);
static GstPadProbeReturn __mmplayer_subtitle_adjust_position_probe(GstPad *pad, GstPadProbeInfo *info, gpointer u_data);
static int __mmplayer_change_selector_pad(mmplayer_t *player, mmplayer_track_type_e type, int index);

static gboolean __mmplayer_check_subtitle(mmplayer_t *player);
static int		__mmplayer_handle_missed_plugin(mmplayer_t *player);
static int		__mmplayer_check_not_supported_codec(mmplayer_t *player, const gchar *factory_class, const gchar *mime);
static void		__mmplayer_add_sink(mmplayer_t *player, GstElement *sink);
static void		__mmplayer_del_sink(mmplayer_t *player, GstElement *sink);
static void		__mmplayer_release_signal_connection(mmplayer_t *player, mmplayer_signal_type_e type);
static gpointer __mmplayer_gapless_play_thread(gpointer data);
static gboolean __mmplayer_add_dump_buffer_probe(mmplayer_t *player, GstElement *element);
static GstPadProbeReturn __mmplayer_dump_buffer_probe_cb(GstPad *pad,  GstPadProbeInfo *info, gpointer u_data);
static void __mmplayer_release_dump_list(GList *dump_list);
static int		__mmplayer_gst_realize(mmplayer_t *player);
static int		__mmplayer_gst_unrealize(mmplayer_t *player);
static int		__mmplayer_gst_adjust_subtitle_position(mmplayer_t *player, int position);
static int		__mmplayer_gst_set_message_callback(mmplayer_t *player, MMMessageCallback callback, gpointer user_param);

/* util */
static gboolean __mmplayer_verify_gapless_play_path(mmplayer_t *player);
static void __mmplayer_check_pipeline(mmplayer_t *player);
static gboolean __mmplayer_deactivate_selector(mmplayer_t *player, mmplayer_track_type_e type);
static void __mmplayer_deactivate_old_path(mmplayer_t *player);
static int __mmplayer_gst_create_plain_text_elements(mmplayer_t *player);
static guint32 _mmplayer_convert_fourcc_string_to_value(const gchar *format_name);
static void		__mmplayer_gst_caps_notify_cb(GstPad *pad, GParamSpec *unused, gpointer data);
static void		__mmplayer_audio_stream_send_data(mmplayer_t *player, mmplayer_audio_stream_buff_t *a_buffer);
static void		__mmplayer_initialize_storage_info(mmplayer_t *player, mmplayer_path_type_e path_type);
static gboolean __mmplayer_update_duration_value(mmplayer_t *player);
static gboolean __mmplayer_update_audio_attrs(mmplayer_t *player, MMHandleType attrs);
static gboolean __mmplayer_update_video_attrs(mmplayer_t *player, MMHandleType attrs);
static gboolean __mmplayer_update_bitrate_attrs(mmplayer_t *player, MMHandleType attrs);

static void __mmplayer_copy_uri_and_set_type(mmplayer_parse_profile_t *data, const char *uri, int uri_type);
static int __mmplayer_set_mem_uri(mmplayer_parse_profile_t *data, char *path, void *param);
static int __mmplayer_set_file_uri(mmplayer_parse_profile_t *data, const char *uri);

static mmplayer_video_decoded_data_info_t *__mmplayer_create_stream_from_pad(GstPad *pad);
static void __mmplayer_zerocopy_set_stride_elevation_bo(mmplayer_video_decoded_data_info_t *stream, GstMemory *mem);
static gboolean __mmplayer_swcodec_set_stride_elevation(mmplayer_video_decoded_data_info_t *stream);
static gboolean __mmplayer_swcodec_set_bo(mmplayer_t *player, mmplayer_video_decoded_data_info_t *stream, GstMemory *mem);

static void __mmplayer_set_pause_state(mmplayer_t *player);
static void __mmplayer_set_playing_state(mmplayer_t *player);
/*===========================================================================================
|																							|
|  FUNCTION DEFINITIONS																		|
|																							|
========================================================================================== */

/* This function should be called after the pipeline goes PAUSED or higher
state. */
gboolean
_mmplayer_update_content_attrs(mmplayer_t *player, enum content_attr_flag flag)
{
	static gboolean has_duration = FALSE;
	static gboolean has_video_attrs = FALSE;
	static gboolean has_audio_attrs = FALSE;
	static gboolean has_bitrate = FALSE;
	gboolean missing_only = FALSE;
	gboolean all = FALSE;
	MMHandleType attrs = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);

	/* check player state here */
	if (MMPLAYER_CURRENT_STATE(player) != MM_PLAYER_STATE_PAUSED &&
		MMPLAYER_CURRENT_STATE(player) != MM_PLAYER_STATE_PLAYING) {
		/* give warning now only */
		LOGW("be careful. content attributes may not available in this state ");
	}

	/* get content attribute first */
	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("cannot get content attribute");
		return FALSE;
	}

	/* get update flag */

	if (flag & ATTR_MISSING_ONLY) {
		missing_only = TRUE;
		LOGD("updating missed attr only");
	}

	if (flag & ATTR_ALL) {
		all = TRUE;
		has_duration = FALSE;
		has_video_attrs = FALSE;
		has_audio_attrs = FALSE;
		has_bitrate = FALSE;

		LOGD("updating all attrs");
	}

	if (missing_only && all) {
		LOGW("cannot use ATTR_MISSING_ONLY and ATTR_ALL. ignoring ATTR_MISSING_ONLY flag!");
		missing_only = FALSE;
	}

	if ((flag & ATTR_DURATION) || (!has_duration && missing_only) || all)
		has_duration = __mmplayer_update_duration_value(player);

	if ((flag & ATTR_AUDIO) || (!has_audio_attrs && missing_only) || all)
		has_audio_attrs = __mmplayer_update_audio_attrs(player, attrs);

	if ((flag & ATTR_VIDEO) || (!has_video_attrs && missing_only) || all)
		has_video_attrs = __mmplayer_update_video_attrs(player, attrs);

	if ((flag & ATTR_BITRATE) || (!has_bitrate && missing_only) || all)
		has_bitrate = __mmplayer_update_bitrate_attrs(player, attrs);

	MMPLAYER_FLEAVE();

	return TRUE;
}

MMStreamingType
_mmplayer_get_stream_service_type(mmplayer_t *player)
{
	MMStreamingType streaming_type = STREAMING_SERVICE_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
			player->pipeline &&
			player->pipeline->mainbin &&
			player->pipeline->mainbin[MMPLAYER_M_SRC].gst,
			STREAMING_SERVICE_NONE);

	/* streaming service type if streaming */
	if (!MMPLAYER_IS_STREAMING(player))
		return STREAMING_SERVICE_NONE;

	streaming_type = (player->duration == 0) ?
		STREAMING_SERVICE_LIVE : STREAMING_SERVICE_VOD;

	switch (streaming_type) {
	case STREAMING_SERVICE_LIVE:
		LOGD("it's live streaming");
		break;
	case STREAMING_SERVICE_VOD:
		LOGD("it's vod streaming");
		break;
	default:
		LOGE("should not get here");
		break;
	}

	MMPLAYER_FLEAVE();

	return streaming_type;
}

/* this function sets the player state and also report
 * it to applicaton by calling callback function
 */
void
_mmplayer_set_state(mmplayer_t *player, int state)
{
	MMMessageParamType msg = {0, };

	MMPLAYER_RETURN_IF_FAIL(player);

	if (MMPLAYER_CURRENT_STATE(player) == state) {
		LOGW("already same state(%s)", MMPLAYER_STATE_GET_NAME(state));
		MMPLAYER_PENDING_STATE(player) = MM_PLAYER_STATE_NONE;
		return;
	}

	/* update player states */
	MMPLAYER_PREV_STATE(player) = MMPLAYER_CURRENT_STATE(player);
	MMPLAYER_CURRENT_STATE(player) = state;

	if (MMPLAYER_CURRENT_STATE(player) == MMPLAYER_PENDING_STATE(player))
		MMPLAYER_PENDING_STATE(player) = MM_PLAYER_STATE_NONE;

	/* print state */
	MMPLAYER_PRINT_STATE(player);

	switch (MMPLAYER_CURRENT_STATE(player)) {
	case MM_PLAYER_STATE_NULL:
	case MM_PLAYER_STATE_READY:
		break;
	case MM_PLAYER_STATE_PAUSED:
		__mmplayer_set_pause_state(player);
		break;
	case MM_PLAYER_STATE_PLAYING:
		__mmplayer_set_playing_state(player);
		break;
	case MM_PLAYER_STATE_NONE:
	default:
		LOGW("invalid target state, there is nothing to do.");
		break;
	}


	/* post message to application */
	if (MMPLAYER_TARGET_STATE(player) == state) {
		/* fill the message with state of player */
		msg.union_type = MM_MSG_UNION_STATE;
		msg.state.previous = MMPLAYER_PREV_STATE(player);
		msg.state.current = MMPLAYER_CURRENT_STATE(player);

		LOGD("player reach the target state (%s)", MMPLAYER_STATE_GET_NAME(MMPLAYER_TARGET_STATE(player)));

		/* state changed by resource callback */
		if (player->interrupted_by_resource)
			MMPLAYER_POST_MSG(player, MM_MESSAGE_STATE_INTERRUPTED, &msg);
		else /* state changed by usecase */
			MMPLAYER_POST_MSG(player, MM_MESSAGE_STATE_CHANGED, &msg);

	} else {
		LOGD("intermediate state, do nothing.");
		MMPLAYER_PRINT_STATE(player);
		return;
	}

	if (MMPLAYER_CURRENT_STATE(player) == MM_PLAYER_STATE_PLAYING
		&& !player->sent_bos) {
		MMPLAYER_POST_MSG(player, MM_MESSAGE_BEGIN_OF_STREAM, NULL);
		player->sent_bos = TRUE;
	}

	return;
}

int
_mmplayer_check_state(mmplayer_t *player, mmplayer_command_state_e command)
{
	mmplayer_state_e current_state = MM_PLAYER_STATE_NUM;
	mmplayer_state_e pending_state = MM_PLAYER_STATE_NUM;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
#ifdef __DEBUG__
	LOGD("incomming command : %d ", command);
#endif
	current_state = MMPLAYER_CURRENT_STATE(player);
	pending_state = MMPLAYER_PENDING_STATE(player);

	MMPLAYER_PRINT_STATE(player);

	switch (command) {
	case MMPLAYER_COMMAND_CREATE:
	{
		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_NULL;

		if (current_state == MM_PLAYER_STATE_NULL ||
			current_state == MM_PLAYER_STATE_READY ||
			current_state == MM_PLAYER_STATE_PAUSED ||
			current_state == MM_PLAYER_STATE_PLAYING)
			goto NO_OP;
	}
	break;

	case MMPLAYER_COMMAND_DESTROY:
	{
		/* destroy can called anytime */

		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_NONE;
	}
	break;

	case MMPLAYER_COMMAND_REALIZE:
	{
		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_READY;

		if (pending_state != MM_PLAYER_STATE_NONE) {
			goto INVALID_STATE;
		} else {
			/* need ready state to realize */
			if (current_state == MM_PLAYER_STATE_READY)
				goto NO_OP;

			if (current_state != MM_PLAYER_STATE_NULL)
				goto INVALID_STATE;
		}
	}
	break;

	case MMPLAYER_COMMAND_UNREALIZE:
	{
		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_NULL;

		if (current_state == MM_PLAYER_STATE_NULL)
			goto NO_OP;
	}
	break;

	case MMPLAYER_COMMAND_START:
	{
		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_PLAYING;

		if (pending_state == MM_PLAYER_STATE_NONE) {
			if (current_state == MM_PLAYER_STATE_PLAYING)
				goto NO_OP;
			else if (current_state  != MM_PLAYER_STATE_READY &&
				current_state != MM_PLAYER_STATE_PAUSED)
				goto INVALID_STATE;
		} else if (pending_state == MM_PLAYER_STATE_PLAYING) {
			goto ALREADY_GOING;
		} else if (pending_state == MM_PLAYER_STATE_PAUSED) {
			LOGD("player is going to paused state, just change the pending state as playing");
		} else {
			goto INVALID_STATE;
		}
	}
	break;

	case MMPLAYER_COMMAND_STOP:
	{
		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_READY;

		if (current_state == MM_PLAYER_STATE_READY)
			goto NO_OP;

		/* need playing/paused state to stop */
		if (current_state != MM_PLAYER_STATE_PLAYING &&
			 current_state != MM_PLAYER_STATE_PAUSED)
			goto INVALID_STATE;
	}
	break;

	case MMPLAYER_COMMAND_PAUSE:
	{
		if (MMPLAYER_IS_LIVE_STREAMING(player))
			goto NO_OP;

		if (player->seek_state == MMPLAYER_SEEK_IN_PROGRESS)
			goto NOT_COMPLETED_SEEK;

		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_PAUSED;

		if (pending_state == MM_PLAYER_STATE_NONE) {
			if (current_state == MM_PLAYER_STATE_PAUSED)
				goto NO_OP;
			else if (current_state != MM_PLAYER_STATE_PLAYING && current_state != MM_PLAYER_STATE_READY) // support loading state of broswer
				goto INVALID_STATE;
		} else if (pending_state == MM_PLAYER_STATE_PAUSED) {
			goto ALREADY_GOING;
		} else if (pending_state == MM_PLAYER_STATE_PLAYING) {
			if (current_state == MM_PLAYER_STATE_PAUSED)
				LOGD("player is PAUSED going to PLAYING, just change the pending state as PAUSED");
			else
				goto INVALID_STATE;
		}
	}
	break;

	case MMPLAYER_COMMAND_RESUME:
	{
		if (player->seek_state == MMPLAYER_SEEK_IN_PROGRESS)
			goto NOT_COMPLETED_SEEK;

		MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_PLAYING;

		if (pending_state == MM_PLAYER_STATE_NONE) {
			if (current_state == MM_PLAYER_STATE_PLAYING)
				goto NO_OP;
			else if (current_state != MM_PLAYER_STATE_PAUSED)
				goto INVALID_STATE;
		} else if (pending_state == MM_PLAYER_STATE_PLAYING) {
			goto ALREADY_GOING;
		} else if (pending_state == MM_PLAYER_STATE_PAUSED) {
			LOGD("player is going to paused state, just change the pending state as playing");
		} else {
			goto INVALID_STATE;
		}
	}
		break;

	default:
		break;
	}
	player->cmd = command;

	return MM_ERROR_NONE;

INVALID_STATE:
	LOGW("since player is in wrong state(%s). it's not able to apply the command(%d)",
		MMPLAYER_STATE_GET_NAME(current_state), command);
	return MM_ERROR_PLAYER_INVALID_STATE;

NOT_COMPLETED_SEEK:
	LOGW("not completed seek");
	return MM_ERROR_PLAYER_DOING_SEEK;

NO_OP:
	LOGW("player is in the desired state(%s). doing noting", MMPLAYER_STATE_GET_NAME(current_state));
	return MM_ERROR_PLAYER_NO_OP;

ALREADY_GOING:
	LOGW("player is already going to %s, doing nothing", MMPLAYER_STATE_GET_NAME(pending_state));
	return MM_ERROR_PLAYER_NO_OP;
}

static int __mmplayer_acquire_hw_resource(mmplayer_t *player, mmplayer_resource_type_e type)
{
	int rm_ret = MM_RESOURCE_MANAGER_ERROR_NONE;
	mm_resource_manager_res_type_e rm_res_type = MM_RESOURCE_MANAGER_RES_TYPE_MAX;

	switch (type) {
		case MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER:
			rm_res_type = MM_RESOURCE_MANAGER_RES_TYPE_VIDEO_DECODER;
			break;
		case MMPLAYER_RESOURCE_TYPE_VIDEO_OVERLAY:
			rm_res_type = MM_RESOURCE_MANAGER_RES_TYPE_VIDEO_OVERLAY;
			break;
		case MMPLAYER_RESOURCE_TYPE_AUDIO_OFFLOAD:
			rm_res_type = MM_RESOURCE_MANAGER_RES_TYPE_AUDIO_OFFLOAD;
			break;
		default:
			LOGE("invalid mmplayer resource type %d", type);
			return MM_ERROR_PLAYER_INTERNAL;
	}

	if (player->hw_resource[type] != NULL) {
		LOGD("[%d type] resource was already acquired", type);
		return MM_ERROR_NONE;
	}

	LOGD("mark for acquire [%d type] resource", type);
	rm_ret = mm_resource_manager_mark_for_acquire(player->resource_manager,
			rm_res_type, MM_RESOURCE_MANAGER_RES_VOLUME_FULL, &player->hw_resource[type]);
	if (rm_ret != MM_RESOURCE_MANAGER_ERROR_NONE) {
		LOGE("failed to mark resource for acquire, ret(0x%x)", rm_ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	rm_ret = mm_resource_manager_commit(player->resource_manager);
	if (rm_ret != MM_RESOURCE_MANAGER_ERROR_NONE) {
		LOGE("failed to commit of resource, ret(0x%x)", rm_ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

static int __mmplayer_release_hw_resource(mmplayer_t *player, mmplayer_resource_type_e type)
{
	int rm_ret = MM_RESOURCE_MANAGER_ERROR_NONE;

	MMPLAYER_FENTER();

	if (player->hw_resource[type] == NULL) {
		LOGD("there is no acquired [%d type] resource", type);
		return MM_ERROR_NONE;
	}

	LOGD("mark for release [%d type] resource", type);
	rm_ret = mm_resource_manager_mark_for_release(player->resource_manager, player->hw_resource[type]);
	if (rm_ret != MM_RESOURCE_MANAGER_ERROR_NONE) {
		LOGE("failed to mark resource for release, ret(0x%x)", rm_ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	player->hw_resource[type] = NULL;

	rm_ret = mm_resource_manager_commit(player->resource_manager);
	if (rm_ret != MM_RESOURCE_MANAGER_ERROR_NONE) {
		LOGE("failed to commit of resource, ret(0x%x)", rm_ret);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

static void
__mmplayer_initialize_gapless_play(mmplayer_t *player)
{
	int i;

	MMPLAYER_FENTER();

	player->smooth_streaming = FALSE;
	player->videodec_linked = 0;
	player->audiodec_linked = 0;
	player->textsink_linked = 0;
	player->is_external_subtitle_present = FALSE;
	player->is_external_subtitle_added_now = FALSE;
	player->not_supported_codec = MISSING_PLUGIN_NONE;
	player->can_support_codec = FOUND_PLUGIN_NONE;
	player->pending_seek.is_pending = false;
	player->pending_seek.pos = 0;
	player->msg_posted = FALSE;
	player->has_many_types = FALSE;
	player->no_more_pad = FALSE;
	player->not_found_demuxer = 0;
	player->seek_state = MMPLAYER_SEEK_NONE;
	player->is_subtitle_force_drop = FALSE;
	player->play_subtitle = FALSE;
	player->adjust_subtitle_pos = 0;

	player->total_bitrate = 0;
	player->total_maximum_bitrate = 0;

	_mmplayer_track_initialize(player);
	__mmplayer_initialize_storage_info(player, MMPLAYER_PATH_MAX);

	for (i = 0; i < MM_PLAYER_STREAM_COUNT_MAX; i++) {
		player->bitrate[i] = 0;
		player->maximum_bitrate[i] = 0;
	}

	if (player->v_stream_caps) {
		gst_caps_unref(player->v_stream_caps);
		player->v_stream_caps = NULL;
	}

	mm_player_set_attribute((MMHandleType)player, NULL, "content_video_found", 0, NULL);

	/* clean found audio decoders */
	if (player->audio_decoders) {
		GList *a_dec = player->audio_decoders;
		for (; a_dec; a_dec = g_list_next(a_dec)) {
			gchar *name = a_dec->data;
			MMPLAYER_FREEIF(name);
		}
		g_list_free(player->audio_decoders);
		player->audio_decoders = NULL;
	}

	__mmplayer_release_hw_resource(player, MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER);

	MMPLAYER_FLEAVE();
}

static gpointer
__mmplayer_gapless_play_thread(gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	mmplayer_gst_element_t *mainbin = NULL;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, NULL);

	MMPLAYER_GAPLESS_PLAY_THREAD_LOCK(player);
	while (!player->gapless_play_thread_exit) {
		LOGD("gapless play thread started. waiting for signal.");
		MMPLAYER_GAPLESS_PLAY_THREAD_WAIT(player);

		LOGD("reconfigure pipeline for gapless play.");

		if (player->gapless_play_thread_exit) {
			if (player->gapless.reconfigure) {
				player->gapless.reconfigure = false;
				MMPLAYER_PLAYBACK_UNLOCK(player);
			}
			LOGD("exiting gapless play thread");
			break;
		}

		mainbin = player->pipeline->mainbin;

		MMPLAYER_RELEASE_ELEMENT(player, mainbin, MMPLAYER_M_MUXED_S_BUFFER);
		MMPLAYER_RELEASE_ELEMENT(player, mainbin, MMPLAYER_M_ID3DEMUX);
		MMPLAYER_RELEASE_ELEMENT(player, mainbin, MMPLAYER_M_AUTOPLUG);
		MMPLAYER_RELEASE_ELEMENT(player, mainbin, MMPLAYER_M_TYPEFIND);
		MMPLAYER_RELEASE_ELEMENT(player, mainbin, MMPLAYER_M_SRC);

		/* Initialize Player values */
		__mmplayer_initialize_gapless_play(player);

		_mmplayer_activate_next_source(player, GST_STATE_PLAYING);
	}
	MMPLAYER_GAPLESS_PLAY_THREAD_UNLOCK(player);

	return NULL;
}

static void
__mmplayer_remove_g_source_from_context(GMainContext *context, guint source_id)
{
	GSource *source = NULL;

	MMPLAYER_FENTER();

	source = g_main_context_find_source_by_id(context, source_id);
	if (source != NULL) {
		LOGW("context: %p, source id: %d, source: %p", context, source_id, source);
		g_source_destroy(source);
	}

	MMPLAYER_FLEAVE();
}

void
_mmplayer_bus_msg_thread_destroy(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	GstMessage *msg = NULL;
	GQueue *queue = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	/* disconnecting bus watch */
	if (player->bus_watcher)
		__mmplayer_remove_g_source_from_context(player->context.thread_default, player->bus_watcher);
	player->bus_watcher = 0;

	/* destroy the gst bus msg thread */
	if (player->bus_msg_thread) {
		MMPLAYER_BUS_MSG_THREAD_LOCK(player);
		player->bus_msg_thread_exit = TRUE;
		MMPLAYER_BUS_MSG_THREAD_SIGNAL(player);
		MMPLAYER_BUS_MSG_THREAD_UNLOCK(player);

		LOGD("gst bus msg thread exit.");
		g_thread_join(player->bus_msg_thread); /* can request cmd lock */
		player->bus_msg_thread = NULL;

		g_mutex_clear(&player->bus_msg_thread_mutex);
		g_cond_clear(&player->bus_msg_thread_cond);
	}

	g_mutex_lock(&player->bus_msg_q_lock);
	queue = player->bus_msg_q;
	while (!g_queue_is_empty(queue)) {
		msg = (GstMessage *)g_queue_pop_head(queue);
		if (!msg) {
			LOGW("msg is null");
			continue;
		}
		LOGW("remove remained %s msg", GST_MESSAGE_TYPE_NAME(msg));
		gst_message_unref(msg);
	}
	g_mutex_unlock(&player->bus_msg_q_lock);

	MMPLAYER_FLEAVE();
}

gboolean
_mmplayer_gst_remove_fakesink(mmplayer_t *player, mmplayer_gst_element_t *fakesink)
{
	GstElement *parent = NULL;

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(fakesink && fakesink->gst, TRUE);

	/* lock */
	MMPLAYER_FSINK_LOCK(player);

	/* get parent of fakesink */
	parent = (GstElement *)gst_object_get_parent((GstObject *)fakesink->gst);
	if (!parent) {
		LOGD("fakesink already removed");
		goto ERROR;
	}

	gst_element_set_locked_state(fakesink->gst, TRUE);

	/* setting the state to NULL never returns async
	 * so no need to wait for completion of state transiton
	 */
	if (GST_STATE_CHANGE_FAILURE == gst_element_set_state(fakesink->gst, GST_STATE_NULL))
		LOGE("fakesink state change failure!");
		/* FIXIT : should I return here? or try to proceed to next? */
		/* return FALSE; */

	/* remove fakesink from it's parent */
	if (!gst_bin_remove(GST_BIN(parent), fakesink->gst)) {
		LOGE("failed to remove fakesink");

		gst_object_unref(parent);

		goto ERROR;
	}

	gst_object_unref(parent);

	LOGD("state-holder removed");

	gst_element_set_locked_state(fakesink->gst, FALSE);

	MMPLAYER_FSINK_UNLOCK(player);
	return TRUE;

ERROR:
	if (fakesink->gst)
		gst_element_set_locked_state(fakesink->gst, FALSE);

	MMPLAYER_FSINK_UNLOCK(player);
	return FALSE;
}

static GstPadProbeReturn
__mmplayer_gst_selector_blocked(GstPad *pad, GstPadProbeInfo *info, gpointer data)
{
	LOGD("pad(%s:%s) is blocked", GST_DEBUG_PAD_NAME(pad));
	return GST_PAD_PROBE_OK;
}

static void
__mmplayer_gst_selector_update_start_time(mmplayer_t *player, mmplayer_track_type_e stream_type)
{
	gint64 stop_running_time = 0;
	gint64 position_running_time = 0;
	gint64 position = 0;
	gint idx = 0;

	for (idx = MM_PLAYER_TRACK_TYPE_AUDIO; idx < MM_PLAYER_TRACK_TYPE_TEXT; idx++) {
		if ((player->gapless.update_segment[idx] == TRUE) ||
			!(player->selector[idx].event_probe_id)) {
#ifdef __DEBUG__
			LOGW("[%d] skip", idx);
#endif
			continue;
		}

		if (GST_CLOCK_TIME_IS_VALID(player->gapless.segment[idx].stop)) {
			stop_running_time =
				gst_segment_to_running_time(&player->gapless.segment[idx],
						GST_FORMAT_TIME, player->gapless.segment[idx].stop);
		} else if (GST_CLOCK_TIME_IS_VALID(player->gapless.segment[idx].duration)) {
			stop_running_time =
				gst_segment_to_running_time(&player->gapless.segment[idx],
						GST_FORMAT_TIME, player->gapless.segment[idx].duration);
		} else {
			LOGD("duration: %"GST_TIME_FORMAT, GST_TIME_ARGS(player->duration));
			stop_running_time =
				gst_segment_to_running_time(&player->gapless.segment[idx],
						GST_FORMAT_TIME, player->duration);
		}

		position_running_time =
			gst_segment_to_running_time(&player->gapless.segment[idx],
			GST_FORMAT_TIME, player->gapless.segment[idx].position);

		LOGD("[type:%d] time info %" GST_TIME_FORMAT " , %"
			GST_TIME_FORMAT" , %" GST_TIME_FORMAT,
			idx,
			GST_TIME_ARGS(stop_running_time),
			GST_TIME_ARGS(position_running_time),
			GST_TIME_ARGS(gst_segment_to_running_time(&player->gapless.segment[idx],
			GST_FORMAT_TIME, player->gapless.segment[idx].start)));

		position_running_time = MAX(position_running_time, stop_running_time);
		position_running_time -= gst_segment_to_running_time(&player->gapless.segment[idx],
										GST_FORMAT_TIME, player->gapless.segment[idx].start);
		position_running_time = MAX(0, position_running_time);
		position = MAX(position, position_running_time);
	}

	if (position != 0) {
		LOGD("[%d]GST_EVENT_STREAM_START: start_time from %" GST_TIME_FORMAT " to %" GST_TIME_FORMAT,
			stream_type, GST_TIME_ARGS(player->gapless.start_time[stream_type]),
			GST_TIME_ARGS(player->gapless.start_time[stream_type] + position));

		player->gapless.start_time[stream_type] += position;
	}

	return;
}

static GstPadProbeReturn
__mmplayer_gst_selector_event_probe(GstPad *pad, GstPadProbeInfo *info, gpointer data)
{
	GstPadProbeReturn ret = GST_PAD_PROBE_OK;
	GstEvent *event = GST_PAD_PROBE_INFO_DATA(info);
	mmplayer_t *player = (mmplayer_t *)data;
	GstCaps *caps = NULL;
	GstStructure *str = NULL;
	const gchar *name = NULL;
	mmplayer_track_type_e stream_type = MM_PLAYER_TRACK_TYPE_VIDEO;
	gboolean caps_ret = TRUE;

	if (GST_EVENT_TYPE(event) != GST_EVENT_STREAM_START &&
		GST_EVENT_TYPE(event) != GST_EVENT_FLUSH_STOP &&
		GST_EVENT_TYPE(event) != GST_EVENT_SEGMENT &&
		GST_EVENT_TYPE(event) != GST_EVENT_EOS &&
		GST_EVENT_TYPE(event) != GST_EVENT_QOS)
		return ret;

	MMPLAYER_GST_GET_CAPS_INFO(pad, caps, str, name, caps_ret);
	if (!caps_ret)
		goto ERROR;

	if (strstr(name, "audio")) {
		stream_type = MM_PLAYER_TRACK_TYPE_AUDIO;
	} else if (strstr(name, "video")) {
		stream_type = MM_PLAYER_TRACK_TYPE_VIDEO;
	} else {
		/* text track is not supportable */
		LOGE("invalid name %s", name);
		goto ERROR;
	}

	switch (GST_EVENT_TYPE(event)) {
	case GST_EVENT_EOS:
		{
			/* in case of gapless, drop eos event not to send it to sink */
			if (player->gapless.reconfigure && !player->msg_posted) {
				LOGD("[%d] %s:%s EOS received but will be drop", stream_type, GST_DEBUG_PAD_NAME(pad));
				ret = GST_PAD_PROBE_DROP;
			}
			break;
		}
	case GST_EVENT_STREAM_START:
		{
			__mmplayer_gst_selector_update_start_time(player, stream_type);
			break;
		}
	case GST_EVENT_FLUSH_STOP:
		{
			LOGD("[%d] GST_EVENT_FLUSH_STOP", stream_type);
			gst_segment_init(&player->gapless.segment[stream_type], GST_FORMAT_UNDEFINED);
			player->gapless.start_time[stream_type] = 0;
			break;
		}
	case GST_EVENT_SEGMENT:
		{
			GstSegment segment;
			GstEvent *tmpev;

			LOGD("[%d] GST_EVENT_SEGMENT", stream_type);
			gst_event_copy_segment(event, &segment);

			if (segment.format != GST_FORMAT_TIME)
				break;

			LOGD("segment base:%" GST_TIME_FORMAT ", offset:%" GST_TIME_FORMAT
				 ", start:%" GST_TIME_FORMAT ", stop: %" GST_TIME_FORMAT
				 ", time: %" GST_TIME_FORMAT ", pos: %" GST_TIME_FORMAT ", dur: %" GST_TIME_FORMAT,
				GST_TIME_ARGS(segment.base), GST_TIME_ARGS(segment.offset),
				GST_TIME_ARGS(segment.start), GST_TIME_ARGS(segment.stop),
				GST_TIME_ARGS(segment.time), GST_TIME_ARGS(segment.position), GST_TIME_ARGS(segment.duration));

			/* keep the all the segment ev to cover the seeking */
			gst_segment_copy_into(&segment, &player->gapless.segment[stream_type]);
			player->gapless.update_segment[stream_type] = TRUE;

			if (!player->gapless.running)
				break;

			player->gapless.segment[stream_type].base = player->gapless.start_time[stream_type];

			LOGD("[%d] new base: %" GST_TIME_FORMAT, stream_type, GST_TIME_ARGS(player->gapless.segment[stream_type].base));

			tmpev = gst_event_new_segment(&player->gapless.segment[stream_type]);
			gst_event_set_seqnum(tmpev, gst_event_get_seqnum(event));
			gst_event_unref(event);
			GST_PAD_PROBE_INFO_DATA(info) = tmpev;

			break;
		}
	case GST_EVENT_QOS:
		{
			gdouble proportion = 0.0;
			GstClockTimeDiff diff = 0;
			GstClockTime timestamp = 0;
			gint64 running_time_diff = -1;
			GstQOSType type = 0;
			GstEvent *tmpev = NULL;

			running_time_diff = player->gapless.segment[stream_type].base;

			if (running_time_diff <= 0) /* don't need to adjust */
				break;

			gst_event_parse_qos(event, &type, &proportion, &diff, &timestamp);
			gst_event_unref(event);

			if (timestamp < running_time_diff) {
				LOGW("QOS event from previous group");
				ret = GST_PAD_PROBE_DROP;
				break;
			}

#ifdef __DEBUG__
			LOGD("[%d] Adjusting QOS event: %" GST_TIME_FORMAT
				 " - %" GST_TIME_FORMAT " = %" GST_TIME_FORMAT,
						stream_type, GST_TIME_ARGS(timestamp),
						GST_TIME_ARGS(running_time_diff),
						GST_TIME_ARGS(timestamp - running_time_diff));
#endif

			timestamp -= running_time_diff;

			/* That case is invalid for QoS events */
			if (diff < 0 && -diff > timestamp) {
				LOGW("QOS event from previous group");
				ret = GST_PAD_PROBE_DROP;
				break;
			}

			tmpev = gst_event_new_qos(GST_QOS_TYPE_UNDERFLOW, proportion, diff, timestamp);
			GST_PAD_PROBE_INFO_DATA(info) = tmpev;

			break;
		}
	default:
		break;
	}

ERROR:
	if (caps)
		gst_caps_unref(caps);
	return ret;
}

/* create fakesink for audio or video path witout audiobin or videobin */
static void
__mmplayer_gst_make_fakesink(mmplayer_t *player, GstPad *pad, const gchar *name)
{
	GstElement *pipeline = NULL;
	GstElement *fakesink = NULL;
	GstPad *sinkpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->mainbin);

	pipeline = player->pipeline->mainbin[MMPLAYER_M_PIPE].gst;

	/* fake sink */
	fakesink = gst_element_factory_make("fakesink", NULL);
	if (fakesink == NULL) {
		LOGE("failed to create fakesink");
		goto EXIT;
	}

	/* store it as it's sink element */
	__mmplayer_add_sink(player, fakesink);

	gst_bin_add(GST_BIN(pipeline), fakesink);

	/* link */
	sinkpad = gst_element_get_static_pad(fakesink, "sink");

	LOGD("pad link %s:%s - %s:%s", GST_DEBUG_PAD_NAME(pad), GST_DEBUG_PAD_NAME(sinkpad));

	if (gst_pad_link(pad, sinkpad) != GST_PAD_LINK_OK) {
		LOGE("failed to link fakesink");
		gst_object_unref(GST_OBJECT(fakesink));
		goto EXIT;
	}

	if (strstr(name, "video")) {
		if (player->v_stream_caps) {
			gst_caps_unref(player->v_stream_caps);
			player->v_stream_caps = NULL;
		}
		if (player->ini.set_dump_element_flag)
			__mmplayer_add_dump_buffer_probe(player, fakesink);
	}

	g_object_set(G_OBJECT(fakesink), "sync", TRUE, NULL);
	gst_element_set_state(fakesink, GST_STATE_PAUSED);

EXIT:
	if (sinkpad)
		gst_object_unref(GST_OBJECT(sinkpad));

	MMPLAYER_FLEAVE();
	return;
}

static GstElement *
__mmplayer_gst_make_selector(mmplayer_t *player, main_element_id_e elem_idx, mmplayer_track_type_e stream_type)
{
	GstElement *pipeline = NULL;
	GstElement *selector = NULL;
	GstPad *srcpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->mainbin, NULL);

	selector = gst_element_factory_make("input-selector", NULL);
	if (!selector) {
		LOGE("failed to create input-selector");
		return NULL;
	}
	g_object_set(selector, "sync-streams", TRUE, NULL);

	player->pipeline->mainbin[elem_idx].id = elem_idx;
	player->pipeline->mainbin[elem_idx].gst = selector;

	/* player->selector[stream_type].active_pad_index = DEFAULT_TRACK; */

	srcpad = gst_element_get_static_pad(selector, "src");

	LOGD("blocking %s:%s", GST_DEBUG_PAD_NAME(srcpad));
	player->selector[stream_type].block_id = gst_pad_add_probe(srcpad, GST_PAD_PROBE_TYPE_BLOCK_DOWNSTREAM,
		__mmplayer_gst_selector_blocked, NULL, NULL);
	player->selector[stream_type].event_probe_id = gst_pad_add_probe(srcpad, GST_PAD_PROBE_TYPE_EVENT_BOTH|GST_PAD_PROBE_TYPE_EVENT_FLUSH,
		__mmplayer_gst_selector_event_probe, player, NULL);

	gst_element_set_state(selector, GST_STATE_PAUSED);

	pipeline = player->pipeline->mainbin[MMPLAYER_M_PIPE].gst;
	gst_bin_add(GST_BIN(pipeline), selector);

	gst_object_unref(GST_OBJECT(srcpad));

	MMPLAYER_FLEAVE();
	return selector;
}

void
_mmplayer_gst_decode_pad_added(GstElement *elem, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	GstElement *selector = NULL;
	GstCaps *caps = NULL;
	GstStructure *str = NULL;
	const gchar *name = NULL;
	GstPad *sinkpad = NULL;
	gboolean first_track = FALSE;
	gboolean caps_ret = TRUE;

	main_element_id_e elem_idx = MMPLAYER_M_NUM;
	mmplayer_track_type_e stream_type = MM_PLAYER_TRACK_TYPE_AUDIO;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(elem && pad);
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->mainbin);

	LOGD("pad-added signal handling");

	/* get mimetype from caps */
	MMPLAYER_GST_GET_CAPS_INFO(pad, caps, str, name, caps_ret);
	if (!caps_ret)
		goto ERROR;

	MMPLAYER_LOG_GST_CAPS_TYPE(caps);
#ifdef __DEBUG__
	LOGD("detected mimetype : %s", name);
#endif

	if (strstr(name, "video")) {
		gint stype = 0;
		gchar *caps_str = NULL;

		caps_str = gst_caps_to_string(caps);
		if (caps_str && (strstr(caps_str, "ST12") || strstr(caps_str, "SN12") ||
			strstr(caps_str, "SN21") || strstr(caps_str, "S420") || strstr(caps_str, "SR32")))
			player->set_mode.video_zc = true;

		MMPLAYER_FREEIF(caps_str);

		mm_player_set_attribute((MMHandleType)player, NULL, "content_video_found", TRUE, NULL);
		mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &stype);

		LOGD("surface type : %d", stype);

		if (MMPLAYER_IS_MS_BUFF_SRC(player)) {
			__mmplayer_gst_create_sinkbin(elem, pad, player);
			goto DONE;
		}

		/* in case of exporting video frame, it requires the 360 video filter.
		 * it will be handled in _no_more_pads(). */
		if ((stype == MM_DISPLAY_SURFACE_NULL) && (!player->set_mode.video_export)) {
			__mmplayer_gst_make_fakesink(player, pad, name);
			goto DONE;
		}

		LOGD("video selector is required");
		elem_idx = MMPLAYER_M_V_INPUT_SELECTOR;
		stream_type = MM_PLAYER_TRACK_TYPE_VIDEO;
	} else if (strstr(name, "audio")) {
		gint samplerate = 0;
		gint channels = 0;

		if (MMPLAYER_IS_MS_BUFF_SRC(player) || player->build_audio_offload) {
			if (player->build_audio_offload)
				player->no_more_pad = TRUE; /* remove state holder */
			__mmplayer_gst_create_sinkbin(elem, pad, player);
			goto DONE;
		}

		gst_structure_get_int(str, "rate", &samplerate);
		gst_structure_get_int(str, "channels", &channels);

		if ((channels > 0 && samplerate == 0)) { /* exclude audio decoding */
			__mmplayer_gst_make_fakesink(player, pad, name);
			goto DONE;
		}

		LOGD("audio selector is required");
		elem_idx = MMPLAYER_M_A_INPUT_SELECTOR;
		stream_type = MM_PLAYER_TRACK_TYPE_AUDIO;

	} else if (strstr(name, "text")) {
		LOGD("text selector is required");
		elem_idx = MMPLAYER_M_T_INPUT_SELECTOR;
		stream_type = MM_PLAYER_TRACK_TYPE_TEXT;
	} else {
		LOGE("invalid caps info");
		goto ERROR;
	}

	/* check selector and create it */
	if (!(selector = player->pipeline->mainbin[elem_idx].gst)) {
		selector = __mmplayer_gst_make_selector(player, elem_idx, stream_type);
		if (!selector)
			goto ERROR;
		first_track = TRUE;
	} else {
		LOGD("input-selector is already created.");
	}

	/* link */
	sinkpad = gst_element_get_request_pad(selector, "sink_%u");

	LOGD("pad link: %s:%s - %s:%s", GST_DEBUG_PAD_NAME(pad), GST_DEBUG_PAD_NAME(sinkpad));

	if (gst_pad_link(pad, sinkpad) != GST_PAD_LINK_OK) {
		LOGE("failed to link selector");
		gst_object_unref(GST_OBJECT(selector));
		goto ERROR;
	}

	if (first_track) {
		LOGD("this track will be activated");
		g_object_set(selector, "active-pad", sinkpad, NULL);
	}

	_mmplayer_track_update_selector_info(player, stream_type, sinkpad);

DONE:
ERROR:

	if (caps)
		gst_caps_unref(caps);

	if (sinkpad) {
		gst_object_unref(GST_OBJECT(sinkpad));
		sinkpad = NULL;
	}

	return;
}

static gboolean
__mmplayer_create_sink_path(mmplayer_t *player, GstElement *selector, mmplayer_track_type_e type)
{
	GstPad *srcpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, FALSE);

	LOGD("type %d", type);

	if (!selector) {
		LOGD("there is no %d track", type);
		return TRUE;
	}

	srcpad = gst_element_get_static_pad(selector, "src");
	if (!srcpad) {
		LOGE("failed to get srcpad from selector");
		return FALSE;
	}

	LOGD("got pad %s:%s from selector", GST_DEBUG_PAD_NAME(srcpad));

	__mmplayer_gst_create_sinkbin(selector, srcpad, player);

	LOGD("unblocking %s:%s", GST_DEBUG_PAD_NAME(srcpad));
	if (player->selector[type].block_id) {
		gst_pad_remove_probe(srcpad, player->selector[type].block_id);
		player->selector[type].block_id = 0;
	}

	if (srcpad) {
		gst_object_unref(GST_OBJECT(srcpad));
		srcpad = NULL;
	}

	MMPLAYER_FLEAVE();
	return TRUE;
}

static void
__mmplayer_set_decode_track_info(mmplayer_t *player, mmplayer_track_type_e type)
{
	gint active_index = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	LOGD("type: %d, the num of track: %d", type, player->selector[type].total_track_num);

	/* change track to active pad */
	active_index = player->selector[type].active_pad_index;
	if ((active_index != DEFAULT_TRACK) &&
		(__mmplayer_change_selector_pad(player, type, active_index) != MM_ERROR_NONE)) {
		LOGW("failed to change %d type track to %d", type, active_index);
		player->selector[type].active_pad_index = DEFAULT_TRACK;
		return;
	}

	if (type == MM_PLAYER_TRACK_TYPE_TEXT)
		mm_player_set_attribute((MMHandleType)player, NULL,
				"content_text_track_num", player->selector[type].total_track_num,
				"current_text_track_index", player->selector[type].active_pad_index, NULL);

	MMPLAYER_FLEAVE();
	return;
}

static gboolean
__mmplayer_create_audio_sink_path(mmplayer_t *player, GstElement *audio_selector)
{
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, FALSE);

	if (!audio_selector) {
		LOGD("there is no audio track, num_dynamic_pad %d", player->num_dynamic_pad);

		/* in case the source is changed, output can be changed. */
		if ((player->pipeline->audiobin) && (player->pipeline->audiobin[MMPLAYER_A_BIN].gst)) {
			LOGD("remove previous audiobin if it exist");

			__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_AUDIOBIN);
			__mmplayer_del_sink(player, player->pipeline->audiobin[MMPLAYER_A_SINK].gst);

			MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->audiobin, MMPLAYER_A_BIN);
			MMPLAYER_FREEIF(player->pipeline->audiobin);
		}

		if (player->num_dynamic_pad == 0) /* FIXME: num_dynamic_pad is only for rtsp? */
			__mmplayer_pipeline_complete(NULL, player);

		return TRUE;
	}

	/* apply the audio track information */
	__mmplayer_set_decode_track_info(player, MM_PLAYER_TRACK_TYPE_AUDIO);

	/* create audio sink path */
	if (!__mmplayer_create_sink_path(player, audio_selector, MM_PLAYER_TRACK_TYPE_AUDIO)) {
		LOGE("failed to create audio sink path");
		return FALSE;
	}

	MMPLAYER_FLEAVE();
	return TRUE;
}

static gboolean
__mmplayer_create_text_sink_path(mmplayer_t *player, GstElement *text_selector)
{
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && text_selector, FALSE);

	if (MMPLAYER_IS_MS_BUFF_SRC(player)) {
		LOGD("text path is not supproted");
		return TRUE;
	}

	/* apply the text track information */
	__mmplayer_set_decode_track_info(player, MM_PLAYER_TRACK_TYPE_TEXT);

	if (player->selector[MM_PLAYER_TRACK_TYPE_TEXT].total_track_num > 0)
		player->has_closed_caption = TRUE;

	/* create text decode path */
	player->no_more_pad = TRUE;

	if (!__mmplayer_create_sink_path(player, text_selector, MM_PLAYER_TRACK_TYPE_TEXT)) {
		LOGE("failed to create text sink path");
		return FALSE;
	}

	MMPLAYER_FLEAVE();
	return TRUE;
}

static gboolean
__mmplayer_gst_set_queue2_buffering(mmplayer_t *player)
{
	gint64 dur_bytes = 0L;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline &&
		player->pipeline->mainbin && player->streamer, FALSE);

	if (!gst_element_query_duration(player->pipeline->mainbin[MMPLAYER_M_SRC].gst, GST_FORMAT_BYTES, &dur_bytes))
		LOGE("fail to get duration.");

	/* there is no mq, enable use-buffering on queue2 (ex) wav streaming
	 * use file information was already set on Q2 when it was created. */
	_mm_player_streaming_set_queue2(player->streamer,
					player->pipeline->mainbin[MMPLAYER_M_MUXED_S_BUFFER].gst,
					TRUE,								/* use_buffering */
					MUXED_BUFFER_TYPE_MAX,				/* use previous buffer type setting */
					((dur_bytes > 0) ? ((guint64)dur_bytes) : 0));

	MMPLAYER_FLEAVE();
	return TRUE;
}

static void
__mmplayer_gst_decode_no_more_pads(GstElement *elem, gpointer data)
{
	mmplayer_t *player = NULL;
	GstElement *video_selector = NULL;
	GstElement *audio_selector = NULL;
	GstElement *text_selector = NULL;

	MMPLAYER_FENTER();
	player = (mmplayer_t *)data;

	LOGD("no-more-pad signal handling");

	if ((player->cmd == MMPLAYER_COMMAND_DESTROY) ||
		(player->cmd == MMPLAYER_COMMAND_UNREALIZE)) {
		LOGW("player is shutting down");
		goto EXIT;
	}

	if ((MMPLAYER_IS_HTTP_STREAMING(player)) &&
		(!player->pipeline->mainbin[MMPLAYER_M_DEMUXED_S_BUFFER].gst) &&
		(player->pipeline->mainbin[MMPLAYER_M_MUXED_S_BUFFER].gst)) {
		if (!__mmplayer_gst_set_queue2_buffering(player)) {
			LOGE("failed to set queue2 buffering");
			goto EXIT;
		}
	}

	video_selector = player->pipeline->mainbin[MMPLAYER_M_V_INPUT_SELECTOR].gst;
	audio_selector = player->pipeline->mainbin[MMPLAYER_M_A_INPUT_SELECTOR].gst;
	text_selector = player->pipeline->mainbin[MMPLAYER_M_T_INPUT_SELECTOR].gst;

	if (!video_selector && !audio_selector && !text_selector) {
		LOGW("there is no selector");
		player->no_more_pad = TRUE;
		goto EXIT;
	}

	/* create video path followed by video-select */
	if (video_selector && !audio_selector && !text_selector)
		player->no_more_pad = TRUE;

	if (!__mmplayer_create_sink_path(player, video_selector, MM_PLAYER_TRACK_TYPE_VIDEO))
		goto EXIT;

	/* create audio path followed by audio-select */
	if (audio_selector && !text_selector)
		player->no_more_pad = TRUE;

	if (!__mmplayer_create_audio_sink_path(player, audio_selector))
		goto EXIT;

	/* create text path followed by text-select */
	__mmplayer_create_text_sink_path(player, text_selector);

EXIT:
	if (player->gapless.reconfigure) {
		player->gapless.reconfigure = FALSE;
		MMPLAYER_PLAYBACK_UNLOCK(player);
	}

	MMPLAYER_FLEAVE();
}

static gboolean
__mmplayer_gst_add_sinkbin_to_pipeline(mmplayer_t *player, GstElement *sinkbin, GstPad *pad, gboolean reusing, gchar *sink_pad_name)
{
	gboolean ret = FALSE;
	GstElement *pipeline = NULL;
	GstPad *sinkpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(sinkbin && pad, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->mainbin, FALSE);

	pipeline = player->pipeline->mainbin[MMPLAYER_M_PIPE].gst;

	sinkpad = gst_element_get_static_pad(GST_ELEMENT(sinkbin), sink_pad_name);
	if (!sinkpad) {
		LOGE("failed to get pad from sinkbin");
		goto EXIT;
	}

	if (reusing) {
		/* link only */
		if (gst_pad_link(pad, sinkpad) != GST_PAD_LINK_OK) {
			LOGE("failed to link sinkbin for reusing");
			goto EXIT; /* exit either pass or fail */
		}
	} else {
		/* warm up */
		if (gst_element_set_state(sinkbin, GST_STATE_READY) == GST_STATE_CHANGE_FAILURE) {
			LOGE("failed to set state(READY) to sinkbin");
			goto EXIT;
		}

		/* add */
		if (!gst_bin_add(GST_BIN(pipeline), sinkbin)) {
			LOGE("failed to add sinkbin to pipeline");
			goto EXIT;
		}

		/* link */
		if (gst_pad_link(pad, sinkpad) != GST_PAD_LINK_OK) {
			LOGE("failed to link %s:%s - %s:%s", GST_DEBUG_PAD_NAME(pad), GST_DEBUG_PAD_NAME(sinkpad));
			goto EXIT;
		}

		/* run */
		if (gst_element_set_state(sinkbin, GST_STATE_PAUSED) == GST_STATE_CHANGE_FAILURE) {
			LOGE("failed to set state(PAUSED) to sinkbin");
			goto EXIT;
		}
	}

	ret = TRUE;

EXIT:
	if (sinkpad)
		gst_object_unref(GST_OBJECT(sinkpad));
	sinkpad = NULL;

	MMPLAYER_FLEAVE();
	return ret;
}

static void
__mmplayer_gst_create_sinkbin(GstElement *elem, GstPad *pad, gpointer data)
{
	mmplayer_t *player = NULL;
	GstCaps *caps = NULL;
	gchar *caps_str = NULL;
	GstStructure *str = NULL;
	const gchar *name = NULL;
	GstElement *sinkbin = NULL;
	gboolean reusing = FALSE;
	gboolean caps_ret = TRUE;
	gchar *sink_pad_name = "sink";

	/* check handles */
	player = (mmplayer_t *)data;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(elem && pad);
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && MMPLAYER_GET_ATTRS(player));

	MMPLAYER_GST_GET_CAPS_INFO(pad, caps, str, name, caps_ret);
	if (!caps_ret)
		goto ERROR;

	caps_str = gst_caps_to_string(caps);
#ifdef __DEBUG__
	LOGD("detected mimetype : %s", name);
#endif
	if (strstr(name, "audio")) {
		if (player->pipeline->audiobin == NULL) {
			const gchar *audio_format = gst_structure_get_string(str, "format");
			if (audio_format) {
				LOGD("original audio format %s", audio_format);
				mm_player_set_attribute((MMHandleType)player, NULL,
						"content_audio_format", audio_format, strlen(audio_format), NULL);
			}

			if (__mmplayer_gst_create_audio_sink_bin(player) != MM_ERROR_NONE) {
				LOGE("failed to create audiobin. continuing without audio");
				goto ERROR;
			}

			sinkbin = player->pipeline->audiobin[MMPLAYER_A_BIN].gst;
			LOGD("creating audiobin success");
		} else {
			reusing = TRUE;
			sinkbin = player->pipeline->audiobin[MMPLAYER_A_BIN].gst;
			LOGD("reusing audiobin");
			_mmplayer_update_content_attrs(player, ATTR_AUDIO);
		}
	} else if (strstr(name, "video")) {
		/* 1. zero copy is updated at _decode_pad_added()
		 * 2. NULL surface type is handled in _decode_pad_added() */
		LOGD("zero copy %d", player->set_mode.video_zc);
		if (player->pipeline->videobin == NULL) {
			int surface_type = 0;
			mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &surface_type);
			LOGD("display_surface_type (%d)", surface_type);

			if ((surface_type == MM_DISPLAY_SURFACE_OVERLAY) &&
				(__mmplayer_acquire_hw_resource(player, MMPLAYER_RESOURCE_TYPE_VIDEO_OVERLAY) != MM_ERROR_NONE)) {
				LOGE("failed to acquire video overlay resource");
				goto ERROR;
			}

			player->interrupted_by_resource = FALSE;

			if (__mmplayer_gst_create_video_sink_bin(player, caps, surface_type) != MM_ERROR_NONE) {
				LOGE("failed to create videobin. continuing without video");
				goto ERROR;
			}

			sinkbin = player->pipeline->videobin[MMPLAYER_V_BIN].gst;
			LOGD("creating videosink bin success");
		} else {
			reusing = TRUE;
			sinkbin = player->pipeline->videobin[MMPLAYER_V_BIN].gst;
			LOGD("re-using videobin");
			_mmplayer_update_content_attrs(player, ATTR_VIDEO);
		}
	} else if (strstr(name, "text")) {
		if (player->pipeline->textbin == NULL) {
			if (__mmplayer_gst_create_text_sink_bin(player) != MM_ERROR_NONE) {
				LOGE("failed to create text sink bin. continuing without text");
				goto ERROR;
			}

			sinkbin = player->pipeline->textbin[MMPLAYER_T_BIN].gst;
			player->textsink_linked  = 1;
			LOGD("creating textsink bin success");
		} else {
			if (!player->textsink_linked) {
				LOGD("re-using textbin");
				reusing = TRUE;
				sinkbin = player->pipeline->textbin[MMPLAYER_T_BIN].gst;
				player->textsink_linked  = 1;
			} else {
				/* linked textbin exist which means that the external subtitle path exist already */
				LOGW("ignoring internal subtutle since external subtitle is available");
			}
		}
		sink_pad_name = "text_sink";
	} else {
		LOGW("unknown mime type %s, ignoring it", name);
		goto ERROR;
	}

	if (!__mmplayer_gst_add_sinkbin_to_pipeline(player, sinkbin, pad, reusing, sink_pad_name))
		goto ERROR;

	LOGD("[handle: %p] success to create and link sink bin", player);

	/* FIXIT : we cannot hold callback for 'no-more-pad' signal because signal was emitted in
	 * streaming task. if the task blocked, then buffer will not flow to the next element
	 *(autoplugging element). so this is special hack for streaming. please try to remove it
	 */
	/* dec stream count. we can remove fakesink if it's zero */
	if (player->num_dynamic_pad)
		player->num_dynamic_pad--;

	LOGD("no more pads: %d, stream count dec : %d(num of dynamic pad)", player->no_more_pad, player->num_dynamic_pad);

	if ((player->no_more_pad) && (player->num_dynamic_pad == 0))
		__mmplayer_pipeline_complete(NULL, player);

ERROR:

	MMPLAYER_FREEIF(caps_str);

	if (caps)
		gst_caps_unref(caps);

	return;
}

static gboolean
__mmplayer_get_property_value_for_rotation(mmplayer_t *player, int display_angle, int orientation, int *value)
{
	int required_angle = 0; /* Angle required for straight view */
	int rotation_angle = 0;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(value, FALSE);

	/* Counter clockwise */
	switch (orientation) {
	case 0:
		required_angle = 0;
		break;
	case 90:
		required_angle = 270;
		break;
	case 180:
		required_angle = 180;
		break;
	case 270:
		required_angle = 90;
		break;
	}

	rotation_angle = display_angle + required_angle;
	if (rotation_angle >= 360)
		rotation_angle -= 360;

	/* chech if supported or not */
	if (rotation_angle % 90) {
		LOGD("not supported rotation angle = %d", rotation_angle);
		return FALSE;
	}

	switch (rotation_angle) {
	case 0:
		*value = MM_DISPLAY_ROTATION_NONE;
		break;
	case 90:
		*value = MM_DISPLAY_ROTATION_90;
		break;
	case 180:
		*value = MM_DISPLAY_ROTATION_180;
		break;
	case 270:
		*value = MM_DISPLAY_ROTATION_270;
		break;
	}

	LOGD("setting rotation property value : %d", *value);

	return TRUE;
}

int
_mmplayer_get_video_angle(mmplayer_t *player, int *display_angle, int *orientation)
{
	int display_rotation = 0;
	gchar *org_orient = NULL;
	MMHandleType attrs = MMPLAYER_GET_ATTRS(player);

	if (!attrs) {
		LOGE("cannot get content attribute");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	if (display_angle) {
		/* update user roation */
		mm_attrs_get_int_by_name(attrs, "display_rotation", &display_rotation);

		/* Counter clockwise */
		switch (display_rotation) {
		case MM_DISPLAY_ROTATION_NONE:
			*display_angle = 0;
			break;
		case MM_DISPLAY_ROTATION_90:
			*display_angle = 90;
			break;
		case MM_DISPLAY_ROTATION_180:
			*display_angle = 180;
			break;
		case MM_DISPLAY_ROTATION_270:
			*display_angle = 270;
			break;
		default:
			LOGW("wrong angle type : %d", display_rotation);
			break;
		}
		LOGD("check user angle: %d", *display_angle);
	}

	if (orientation) {
		/* Counter clockwise */
		mm_attrs_get_string_by_name(attrs, "content_video_orientation", &org_orient);

		if (org_orient) {
			if (!strcmp(org_orient, "rotate-90"))
				*orientation = 90;
			else if (!strcmp(org_orient, "rotate-180"))
				*orientation = 180;
			else if (!strcmp(org_orient, "rotate-270"))
				*orientation = 270;
			else
				LOGD("original rotation is %s", org_orient);
		} else {
			LOGD("content_video_orientation get fail");
		}

		LOGD("check orientation: %d", *orientation);
	}

	return MM_ERROR_NONE;
}

static void __mmplayer_video_param_set_display_rotation(mmplayer_t *player)
{
	int rotation_value = 0;
	int orientations = 0; // current supported angle values are 0, 90, 180, 270
	int display_angle = 0;
	MMPLAYER_FENTER();

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	_mmplayer_get_video_angle(player, &display_angle, &orientations);

	/* get rotation value to set */
	__mmplayer_get_property_value_for_rotation(player, display_angle, orientations, &rotation_value);
	g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "rotate", rotation_value, NULL);
	LOGD("set video param : rotate %d", rotation_value);
}

static void __mmplayer_video_param_set_display_visible(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	int visible = 0;
	MMPLAYER_FENTER();

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_IF_FAIL(attrs);

	mm_attrs_get_int_by_name(attrs, "display_visible", &visible);
	g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "visible", visible, NULL);
	LOGD("set video param : visible %d", visible);
}

static void __mmplayer_video_param_set_display_method(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	int display_method = 0;
	MMPLAYER_FENTER();

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_IF_FAIL(attrs);

	mm_attrs_get_int_by_name(attrs, "display_method", &display_method);
	g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "display-geometry-method", display_method, NULL);
	LOGD("set video param : method %d", display_method);
}

static void __mmplayer_video_param_set_video_roi_area(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	int handle = 0;
	MMPLAYER_FENTER();

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_IF_FAIL(attrs);

	mm_attrs_get_int_by_name(attrs, "display_overlay", &handle);
	MMPLAYER_RETURN_IF_FAIL(handle);

	gst_video_overlay_set_video_roi_area(
		 GST_VIDEO_OVERLAY(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
		 player->video_roi.scale_x, player->video_roi.scale_y, player->video_roi.scale_width, player->video_roi.scale_height);
	LOGD("set video param : video roi area scale value: x(%f) y(%f) width(%f) height(%f)",
		player->video_roi.scale_x, player->video_roi.scale_y, player->video_roi.scale_width, player->video_roi.scale_height);
}

static void __mmplayer_video_param_set_roi_area(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	int handle = 0;
	/*set wl_display*/
	int win_roi_x = 0;
	int win_roi_y = 0;
	int win_roi_width = 0;
	int win_roi_height = 0;
	MMPLAYER_FENTER();

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_IF_FAIL(attrs);

	mm_attrs_get_int_by_name(attrs, "display_overlay", &handle);
	MMPLAYER_RETURN_IF_FAIL(handle);

	/* It should be set after setting window */
	mm_attrs_multiple_get(attrs, NULL,
				"display_win_roi_x", &win_roi_x,
				"display_win_roi_y", &win_roi_y,
				"display_win_roi_width", &win_roi_width,
				"display_win_roi_height", &win_roi_height, NULL);

	/* After setting window handle, set display roi area */
	gst_video_overlay_set_display_roi_area(
		 GST_VIDEO_OVERLAY(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
		 win_roi_x, win_roi_y, win_roi_width, win_roi_height);
	LOGD("set video param : roi area : x(%d) y(%d) width(%d) height(%d)",
		win_roi_x, win_roi_y, win_roi_width, win_roi_height);
}

static void __mmplayer_video_param_set_display_overlay(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	int handle = 0;

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY))
		return;

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_IF_FAIL(attrs);

	/* common case if using overlay surface */
	mm_attrs_get_int_by_name(attrs, "display_overlay", &handle);
	MMPLAYER_RETURN_IF_FAIL(handle);

	/* default is using wl_surface_id */
	LOGD("set video param : wl_surface_id %d", handle);
	gst_video_overlay_set_wl_window_wl_surface_id(
			GST_VIDEO_OVERLAY(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
			handle);
}

int
_mmplayer_update_video_overlay_param(mmplayer_t *player, const char *param_name)
{
	gboolean update_all_param = FALSE;

	MMPLAYER_FENTER();

	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_OVERLAY)) {
		LOGW("videosink is not ready yet");
		return MM_ERROR_PLAYER_NOT_INITIALIZED;
	}

	if (strcmp(player->ini.videosink_element_overlay, "tizenwlsink")) {
		LOGE("invalid videosink [%s]", player->ini.videosink_element_overlay);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	LOGD("param_name : %s", param_name);
	if (!g_strcmp0(param_name, "update_all_param"))
		update_all_param = TRUE;

	if (update_all_param || !g_strcmp0(param_name, "display_overlay"))
		__mmplayer_video_param_set_display_overlay(player);
	if (update_all_param || !g_strcmp0(param_name, "display_method"))
		__mmplayer_video_param_set_display_method(player);
	if (update_all_param || !g_strcmp0(param_name, "display_visible"))
		__mmplayer_video_param_set_display_visible(player);
	if (update_all_param || !g_strcmp0(param_name, "display_rotation"))
		__mmplayer_video_param_set_display_rotation(player);
	if (update_all_param || !g_strcmp0(param_name, "display_win_roi_x"))
		__mmplayer_video_param_set_roi_area(player);
	if (update_all_param)
		__mmplayer_video_param_set_video_roi_area(player);


	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_set_audio_only(MMHandleType hplayer, bool audio_only)
{
	gboolean disable_overlay = FALSE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(player->pipeline->videobin &&
								player->pipeline->videobin[MMPLAYER_V_SINK].gst,
								MM_ERROR_PLAYER_NO_OP); /* invalid op */

	if (!g_object_class_find_property(G_OBJECT_GET_CLASS(player->pipeline->videobin[MMPLAYER_V_SINK].gst), "disable-overlay")) {
		LOGW("Display control is not supported");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	g_object_get(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "disable-overlay", &disable_overlay, NULL);

	if (audio_only == (bool)disable_overlay) {
		LOGE("It's the same with current setting: (%d)", audio_only);
		return MM_ERROR_NONE;
	}

	if (audio_only) {
		LOGE("disable overlay");
		g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "disable-overlay", TRUE, NULL);

		/* release overlay resource */
		if (__mmplayer_release_hw_resource(player, MMPLAYER_RESOURCE_TYPE_VIDEO_OVERLAY) != MM_ERROR_NONE) {
			LOGE("failed to release overlay resource");
			goto ERROR;
		}
	} else {
		if (__mmplayer_acquire_hw_resource(player, MMPLAYER_RESOURCE_TYPE_VIDEO_OVERLAY) != MM_ERROR_NONE) {
			LOGE("failed to acquire video overlay resource");
			goto ERROR;
		}
		player->interrupted_by_resource = FALSE;

		LOGD("enable overlay");
		__mmplayer_video_param_set_display_overlay(player);
		g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "disable-overlay", FALSE, NULL);
	}

ERROR:
	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_get_audio_only(MMHandleType hplayer, bool *paudio_only)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gboolean disable_overlay = FALSE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(paudio_only, MM_ERROR_INVALID_ARGUMENT);
	MMPLAYER_RETURN_VAL_IF_FAIL(player->pipeline->videobin &&
								player->pipeline->videobin[MMPLAYER_V_SINK].gst,
								MM_ERROR_PLAYER_NO_OP); /* invalid op */

	if (!g_object_class_find_property(G_OBJECT_GET_CLASS(player->pipeline->videobin[MMPLAYER_V_SINK].gst), "disable-overlay")) {
		LOGW("Display control is not supported");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	g_object_get(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "disable-overlay", &disable_overlay, NULL);

	*paudio_only = (bool)disable_overlay;

	LOGD("audio_only : %d", *paudio_only);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_gst_element_link_bucket(GList *element_bucket)
{
	GList *bucket = element_bucket;
	mmplayer_gst_element_t *element = NULL;
	mmplayer_gst_element_t *prv_element = NULL;
	GstElement *tee_element = NULL;
	gint successful_link_count = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(element_bucket, -1);

	prv_element = (mmplayer_gst_element_t *)bucket->data;
	bucket = bucket->next;

	for (; bucket; bucket = bucket->next) {
		element = (mmplayer_gst_element_t *)bucket->data;

		if (element && element->gst) {
			if (prv_element && prv_element->gst) {
				if (strstr(GST_ELEMENT_NAME(element->gst), "audio-tee-queue") && strcmp(GST_ELEMENT_NAME(prv_element->gst), "audio-tee")) {
					if (tee_element) {
						prv_element->gst = tee_element;
					} else {
						LOGD("failed to make new audio branch - linking [%s] to [%s] is not supported",
							GST_ELEMENT_NAME(GST_ELEMENT(prv_element->gst)),
							GST_ELEMENT_NAME(GST_ELEMENT(element->gst)));
						return -1;
					}
				}
				if (gst_element_link(GST_ELEMENT(prv_element->gst), GST_ELEMENT(element->gst))) {
					LOGD("linking [%s] to [%s] success",
						GST_ELEMENT_NAME(GST_ELEMENT(prv_element->gst)),
						GST_ELEMENT_NAME(GST_ELEMENT(element->gst)));
					successful_link_count++;
					if (!strcmp(GST_ELEMENT_NAME(prv_element->gst), "audio-tee")) {
						LOGD("keep audio-tee element for next audio pipeline branch");
						tee_element = prv_element->gst;
					}
				} else {
					LOGD("linking [%s] to [%s] failed",
						GST_ELEMENT_NAME(GST_ELEMENT(prv_element->gst)),
						GST_ELEMENT_NAME(GST_ELEMENT(element->gst)));
					return -1;
				}
			}
		}

		prv_element = element;
	}

	MMPLAYER_FLEAVE();

	return successful_link_count;
}

int
_mmplayer_gst_element_add_bucket_to_bin(GstBin *bin, GList *element_bucket)
{
	GList *bucket = element_bucket;
	mmplayer_gst_element_t *element = NULL;
	int successful_add_count = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(element_bucket, 0);
	MMPLAYER_RETURN_VAL_IF_FAIL(bin, 0);

	for (; bucket; bucket = bucket->next) {
		element = (mmplayer_gst_element_t *)bucket->data;

		if (element && element->gst) {
			if (!gst_bin_add(bin, GST_ELEMENT(element->gst))) {
				LOGD("_mmplayer_gst_element_link_bucket : Adding element [%s]  to bin [%s] failed",
					GST_ELEMENT_NAME(GST_ELEMENT(element->gst)),
					GST_ELEMENT_NAME(GST_ELEMENT(bin)));
				return 0;
			}
			successful_add_count++;
		}
	}

	MMPLAYER_FLEAVE();

	return successful_add_count;
}

static void
__mmplayer_gst_caps_notify_cb(GstPad *pad, GParamSpec *unused, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	GstCaps *caps = NULL;
	GstStructure *str = NULL;
	const char *name;
	gboolean caps_ret = TRUE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(pad);
	MMPLAYER_RETURN_IF_FAIL(unused);
	MMPLAYER_RETURN_IF_FAIL(data);

	caps = gst_pad_get_current_caps(pad);
	if (!caps)
		return;

	MMPLAYER_GST_GET_CAPS_INFO(pad, caps, str, name, caps_ret);
	if (!caps_ret)
		goto ERROR;

	LOGD("name = %s", name);

	if (strstr(name, "audio")) {
		_mmplayer_update_content_attrs(player, ATTR_AUDIO);

		if (player->audio_stream_changed_cb) {
			LOGE("call the audio stream changed cb");
			player->audio_stream_changed_cb(player->audio_stream_changed_cb_user_param);
		}
	} else if (strstr(name, "video")) {
		if ((name = gst_structure_get_string(str, "format")))
			player->set_mode.video_zc = name[0] == 'S';

		_mmplayer_update_content_attrs(player, ATTR_VIDEO);
		MMPLAYER_POST_MSG(player, MM_MESSAGE_VIDEO_STREAM_CHANGED, NULL);
	} else {
		LOGW("invalid caps info");
	}

ERROR:
	if (caps)
		gst_caps_unref(caps);

	MMPLAYER_FLEAVE();

	return;
}

void
_mmplayer_audio_stream_clear_buffer(mmplayer_t *player, gboolean send_all)
{
	GList *l = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	if (player->audio_stream_buff_list) {
		for (l = g_list_first(player->audio_stream_buff_list); l; l = g_list_next(l)) {
			mmplayer_audio_stream_buff_t *tmp = (mmplayer_audio_stream_buff_t *)l->data;
			if (tmp) {
				if (send_all) {
					LOGD("[%"G_GUINT64_FORMAT"] send remained data.", tmp->channel_mask);
					__mmplayer_audio_stream_send_data(player, tmp);
				}
				MMPLAYER_FREEIF(tmp->pcm_data);
				MMPLAYER_FREEIF(tmp);
			}
		}
		g_list_free(player->audio_stream_buff_list);
		player->audio_stream_buff_list = NULL;
	}

	MMPLAYER_FLEAVE();
}

static void
__mmplayer_audio_stream_send_data(mmplayer_t *player, mmplayer_audio_stream_buff_t *a_buffer)
{
	mmplayer_audio_decoded_data_info_t audio_stream = { 0, };

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->audio_decoded_cb);

	audio_stream.bitrate = a_buffer->bitrate;
	audio_stream.channel = a_buffer->channel;
	audio_stream.channel_mask = a_buffer->channel_mask;
	audio_stream.data_size = a_buffer->data_size;
	audio_stream.data = a_buffer->pcm_data;
	audio_stream.pcm_format = a_buffer->pcm_format;
#ifdef __DEBUG__
	LOGD("[%"G_GUINT64_FORMAT"] send data size:%d, %p", audio_stream.channel_mask, audio_stream.data_size, player->audio_decoded_cb_user_param);
#endif
	player->audio_decoded_cb(&audio_stream, player->audio_decoded_cb_user_param);

	MMPLAYER_FLEAVE();
}

static void
__mmplayer_audio_stream_decoded_render_cb(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	const gchar *pcm_format = NULL;
	gint channel = 0;
	gint rate = 0;
	guint64 channel_mask = 0;
	void *a_data = NULL;
	gint a_size = 0;
	mmplayer_audio_stream_buff_t *a_buffer = NULL;
	GstMapInfo mapinfo = GST_MAP_INFO_INIT;
	GList *l = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->audio_decoded_cb);

	gst_buffer_map(buffer, &mapinfo, GST_MAP_READ);
	a_data = mapinfo.data;
	a_size = mapinfo.size;

	GstCaps *caps = gst_pad_get_current_caps(pad);
	GstStructure *structure = gst_caps_get_structure(caps, 0);
#ifdef __DEBUG__
	MMPLAYER_LOG_GST_CAPS_TYPE(caps);
#endif
	pcm_format = gst_structure_get_string(structure, "format");
	gst_structure_get_int(structure, "rate", &rate);
	gst_structure_get_int(structure, "channels", &channel);
	gst_structure_get(structure, "channel-mask", GST_TYPE_BITMASK, &channel_mask, NULL);
	gst_caps_unref(GST_CAPS(caps));

	/* In case of the sync is false, use buffer list.              *
	 * The num of buffer list depends on the num of audio channels */
	if (player->audio_stream_buff_list) {
		for (l = g_list_first(player->audio_stream_buff_list); l; l = g_list_next(l)) {
			mmplayer_audio_stream_buff_t *tmp = (mmplayer_audio_stream_buff_t *)l->data;
			if (tmp) {
				if (channel_mask == tmp->channel_mask) {
#ifdef __DEBUG__
					LOGD("[%"G_GUINT64_FORMAT"] total: %d, data: %d, buffer: %d", channel_mask, tmp->data_size, a_size, tmp->buff_size);
#endif
					if (tmp->data_size + a_size < tmp->buff_size) {
						memcpy(tmp->pcm_data + tmp->data_size, a_data, a_size);
						tmp->data_size += a_size;
					} else {
						/* send data to client */
						__mmplayer_audio_stream_send_data(player, tmp);

						if (a_size > tmp->buff_size) {
							LOGD("[%"G_GUINT64_FORMAT"] adj buffer size %d -> %d", channel_mask, tmp->buff_size, a_size);
							tmp->pcm_data = g_realloc(tmp->pcm_data, a_size);
							if (tmp->pcm_data == NULL) {
								LOGE("failed to realloc data.");
								goto DONE;
							}
							tmp->buff_size = a_size;
						}
						memset(tmp->pcm_data, 0x00, tmp->buff_size);
						memcpy(tmp->pcm_data, a_data, a_size);
						tmp->data_size = a_size;
					}
					goto DONE;
				}
			} else {
				LOGE("data is empty in list.");
				goto DONE;
			}
		}
	}

	/* create new audio stream data for newly found audio channel */
	a_buffer = (mmplayer_audio_stream_buff_t *)g_try_malloc0(sizeof(mmplayer_audio_stream_buff_t));
	if (a_buffer == NULL) {
		LOGE("failed to alloc data.");
		goto DONE;
	}
	a_buffer->bitrate = rate;
	a_buffer->channel = channel;
	a_buffer->channel_mask = channel_mask;
	a_buffer->data_size = a_size;
	a_buffer->pcm_format = _mmplayer_convert_audio_pcm_str_to_media_format_mime(pcm_format);

	if (player->audio_extract_opt & MM_PLAYER_AUDIO_EXTRACT_NO_SYNC_WITH_CLOCK) {
		/* If sync is FALSE, use buffer list to reduce the IPC. */
		a_buffer->buff_size = (a_size > player->ini.pcm_buffer_size) ? (a_size) : (player->ini.pcm_buffer_size);
		a_buffer->pcm_data = g_try_malloc(a_buffer->buff_size);
		if (a_buffer->pcm_data == NULL) {
			LOGE("failed to alloc data.");
			MMPLAYER_FREEIF(a_buffer);
			goto DONE;
		}
		memcpy(a_buffer->pcm_data, a_data, a_size);
#ifdef __DEBUG__
		LOGD("new [%"G_GUINT64_FORMAT"] total:%d buff:%d", channel_mask, a_buffer->data_size, a_buffer->buff_size);
#endif
		player->audio_stream_buff_list = g_list_append(player->audio_stream_buff_list, a_buffer);
	} else {
		/* If sync is TRUE, send data directly. */
		a_buffer->pcm_data = a_data;
		__mmplayer_audio_stream_send_data(player, a_buffer);
		MMPLAYER_FREEIF(a_buffer);
	}

DONE:
	gst_buffer_unmap(buffer, &mapinfo);
	MMPLAYER_FLEAVE();
}

static void
__mmplayer_gst_audio_deinterleave_pad_added(GstElement *elem, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	mmplayer_gst_element_t *audiobin = player->pipeline->audiobin;
	GstPad *sinkpad = NULL;
	GstElement *queue = NULL, *sink = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->mainbin);

	queue = gst_element_factory_make("queue", NULL);
	if (queue == NULL) {
		LOGD("fail make queue");
		goto ERROR;
	}

	sink = gst_element_factory_make("fakesink", NULL);
	if (sink == NULL) {
		LOGD("fail make fakesink");
		goto ERROR;
	}

	gst_bin_add_many(GST_BIN(audiobin[MMPLAYER_A_BIN].gst), queue, sink, NULL);

	if (!gst_element_link_pads_full(queue, "src", sink, "sink", GST_PAD_LINK_CHECK_NOTHING)) {
		LOGW("failed to link queue & sink");
		goto ERROR;
	}

	sinkpad = gst_element_get_static_pad(queue, "sink");

	if (GST_PAD_LINK_OK != gst_pad_link(pad, sinkpad)) {
		LOGW("failed to link [%s:%s] to queue", GST_DEBUG_PAD_NAME(pad));
		goto ERROR;
	}

	LOGE("audio_extract_opt : 0x%X", player->audio_extract_opt);

	gst_object_unref(sinkpad);
	if (!(player->audio_extract_opt & MM_PLAYER_AUDIO_EXTRACT_NO_SYNC_WITH_CLOCK))
		g_object_set(sink, "sync", TRUE, NULL);
	g_object_set(sink, "signal-handoffs", TRUE, NULL);

	/* keep the first sink reference only */
	if (!audiobin[MMPLAYER_A_SINK].gst) {
		audiobin[MMPLAYER_A_SINK].id = MMPLAYER_A_SINK;
		audiobin[MMPLAYER_A_SINK].gst = sink;
	}


	_mmplayer_add_signal_connection(player,
		G_OBJECT(sink),
		MM_PLAYER_SIGNAL_TYPE_AUDIOBIN,
		"handoff",
		G_CALLBACK(__mmplayer_audio_stream_decoded_render_cb),
		(gpointer)player);

	__mmplayer_add_sink(player, sink);

	if (gst_element_sync_state_with_parent(queue) == GST_STATE_CHANGE_FAILURE) {
		LOGE("failed to sync state");
		goto ERROR;
	}

	if (gst_element_sync_state_with_parent(sink) == GST_STATE_CHANGE_FAILURE) {
		LOGE("failed to sync state");
		goto ERROR;
	}

	MMPLAYER_FLEAVE();
	return;

ERROR:
	LOGE("__mmplayer_gst_audio_deinterleave_pad_added ERROR");
	if (queue) {
		gst_object_unref(GST_OBJECT(queue));
		queue = NULL;
	}
	if (sink) {
		gst_object_unref(GST_OBJECT(sink));
		sink = NULL;
	}
	if (sinkpad) {
		gst_object_unref(GST_OBJECT(sinkpad));
		sinkpad = NULL;
	}

	return;
}

void
__mmplayer_gst_audio_deinterleave_no_more_pads(GstElement* object, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->mainbin);

	player->no_more_pad = TRUE;
	__mmplayer_pipeline_complete(NULL, player);

	MMPLAYER_FLEAVE();
	return;
}

void
__mmplayer_gst_set_pulsesink_property(mmplayer_t *player)
{
	#define MAX_PROPS_LEN 128
	mmplayer_gst_element_t *audiobin = NULL;
	gint latency_mode = 0;
	gchar *stream_type = NULL;
	gchar *latency = NULL;
	gint stream_id = 0;
	gchar stream_props[MAX_PROPS_LEN] = {0,};
	GstStructure *props = NULL;

	/* set volume table
	 * It should be set after player creation through attribute.
	 * But, it can not be changed during playing.
	 */
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->audiobin);

	audiobin = player->pipeline->audiobin;

	g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "volume", player->sound.volume, NULL);
	if (player->sound.mute) {
		LOGD("mute enabled");
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "mute", player->sound.mute, NULL);
	}

	mm_attrs_get_int_by_name(player->attrs, "sound_stream_index", &stream_id);
	mm_attrs_get_string_by_name(player->attrs, "sound_stream_type", &stream_type);

	if (!stream_type)
		snprintf(stream_props, sizeof(stream_props) - 1,
				"props,application.process.id.origin=%d", player->client_pid);
	else
		snprintf(stream_props, sizeof(stream_props) - 1,
				"props,media.role=%s, media.parent_id=%d, application.process.id.origin=%d",
				stream_type, stream_id, player->client_pid);

	props = gst_structure_from_string(stream_props, NULL);
	g_object_set(audiobin[MMPLAYER_A_SINK].gst, "stream-properties", props, NULL);
	LOGI("props result[%s].", stream_props);
	gst_structure_free(props);

	mm_attrs_get_int_by_name(player->attrs, "sound_latency_mode", &latency_mode);

	switch (latency_mode) {
	case AUDIO_LATENCY_MODE_LOW:
		latency = g_strndup("low", 3);
		break;
	case AUDIO_LATENCY_MODE_MID:
		latency = g_strndup("mid", 3);
		break;
	case AUDIO_LATENCY_MODE_HIGH:
		latency = g_strndup("high", 4);
		break;
	};

	g_object_set(audiobin[MMPLAYER_A_SINK].gst, "latency", latency, NULL);

	LOGD("audiosink property - latency=%s", latency);

	MMPLAYER_FREEIF(latency);

	MMPLAYER_FLEAVE();
}

void
__mmplayer_gst_set_openalsink_property(mmplayer_t *player)
{
	mmplayer_gst_element_t *audiobin = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->audiobin);

	audiobin = player->pipeline->audiobin;

	g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "source-ambisonics-type", 1, NULL);
	sound_manager_create_stream_information(SOUND_STREAM_TYPE_MEDIA, NULL, NULL, &stream_info);
	g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "stream-info", stream_info, NULL);

	if (player->video360_yaw_radians <= M_PI &&
			player->video360_yaw_radians >= -M_PI &&
			player->video360_pitch_radians <= M_PI_2 &&
			player->video360_pitch_radians >= -M_PI_2) {
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst),
				"source-orientation-y", (int)(player->video360_yaw_radians * 180.0 / M_PI),
				"source-orientation-x", (int)(player->video360_pitch_radians * 180.0 / M_PI), NULL);
	} else if (player->video360_metadata.init_view_heading || player->video360_metadata.init_view_pitch) {
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst),
				"source-orientation-y", player->video360_metadata.init_view_heading,
				"source-orientation-x", player->video360_metadata.init_view_pitch, NULL);
	}

	MMPLAYER_FLEAVE();
}

static int
__mmplayer_gst_make_audio_playback_sink(mmplayer_t *player, GList **bucket)
{
	mmplayer_gst_element_t *audiobin = NULL;
	GstPad *sink_pad = NULL;
	GstCaps *acaps = NULL;
	gint channels = 0;
	int pitch_control = 0;
	double pitch_value = 1.0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline &&
				player->pipeline->audiobin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	audiobin = player->pipeline->audiobin;

	LOGD("make element for normal audio playback");

	/* audio bin structure for playback. {} means optional.
	   optional : pitch, audioeq, custom audioeq, openalsink for 360 audio content

	 * src - ... - {aconv - pitch} - aconv - rgvolume - resample - volume -
			{audioeq} - {custom audioeq} - pulsesink or {aconv - capsfilter - openalsink}
	 */

	/* for pitch control */
	mm_attrs_multiple_get(player->attrs, NULL,
				MM_PLAYER_PITCH_CONTROL, &pitch_control,
				MM_PLAYER_PITCH_VALUE, &pitch_value,
				NULL);

	LOGD("pitch %d / %1.3f", pitch_control, pitch_value);
	if (pitch_control && (player->videodec_linked == 0)) {
		GstElementFactory *factory;

		factory = gst_element_factory_find("pitch");
		if (factory) {
			gst_object_unref(factory);

			/* converter */
			MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_CONV_PITCH, "audioconvert", "audio convert pitch", *bucket, player);

			/* pitch */
			MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_PITCH, "pitch", "audio pitch", *bucket, player);
			g_object_set(G_OBJECT(audiobin[MMPLAYER_A_PITCH].gst), "pitch", (gdouble)pitch_value, NULL);
		} else {
			LOGW("there is no pitch element");
		}
	}

	/* converter */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_CONV, "audioconvert", "audio converter", *bucket, player);

	/* replaygain volume */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_RGVOL, "rgvolume", "audio rgvolume", *bucket, player);
	if (player->sound.rg_enable)
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_RGVOL].gst), "enable-rgvolume", TRUE, NULL);
	else
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_RGVOL].gst), "enable-rgvolume", FALSE, NULL);

	/* resampler */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_RESAMPLER,  player->ini.audioresampler_element, "audio resampler", *bucket, player);

	if (g_strrstr(player->ini.audiosink_element, "openalsink")) {
		/* currently, only openalsink uses volume element */
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_VOL, "volume", "volume", *bucket, player);
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_VOL].gst), "volume", player->sound.volume, NULL);

		if (player->sound.mute) {
			LOGD("mute enabled");
			g_object_set(G_OBJECT(audiobin[MMPLAYER_A_VOL].gst), "mute", player->sound.mute, NULL);
		}
	}

	mm_attrs_get_int_by_name(player->attrs, "content_audio_channels", &channels);

	/* audio effect element. if audio effect is enabled */
	if ((strcmp(player->ini.audioeffect_element, ""))
		&& (channels <= 2)
		&& (player->ini.use_audio_effect_preset || player->ini.use_audio_effect_custom)) {
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_FILTER, player->ini.audioeffect_element, "audio effect filter", *bucket, player);

		LOGD("audio effect config. bypass = %d, effect type = %d", player->bypass_audio_effect, player->audio_effect_info.effect_type);

		if ((!player->bypass_audio_effect)
			&& (player->ini.use_audio_effect_preset || player->ini.use_audio_effect_custom)) {
			if (player->audio_effect_info.effect_type == MM_AUDIO_EFFECT_TYPE_CUSTOM) {
				if (!_mmplayer_audio_effect_custom_apply(player))
					LOGI("apply audio effect(custom) setting success");
			}
		}

		if ((strcmp(player->ini.audioeffect_element_custom, ""))
			&& (player->set_mode.rich_audio)) {
			MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_FILTER_SEC, player->ini.audioeffect_element_custom, "audio effect filter custom", *bucket, player);
		}
	}

	/* create audio sink */
	LOGD("spherical %d, channels %d, ambisonic type %d, format %d, order %d",
			player->is_content_spherical, channels, player->video360_metadata.ambisonic_type,
			player->video360_metadata.ambisonic_format, player->video360_metadata.ambisonic_order);

	/* Note: qtdemux converts audio metadata defaults to openalsink defaults. */
	if (player->is_360_feature_enabled &&
		player->is_content_spherical &&
		channels == 4 &&
		player->video360_metadata.ambisonic_type == MMFILE_AMBISONIC_TYPE_PERIPHONIC &&
		player->video360_metadata.ambisonic_format == MMFILE_AMBISONIC_FORMAT_AMB &&
		player->video360_metadata.ambisonic_order == MMFILE_AMBISONIC_ORDER_FOA) {

		strncpy(player->ini.audiosink_element, "openalsink", PLAYER_INI_MAX_STRLEN - 1);

		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_CONV_BFORMAT, "audioconvert", "audio-converter-bformat", *bucket, player);

		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_CAPS_360, "capsfilter", "audio-caps-filter", *bucket, player);
		acaps = gst_caps_from_string(SPATIAL_AUDIO_CAPS);
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_CAPS_360].gst), "caps", acaps, NULL);
		gst_caps_unref(acaps);

		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_SINK, "openalsink", "audiosink", *bucket, player);

		player->is_openal_plugin_used = TRUE;
	} else {
		if (player->is_360_feature_enabled && player->is_content_spherical)
			LOGW("Audio track isn't of the ambisonic type and can't be played back as a spatial sound.");
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_SINK, player->ini.audiosink_element, "audiosink", *bucket, player);
	}

	if ((MMPLAYER_IS_RTSP_STREAMING(player)) ||
		(player->videodec_linked && player->ini.use_system_clock)) {
		LOGD("system clock will be used.");
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "provide-clock", FALSE,  NULL);
	}

	if (g_strrstr(player->ini.audiosink_element, "pulsesink"))
		__mmplayer_gst_set_pulsesink_property(player);
	else if (g_strrstr(player->ini.audiosink_element, "openalsink"))
		__mmplayer_gst_set_openalsink_property(player);

	/* qos on */
	g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "qos", TRUE, NULL);	/* qos on */
	g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "slave-method", GST_AUDIO_BASE_SINK_SLAVE_NONE, NULL);

	sink_pad = gst_element_get_static_pad(audiobin[MMPLAYER_A_SINK].gst, "sink");
	_mmplayer_add_signal_connection(player, G_OBJECT(sink_pad), MM_PLAYER_SIGNAL_TYPE_AUDIOBIN,
				"notify::caps", G_CALLBACK(__mmplayer_gst_caps_notify_cb), (gpointer)player);
	gst_object_unref(GST_OBJECT(sink_pad));

	__mmplayer_add_sink(player, audiobin[MMPLAYER_A_SINK].gst);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;

ERROR: /* MMPLAYER_CREATE_ELEMENT */
	MMPLAYER_FLEAVE();
	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_make_audio_extract_sink(mmplayer_t *player, GList **bucket)
{
	mmplayer_gst_element_t *audiobin = NULL;
	enum audio_element_id extract_sink_id = MMPLAYER_A_SINK;

	gchar *dst_format = NULL;
	int dst_len = 0;
	int dst_samplerate = 0;
	int dst_channels = 0;
	GstCaps *caps = NULL;
	char *caps_str = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline &&
				player->pipeline->audiobin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	audiobin = player->pipeline->audiobin;

	LOGD("make element for audio extract, option = 0x%X", player->audio_extract_opt);

	/* audio bin structure according to the mmplayer_audio_extract_opt_e.

	   [case 1] extract interleave audio pcm without playback
				: MM_PLAYER_AUDIO_EXTRACT_DEFAULT (sync)
				  MM_PLAYER_AUDIO_EXTRACT_NO_SYNC_WITH_CLOCK (non sync)

				* src - ... - aconv - resample - capsfilter - fakesink (sync or not)

	   [case 2] deinterleave for each channel without playback
				: MM_PLAYER_AUDIO_EXTRACT_DEINTERLEAVE (sync)
				  MM_PLAYER_AUDIO_EXTRACT_NO_SYNC_AND_DEINTERLEAVE (non sync)

				* src - ... - aconv - resample - capsfilter - deinterleave - fakesink (sync or not)
																		   - fakesink (sync or not)
																		   - ...      (sync or not)

	   [case 3] [case 1(sync only)] + playback
				: MM_PLAYER_AUDIO_EXTRACT_WITH_PLAYBACK

				* src - ... - tee - queue1 - playback path
								  - queue2 - [case1 pipeline with sync]

	   [case 4] [case 2(sync only)] + playback
				: MM_PLAYER_AUDIO_EXTRACT_DEINTERLEAVE_WITH_PLAYBACK

				* src - ... - tee - queue1 - playback path
								  - queue2 - [case2 pipeline with sync]

	 */

	/* 1. create tee and playback path
	      'tee' should be added at first to copy the decoded stream
	 */
	if (player->audio_extract_opt & MM_PLAYER_AUDIO_EXTRACT_WITH_PLAYBACK) {
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_TEE, "tee", "audio-tee", *bucket, player);
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_TEE].gst), "num-src-pads", 2, NULL);

		/* tee - path 1 : for playback path */
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_TEE_Q1, "queue", "audio-tee-queue1", *bucket, player);
		__mmplayer_gst_make_audio_playback_sink(player, bucket);

		/* tee - path 2 : for extract path */
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_TEE_Q2, "queue", "audio-tee-queue2", *bucket, player);
		extract_sink_id = MMPLAYER_A_EXTRACT_SINK; /* there is another playback sink */
	}

	/* if there is tee, 'tee - path 2' is linked here */
	/* converter */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_EXTRACT_CONV, "audioconvert", "audio-ext-conv", *bucket, player);

	/* resampler */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_EXTRACT_RESAMPLER,  player->ini.audioresampler_element, "audio-ext-resampler", *bucket, player);

	/* 2. decide the extract pcm format */
	mm_attrs_multiple_get(player->attrs, NULL,
				MM_PLAYER_PCM_EXT_FORMAT, &dst_format, &dst_len,
				MM_PLAYER_PCM_EXT_SAMPLERATE, &dst_samplerate,
				MM_PLAYER_PCM_EXT_CHANNELS, &dst_channels,
				NULL);

	LOGD("required extract pcm format - format: %s(%d), samplerate : %d, channel: %d",
			dst_format, dst_len, dst_samplerate, dst_channels);

	if (dst_format == NULL || dst_len == 0 || dst_samplerate == 0 || dst_channels == 0) {
		mm_attrs_multiple_get(player->attrs, NULL,
					"content_audio_format", &dst_format, &dst_len, /* get string and len */
					"content_audio_samplerate", &dst_samplerate,
					"content_audio_channels", &dst_channels,
					NULL);

		LOGD("apply the decoded pcm format - format: %s(%d), samplerate : %d, channel: %d",
				dst_format, dst_len, dst_samplerate, dst_channels);

		/* If there is no enough information, set it to platform default value. */
		if (dst_format == NULL || _mmplayer_convert_audio_pcm_str_to_media_format_mime(dst_format) == MEDIA_FORMAT_MAX) {
			LOGD("set platform default format");
			dst_format = DEFAULT_PCM_OUT_FORMAT;
		}
		if (dst_samplerate <= 0) dst_samplerate = DEFAULT_PCM_OUT_SAMPLERATE;
		if (dst_channels <= 0)	 dst_channels = DEFAULT_PCM_OUT_CHANNEL;
	}

	/* 3. create capsfilter */
	MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_EXTRACT_CAPS, "capsfilter", "audio-ext-caps", *bucket, player);
	caps = gst_caps_new_simple("audio/x-raw",
			"format", G_TYPE_STRING, dst_format,
			"rate", G_TYPE_INT, dst_samplerate,
			"channels", G_TYPE_INT, dst_channels,
			NULL);

	caps_str = gst_caps_to_string(caps);
	LOGD("new caps : %s", caps_str);

	g_object_set(GST_ELEMENT(audiobin[MMPLAYER_A_EXTRACT_CAPS].gst), "caps", caps, NULL);

	/* clean */
	gst_caps_unref(caps);
	MMPLAYER_FREEIF(caps_str);

	/* 4-1. create deinterleave to extract pcm for each channel */
	if (player->audio_extract_opt & MM_PLAYER_AUDIO_EXTRACT_DEINTERLEAVE) {
		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_EXTRACT_DEINTERLEAVE, "deinterleave", "deinterleave", *bucket, player);
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_EXTRACT_DEINTERLEAVE].gst), "keep-positions", TRUE, NULL);

		/* audiosink will be added after getting signal for each channel */
		_mmplayer_add_signal_connection(player, G_OBJECT(audiobin[MMPLAYER_A_EXTRACT_DEINTERLEAVE].gst),
				MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "pad-added", G_CALLBACK(__mmplayer_gst_audio_deinterleave_pad_added), (gpointer)player);
		_mmplayer_add_signal_connection(player, G_OBJECT(audiobin[MMPLAYER_A_EXTRACT_DEINTERLEAVE].gst),
				MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "no-more-pads", G_CALLBACK(__mmplayer_gst_audio_deinterleave_no_more_pads), (gpointer)player);
		player->no_more_pad = FALSE;
	} else {
	/* 4-2. create fakesink to extract interlevaed pcm */
		LOGD("add audio fakesink for interleaved audio");
		MMPLAYER_CREATE_ELEMENT(audiobin, extract_sink_id, "fakesink", "fakeaudiosink", *bucket, player);
		if (!(player->audio_extract_opt & MM_PLAYER_AUDIO_EXTRACT_NO_SYNC_WITH_CLOCK))
			g_object_set(G_OBJECT(audiobin[extract_sink_id].gst), "sync", TRUE, NULL);
		g_object_set(G_OBJECT(audiobin[extract_sink_id].gst), "signal-handoffs", TRUE, NULL);

		_mmplayer_add_signal_connection(player,
			G_OBJECT(audiobin[extract_sink_id].gst),
			MM_PLAYER_SIGNAL_TYPE_AUDIOBIN,
			"handoff",
			G_CALLBACK(__mmplayer_audio_stream_decoded_render_cb),
			(gpointer)player);

		__mmplayer_add_sink(player, audiobin[extract_sink_id].gst);
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;

ERROR: /* MMPLAYER_CREATE_ELEMENT */
	MMPLAYER_FLEAVE();
	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_make_audio_bin_element(mmplayer_t *player, GList **bucket)
{
	int ret = MM_ERROR_NONE;
	mmplayer_gst_element_t *audiobin = NULL;
	GList *element_bucket = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline &&
				player->pipeline->audiobin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	audiobin = player->pipeline->audiobin;

	if (player->build_audio_offload) { /* skip all the audio filters */
		LOGD("create audio offload sink : %s", player->ini.audio_offload_sink_element);

		MMPLAYER_CREATE_ELEMENT(audiobin, MMPLAYER_A_SINK, player->ini.audio_offload_sink_element, "audiosink", element_bucket, player);
		g_object_set(G_OBJECT(audiobin[MMPLAYER_A_SINK].gst), "sync", TRUE,
				"volume", player->sound.volume, "mute", player->sound.mute, NULL);

		__mmplayer_add_sink(player, audiobin[MMPLAYER_A_SINK].gst);
		goto DONE;
	}

	/* FIXME: need to mention the supportable condition at API reference */
	if (player->audio_decoded_cb && (!MMPLAYER_IS_RTSP_STREAMING(player)))
		ret = __mmplayer_gst_make_audio_extract_sink(player, &element_bucket);
	else
		ret = __mmplayer_gst_make_audio_playback_sink(player, &element_bucket);

	if (ret != MM_ERROR_NONE)
		goto ERROR;
DONE:
	LOGD("success to make audio bin element");
	*bucket = element_bucket;

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;

ERROR:
	LOGE("failed to make audio bin element");
	g_list_free(element_bucket);

	*bucket = NULL;
	MMPLAYER_FLEAVE();
	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_create_audio_sink_bin(mmplayer_t *player)
{
	mmplayer_gst_element_t *first_element = NULL;
	mmplayer_gst_element_t *audiobin = NULL;
	GstPad *pad = NULL;
	GstPad *ghostpad = NULL;
	GList *element_bucket = NULL;
	int i = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* alloc handles */
	audiobin = (mmplayer_gst_element_t *)g_try_malloc0(sizeof(mmplayer_gst_element_t) * MMPLAYER_A_NUM);
	if (!audiobin) {
		LOGE("failed to allocate memory for audiobin");
		return MM_ERROR_PLAYER_NO_FREE_SPACE;
	}

	/* create bin */
	audiobin[MMPLAYER_A_BIN].id = MMPLAYER_A_BIN;
	audiobin[MMPLAYER_A_BIN].gst = gst_bin_new("audiobin");
	if (!audiobin[MMPLAYER_A_BIN].gst) {
		LOGE("failed to create audiobin");
		goto ERROR;
	}

	/* take it */
	player->pipeline->audiobin = audiobin;

	/* create audio filters and audiosink */
	if (__mmplayer_gst_make_audio_bin_element(player, &element_bucket) != MM_ERROR_NONE)
		goto ERROR;

	/* adding created elements to bin */
	LOGD("adding created elements to bin");
	if (!_mmplayer_gst_element_add_bucket_to_bin(GST_BIN(audiobin[MMPLAYER_A_BIN].gst), element_bucket))
		goto ERROR;

	/* linking elements in the bucket by added order. */
	LOGD("Linking elements in the bucket by added order.");
	if (_mmplayer_gst_element_link_bucket(element_bucket) == -1)
		goto ERROR;

	/* get first element's sinkpad for creating ghostpad */
	first_element = (mmplayer_gst_element_t *)element_bucket->data;
	if (!first_element) {
		LOGE("failed to get first elem");
		goto ERROR;
	}

	pad = gst_element_get_static_pad(GST_ELEMENT(first_element->gst), "sink");
	if (!pad) {
		LOGE("failed to get pad from first element of audiobin");
		goto ERROR;
	}

	ghostpad = gst_ghost_pad_new("sink", pad);
	if (!ghostpad) {
		LOGE("failed to create ghostpad");
		goto ERROR;
	}

	if (!gst_element_add_pad(audiobin[MMPLAYER_A_BIN].gst, ghostpad)) {
		LOGE("failed to add ghostpad to audiobin");
		goto ERROR;
	}

	gst_object_unref(pad);

	g_list_free(element_bucket);
	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;

ERROR:
	LOGD("ERROR : releasing audiobin");

	if (pad)
		gst_object_unref(GST_OBJECT(pad));

	if (ghostpad)
		gst_object_unref(GST_OBJECT(ghostpad));

	if (element_bucket)
		g_list_free(element_bucket);

	/* release element which are not added to bin */
	for (i = 1; i < MMPLAYER_A_NUM; i++) {
		/* NOTE : skip bin */
		if (audiobin[i].gst) {
			GstObject *parent = NULL;
			parent = gst_element_get_parent(audiobin[i].gst);

			if (!parent) {
				gst_object_unref(GST_OBJECT(audiobin[i].gst));
				audiobin[i].gst = NULL;
			} else
				gst_object_unref(GST_OBJECT(parent));
		}
	}

	/* release audiobin with it's childs */
	if (audiobin[MMPLAYER_A_BIN].gst)
		gst_object_unref(GST_OBJECT(audiobin[MMPLAYER_A_BIN].gst));

	MMPLAYER_FREEIF(audiobin);

	player->pipeline->audiobin = NULL;

	return MM_ERROR_PLAYER_INTERNAL;
}

static guint32
_mmplayer_convert_fourcc_string_to_value(const gchar *format_name)
{
	return format_name[0] | (format_name[1] << 8) | (format_name[2] << 16) | (format_name[3] << 24);
}

int
_mmplayer_video_stream_release_bo(mmplayer_t *player, void *bo)
{
	int ret = MM_ERROR_NONE;
	GList *l = NULL;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(bo, MM_ERROR_INVALID_ARGUMENT);

	MMPLAYER_VIDEO_BO_LOCK(player);

	if (player->video_bo_list) {
		for (l = g_list_first(player->video_bo_list); l; l = g_list_next(l)) {
			mmplayer_video_bo_info_t *tmp = (mmplayer_video_bo_info_t *)l->data;
			if (tmp && tmp->bo == bo) {
				tmp->used = FALSE;
				LOGD("release bo %p", bo);
				tbm_bo_unref(tmp->bo);
				MMPLAYER_VIDEO_BO_UNLOCK(player);
				MMPLAYER_VIDEO_BO_SIGNAL(player);
				return ret;
			}
		}
	} else {
		/* hw codec is running or the list was reset for DRC. */
		LOGW("there is no bo list.");
	}
	MMPLAYER_VIDEO_BO_UNLOCK(player);

	LOGW("failed to find bo %p", bo);
	return ret;
}

static void
__mmplayer_video_stream_destroy_bo_list(mmplayer_t *player)
{
	GList *l = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	MMPLAYER_VIDEO_BO_LOCK(player);
	if (player->video_bo_list) {
		LOGD("destroy video_bo_list : %d", g_list_length(player->video_bo_list));
		for (l = g_list_first(player->video_bo_list); l; l = g_list_next(l)) {
			mmplayer_video_bo_info_t *tmp = (mmplayer_video_bo_info_t *)l->data;
			if (tmp) {
				if (tmp->bo)
					tbm_bo_unref(tmp->bo);
				g_free(tmp);
			}
		}
		g_list_free(player->video_bo_list);
		player->video_bo_list = NULL;
	}
	player->video_bo_size = 0;
	MMPLAYER_VIDEO_BO_UNLOCK(player);

	MMPLAYER_FLEAVE();
	return;
}

static void *
__mmplayer_video_stream_get_bo(mmplayer_t *player, int size)
{
	GList *l = NULL;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, NULL);
	gboolean ret = TRUE;

	/* check DRC, if it is, destroy the prev bo list to create again */
	if (player->video_bo_size != size) {
		LOGD("video size is changed: %d -> %d", player->video_bo_size, size);
		__mmplayer_video_stream_destroy_bo_list(player);
		player->video_bo_size = size;
	}

	MMPLAYER_VIDEO_BO_LOCK(player);

	if ((!player->video_bo_list) ||
		(g_list_length(player->video_bo_list) < player->ini.num_of_video_bo)) {

		/* create bo list */
		int idx = 0;
		LOGD("Create bo list for decoded video stream(num:%d)", player->ini.num_of_video_bo);

		if (player->video_bo_list) {
			/* if bo list did not created all, try it again. */
			idx = g_list_length(player->video_bo_list);
			LOGD("bo list exist(len: %d)", idx);
		}

		for (; idx < player->ini.num_of_video_bo; idx++) {
			mmplayer_video_bo_info_t *bo_info = g_new(mmplayer_video_bo_info_t, 1);
			if (!bo_info) {
				LOGE("Fail to alloc bo_info.");
				break;
			}
			bo_info->bo = tbm_bo_alloc(player->bufmgr, size, TBM_BO_DEFAULT);
			if (!bo_info->bo) {
				LOGE("Fail to tbm_bo_alloc.");
				MMPLAYER_FREEIF(bo_info);
				break;
			}
			bo_info->used = FALSE;
			player->video_bo_list = g_list_append(player->video_bo_list, bo_info);
		}

		/* update video num buffers */
		LOGD("video_num_buffers : %d", idx);
		mm_player_set_attribute((MMHandleType)player, NULL,
				MM_PLAYER_VIDEO_BUFFER_TOTAL_SIZE, idx,
				MM_PLAYER_VIDEO_BUFFER_EXTRA_SIZE, MAX(DEFAULT_NUM_OF_V_OUT_BUFFER, (idx / 2)),
				NULL);

		if (idx == 0) {
			MMPLAYER_VIDEO_BO_UNLOCK(player);
			return NULL;
		}
	}

	while (TRUE) {
		/* get bo from list*/
		for (l = g_list_first(player->video_bo_list); l; l = g_list_next(l)) {
			mmplayer_video_bo_info_t *tmp = (mmplayer_video_bo_info_t *)l->data;
			if (tmp && (tmp->used == FALSE)) {
				LOGD("found bo %p to use", tmp->bo);
				tmp->used = TRUE;
				MMPLAYER_VIDEO_BO_UNLOCK(player);
				return tbm_bo_ref(tmp->bo);
			}
		}
		if (!ret) {
			LOGE("failed to get bo in %d timeout", player->ini.video_bo_timeout);
			MMPLAYER_VIDEO_BO_UNLOCK(player);
			return NULL;
		}

		if (player->ini.video_bo_timeout <= 0) {
			MMPLAYER_VIDEO_BO_WAIT(player);
		} else {
			gint64 timeout = g_get_monotonic_time() + player->ini.video_bo_timeout * G_TIME_SPAN_SECOND;
			ret = MMPLAYER_VIDEO_BO_WAIT_UNTIL(player, timeout);
		}
		continue;
	}
}

static void
__mmplayer_video_stream_decoded_preroll_cb(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->video_decoded_cb);

	/* send prerolled pkt */
	player->video_stream_prerolled = false;

	__mmplayer_video_stream_decoded_render_cb(object, buffer, pad, data);

	/* not to send prerolled pkt again */
	player->video_stream_prerolled = true;
}

static void
__mmplayer_video_stream_decoded_render_cb(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	mmplayer_video_decoded_data_info_t *stream = NULL;
	GstMemory *mem = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);
	MMPLAYER_RETURN_IF_FAIL(player->video_decoded_cb);

	if (player->video_stream_prerolled) {
		player->video_stream_prerolled = false;
		LOGD("skip the prerolled pkt not to send it again");
		return;
	}

	/* clear stream data structure */
	stream = __mmplayer_create_stream_from_pad(pad);
	if (!stream) {
		LOGE("failed to alloc stream");
		return;
	}

	_mmplayer_get_video_angle(player, NULL, &stream->orientation);

	/* set size and timestamp */
	mem = gst_buffer_peek_memory(buffer, 0);
	stream->length_total = gst_memory_get_sizes(mem, NULL, NULL);
	stream->timestamp = (unsigned int)(GST_TIME_AS_MSECONDS(GST_BUFFER_PTS(buffer))); /* nano sec -> mili sec */

	/* check zero-copy */
	if (player->set_mode.video_zc &&
		player->set_mode.video_export &&
		gst_is_tizen_memory(mem)) {
		__mmplayer_zerocopy_set_stride_elevation_bo(stream, mem);
		stream->internal_buffer = gst_buffer_ref(buffer);
	} else { /* sw codec */
		if (!__mmplayer_swcodec_set_stride_elevation(stream))
			goto ERROR;

		if (!__mmplayer_swcodec_set_bo(player, stream, mem))
			goto ERROR;
	}

	if (!player->video_decoded_cb(stream, player->video_decoded_cb_user_param)) {
		LOGE("failed to send video decoded data.");
		goto ERROR;
	}

	return;

ERROR:
	LOGE("release video stream resource.");
	if (gst_is_tizen_memory(mem)) {
		int i = 0;
		for (i = 0 ; i < MM_VIDEO_BUFFER_PLANE_MAX ; i++) {
			if (stream->bo[i])
				tbm_bo_unref(stream->bo[i]);
		}

		/* unref gst buffer */
		if (stream->internal_buffer)
			gst_buffer_unref(stream->internal_buffer);
	} else {
		if (stream->bo[0])
			_mmplayer_video_stream_release_bo(player, stream->bo[0]);
	}
	MMPLAYER_FREEIF(stream);
	return;
}

static void
__mmplayer_gst_set_video360_property(mmplayer_t *player)
{
	mmplayer_gst_element_t *videobin = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->pipeline && player->pipeline->videobin);

	videobin = player->pipeline->videobin;

	/* Set spatial media metadata and/or user settings to the element.
	 * */
	g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
			"projection-type", player->video360_metadata.projection_type, NULL);

	g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
			"stereo-mode", player->video360_metadata.stereo_mode, NULL);

	if (player->video360_metadata.full_pano_width_pixels &&
			player->video360_metadata.full_pano_height_pixels &&
			player->video360_metadata.cropped_area_image_width &&
			player->video360_metadata.cropped_area_image_height) {
		g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
				"projection-bounds-top", player->video360_metadata.cropped_area_top,
				"projection-bounds-bottom", player->video360_metadata.full_pano_height_pixels -
						player->video360_metadata.cropped_area_top - player->video360_metadata.cropped_area_image_height,
				"projection-bounds-left", player->video360_metadata.cropped_area_left,
				"projection-bounds-right", player->video360_metadata.full_pano_width_pixels -
						player->video360_metadata.cropped_area_left - player->video360_metadata.cropped_area_image_width,
				NULL);
	}

	if (player->video360_horizontal_fov && player->video360_vertical_fov) {
		g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
				"horizontal-fov", player->video360_horizontal_fov,
				"vertical-fov", player->video360_vertical_fov, NULL);
	}

	if (player->video360_zoom <= VIDEO360_MAX_ZOOM && player->video360_zoom > 1.0f) {
		g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
				"zoom", 1.0f / player->video360_zoom, NULL);
	}

	if (player->video360_yaw_radians <= M_PI &&
			player->video360_yaw_radians >= -M_PI &&
			player->video360_pitch_radians <= M_PI_2 &&
			player->video360_pitch_radians >= -M_PI_2) {
		g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
				"pose-yaw", (int)(player->video360_yaw_radians * 180.0 / M_PI),
				"pose-pitch", (int)(player->video360_pitch_radians * 180.0 / M_PI), NULL);
	} else if (player->video360_metadata.init_view_heading || player->video360_metadata.init_view_pitch) {
		g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
				"pose-yaw", player->video360_metadata.init_view_heading,
				"pose-pitch", player->video360_metadata.init_view_pitch, NULL);
	}

	g_object_set(G_OBJECT(videobin[MMPLAYER_V_360].gst),
			"passthrough", !player->is_video360_enabled, NULL);


	MMPLAYER_FLEAVE();
}

static int
__mmplayer_gst_create_video_filters(mmplayer_t *player, MMDisplaySurfaceType surface_type, GList **bucket)
{
	gchar *video_csc = "videoconvert"; /* default colorspace converter */
	GList *element_bucket = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->videobin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* create video360 filter */
	if (player->is_360_feature_enabled && player->is_content_spherical) {
		LOGD("create video360 element");
		MMPLAYER_CREATE_ELEMENT(player->pipeline->videobin, MMPLAYER_V_360, "video360", "video-360", element_bucket, player);
		__mmplayer_gst_set_video360_property(player);
		goto EXIT;
	}

	if (surface_type != MM_DISPLAY_SURFACE_OVERLAY || player->set_mode.video_zc) {
		LOGD("skip creating the videoconv and rotator");
		return MM_ERROR_NONE;
	}

	/* in case of sw codec & overlay surface type, except 360 playback.
	 * if libav video decoder is selected, videoconvert is required to render the shm wl-buffer which support RGB only via tizenwlsink. */
	LOGD("create video converter: %s", video_csc);
	MMPLAYER_CREATE_ELEMENT(player->pipeline->videobin, MMPLAYER_V_CONV, video_csc, "video converter", element_bucket, player);

EXIT:
	*bucket = element_bucket;
	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;

ERROR: /* refer MMPLAYER_CREATE_ELEMENT */
	g_list_free(element_bucket);

	*bucket = NULL;
	MMPLAYER_FLEAVE();
	return MM_ERROR_PLAYER_INTERNAL;
}

static gchar *
__mmplayer_get_videosink_factory_name(mmplayer_t *player, MMDisplaySurfaceType surface_type)
{
	gchar *factory_name = NULL;

	switch (surface_type) {
	case MM_DISPLAY_SURFACE_OVERLAY:
		if (strlen(player->ini.videosink_element_overlay) > 0)
			factory_name = player->ini.videosink_element_overlay;
		break;
	case MM_DISPLAY_SURFACE_REMOTE:
	case MM_DISPLAY_SURFACE_NULL:
		if (strlen(player->ini.videosink_element_fake) > 0)
			factory_name = player->ini.videosink_element_fake;
		break;
	default:
		LOGE("unidentified surface type");
		break;
	}

	LOGD("surface_type %d, videosink is %s", surface_type, factory_name);
	return factory_name;
}

static int
__mmplayer_gst_set_videosink_property(mmplayer_t *player, MMDisplaySurfaceType surface_type)
{
	gchar *factory_name = NULL;
	mmplayer_gst_element_t *videobin = NULL;
	MMHandleType attrs;
	int gapless = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->videobin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	videobin = player->pipeline->videobin;
	factory_name = GST_OBJECT_NAME(gst_element_get_factory(videobin[MMPLAYER_V_SINK].gst));

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("cannot get content attribute");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	if (surface_type == MM_DISPLAY_SURFACE_OVERLAY) {
		bool use_tbm = (player->set_mode.video_zc || (player->is_360_feature_enabled && player->is_content_spherical));
		if (strncmp(factory_name, "tizenwlsink", strlen(factory_name)) == 0) {
			g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst,
				"use-tbm", use_tbm, NULL);
		}

		if (_mmplayer_update_video_overlay_param(player, "update_all_param") != MM_ERROR_NONE)
			return MM_ERROR_PLAYER_INTERNAL;

		LOGI("videosink factory name is %s use-tbm : %d", factory_name, use_tbm);

	} else {
		g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
										"sync", TRUE, "max-lateness", FAKE_SINK_MAX_LATENESS, NULL);
	}

	mm_attrs_get_int_by_name(attrs, MM_PLAYER_GAPLESS_MODE, &gapless);
	if (gapless > 0) {
		LOGD("disable last-sample");
		g_object_set(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "enable-last-sample", FALSE, NULL);
	}

	if (player->set_mode.video_export) {
		int enable = 0;
		mm_attrs_get_int_by_name(player->attrs, "enable_video_decoded_cb", &enable);
		if (enable || (surface_type == MM_DISPLAY_SURFACE_REMOTE) || (surface_type == MM_DISPLAY_SURFACE_NULL))
			g_object_set(G_OBJECT(player->pipeline->videobin[MMPLAYER_V_SINK].gst), "signal-handoffs", TRUE, NULL);

		_mmplayer_add_signal_connection(player,
								G_OBJECT(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
								MM_PLAYER_SIGNAL_TYPE_VIDEOBIN,
								"handoff",
								G_CALLBACK(__mmplayer_video_stream_decoded_render_cb),
								(gpointer)player);

		_mmplayer_add_signal_connection(player,
								G_OBJECT(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
								MM_PLAYER_SIGNAL_TYPE_VIDEOBIN,
								"preroll-handoff",
								G_CALLBACK(__mmplayer_video_stream_decoded_preroll_cb),
								(gpointer)player);
	}

	if (videobin[MMPLAYER_V_SINK].gst) {
		GstPad *sink_pad = NULL;
		sink_pad = gst_element_get_static_pad(videobin[MMPLAYER_V_SINK].gst, "sink");
		if (sink_pad) {
			_mmplayer_add_signal_connection(player, G_OBJECT(sink_pad), MM_PLAYER_SIGNAL_TYPE_VIDEOBIN,
					"notify::caps", G_CALLBACK(__mmplayer_gst_caps_notify_cb), (gpointer)player);
			gst_object_unref(GST_OBJECT(sink_pad));
		} else {
			LOGE("failed to get sink pad from videosink");
		}
	}

	return MM_ERROR_NONE;
}

/**
 * VIDEO BIN
 * - video overlay surface(arm/x86) : tizenwlsink
 */
static int
__mmplayer_gst_create_video_sink_bin(mmplayer_t *player, GstCaps *caps, MMDisplaySurfaceType surface_type)
{
	GstPad *pad = NULL;
	GList *element_bucket = NULL;
	mmplayer_gst_element_t *first_element = NULL;
	mmplayer_gst_element_t *videobin = NULL;
	gchar *videosink_factory_name = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* alloc handles */
	videobin = (mmplayer_gst_element_t *)g_try_malloc0(sizeof(mmplayer_gst_element_t) * MMPLAYER_V_NUM);
	if (!videobin)
		return MM_ERROR_PLAYER_NO_FREE_SPACE;

	player->pipeline->videobin = videobin;

	/* create bin */
	videobin[MMPLAYER_V_BIN].id = MMPLAYER_V_BIN;
	videobin[MMPLAYER_V_BIN].gst = gst_bin_new("videobin");
	if (!videobin[MMPLAYER_V_BIN].gst) {
		LOGE("failed to create videobin");
		goto ERROR;
	}

	if (__mmplayer_gst_create_video_filters(player, surface_type, &element_bucket) != MM_ERROR_NONE)
		goto ERROR;

	videosink_factory_name = __mmplayer_get_videosink_factory_name(player, surface_type);
	MMPLAYER_CREATE_ELEMENT(videobin, MMPLAYER_V_SINK, videosink_factory_name, "videosink", element_bucket, player);

	/* additional setting for sink plug-in */
	if (__mmplayer_gst_set_videosink_property(player, surface_type) != MM_ERROR_NONE) {
		LOGE("failed to set video property");
		goto ERROR;
	}

	/* store it as it's sink element */
	__mmplayer_add_sink(player, videobin[MMPLAYER_V_SINK].gst);

	/* adding created elements to bin */
	if (!_mmplayer_gst_element_add_bucket_to_bin(GST_BIN(videobin[MMPLAYER_V_BIN].gst), element_bucket)) {
		LOGE("failed to add elements");
		goto ERROR;
	}

	/* Linking elements in the bucket by added order */
	if (_mmplayer_gst_element_link_bucket(element_bucket) == -1) {
		LOGE("failed to link elements");
		goto ERROR;
	}

	/* get first element's sinkpad for creating ghostpad */
	first_element = (mmplayer_gst_element_t *)element_bucket->data;
	if (!first_element) {
		LOGE("failed to get first element from bucket");
		goto ERROR;
	}

	pad = gst_element_get_static_pad(GST_ELEMENT(first_element->gst), "sink");
	if (!pad) {
		LOGE("failed to get pad from first element");
		goto ERROR;
	}

	/* create ghostpad */
	player->ghost_pad_for_videobin = gst_ghost_pad_new("sink", pad);
	if (!gst_element_add_pad(videobin[MMPLAYER_V_BIN].gst, player->ghost_pad_for_videobin)) {
		LOGE("failed to add ghostpad to videobin");
		goto ERROR;
	}
	gst_object_unref(pad);

	/* done. free allocated variables */
	g_list_free(element_bucket);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;

ERROR:
	LOGE("ERROR : releasing videobin");
	g_list_free(element_bucket);

	if (pad)
		gst_object_unref(GST_OBJECT(pad));

	/* release videobin with it's childs */
	if (videobin[MMPLAYER_V_BIN].gst)
		gst_object_unref(GST_OBJECT(videobin[MMPLAYER_V_BIN].gst));

	MMPLAYER_FREEIF(videobin);
	player->pipeline->videobin = NULL;

	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_create_plain_text_elements(mmplayer_t *player)
{
	GList *element_bucket = NULL;
	mmplayer_gst_element_t *textbin = player->pipeline->textbin;

	MMPLAYER_CREATE_ELEMENT(textbin, MMPLAYER_T_QUEUE, "queue", "text_queue", element_bucket, player);
	MMPLAYER_CREATE_ELEMENT(textbin, MMPLAYER_T_IDENTITY, "identity", "text_identity", element_bucket, player);
	g_object_set(G_OBJECT(textbin[MMPLAYER_T_IDENTITY].gst),
							"signal-handoffs", FALSE,
							NULL);

	MMPLAYER_CREATE_ELEMENT(textbin, MMPLAYER_T_FAKE_SINK, "fakesink", "text_fakesink", element_bucket, player);
	_mmplayer_add_signal_connection(player,
							G_OBJECT(textbin[MMPLAYER_T_FAKE_SINK].gst),
							MM_PLAYER_SIGNAL_TYPE_TEXTBIN,
							"handoff",
							G_CALLBACK(__mmplayer_update_subtitle),
							(gpointer)player);

	g_object_set(G_OBJECT(textbin[MMPLAYER_T_FAKE_SINK].gst), "sync", TRUE,
						"signal-handoffs", TRUE, "max-lateness", FAKE_SINK_MAX_LATENESS, NULL);

	if (!player->play_subtitle) {
		LOGD("add textbin sink as sink element of whole pipeline.");
		__mmplayer_add_sink(player, GST_ELEMENT(textbin[MMPLAYER_T_FAKE_SINK].gst));
	}

	/* adding created elements to bin */
	LOGD("adding created elements to bin");
	if (!_mmplayer_gst_element_add_bucket_to_bin(GST_BIN(textbin[MMPLAYER_T_BIN].gst), element_bucket)) {
		LOGE("failed to add elements");
		goto ERROR;
	}

	/* unset sink flag from textbin. not to hold eos when video data is shorter than subtitle */
	GST_OBJECT_FLAG_UNSET(textbin[MMPLAYER_T_BIN].gst, GST_ELEMENT_FLAG_SINK);
	GST_OBJECT_FLAG_UNSET(textbin[MMPLAYER_T_FAKE_SINK].gst, GST_ELEMENT_FLAG_SINK);

	/* linking elements in the bucket by added order. */
	LOGD("Linking elements in the bucket by added order.");
	if (_mmplayer_gst_element_link_bucket(element_bucket) == -1) {
		LOGE("failed to link elements");
		goto ERROR;
	}

	/* done. free allocated variables */
	g_list_free(element_bucket);

	if (textbin[MMPLAYER_T_QUEUE].gst) {
		GstPad *pad = NULL;
		GstPad *ghostpad = NULL;

		pad = gst_element_get_static_pad(GST_ELEMENT(textbin[MMPLAYER_T_QUEUE].gst), "sink");
		if (!pad) {
			LOGE("failed to get sink pad of text queue");
			goto ERROR;
		}

		ghostpad = gst_ghost_pad_new("text_sink", pad);
		gst_object_unref(pad);

		if (!ghostpad) {
			LOGE("failed to create ghostpad of textbin");
			goto ERROR;
		}

		if (!gst_element_add_pad(textbin[MMPLAYER_T_BIN].gst, ghostpad)) {
			LOGE("failed to add ghostpad to textbin");
			gst_object_unref(ghostpad);
			goto ERROR;
		}
	}

	return MM_ERROR_NONE;

ERROR:
	g_list_free(element_bucket);

	if (!player->play_subtitle && textbin[MMPLAYER_T_FAKE_SINK].gst) {
		LOGE("remove textbin sink from sink list");
		__mmplayer_del_sink(player, textbin[MMPLAYER_T_FAKE_SINK].gst);
	}

	/* release element at __mmplayer_gst_create_text_sink_bin */
	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_create_text_sink_bin(mmplayer_t *player)
{
	mmplayer_gst_element_t *textbin = NULL;
	GList *element_bucket = NULL;
	int surface_type = 0;
	gint i = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* alloc handles */
	textbin = (mmplayer_gst_element_t *)g_try_malloc0(sizeof(mmplayer_gst_element_t) * MMPLAYER_T_NUM);
	if (!textbin) {
		LOGE("failed to allocate memory for textbin");
		return MM_ERROR_PLAYER_NO_FREE_SPACE;
	}

	/* create bin */
	textbin[MMPLAYER_T_BIN].id = MMPLAYER_T_BIN;
	textbin[MMPLAYER_T_BIN].gst = gst_bin_new("textbin");
	if (!textbin[MMPLAYER_T_BIN].gst) {
		LOGE("failed to create textbin");
		goto ERROR;
	}

	/* take it */
	player->pipeline->textbin = textbin;

	/* fakesink */
	mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &surface_type);
	LOGD("surface type for subtitle : %d", surface_type);
	switch (surface_type) {
	case MM_DISPLAY_SURFACE_OVERLAY:
	case MM_DISPLAY_SURFACE_NULL:
	case MM_DISPLAY_SURFACE_REMOTE:
		if (__mmplayer_gst_create_plain_text_elements(player) != MM_ERROR_NONE) {
			LOGE("failed to make plain text elements");
			goto ERROR;
		}
		break;
	default:
		goto ERROR;
		break;
	}

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;

ERROR:

	LOGD("ERROR : releasing textbin");

	g_list_free(element_bucket);

	/* release signal */
	__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_TEXTBIN);

	/* release element which are not added to bin */
	for (i = 1; i < MMPLAYER_T_NUM; i++) {
		/* NOTE : skip bin */
		if (textbin[i].gst) {
			GstObject *parent = NULL;
			parent = gst_element_get_parent(textbin[i].gst);

			if (!parent) {
				gst_object_unref(GST_OBJECT(textbin[i].gst));
				textbin[i].gst = NULL;
			} else {
				gst_object_unref(GST_OBJECT(parent));
			}
		}
	}

	/* release textbin with it's childs */
	if (textbin[MMPLAYER_T_BIN].gst)
		gst_object_unref(GST_OBJECT(textbin[MMPLAYER_T_BIN].gst));

	MMPLAYER_FREEIF(player->pipeline->textbin);
	player->pipeline->textbin = NULL;

	MMPLAYER_FLEAVE();
	return MM_ERROR_PLAYER_INTERNAL;
}

static int
__mmplayer_gst_create_text_pipeline(mmplayer_t *player)
{
	mmplayer_gst_element_t *mainbin = NULL;
	mmplayer_gst_element_t *textbin = NULL;
	MMHandleType attrs = 0;
	GstElement *subsrc = NULL;
	GstElement *subparse = NULL;
	gchar *subtitle_uri = NULL;
	const gchar *charset = NULL;
	GstPad *pad = NULL;

	MMPLAYER_FENTER();

	/* get mainbin */
	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
								player->pipeline &&
								player->pipeline->mainbin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	mainbin = player->pipeline->mainbin;

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("cannot get content attribute");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	mm_attrs_get_string_by_name(attrs, "subtitle_uri", &subtitle_uri);
	if (!subtitle_uri || strlen(subtitle_uri) < 1) {
		LOGE("subtitle uri is not proper filepath.");
		return MM_ERROR_PLAYER_INVALID_URI;
	}

	if (!_mmplayer_get_storage_info(subtitle_uri, &player->storage_info[MMPLAYER_PATH_TEXT])) {
		LOGE("failed to get storage info of subtitle path");
		return MM_ERROR_PLAYER_INVALID_URI;
	}

	SECURE_LOGD("subtitle file path is [%s].", subtitle_uri);

	MMPLAYER_SUBTITLE_INFO_LOCK(player);
	player->subtitle_language_list = NULL;
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);

	/* create the subtitle source */
	subsrc = gst_element_factory_make("filesrc", "subtitle_source");
	if (!subsrc) {
		LOGE("failed to create filesrc element");
		goto ERROR;
	}
	g_object_set(G_OBJECT(subsrc), "location", subtitle_uri, NULL);

	mainbin[MMPLAYER_M_SUBSRC].id = MMPLAYER_M_SUBSRC;
	mainbin[MMPLAYER_M_SUBSRC].gst = subsrc;

	if (!gst_bin_add(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), subsrc)) {
		LOGW("failed to add queue");
		gst_object_unref(mainbin[MMPLAYER_M_SUBSRC].gst);
		mainbin[MMPLAYER_M_SUBSRC].gst = NULL;
		mainbin[MMPLAYER_M_SUBSRC].id = MMPLAYER_M_NUM;
		goto ERROR;
	}

	/* subparse */
	subparse = gst_element_factory_make("subparse", "subtitle_parser");
	if (!subparse) {
		LOGE("failed to create subparse element");
		goto ERROR;
	}

	charset = _mmplayer_get_charset(subtitle_uri);
	if (charset) {
		LOGD("detected charset is %s", charset);
		g_object_set(G_OBJECT(subparse), "subtitle-encoding", charset, NULL);
	}

	mainbin[MMPLAYER_M_SUBPARSE].id = MMPLAYER_M_SUBPARSE;
	mainbin[MMPLAYER_M_SUBPARSE].gst = subparse;

	if (!gst_bin_add(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), subparse)) {
		LOGW("failed to add subparse");
		gst_object_unref(mainbin[MMPLAYER_M_SUBPARSE].gst);
		mainbin[MMPLAYER_M_SUBPARSE].gst = NULL;
		mainbin[MMPLAYER_M_SUBPARSE].id = MMPLAYER_M_NUM;
		goto ERROR;
	}

	if (!gst_element_link_pads(subsrc, "src", subparse, "sink")) {
		LOGW("failed to link subsrc and subparse");
		goto ERROR;
	}

	player->play_subtitle = TRUE;
	player->adjust_subtitle_pos = 0;

	LOGD("play subtitle using subtitle file");

	if (player->pipeline->textbin == NULL) {
		if (MM_ERROR_NONE != __mmplayer_gst_create_text_sink_bin(player)) {
			LOGE("failed to create text sink bin. continuing without text");
			goto ERROR;
		}

		textbin = player->pipeline->textbin;

		if (!gst_bin_add(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), GST_ELEMENT(textbin[MMPLAYER_T_BIN].gst))) {
			LOGW("failed to add textbin");

			/* release signal */
			__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_TEXTBIN);

			/* release textbin with it's childs */
			gst_object_unref(GST_OBJECT(textbin[MMPLAYER_T_BIN].gst));
			MMPLAYER_FREEIF(player->pipeline->textbin);
			player->pipeline->textbin = textbin = NULL;
			goto ERROR;
		}

		LOGD("link text input selector and textbin ghost pad");

		player->textsink_linked = 1;
		player->external_text_idx = 0;
		LOGI("textsink is linked");
	} else {
		textbin = player->pipeline->textbin;
		LOGD("text bin has been created. reuse it.");
		player->external_text_idx = 1;
	}

	if (!gst_element_link_pads(subparse, "src", textbin[MMPLAYER_T_BIN].gst, "text_sink")) {
		LOGW("failed to link subparse and textbin");
		goto ERROR;
	}

	pad = gst_element_get_static_pad(textbin[MMPLAYER_T_FAKE_SINK].gst, "sink");
	if (!pad) {
		LOGE("failed to get sink pad from textsink to probe data");
		goto ERROR;
	}

	gst_pad_add_probe(pad, GST_PAD_PROBE_TYPE_BUFFER,
				__mmplayer_subtitle_adjust_position_probe, player, NULL);

	gst_object_unref(pad);
	pad = NULL;

	/* create dot. for debugging */
	MMPLAYER_GENERATE_DOT_IF_ENABLED(player, "pipeline-with-subtitle");
	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;

ERROR:
	/* release text pipeline resource */
	player->textsink_linked = 0;

	/* release signal */
	__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_TEXTBIN);

	if (player->pipeline->textbin) {
		LOGE("remove textbin");

		/* release textbin with it's childs */
		MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->textbin, MMPLAYER_T_BIN);
		MMPLAYER_FREEIF(player->pipeline->textbin);
		player->pipeline->textbin = NULL;

	}

	/* release subtitle elem */
	MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->mainbin, MMPLAYER_M_SUBSRC);
	MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->mainbin, MMPLAYER_M_SUBPARSE);

	return MM_ERROR_PLAYER_INTERNAL;
}

gboolean
__mmplayer_update_subtitle(GstElement *object, GstBuffer *buffer, GstPad *pad, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	MMMessageParamType msg = {0, };
	GstClockTime duration = 0;
	gpointer text = NULL;
	guint text_size = 0;
	gboolean ret = TRUE;
	GstMapInfo mapinfo = GST_MAP_INFO_INIT;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(buffer, FALSE);

	if (player->is_subtitle_force_drop) {
		LOGW("subtitle is dropped forcedly.");
		return ret;
	}

	gst_buffer_map(buffer, &mapinfo, GST_MAP_READ);
	text = mapinfo.data;
	text_size = mapinfo.size;

	if (player->set_mode.subtitle_off) {
		LOGD("subtitle is OFF.");
		return TRUE;
	}

	if (!text || (text_size == 0)) {
		LOGD("There is no subtitle to be displayed.");
		return TRUE;
	}

	msg.data = (void *)text;

	duration = GST_BUFFER_DURATION(buffer);

	if (!GST_CLOCK_TIME_IS_VALID(duration)) {
		if (player->duration > GST_BUFFER_PTS(buffer))
			duration = player->duration - GST_BUFFER_PTS(buffer);
		else
			duration = 0;
		LOGI("subtitle duration is invalid, subtitle duration change "
			"GST_CLOCK_TIME_NONE -> %" GST_TIME_FORMAT, GST_TIME_ARGS(duration));
	}
	msg.subtitle.duration = GST_TIME_AS_MSECONDS(duration);

	LOGD("update subtitle : [%ld msec] %s", msg.subtitle.duration, (char *)msg.data);

	MMPLAYER_POST_MSG(player, MM_MESSAGE_UPDATE_SUBTITLE, &msg);
	gst_buffer_unmap(buffer, &mapinfo);

	MMPLAYER_FLEAVE();

	return ret;
}

static GstPadProbeReturn
__mmplayer_subtitle_adjust_position_probe(GstPad *pad, GstPadProbeInfo *info, gpointer u_data)
{
	mmplayer_t *player = (mmplayer_t *)u_data;
	GstClockTime cur_timestamp = 0;
	gint64 adjusted_timestamp = 0;
	GstBuffer *buffer = gst_pad_probe_info_get_buffer(info);

	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);

	if (player->set_mode.subtitle_off) {
		LOGD("subtitle is OFF.");
		return TRUE;
	}

	if (player->adjust_subtitle_pos == 0) {
		LOGD("nothing to do");
		return TRUE;
	}

	cur_timestamp = GST_BUFFER_TIMESTAMP(buffer);
	adjusted_timestamp = (gint64)cur_timestamp + ((gint64)player->adjust_subtitle_pos * G_GINT64_CONSTANT(1000000));

	if (adjusted_timestamp < 0) {
		LOGD("adjusted_timestamp under zero");
		MMPLAYER_FLEAVE();
		return FALSE;
	}

	GST_BUFFER_TIMESTAMP(buffer) = (GstClockTime) adjusted_timestamp;
	LOGD("buffer timestamp changed %" GST_TIME_FORMAT " -> %" GST_TIME_FORMAT "",
				GST_TIME_ARGS(cur_timestamp),
				GST_TIME_ARGS(GST_BUFFER_TIMESTAMP(buffer)));

	return GST_PAD_PROBE_OK;
}

static int
__mmplayer_gst_adjust_subtitle_position(mmplayer_t *player, int position)
{
	MMPLAYER_FENTER();

	/* check player and subtitlebin are created */
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(player->play_subtitle, MM_ERROR_NOT_SUPPORT_API);

	if (position == 0) {
		LOGD("nothing to do");
		MMPLAYER_FLEAVE();
		return MM_ERROR_NONE;
	}

	/* check current postion */
	player->adjust_subtitle_pos = position;

	LOGD("save adjust_subtitle_pos in player");

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

/**
 * This function is to create  audio or video pipeline for playing.
 *
 * @param	player		[in]	handle of player
 *
 * @return	This function returns zero on success.
 * @remark
 * @see
 */
static int
__mmplayer_gst_create_pipeline(mmplayer_t *player)
{
	int ret = MM_ERROR_NONE;
	mmplayer_gst_element_t *mainbin = NULL;
	MMHandleType attrs = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* get profile attribute */
	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("failed to get content attribute");
		goto INIT_ERROR;
	}

	/* create pipeline handles */
	if (player->pipeline) {
		LOGE("pipeline should be released before create new one");
		goto INIT_ERROR;
	}

	player->pipeline = (mmplayer_pipeline_info_t *)g_malloc0(sizeof(mmplayer_pipeline_info_t));

	/* create mainbin */
	mainbin = (mmplayer_gst_element_t *)g_try_malloc0(sizeof(mmplayer_gst_element_t) * MMPLAYER_M_NUM);
	if (mainbin == NULL)
		goto INIT_ERROR;

	/* create pipeline */
	mainbin[MMPLAYER_M_PIPE].id = MMPLAYER_M_PIPE;
	mainbin[MMPLAYER_M_PIPE].gst = gst_pipeline_new("player");
	if (!mainbin[MMPLAYER_M_PIPE].gst) {
		LOGE("failed to create pipeline");
		g_free(mainbin);
		goto INIT_ERROR;
	}

	player->pipeline->mainbin = mainbin;

	/* create the source and decoder elements */
	if (MMPLAYER_IS_MS_BUFF_SRC(player))
		ret = _mmplayer_gst_build_es_pipeline(player);
	else
		ret = _mmplayer_gst_build_pipeline(player);

	if (ret != MM_ERROR_NONE) {
		LOGE("failed to create some elements");
		goto INIT_ERROR;
	}

	/* Note : check whether subtitle atrribute uri is set. If uri is set, then try to play subtitle file */
	if (__mmplayer_check_subtitle(player)
		&& (__mmplayer_gst_create_text_pipeline(player) != MM_ERROR_NONE))
		LOGE("failed to create text pipeline");

	/* add bus watch */
	ret = _mmplayer_gst_add_bus_watch(player);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to add bus watch");
		goto INIT_ERROR;
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;

INIT_ERROR:
	__mmplayer_gst_destroy_pipeline(player);
	return MM_ERROR_PLAYER_INTERNAL;
}

static void
__mmplayer_reset_gapless_state(mmplayer_t *player)
{
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player
		&& player->pipeline
		&& player->pipeline->audiobin
		&& player->pipeline->audiobin[MMPLAYER_A_BIN].gst);

	memset(&player->gapless, 0, sizeof(mmplayer_gapless_t));

	MMPLAYER_FLEAVE();
	return;
}

static int
__mmplayer_gst_destroy_pipeline(mmplayer_t *player)
{
	gint timeout = 0;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_INVALID_HANDLE);

	/* cleanup stuffs */
	MMPLAYER_FREEIF(player->type);
	player->no_more_pad = FALSE;
	player->num_dynamic_pad = 0;
	player->demux_pad_index = 0;

	MMPLAYER_SUBTITLE_INFO_LOCK(player);
	player->subtitle_language_list = NULL;
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);

	__mmplayer_reset_gapless_state(player);

	if (player->streamer) {
		_mm_player_streaming_initialize(player->streamer, FALSE);
		_mm_player_streaming_destroy(player->streamer);
		player->streamer = NULL;
	}

	/* cleanup unlinked mime type */
	MMPLAYER_FREEIF(player->unlinked_audio_mime);
	MMPLAYER_FREEIF(player->unlinked_video_mime);
	MMPLAYER_FREEIF(player->unlinked_demuxer_mime);

	/* cleanup running stuffs */
	_mmplayer_cancel_eos_timer(player);

	/* cleanup gst stuffs */
	if (player->pipeline) {
		mmplayer_gst_element_t *mainbin = player->pipeline->mainbin;
		GstTagList *tag_list = player->pipeline->tag_list;

		/* first we need to disconnect all signal hander */
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_ALL);

		if (mainbin) {
			mmplayer_gst_element_t *audiobin = player->pipeline->audiobin;
			mmplayer_gst_element_t *videobin = player->pipeline->videobin;
			mmplayer_gst_element_t *textbin = player->pipeline->textbin;
			GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(mainbin[MMPLAYER_M_PIPE].gst));
			gst_bus_set_sync_handler(bus, NULL, NULL, NULL);
			gst_object_unref(bus);

			timeout = MMPLAYER_STATE_CHANGE_TIMEOUT(player);
			ret = _mmplayer_gst_set_state(player, mainbin[MMPLAYER_M_PIPE].gst, GST_STATE_NULL, FALSE, timeout);
			if (ret != MM_ERROR_NONE) {
				LOGE("fail to change state to NULL");
				return MM_ERROR_PLAYER_INTERNAL;
			}

			LOGW("succeeded in changing state to NULL");

			gst_object_unref(GST_OBJECT(mainbin[MMPLAYER_M_PIPE].gst));

			/* free fakesink */
			if (mainbin[MMPLAYER_M_SRC_FAKESINK].gst)
				gst_object_unref(GST_OBJECT(mainbin[MMPLAYER_M_SRC_FAKESINK].gst));

			/* free avsysaudiosink
			   avsysaudiosink should be unref when destory pipeline just after start play with BT.
			   Because audiosink is created but never added to bin, and therefore it will not be unref when pipeline is destroyed.
			*/
			MMPLAYER_FREEIF(audiobin);
			MMPLAYER_FREEIF(videobin);
			MMPLAYER_FREEIF(textbin);
			MMPLAYER_FREEIF(mainbin);
		}

		if (tag_list)
			gst_tag_list_unref(tag_list);

		MMPLAYER_FREEIF(player->pipeline);
	}
	MMPLAYER_FREEIF(player->album_art);

	if (player->v_stream_caps) {
		gst_caps_unref(player->v_stream_caps);
		player->v_stream_caps = NULL;
	}

	if (player->a_stream_caps) {
		gst_caps_unref(player->a_stream_caps);
		player->a_stream_caps = NULL;
	}

	if (player->s_stream_caps) {
		gst_caps_unref(player->s_stream_caps);
		player->s_stream_caps = NULL;
	}
	_mmplayer_track_destroy(player);

	if (player->sink_elements)
		g_list_free(player->sink_elements);
	player->sink_elements = NULL;

	if (player->bufmgr) {
		tbm_bufmgr_deinit(player->bufmgr);
		player->bufmgr = NULL;
	}

	LOGW("finished destroy pipeline");

	MMPLAYER_FLEAVE();

	return ret;
}

static int
__mmplayer_gst_realize(mmplayer_t *player)
{
	gint timeout = 0;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_PENDING_STATE(player) = MM_PLAYER_STATE_READY;

	ret = __mmplayer_gst_create_pipeline(player);
	if (ret) {
		LOGE("failed to create pipeline");
		return ret;
	}

	/* set pipeline state to READY */
	/* NOTE : state change to READY must be performed sync. */
	timeout = MMPLAYER_STATE_CHANGE_TIMEOUT(player);
	ret = _mmplayer_gst_set_state(player,
				player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, GST_STATE_READY, FALSE, timeout);

	if (ret != MM_ERROR_NONE) {
		/* return error if failed to set state */
		LOGE("failed to set READY state");
		return ret;
	}

	MMPLAYER_SET_STATE(player, MM_PLAYER_STATE_READY);

	/* create dot before error-return. for debugging */
	MMPLAYER_GENERATE_DOT_IF_ENABLED(player, "pipeline-status-realize");

	MMPLAYER_FLEAVE();

	return ret;
}

static int
__mmplayer_gst_unrealize(mmplayer_t *player)
{
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_PENDING_STATE(player) = MM_PLAYER_STATE_NULL;
	MMPLAYER_PRINT_STATE(player);

	/* release miscellaneous information */
	__mmplayer_release_misc(player);

	/* destroy pipeline */
	ret = __mmplayer_gst_destroy_pipeline(player);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to destory pipeline");
		return ret;
	}

	/* release miscellaneous information.
	   these info needs to be released after pipeline is destroyed. */
	__mmplayer_release_misc_post(player);

	MMPLAYER_SET_STATE(player, MM_PLAYER_STATE_NULL);

	MMPLAYER_FLEAVE();

	return ret;
}

static int
__mmplayer_gst_set_message_callback(mmplayer_t *player, MMMessageCallback callback, gpointer user_param)
{
	MMPLAYER_FENTER();

	if (!player) {
		LOGW("set_message_callback is called with invalid player handle");
		return MM_ERROR_PLAYER_NOT_INITIALIZED;
	}

	player->msg_cb = callback;
	player->msg_cb_param = user_param;

	LOGD("msg_cb : %p     msg_cb_param : %p", callback, user_param);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_parse_profile(const char *uri, void *param, mmplayer_parse_profile_t *data)
{
	int ret = MM_ERROR_NONE;
	char *path = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(uri, MM_ERROR_PLAYER_INVALID_URI);
	MMPLAYER_RETURN_VAL_IF_FAIL(data, MM_ERROR_PLAYER_INTERNAL);
	MMPLAYER_RETURN_VAL_IF_FAIL((strlen(uri) <= MM_MAX_URL_LEN), MM_ERROR_PLAYER_INVALID_URI);

	memset(data, 0, sizeof(mmplayer_parse_profile_t));

	if (strstr(uri, "es_buff://")) {
		__mmplayer_copy_uri_and_set_type(data, uri, MM_PLAYER_URI_TYPE_MS_BUFF);
	} else if (strstr(uri, "rtsp://") || strstr(uri, "rtsps://") || strstr(uri, "rtspu://")) {
		__mmplayer_copy_uri_and_set_type(data, uri, MM_PLAYER_URI_TYPE_URL_RTSP);
	} else if (strstr(uri, "http://") || strstr(uri, "https://")) {
		gchar *tmp = NULL;
		tmp = g_ascii_strdown(uri, strlen(uri));
		if (tmp && (g_str_has_suffix(tmp, ".ism/manifest") || g_str_has_suffix(tmp, ".isml/manifest")))
			__mmplayer_copy_uri_and_set_type(data, uri, MM_PLAYER_URI_TYPE_SS);
		else
			__mmplayer_copy_uri_and_set_type(data, uri, MM_PLAYER_URI_TYPE_URL_HTTP);
		g_free(tmp);
	} else if (strstr(uri, "mms://")) {
		__mmplayer_copy_uri_and_set_type(data, uri, MM_PLAYER_URI_TYPE_URL_MMS);
	} else if ((path = strstr(uri, "mem://"))) {
		ret = __mmplayer_set_mem_uri(data, path, param);
	} else {
		ret = __mmplayer_set_file_uri(data, uri);
	}

	if (data->uri_type == MM_PLAYER_URI_TYPE_NONE)
		ret = MM_ERROR_PLAYER_FILE_NOT_FOUND;
	else if (data->uri_type == MM_PLAYER_URI_TYPE_NO_PERMISSION)
		ret = MM_ERROR_PLAYER_PERMISSION_DENIED;

	/* dump parse result */
	SECURE_LOGW("incoming uri : %s", uri);
	LOGD("uri_type : %d, mem : %p, mem_size : %d, urgent : %s",
		data->uri_type, data->input_mem.buf, data->input_mem.len, data->urgent);

	MMPLAYER_FLEAVE();

	return ret;
}

static gboolean
__mmplayer_can_do_interrupt(mmplayer_t *player)
{
	if (!player || !player->pipeline || !player->attrs) {
		LOGW("not initialized");
		goto FAILED;
	}

	if (player->audio_decoded_cb) {
		LOGW("not support in pcm extraction mode");
		goto FAILED;
	}

	/* check if seeking */
	if (player->seek_state != MMPLAYER_SEEK_NONE) {
		MMMessageParamType msg_param;
		memset(&msg_param, 0, sizeof(MMMessageParamType));
		msg_param.code = MM_ERROR_PLAYER_SEEK;
		player->seek_state = MMPLAYER_SEEK_NONE;
		MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);
		goto FAILED;
	}

	/* check other thread */
	if (!MMPLAYER_CMD_TRYLOCK(player)) {
		LOGW("locked already, cmd state : %d", player->cmd);

		/* check application command */
		if (player->cmd == MMPLAYER_COMMAND_START || player->cmd == MMPLAYER_COMMAND_RESUME) {
			LOGW("playing.. should wait cmd lock then, will be interrupted");

			/* lock will be released at mrp_resource_release_cb() */
			MMPLAYER_CMD_LOCK(player);
			goto INTERRUPT;
		}
		LOGW("nothing to do");
		goto FAILED;
	} else {
		LOGW("can interrupt immediately");
		goto INTERRUPT;
	}

FAILED:    /* with CMD UNLOCKED */
	return FALSE;

INTERRUPT: /* with CMD LOCKED, will do UNLOCK at __resource_release_cb() */
	return TRUE;
}

static int
__resource_release_cb(mm_resource_manager_h rm, mm_resource_manager_res_h res,
		void *user_data)
{
	mmplayer_t *player = NULL;
	MMMessageParamType msg = {0, };
	gint64 pos = 0;
	mmplayer_resource_type_e res_idx = MMPLAYER_RESOURCE_TYPE_MAX;

	MMPLAYER_FENTER();

	if (!user_data) {
		LOGE("user_data is null");
		return FALSE;
	}
	player = (mmplayer_t *)user_data;

	if (!__mmplayer_can_do_interrupt(player)) {
		LOGW("no need to interrupt, so leave");
		/* FIXME: there is no way to avoid releasing resource. */
		return FALSE;
	}

	player->interrupted_by_resource = TRUE;

	/* get last play position */
	if (_mmplayer_gst_get_position(player, &pos) == MM_ERROR_NONE) {
		msg.union_type = MM_MSG_UNION_TIME;
		msg.time.elapsed = pos;
		MMPLAYER_POST_MSG(player, MM_MESSAGE_PLAY_POSITION, &msg);
	} else {
		LOGW("failed to get play position.");
	}

	LOGD("video resource conflict so, resource will be freed by unrealizing");
	if (_mmplayer_unrealize((MMHandleType)player) != MM_ERROR_NONE)
		LOGE("failed to unrealize");

	/* lock is called in __mmplayer_can_do_interrupt() */
	MMPLAYER_CMD_UNLOCK(player);

	for (res_idx = MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER; res_idx < MMPLAYER_RESOURCE_TYPE_MAX; res_idx++) {
		player->hw_resource[res_idx] = NULL;
	}

	MMPLAYER_FLEAVE();
	return TRUE; /* release all the resources */
}

static void
__mmplayer_initialize_video_roi(mmplayer_t *player)
{
	player->video_roi.scale_x = 0.0;
	player->video_roi.scale_y = 0.0;
	player->video_roi.scale_width = 1.0;
	player->video_roi.scale_height = 1.0;
}

int
_mmplayer_create_player(MMHandleType handle)
{
	int ret = MM_ERROR_PLAYER_INTERNAL;
	bool enabled = false;

	mmplayer_t *player = MM_PLAYER_CAST(handle);

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* initialize player state */
	MMPLAYER_CURRENT_STATE(player) = MM_PLAYER_STATE_NONE;
	MMPLAYER_PREV_STATE(player) = MM_PLAYER_STATE_NONE;
	MMPLAYER_PENDING_STATE(player) = MM_PLAYER_STATE_NONE;
	MMPLAYER_TARGET_STATE(player) = MM_PLAYER_STATE_NONE;

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_CREATE);

	/* construct attributes */
	player->attrs = _mmplayer_construct_attribute(handle);

	if (!player->attrs) {
		LOGE("Failed to construct attributes");
		return ret;
	}

	/* initialize gstreamer with configured parameter */
	if (!__mmplayer_init_gstreamer(player)) {
		LOGE("Initializing gstreamer failed");
		_mmplayer_deconstruct_attribute(handle);
		return ret;
	}

	/* create lock. note that g_tread_init() has already called in gst_init() */
	g_mutex_init(&player->fsink_lock);

	/* create update tag lock */
	g_mutex_init(&player->update_tag_lock);

	/* create gapless play mutex */
	g_mutex_init(&player->gapless_play_thread_mutex);

	/* create gapless play cond */
	g_cond_init(&player->gapless_play_thread_cond);

	/* create gapless play thread */
	player->gapless_play_thread =
		g_thread_try_new("gapless_play_thread", __mmplayer_gapless_play_thread, (gpointer)player, NULL);
	if (!player->gapless_play_thread) {
		LOGE("failed to create gapless play thread");
		ret = MM_ERROR_PLAYER_RESOURCE_LIMIT;
		g_mutex_clear(&player->gapless_play_thread_mutex);
		g_cond_clear(&player->gapless_play_thread_cond);
		goto ERROR;
	}

	player->bus_msg_q = g_queue_new();
	if (!player->bus_msg_q) {
		LOGE("failed to create queue for bus_msg");
		ret = MM_ERROR_PLAYER_RESOURCE_LIMIT;
		goto ERROR;
	}

	ret = _mmplayer_initialize_video_capture(player);
	if (ret != MM_ERROR_NONE) {
		LOGE("failed to initialize video capture");
		goto ERROR;
	}

	/* initialize resource manager */
	if (mm_resource_manager_create(MM_RESOURCE_MANAGER_APP_CLASS_MEDIA,
		__resource_release_cb, player, &player->resource_manager)
		!= MM_RESOURCE_MANAGER_ERROR_NONE) {
		LOGE("failed to initialize resource manager");
		ret = MM_ERROR_PLAYER_INTERNAL;
		goto ERROR;
	}

	/* create video bo lock and cond */
	g_mutex_init(&player->video_bo_mutex);
	g_cond_init(&player->video_bo_cond);

	/* create subtitle info lock and cond */
	g_mutex_init(&player->subtitle_info_mutex);
	g_cond_init(&player->subtitle_info_cond);

	player->streaming_type = STREAMING_SERVICE_NONE;

	/* give default value of audio effect setting */
	player->sound.volume = MM_VOLUME_FACTOR_DEFAULT;
	player->sound.rg_enable = false;
	player->playback_rate = DEFAULT_PLAYBACK_RATE;

	player->play_subtitle = FALSE;
	player->has_closed_caption = FALSE;
	player->pending_resume = FALSE;
	if (player->ini.dump_element_keyword[0][0] == '\0')
		player->ini.set_dump_element_flag = FALSE;
	else
		player->ini.set_dump_element_flag = TRUE;

	player->adaptive_info.limit.bandwidth = ADAPTIVE_VARIANT_DEFAULT_VALUE;
	player->adaptive_info.limit.width = ADAPTIVE_VARIANT_DEFAULT_VALUE;
	player->adaptive_info.limit.height = ADAPTIVE_VARIANT_DEFAULT_VALUE;

	/* Set video360 settings to their defaults for just-created player.
	 * */

	player->is_360_feature_enabled = FALSE;
	if (SYSTEM_INFO_ERROR_NONE == system_info_get_platform_bool(FEATURE_NAME_SPHERICAL_VIDEO, &enabled)) {
		LOGI("spherical feature info: %d", enabled);
		if (enabled)
			player->is_360_feature_enabled = TRUE;
	} else {
		LOGE("failed to get spherical feature info");
	}

	player->is_content_spherical = FALSE;
	player->is_video360_enabled = TRUE;
	player->video360_metadata.projection_type = VIDEO360_PROJECTION_TYPE_EQUIRECTANGULAR;
	player->video360_metadata.stereo_mode = VIDEO360_MODE_MONOSCOPIC;
	player->video360_yaw_radians = PLAYER_SPHERICAL_DEFAULT_YAW;
	player->video360_pitch_radians = PLAYER_SPHERICAL_DEFAULT_PITCH;
	player->video360_zoom = 1.0f;
	player->video360_horizontal_fov = PLAYER_SPHERICAL_DEFAULT_H_FOV;
	player->video360_vertical_fov = PLAYER_SPHERICAL_DEFAULT_V_FOV;

	__mmplayer_initialize_video_roi(player);

	/* set player state to null */
	MMPLAYER_STATE_CHANGE_TIMEOUT(player) = player->ini.localplayback_state_change_timeout;
	MMPLAYER_SET_STATE(player, MM_PLAYER_STATE_NULL);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;

ERROR:
	/* free lock */
	g_mutex_clear(&player->fsink_lock);
	/* free update tag lock */
	g_mutex_clear(&player->update_tag_lock);
	g_queue_free(player->bus_msg_q);
	player->bus_msg_q = NULL;
	/* free gapless play thread */
	if (player->gapless_play_thread) {
		MMPLAYER_GAPLESS_PLAY_THREAD_LOCK(player);
		player->gapless_play_thread_exit = TRUE;
		MMPLAYER_GAPLESS_PLAY_THREAD_SIGNAL(player);
		MMPLAYER_GAPLESS_PLAY_THREAD_UNLOCK(player);

		g_thread_join(player->gapless_play_thread);
		player->gapless_play_thread = NULL;

		g_mutex_clear(&player->gapless_play_thread_mutex);
		g_cond_clear(&player->gapless_play_thread_cond);
	}

	/* release attributes */
	_mmplayer_deconstruct_attribute(handle);

	MMPLAYER_FLEAVE();

	return ret;
}

static gboolean
__mmplayer_init_gstreamer(mmplayer_t *player)
{
	static gboolean initialized = FALSE;
	static const int max_argc = 50;
	gint *argc = NULL;
	gchar **argv = NULL;
	gchar **argv2 = NULL;
	GError *err = NULL;
	int i = 0;
	int arg_count = 0;

	if (initialized) {
		LOGD("gstreamer already initialized.");
		return TRUE;
	}

	/* alloc */
	argc = malloc(sizeof(int));
	argv = malloc(sizeof(gchar *) * max_argc);
	argv2 = malloc(sizeof(gchar *) * max_argc);

	if (!argc || !argv || !argv2)
		goto ERROR;

	memset(argv, 0, sizeof(gchar *) * max_argc);
	memset(argv2, 0, sizeof(gchar *) * max_argc);

	/* add initial */
	*argc = 1;
	argv[0] = g_strdup("mmplayer");

	/* add gst_param */
	for (i = 0; i < 5; i++) {
		/* FIXIT : num of param is now fixed to 5. make it dynamic */
		if (strlen(player->ini.gst_param[i]) > 0) {
			argv[*argc] = g_strdup(player->ini.gst_param[i]);
			(*argc)++;
		}
	}

	/* we would not do fork for scanning plugins */
	argv[*argc] = g_strdup("--gst-disable-registry-fork");
	(*argc)++;

	/* check disable registry scan */
	if (player->ini.skip_rescan) {
		argv[*argc] = g_strdup("--gst-disable-registry-update");
		(*argc)++;
	}

	/* check disable segtrap */
	if (player->ini.disable_segtrap) {
		argv[*argc] = g_strdup("--gst-disable-segtrap");
		(*argc)++;
	}

	LOGD("initializing gstreamer with following parameter");
	LOGD("argc : %d", *argc);
	arg_count = *argc;

	for (i = 0; i < arg_count; i++) {
		argv2[i] = argv[i];
		LOGD("argv[%d] : %s", i, argv2[i]);
	}

	/* initializing gstreamer */
	if (!gst_init_check(argc, &argv, &err)) {
		LOGE("Could not initialize GStreamer: %s", err ? err->message : "unknown error occurred");
		if (err)
			g_error_free(err);

		goto ERROR;
	}
	/* release */
	for (i = 0; i < arg_count; i++) {
#ifdef __DEBUG__
		LOGD("release - argv[%d] : %s", i, argv2[i]);
#endif
		MMPLAYER_FREEIF(argv2[i]);
	}

	MMPLAYER_FREEIF(argv);
	MMPLAYER_FREEIF(argv2);
	MMPLAYER_FREEIF(argc);

	/* done */
	initialized = TRUE;

	return TRUE;

ERROR:

	/* release */
	for (i = 0; i < arg_count; i++) {
		LOGD("free[%d] : %s", i, argv2[i]);
		MMPLAYER_FREEIF(argv2[i]);
	}

	MMPLAYER_FREEIF(argv);
	MMPLAYER_FREEIF(argv2);
	MMPLAYER_FREEIF(argc);

	return FALSE;
}

static void
__mmplayer_check_async_state_transition(mmplayer_t *player)
{
	GstState element_state = GST_STATE_VOID_PENDING;
	GstState element_pending_state = GST_STATE_VOID_PENDING;
	GstStateChangeReturn ret = GST_STATE_CHANGE_FAILURE;
	GstElement *element = NULL;
	gboolean async = FALSE;

	/* check player handle */
	MMPLAYER_RETURN_IF_FAIL(player &&
						player->pipeline &&
						player->pipeline->mainbin &&
						player->pipeline->mainbin[MMPLAYER_M_PIPE].gst);

	if (player->attrs)
		mm_attrs_get_int_by_name(player->attrs, "profile_prepare_async", &async);

	if (!MMPLAYER_IS_MS_BUFF_SRC(player) && (async == FALSE)) {
		LOGD("don't need to check the pipeline state");
		return;
	}

	MMPLAYER_PRINT_STATE(player);

	/* wait for state transition */
	element = player->pipeline->mainbin[MMPLAYER_M_PIPE].gst;
	ret = gst_element_get_state(element, &element_state, &element_pending_state, 1 * GST_SECOND);

	if (ret == GST_STATE_CHANGE_FAILURE) {
		LOGE(" [%s] state : %s   pending : %s",
			GST_ELEMENT_NAME(element),
			gst_element_state_get_name(element_state),
			gst_element_state_get_name(element_pending_state));

		/* dump state of all element */
		_mmplayer_dump_pipeline_state(player);

		return;
	}

	LOGD("[%s] element state has changed", GST_ELEMENT_NAME(element));
	return;
}

int
_mmplayer_destroy(MMHandleType handle)
{
	mmplayer_t *player = MM_PLAYER_CAST(handle);

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* destroy can called at anytime */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_DESTROY);

	/* check async state transition */
	__mmplayer_check_async_state_transition(player);

	/* release gapless play thread */
	if (player->gapless_play_thread) {
		MMPLAYER_GAPLESS_PLAY_THREAD_LOCK(player);
		player->gapless_play_thread_exit = TRUE;
		MMPLAYER_GAPLESS_PLAY_THREAD_SIGNAL(player);
		MMPLAYER_GAPLESS_PLAY_THREAD_UNLOCK(player);

		LOGD("waitting for gapless play thread exit");
		g_thread_join(player->gapless_play_thread);
		g_mutex_clear(&player->gapless_play_thread_mutex);
		g_cond_clear(&player->gapless_play_thread_cond);
		LOGD("gapless play thread released");
	}

	_mmplayer_release_video_capture(player);

	/* de-initialize resource manager */
	if (MM_RESOURCE_MANAGER_ERROR_NONE != mm_resource_manager_destroy(
			player->resource_manager))
		LOGE("failed to deinitialize resource manager");

	/* release pipeline */
	if (MM_ERROR_NONE != __mmplayer_gst_destroy_pipeline(player)) {
		LOGE("failed to destory pipeline");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	g_queue_free(player->bus_msg_q);

	/* release subtitle info lock and cond */
	g_mutex_clear(&player->subtitle_info_mutex);
	g_cond_clear(&player->subtitle_info_cond);

	__mmplayer_release_dump_list(player->dump_list);

	/* release miscellaneous information */
	__mmplayer_release_misc(player);

	/* release miscellaneous information.
	   these info needs to be released after pipeline is destroyed. */
	__mmplayer_release_misc_post(player);

	/* release attributes */
	_mmplayer_deconstruct_attribute(handle);

	/* release lock */
	g_mutex_clear(&player->fsink_lock);

	/* release lock */
	g_mutex_clear(&player->update_tag_lock);

	/* release video bo lock and cond */
	g_mutex_clear(&player->video_bo_mutex);
	g_cond_clear(&player->video_bo_cond);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_realize(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;
	char *uri = NULL;
	void *param = NULL;
	MMHandleType attrs = 0;
	int video_codec_type = 0;
	int audio_codec_type = 0;
	int default_codec_type = 0;
	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_REALIZE);

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("fail to get attributes.");
		return MM_ERROR_PLAYER_INTERNAL;
	}
	mm_attrs_get_string_by_name(attrs, "profile_uri", &uri);
	mm_attrs_get_data_by_name(attrs, "profile_user_param", &param);

	if (player->profile.uri_type == MM_PLAYER_URI_TYPE_NONE) {
		ret = _mmplayer_parse_profile((const char *)uri, param, &player->profile);

		if (ret != MM_ERROR_NONE) {
			LOGE("failed to parse profile");
			return ret;
		}
	}

	if (uri && (strstr(uri, "es_buff://"))) {
		if (strstr(uri, "es_buff://push_mode"))
			player->es_player_push_mode = TRUE;
		else
			player->es_player_push_mode = FALSE;
	}

	if (player->profile.uri_type == MM_PLAYER_URI_TYPE_URL_MMS) {
		LOGW("mms protocol is not supported format.");
		return MM_ERROR_PLAYER_NOT_SUPPORTED_FORMAT;
	}

	if (MMPLAYER_IS_STREAMING(player))
		MMPLAYER_STATE_CHANGE_TIMEOUT(player) = player->ini.live_state_change_timeout;
	else
		MMPLAYER_STATE_CHANGE_TIMEOUT(player) = player->ini.localplayback_state_change_timeout;

	player->smooth_streaming = FALSE;
	player->videodec_linked  = 0;
	player->audiodec_linked  = 0;
	player->textsink_linked = 0;
	player->is_external_subtitle_present = FALSE;
	player->is_external_subtitle_added_now = FALSE;
	player->is_subtitle_off = FALSE; /* set the subtitle ON default */
	player->video360_metadata.is_spherical = -1;
	player->is_openal_plugin_used = FALSE;
	player->demux_pad_index = 0;
	player->subtitle_language_list = NULL;
	player->is_subtitle_force_drop = FALSE;

	_mmplayer_track_initialize(player);
	__mmplayer_initialize_storage_info(player, MMPLAYER_PATH_MAX);

	if ((MMPLAYER_IS_STREAMING(player)) && (player->streamer == NULL)) {
		gint prebuffer_ms = 0, rebuffer_ms = 0;

		player->streamer = _mm_player_streaming_create();
		_mm_player_streaming_initialize(player->streamer, TRUE);

		mm_attrs_multiple_get(player->attrs, NULL,
				MM_PLAYER_PREBUFFER_MS, &prebuffer_ms,
				MM_PLAYER_REBUFFER_MS, &rebuffer_ms, NULL);

		if (prebuffer_ms > 0) {
			prebuffer_ms = MAX(prebuffer_ms, 1000);
			player->streamer->buffering_req.prebuffer_time = prebuffer_ms;
		}

		if (rebuffer_ms > 0) {
			player->streamer->buffering_req.mode = MM_PLAYER_BUFFERING_MODE_FIXED;
			rebuffer_ms = MAX(rebuffer_ms, 1000);
			player->streamer->buffering_req.rebuffer_time = rebuffer_ms;
		}

		LOGD("buffering time %d ms, %d ms", player->streamer->buffering_req.prebuffer_time,
								player->streamer->buffering_req.rebuffer_time);
	}

	mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_AUDIO_CODEC_TYPE, &audio_codec_type);
	if (!strcmp(player->ini.audiocodec_default_type, "hw"))
		default_codec_type = MM_PLAYER_CODEC_TYPE_HW;
	else
		default_codec_type = MM_PLAYER_CODEC_TYPE_SW;

	if (audio_codec_type != default_codec_type) {
		LOGD("audio dec sorting is required");
		player->need_audio_dec_sorting = TRUE;
	}

	mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_VIDEO_CODEC_TYPE, &video_codec_type);
	if (video_codec_type != MM_PLAYER_CODEC_TYPE_DEFAULT) {
		LOGD("video dec sorting is required");
		player->need_video_dec_sorting = TRUE;
	}

	/* realize pipeline */
	ret = __mmplayer_gst_realize(player);
	if (ret != MM_ERROR_NONE)
		LOGE("fail to realize the player.");

	MMPLAYER_BUS_MSG_THREAD_SIGNAL(player);

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_unrealize(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	MMPLAYER_CMD_UNLOCK(player);
	/* destroy the gst bus msg thread which is created during realize.
	   this funct have to be called before getting cmd lock. */
	_mmplayer_bus_msg_thread_destroy(player);
	MMPLAYER_CMD_LOCK(player);

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_UNREALIZE);

	/* check async state transition */
	__mmplayer_check_async_state_transition(player);

	/* unrealize pipeline */
	ret = __mmplayer_gst_unrealize(player);

	if (!player->interrupted_by_resource) {
		int rm_ret = MM_ERROR_NONE;
		mmplayer_resource_type_e res_idx = MMPLAYER_RESOURCE_TYPE_MAX;

		for (res_idx = MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER; res_idx < MMPLAYER_RESOURCE_TYPE_MAX; res_idx++) {
			rm_ret = __mmplayer_release_hw_resource(player, res_idx);
			if (rm_ret != MM_ERROR_NONE)
				LOGE("failed to release [%d] resources", res_idx);
		}
	}

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_set_message_callback(MMHandleType hplayer, MMMessageCallback callback, gpointer user_param)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	return __mmplayer_gst_set_message_callback(player, callback, user_param);
}

int
_mmplayer_get_state(MMHandleType hplayer, int *state)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_RETURN_VAL_IF_FAIL(state, MM_ERROR_INVALID_ARGUMENT);

	*state = MMPLAYER_CURRENT_STATE(player);

	return MM_ERROR_NONE;
}

static int
__mmplayer_gst_set_volume_property(mmplayer_t *player, const char *prop_name)
{
	GstElement *vol_element = NULL;
	enum audio_element_id volume_elem_id = MMPLAYER_A_VOL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(prop_name, MM_ERROR_INVALID_ARGUMENT);

	/* check pipeline handle */
	if (!player->pipeline || !player->pipeline->audiobin) {
		LOGD("'%s' will be applied when audiobin is created", prop_name);

		/* NOTE : stored value will be used in create_audiobin
		 * returning MM_ERROR_NONE here makes application to able to
		 * set audio volume or mute at anytime.
		 */
		return MM_ERROR_NONE;
	}

	if (player->build_audio_offload || g_strrstr(player->ini.audiosink_element, "pulsesink"))
		volume_elem_id = MMPLAYER_A_SINK;

	vol_element = player->pipeline->audiobin[volume_elem_id].gst;
	if (!vol_element) {
		LOGE("failed to get vol element %d", volume_elem_id);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	LOGD("set '%s' property to element[%s]", prop_name, GST_ELEMENT_NAME(vol_element));

	if (!g_object_class_find_property(G_OBJECT_GET_CLASS(vol_element), prop_name)) {
		LOGE("there is no '%s' property", prop_name);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	if (!strcmp(prop_name, "volume")) {
		g_object_set(vol_element, "volume", player->sound.volume, NULL);
	} else if (!strcmp(prop_name, "mute")) {
		g_object_set(vol_element, "mute", player->sound.mute, NULL);
	} else {
		LOGE("invalid property %s", prop_name);
		return MM_ERROR_PLAYER_INTERNAL;
	}

	return MM_ERROR_NONE;
}

int
_mmplayer_set_volume(MMHandleType hplayer, float volume)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("volume = %f", volume);

	/* invalid factor range or not */
	if (volume < MM_VOLUME_FACTOR_MIN || volume > MM_VOLUME_FACTOR_MAX) {
		LOGE("Invalid volume value");
		return MM_ERROR_INVALID_ARGUMENT;
	}

	player->sound.volume = volume;

	ret = __mmplayer_gst_set_volume_property(player, "volume");

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_volume(MMHandleType hplayer, float *volume)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(volume, MM_ERROR_INVALID_ARGUMENT);

	*volume = player->sound.volume;

	LOGD("current vol = %f", *volume);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_set_mute(MMHandleType hplayer, bool mute)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("mute = %d", mute);

	player->sound.mute = mute;

	ret = __mmplayer_gst_set_volume_property(player, "mute");

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_mute(MMHandleType hplayer, bool *mute)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(mute, MM_ERROR_INVALID_ARGUMENT);

	*mute = player->sound.mute;

	LOGD("current mute = %d", *mute);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_set_audiostream_changed_cb(MMHandleType hplayer, mm_player_stream_changed_callback callback, void *user_param)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	player->audio_stream_changed_cb = callback;
	player->audio_stream_changed_cb_user_param = user_param;
	LOGD("Handle value is %p : %p", player, player->audio_stream_changed_cb);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_set_audio_decoded_cb(MMHandleType hplayer, mmplayer_audio_extract_opt_e opt, mm_player_audio_decoded_callback callback, void *user_param)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	player->audio_decoded_cb = callback;
	player->audio_decoded_cb_user_param = user_param;
	player->audio_extract_opt = opt;
	LOGD("handle: %p, cb: %p, opt: 0x%X", player, player->audio_decoded_cb, player->audio_extract_opt);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_set_video_decoded_cb(MMHandleType hplayer, mm_player_video_decoded_callback callback, void *user_param)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if (callback && !player->bufmgr)
		player->bufmgr = tbm_bufmgr_init(-1);

	player->set_mode.video_export = (callback) ? true : false;
	player->video_decoded_cb = callback;
	player->video_decoded_cb_user_param = user_param;

	LOGD("Stream cb Handle value is %p : %p, enable:%d", player, player->video_decoded_cb, player->set_mode.video_export);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_start(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gint ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_START);

	/* start pipeline */
	ret = _mmplayer_gst_start(player);
	if (ret != MM_ERROR_NONE)
		LOGE("failed to start player.");

	if ((player->streamer) && (player->streamer->buffering_state & MM_PLAYER_BUFFERING_IN_PROGRESS)) {
		LOGD("force playing start even during buffering");
		player->streamer->buffering_state = MM_PLAYER_BUFFERING_ABORT;
	}

	MMPLAYER_FLEAVE();

	return ret;
}

/* NOTE: post "not supported codec message" to application
 * when one codec is not found during AUTOPLUGGING in MSL.
 * So, it's separated with error of __mmplayer_gst_bus_msg_callback().
 * And, if any codec is not found, don't send message here.
 * Because GST_ERROR_MESSAGE is posted by other plugin internally.
 */
int
__mmplayer_handle_missed_plugin(mmplayer_t *player)
{
	MMMessageParamType msg_param;
	memset(&msg_param, 0, sizeof(MMMessageParamType));
	gboolean post_msg_direct = FALSE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("not_supported_codec = 0x%02x, can_support_codec = 0x%02x",
			player->not_supported_codec, player->can_support_codec);

	if (player->not_found_demuxer) {
		msg_param.code = MM_ERROR_PLAYER_CODEC_NOT_FOUND;
		msg_param.data = g_strdup_printf("%s", player->unlinked_demuxer_mime);

		MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);
		MMPLAYER_FREEIF(msg_param.data);

		return MM_ERROR_NONE;
	}

	if (player->not_supported_codec) {
		if (player->can_support_codec) {
			// There is one codec to play
			post_msg_direct = TRUE;
		} else {
			if (player->pipeline->audiobin) // Some content has only PCM data in container.
				post_msg_direct = TRUE;
		}

		if (post_msg_direct) {
			MMMessageParamType msg_param;
			memset(&msg_param, 0, sizeof(MMMessageParamType));

			if (player->not_supported_codec ==  MISSING_PLUGIN_AUDIO) {
				LOGW("not found AUDIO codec, posting error code to application.");

				msg_param.code = MM_ERROR_PLAYER_AUDIO_CODEC_NOT_FOUND;
				msg_param.data = g_strdup_printf("%s", player->unlinked_audio_mime);
			} else if (player->not_supported_codec ==  MISSING_PLUGIN_VIDEO) {
				LOGW("not found VIDEO codec, posting error code to application.");

				msg_param.code = MM_ERROR_PLAYER_VIDEO_CODEC_NOT_FOUND;
				msg_param.data = g_strdup_printf("%s", player->unlinked_video_mime);
			}

			MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);

			MMPLAYER_FREEIF(msg_param.data);

			return MM_ERROR_NONE;
		} else {
			// no any supported codec case
			LOGW("not found any codec, posting error code to application.");

			if (player->not_supported_codec ==  MISSING_PLUGIN_AUDIO) {
				msg_param.code = MM_ERROR_PLAYER_AUDIO_CODEC_NOT_FOUND;
				msg_param.data = g_strdup_printf("%s", player->unlinked_audio_mime);
			} else {
				msg_param.code = MM_ERROR_PLAYER_CODEC_NOT_FOUND;
				msg_param.data = g_strdup_printf("%s, %s", player->unlinked_video_mime, player->unlinked_audio_mime);
			}

			MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);

			MMPLAYER_FREEIF(msg_param.data);
		}
	}

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

static void
__mmplayer_check_pipeline(mmplayer_t *player)
{
	GstState element_state = GST_STATE_VOID_PENDING;
	GstState element_pending_state = GST_STATE_VOID_PENDING;
	gint timeout = 0;
	int ret = MM_ERROR_NONE;

	if (!player->gapless.reconfigure)
		return;

	LOGW("pipeline is under construction.");

	MMPLAYER_PLAYBACK_LOCK(player);
	MMPLAYER_PLAYBACK_UNLOCK(player);

	timeout = MMPLAYER_STATE_CHANGE_TIMEOUT(player);

	/* wait for state transition */
	ret = gst_element_get_state(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, &element_state, &element_pending_state, timeout * GST_SECOND);
	if (ret == GST_STATE_CHANGE_FAILURE)
		LOGE("failed to change pipeline state within %d sec", timeout);
}

/* NOTE : it should be able to call 'stop' anytime*/
int
_mmplayer_stop(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_STOP);

	/* check pipline building state */
	__mmplayer_check_pipeline(player);
	__mmplayer_reset_gapless_state(player);

	/* NOTE : application should not wait for EOS after calling STOP */
	_mmplayer_cancel_eos_timer(player);

	/* reset */
	player->seek_state = MMPLAYER_SEEK_NONE;

	/* stop pipeline */
	ret = _mmplayer_gst_stop(player);

	if (ret != MM_ERROR_NONE)
		LOGE("failed to stop player.");

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_pause(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gint64 pos_nsec = 0;
	gboolean async = FALSE;
	gint ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_PAUSE);

	/* check pipline building state */
	__mmplayer_check_pipeline(player);

	switch (MMPLAYER_CURRENT_STATE(player)) {
	case MM_PLAYER_STATE_READY:
		{
			/* check prepare async or not.
			 * In the case of streaming playback, it's recommned to avoid blocking wait.
			 */
			mm_attrs_get_int_by_name(player->attrs, "profile_prepare_async", &async);
			LOGD("prepare working mode : %s", (async ? "async" : "sync"));

			/* Changing back sync of rtspsrc to async */
			if (MMPLAYER_IS_RTSP_STREAMING(player)) {
				LOGD("async prepare working mode for rtsp");
				async = TRUE;
			}
		}
		break;

	case MM_PLAYER_STATE_PLAYING:
		{
			/* NOTE : store current point to overcome some bad operation
			*(returning zero when getting current position in paused state) of some
			* elements
			*/
			if (!gst_element_query_position(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &pos_nsec))
				LOGW("getting current position failed in paused");

			player->last_position = pos_nsec;

			/* For rtspsrc stack , npt-start value coming from server is used for finding the current position.
			   But when a rtsp clip (especially from Youtube Desktop View) is paused and kept for sometime,npt-start is still increasing.
			   This causes problem is position calculation during normal pause resume scenarios also.
			   Currently during pause , we are sending the current position to rtspsrc module for position saving. */
			if ((MMPLAYER_IS_RTSP_STREAMING(player)) &&
				(_mmplayer_get_stream_service_type(player) == STREAMING_SERVICE_VOD)) {
				g_object_set(player->pipeline->mainbin[MMPLAYER_M_SRC].gst, "resume-position", player->last_position, NULL);
			}
		}
		break;
	}

	if (MMPLAYER_IS_MS_BUFF_SRC(player)) {
		LOGD("doing async pause in case of ms buff src");
		async = TRUE;
	}

	/* pause pipeline */
	ret = _mmplayer_gst_pause(player, async);

	if (ret != MM_ERROR_NONE)
		LOGE("failed to pause player. ret : 0x%x", ret);

	if (MMPLAYER_PREV_STATE(player) == MM_PLAYER_STATE_READY && MMPLAYER_CURRENT_STATE(player) == MM_PLAYER_STATE_PAUSED) {
		if (_mmplayer_update_video_overlay_param(player, "display_rotation") != MM_ERROR_NONE)
			LOGE("failed to update display_rotation");
	}

	MMPLAYER_FLEAVE();

	return ret;
}

/* in case of streaming, pause could take long time.*/
int
_mmplayer_abort_pause(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
						player->pipeline &&
						player->pipeline->mainbin,
						MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("set the pipeline state to READY");

	/* set state to READY */
	ret = _mmplayer_gst_set_state(player, player->pipeline->mainbin[MMPLAYER_M_PIPE].gst,
						GST_STATE_READY, FALSE, MMPLAYER_STATE_CHANGE_TIMEOUT(player));
	if (ret != MM_ERROR_NONE) {
		LOGE("fail to change state to READY");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	LOGD("succeeded in changing state to READY");
	return ret;
}

int
_mmplayer_resume(MMHandleType hplayer)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;
	gboolean async = FALSE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if ((MMPLAYER_IS_RTSP_STREAMING(player))) {
		if (player->is_external_subtitle_added_now) { /* after setting external subtitle, seeking and buffering is in progress. */
			player->pending_resume = TRUE; /* will be resumed after finishing the buffering. */
			return ret;
		}

		/* Changing back sync mode rtspsrc to async */
		LOGD("async resume for rtsp case");
		async = TRUE;
	}

	/* check current state */
	MMPLAYER_CHECK_STATE(player, MMPLAYER_COMMAND_RESUME);

	ret = _mmplayer_gst_resume(player, async);
	if (ret != MM_ERROR_NONE)
		LOGE("failed to resume player.");

	if ((player->streamer) && (player->streamer->buffering_state & MM_PLAYER_BUFFERING_IN_PROGRESS)) {
		LOGD("force resume even during buffering");
		player->streamer->buffering_state = MM_PLAYER_BUFFERING_ABORT;
	}

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_set_playspeed(MMHandleType hplayer, float rate, bool streaming)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	gint64 pos_nsec = 0;
	int ret = MM_ERROR_NONE;
	bool mute = false;
	signed long long start = 0, stop = 0;
	mmplayer_state_e current_state = MM_PLAYER_STATE_NONE;
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(streaming || !MMPLAYER_IS_STREAMING(player), MM_ERROR_NOT_SUPPORT_API);

	/* The sound of video is not supported under 0.0 and over 2.0. */
	if (rate >= TRICK_PLAY_MUTE_THRESHOLD_MAX || rate < TRICK_PLAY_MUTE_THRESHOLD_MIN) {
		if (player->can_support_codec & FOUND_PLUGIN_VIDEO)
			mute = true;
	}
	_mmplayer_set_mute(hplayer, mute);

	if (player->playback_rate == rate)
		return MM_ERROR_NONE;

	/* If the position is reached at start potion during fast backward, EOS is posted.
	 * So, This EOS have to be classified with it which is posted at reaching the end of stream.
	 */
	player->playback_rate = rate;

	current_state = MMPLAYER_CURRENT_STATE(player);

	if (current_state != MM_PLAYER_STATE_PAUSED)
		ret = gst_element_query_position(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &pos_nsec);

	LOGD("pos_msec = %"GST_TIME_FORMAT" and ret = %d and state = %d", GST_TIME_ARGS(pos_nsec), ret, current_state);

	if ((current_state == MM_PLAYER_STATE_PAUSED)
		|| (!ret) /*|| (player->last_position != 0 && pos_msec == 0)*/) {
		LOGW("returning last point : %"G_GINT64_FORMAT, player->last_position);
		pos_nsec = player->last_position;
	}

	if (rate >= 0) {
		start = pos_nsec;
		stop = GST_CLOCK_TIME_NONE;
	} else {
		start = GST_CLOCK_TIME_NONE;
		stop = pos_nsec;
	}

	if (!_mmplayer_gst_seek(player, player->pipeline->mainbin[MMPLAYER_M_PIPE].gst,
				player->playback_rate,
				GST_FORMAT_TIME,
				(GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE),
				GST_SEEK_TYPE_SET, start,
				GST_SEEK_TYPE_SET, stop)) {
		LOGE("failed to set speed playback");
		return MM_ERROR_PLAYER_SEEK;
	}

	LOGD("succeeded to set speed playback as %0.1f", rate);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;;
}

int
_mmplayer_set_position(MMHandleType hplayer, gint64 position)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* check pipline building state */
	__mmplayer_check_pipeline(player);

	ret = _mmplayer_gst_set_position(player, position, FALSE);

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_get_duration(MMHandleType hplayer, gint64 *duration)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(duration, MM_ERROR_COMMON_INVALID_ARGUMENT);

	if (g_strrstr(player->type, "video/mpegts"))
		__mmplayer_update_duration_value(player);

	*duration = player->duration;
	return ret;
}

int
_mmplayer_get_buffer_position(MMHandleType hplayer, int *start_pos, int *end_pos)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	ret = _mmplayer_gst_get_buffer_position(player, start_pos, end_pos);

	return ret;
}

int
_mmplayer_adjust_subtitle_postion(MMHandleType hplayer, int position)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	ret = __mmplayer_gst_adjust_subtitle_position(player, position);

	MMPLAYER_FLEAVE();

	return ret;
}

static gboolean
__mmplayer_is_midi_type(gchar *str_caps)
{
	if ((g_strrstr(str_caps, "audio/midi")) ||
		(g_strrstr(str_caps, "application/x-gst_ff-mmf")) ||
		(g_strrstr(str_caps, "application/x-smaf")) ||
		(g_strrstr(str_caps, "audio/x-imelody")) ||
		(g_strrstr(str_caps, "audio/mobile-xmf")) ||
		(g_strrstr(str_caps, "audio/xmf")) ||
		(g_strrstr(str_caps, "audio/mxmf"))) {
		LOGD("midi");
		return TRUE;
	}

	return FALSE;
}

static gboolean
__mmplayer_is_only_mp3_type(gchar *str_caps)
{
	if (g_strrstr(str_caps, "application/x-id3") ||
		(g_strrstr(str_caps, "audio/mpeg") && g_strrstr(str_caps, "mpegversion=(int)1")))
		return TRUE;
	return FALSE;
}

static void
__mmplayer_set_audio_attrs(mmplayer_t *player, GstCaps *caps)
{
	GstStructure *caps_structure = NULL;
	gint samplerate = 0;
	gint channels = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && caps);

	caps_structure = gst_caps_get_structure(caps, 0);

	/* set stream information */
	gst_structure_get_int(caps_structure, "rate", &samplerate);
	gst_structure_get_int(caps_structure, "channels", &channels);

	mm_player_set_attribute((MMHandleType)player, NULL,
			"content_audio_samplerate", samplerate,
			"content_audio_channels", channels, NULL);

	LOGD("audio samplerate : %d	channels : %d", samplerate, channels);
}

static void
__mmplayer_update_content_type_info(mmplayer_t *player)
{
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player && player->type);

	if (__mmplayer_is_midi_type(player->type)) {
		player->bypass_audio_effect = TRUE;
		return;
	}

	if (!player->streamer) {
		LOGD("no need to check streaming type");
		return;
	}

	if (g_strrstr(player->type, "application/x-hls")) {
		/* If it can't know exact type when it parses uri because of redirection case,
		 * it will be fixed by typefinder or when doing autoplugging.
		 */
		player->profile.uri_type = MM_PLAYER_URI_TYPE_HLS;
		player->streamer->is_adaptive_streaming = TRUE;
	} else if (g_strrstr(player->type, "application/dash+xml")) {
		player->profile.uri_type = MM_PLAYER_URI_TYPE_DASH;
		player->streamer->is_adaptive_streaming = TRUE;
	}

	/* in case of TS, fixed buffering mode should be used because player can not get exact duration time */
	if ((player->streamer->is_adaptive_streaming) || (g_strrstr(player->type, "video/mpegts"))) {
		player->streamer->buffering_req.mode = MM_PLAYER_BUFFERING_MODE_FIXED;

		if (player->streamer->buffering_req.rebuffer_time <= MIN_BUFFERING_TIME) { /* if user did not set the rebuffer value */
			if (player->streamer->is_adaptive_streaming)
				player->streamer->buffering_req.rebuffer_time = DEFAULT_ADAPTIVE_REBUFFER_TIME;
			else
				player->streamer->buffering_req.rebuffer_time = DEFAULT_REBUFFERING_TIME;
		}
	}

	LOGD("uri type : %d, %d", player->profile.uri_type, player->streamer->buffering_req.rebuffer_time);
	MMPLAYER_FLEAVE();
}

void
_mmplayer_typefind_have_type(GstElement *tf, guint probability,
	GstCaps *caps, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	GstPad *pad = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player && tf && caps);

	/* store type string */
	MMPLAYER_FREEIF(player->type);
	player->type = gst_caps_to_string(caps);
	if (player->type)
		LOGD("[handle: %p] media type %s found, probability %d%% / %d",
				player, player->type, probability, gst_caps_get_size(caps));

	if ((!MMPLAYER_IS_RTSP_STREAMING(player)) &&
		(g_strrstr(player->type, "audio/x-raw-int"))) {
		LOGE("not support media format");

		if (player->msg_posted == FALSE) {
			MMMessageParamType msg_param;
			memset(&msg_param, 0, sizeof(MMMessageParamType));

			msg_param.code = MM_ERROR_PLAYER_NOT_SUPPORTED_FORMAT;
			MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);

			/* don't post more if one was sent already */
			player->msg_posted = TRUE;
		}
		return;
	}

	__mmplayer_update_content_type_info(player);

	pad = gst_element_get_static_pad(tf, "src");
	if (!pad) {
		LOGE("fail to get typefind src pad.");
		return;
	}

	if (!_mmplayer_gst_create_decoder(player, pad, caps)) {
		gboolean async = FALSE;
		LOGE("failed to autoplug %s", player->type);

		mm_attrs_get_int_by_name(player->attrs, "profile_prepare_async", &async);

		if (async && player->msg_posted == FALSE)
			__mmplayer_handle_missed_plugin(player);

	}

	gst_object_unref(GST_OBJECT(pad));

	MMPLAYER_FLEAVE();

	return;
}

GstElement *
_mmplayer_gst_make_decodebin(mmplayer_t *player)
{
	GstElement *decodebin = NULL;

	MMPLAYER_FENTER();

	/* create decodebin */
	decodebin = gst_element_factory_make("decodebin", NULL);

	if (!decodebin) {
		LOGE("fail to create decodebin");
		goto ERROR;
	}

	/* raw pad handling signal */
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "pad-added",
						G_CALLBACK(_mmplayer_gst_decode_pad_added), (gpointer)player);

	/* no-more-pad pad handling signal */
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "no-more-pads",
						G_CALLBACK(__mmplayer_gst_decode_no_more_pads), (gpointer)player);

	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "pad-removed",
						G_CALLBACK(__mmplayer_gst_decode_pad_removed), (gpointer)player);

	/* This signal is emitted when a pad for which there is no further possible
	   decoding is added to the decodebin.*/
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "unknown-type",
						G_CALLBACK(__mmplayer_gst_decode_unknown_type), (gpointer)player);

	/* This signal is emitted whenever decodebin finds a new stream. It is emitted
	   before looking for any elements that can handle that stream.*/
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "autoplug-continue",
						G_CALLBACK(__mmplayer_gst_decode_autoplug_continue), (gpointer)player);

	if (player->need_video_dec_sorting || player->need_audio_dec_sorting)
		_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "autoplug-sort",
							G_CALLBACK(_mmplayer_gst_decode_autoplug_sort), (gpointer)player);

	/* This signal is emitted whenever decodebin finds a new stream. It is emitted
	   before looking for any elements that can handle that stream.*/
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "autoplug-select",
						G_CALLBACK(_mmplayer_gst_decode_autoplug_select), (gpointer)player);

	/* This signal is emitted once decodebin has finished decoding all the data.*/
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "drained",
						G_CALLBACK(__mmplayer_gst_decode_drained), (gpointer)player);

	/* This signal is emitted when a element is added to the bin.*/
	_mmplayer_add_signal_connection(player, G_OBJECT(decodebin), MM_PLAYER_SIGNAL_TYPE_AUTOPLUG, "element-added",
						G_CALLBACK(_mmplayer_gst_element_added), (gpointer)player);

ERROR:
	return decodebin;
}

static GstElement *
__mmplayer_gst_make_queue2(mmplayer_t *player)
{
	GstElement *queue2 = NULL;
	gint64 dur_bytes = 0L;
	mmplayer_gst_element_t *mainbin = NULL;
	muxed_buffer_type_e type = MUXED_BUFFER_TYPE_MEM_QUEUE;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->mainbin, NULL);

	mainbin = player->pipeline->mainbin;

	queue2 = gst_element_factory_make("queue2", "queue2");
	if (!queue2) {
		LOGE("failed to create buffering queue element");
		return NULL;
	}

	if (!gst_element_query_duration(mainbin[MMPLAYER_M_SRC].gst, GST_FORMAT_BYTES, &dur_bytes))
		LOGW("failed to get duration from source %s", GST_ELEMENT_NAME(mainbin[MMPLAYER_M_SRC].gst));

	LOGD("dur_bytes = %"G_GINT64_FORMAT, dur_bytes);

	/* NOTE : in case of ts streaming, player could not get the correct duration info *
	 *		  skip the pull mode(file or ring buffering) setting. */
	if (dur_bytes > 0) {
		if (!g_strrstr(player->type, "video/mpegts")) {
			type = MUXED_BUFFER_TYPE_MEM_RING_BUFFER;
			player->streamer->ring_buffer_size = player->ini.http_ring_buffer_size;
		}
	} else {
		dur_bytes = 0;
	}

	_mm_player_streaming_set_queue2(player->streamer,
									queue2,
									FALSE,
									type,
									(guint64)dur_bytes); /* no meaning at the moment */

	return queue2;
}

gboolean
_mmplayer_gst_create_decoder(mmplayer_t *player, GstPad *srcpad, const GstCaps *caps)
{
	mmplayer_gst_element_t *mainbin = NULL;
	GstElement *decodebin = NULL;
	GstElement *queue2 = NULL;
	GstPad *sinkpad = NULL;
	GstPad *qsrcpad = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->mainbin, FALSE);

	mainbin = player->pipeline->mainbin;

	if (MMPLAYER_IS_HTTP_STREAMING(player)) {

		if (mainbin[MMPLAYER_M_MUXED_S_BUFFER].gst) {
			LOGW("need to check: muxed buffer is not null");
		}

		queue2 = __mmplayer_gst_make_queue2(player);
		if (!queue2) {
			LOGE("failed to make queue2");
			goto ERROR;
		}

		if (!gst_bin_add(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), queue2)) {
			LOGE("failed to add buffering queue");
			goto ERROR;
		}

		sinkpad = gst_element_get_static_pad(queue2, "sink");
		qsrcpad = gst_element_get_static_pad(queue2, "src");

		if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
			LOGE("failed to link [%s:%s]-[%s:%s]",
						GST_DEBUG_PAD_NAME(srcpad), GST_DEBUG_PAD_NAME(sinkpad));
			goto ERROR;
		}

		if (gst_element_sync_state_with_parent(queue2) == GST_STATE_CHANGE_FAILURE) {
			LOGE("failed to sync queue2 state with parent");
			goto ERROR;
		}

		mainbin[MMPLAYER_M_MUXED_S_BUFFER].id = MMPLAYER_M_MUXED_S_BUFFER;
		mainbin[MMPLAYER_M_MUXED_S_BUFFER].gst = queue2;

		srcpad = qsrcpad;

		gst_object_unref(GST_OBJECT(sinkpad));
		sinkpad = NULL;
	}

	/* create decodebin */
	decodebin = _mmplayer_gst_make_decodebin(player);
	if (!decodebin) {
		LOGE("failed to make decodebin");
		goto ERROR;
	}

	if (!gst_bin_add(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), decodebin)) {
		LOGE("failed to add decodebin");
		goto ERROR;
	}

	/* to force caps on the decodebin element and avoid reparsing stuff by
	* typefind. It also avoids a deadlock in the way typefind activates pads in
	* the state change */
	g_object_set(decodebin, "sink-caps", caps, NULL);

	sinkpad = gst_element_get_static_pad(decodebin, "sink");

	if (GST_PAD_LINK_OK != gst_pad_link(srcpad, sinkpad)) {
		LOGE("failed to link [%s:%s]-[%s:%s]",
				GST_DEBUG_PAD_NAME(srcpad), GST_DEBUG_PAD_NAME(sinkpad));
		goto ERROR;
	}

	gst_object_unref(GST_OBJECT(sinkpad));
	sinkpad = NULL;
	gst_object_unref(GST_OBJECT(qsrcpad));
	qsrcpad = NULL;

	mainbin[MMPLAYER_M_AUTOPLUG].id = MMPLAYER_M_AUTOPLUG;
	mainbin[MMPLAYER_M_AUTOPLUG].gst = decodebin;

	/* set decodebin property about buffer in streaming playback. *
	 * in case of HLS/DASH, it does not need to have big buffer   *
	 * because it is kind of adaptive streaming.                  */
	if (MMPLAYER_IS_HTTP_STREAMING(player) || MMPLAYER_IS_HTTP_LIVE_STREAMING(player) || MMPLAYER_IS_DASH_STREAMING(player)) {
		gint init_buffering_time = DEFAULT_PREBUFFERING_TIME;
		gint high_percent = 0;

		if (player->streamer->buffering_req.prebuffer_time > MIN_BUFFERING_TIME)
			init_buffering_time = player->streamer->buffering_req.prebuffer_time;

		high_percent = (gint)ceil((gdouble)(init_buffering_time * 100) / MAX_BUFFER_SIZE_TIME);

		LOGD("buffering time %d, per: 1~%d", init_buffering_time, high_percent);

		g_object_set(G_OBJECT(decodebin), "use-buffering", TRUE,
											"high-percent", high_percent,
											"max-size-bytes", MAX_BUFFER_SIZE_BYTES,
											"max-size-time", (guint64)(MAX_BUFFER_SIZE_TIME * GST_MSECOND),
											"max-size-buffers", 0, NULL);  // disable or automatic
	}

	if (gst_element_sync_state_with_parent(decodebin) == GST_STATE_CHANGE_FAILURE) {
		LOGE("failed to sync decodebin state with parent");
		goto ERROR;
	}

	MMPLAYER_FLEAVE();

	return TRUE;

ERROR:

	if (sinkpad)
		gst_object_unref(GST_OBJECT(sinkpad));

	if (qsrcpad)
		gst_object_unref(GST_OBJECT(qsrcpad));

	if (queue2) {
		/* NOTE : Trying to dispose element queue0, but it is in READY instead of the NULL state.
		 * You need to explicitly set elements to the NULL state before
		 * dropping the final reference, to allow them to clean up.
		 */
		gst_element_set_state(queue2, GST_STATE_NULL);

		/* And, it still has a parent "player".
		 * You need to let the parent manage the object instead of unreffing the object directly.
		 */
		gst_bin_remove(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), queue2);
		gst_object_unref(queue2);
		queue2 = NULL;
	}

	if (decodebin) {
		/* NOTE : Trying to dispose element queue0, but it is in READY instead of the NULL state.
		 * You need to explicitly set elements to the NULL state before
		 * dropping the final reference, to allow them to clean up.
		 */
		gst_element_set_state(decodebin, GST_STATE_NULL);

		/* And, it still has a parent "player".
		 * You need to let the parent manage the object instead of unreffing the object directly.
		 */

		gst_bin_remove(GST_BIN(mainbin[MMPLAYER_M_PIPE].gst), decodebin);
		gst_object_unref(decodebin);
		decodebin = NULL;
	}

	return FALSE;
}

static int
__mmplayer_check_not_supported_codec(mmplayer_t *player, const gchar *factory_class, const gchar *mime)
{
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(mime, MM_ERROR_INVALID_ARGUMENT);

	LOGD("class : %s, mime : %s", factory_class, mime);

	/* add missing plugin */
	/* NOTE : msl should check missing plugin for image mime type.
	 * Some motion jpeg clips can have playable audio track.
	 * So, msl have to play audio after displaying popup written video format not supported.
	 */
	if (!(player->pipeline->mainbin[MMPLAYER_M_DEMUX].gst)) {
		if (!(player->can_support_codec | player->videodec_linked | player->audiodec_linked)) {
			LOGD("not found demuxer");
			player->not_found_demuxer = TRUE;
			player->unlinked_demuxer_mime = g_strdup_printf("%s", mime);

			goto DONE;
		}
	}

	if (!g_strrstr(factory_class, "Demuxer")) {
		if ((g_str_has_prefix(mime, "video")) || (g_str_has_prefix(mime, "image"))) {
			LOGD("can support codec=0x%X, vdec_linked=%d, adec_linked=%d",
				player->can_support_codec, player->videodec_linked, player->audiodec_linked);

			/* check that clip have multi tracks or not */
			if ((player->can_support_codec & FOUND_PLUGIN_VIDEO) && (player->videodec_linked)) {
				LOGD("video plugin is already linked");
			} else {
				LOGW("add VIDEO to missing plugin");
				player->not_supported_codec |= MISSING_PLUGIN_VIDEO;
				player->unlinked_video_mime = g_strdup_printf("%s", mime);
			}
		} else if (g_str_has_prefix(mime, "audio")) {
			if ((player->can_support_codec & FOUND_PLUGIN_AUDIO) && (player->audiodec_linked)) {
				LOGD("audio plugin is already linked");
			} else {
				LOGW("add AUDIO to missing plugin");
				player->not_supported_codec |= MISSING_PLUGIN_AUDIO;
				player->unlinked_audio_mime = g_strdup_printf("%s", mime);
			}
		}
	}

DONE:
	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

static void
__mmplayer_pipeline_complete(GstElement *decodebin,  gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player);

	/* remove fakesink. */
	if (!_mmplayer_gst_remove_fakesink(player,
				&player->pipeline->mainbin[MMPLAYER_M_SRC_FAKESINK])) {
		/* NOTE : __mmplayer_pipeline_complete() can be called several time. because
		 * signaling mechanism(pad-added, no-more-pad, new-decoded-pad) from various
		 * source element are not same. To overcome this situation, this function will called
		 * several places and several times. Therefore, this is not an error case.
		 */
		return;
	}

	LOGD("[handle: %p] pipeline has completely constructed", player);

	if ((player->msg_posted == FALSE) &&
		(player->cmd >= MMPLAYER_COMMAND_START))
		__mmplayer_handle_missed_plugin(player);

	MMPLAYER_GENERATE_DOT_IF_ENABLED(player, "pipeline-status-complete");
}

static int
__mmplayer_check_profile(void)
{
	char *profileName;
	static int profile_tv = -1;

	if (__builtin_expect(profile_tv != -1, 1))
		return profile_tv;

	system_info_get_platform_string("http://tizen.org/feature/profile", &profileName);
	switch (*profileName) {
	case 't':
	case 'T':
		profile_tv = 1;
		break;
	default:
		profile_tv = 0;
		break;
	}
	free(profileName);

	return profile_tv;
}

static gboolean
__mmplayer_get_next_uri(mmplayer_t *player)
{
	mmplayer_parse_profile_t profile;
	gint uri_idx = 0;
	guint num_of_list = 0;
	char *uri = NULL;

	num_of_list = g_list_length(player->uri_info.uri_list);
	uri_idx = player->uri_info.uri_idx;

	LOGD("num of uri list = %d, current uri idx %d", num_of_list, uri_idx);
	for (uri_idx++; uri_idx < num_of_list; uri_idx++) {
		uri = g_list_nth_data(player->uri_info.uri_list, uri_idx);
		if (!uri) {
			LOGW("next uri does not exist");
			continue;
		}

		if (_mmplayer_parse_profile((const char *)uri, NULL, &profile) != MM_ERROR_NONE) {
			LOGE("failed to parse profile");
			continue;
		}

		if ((profile.uri_type != MM_PLAYER_URI_TYPE_FILE) &&
			(profile.uri_type != MM_PLAYER_URI_TYPE_URL_HTTP)) {
			LOGW("uri type is not supported(%d)", profile.uri_type);
			continue;
		}

		LOGD("success to find next uri %d", uri_idx);
		break;
	}

	if (!uri || uri_idx == num_of_list) {
		LOGE("failed to find next uri");
		return FALSE;
	}

	player->uri_info.uri_idx = uri_idx;
	if (mm_player_set_attribute((MMHandleType)player, NULL,
			"profile_uri", uri, strlen(uri), NULL) != MM_ERROR_NONE) {
		LOGE("failed to set attribute");
		return FALSE;
	}

	SECURE_LOGD("next playback uri: %s", uri);
	return TRUE;
}

static gboolean
__mmplayer_verify_gapless_play_path(mmplayer_t *player)
{
#define REPEAT_COUNT_INFINITE -1
#define REPEAT_COUNT_MIN 2
#define ORIGINAL_URI_ONLY 1

	MMHandleType attrs = 0;
	gint video = 0;
	gint count = 0;
	gint gapless = 0;
	guint num_of_uri = 0;
	int profile_tv = -1;

	MMPLAYER_FENTER();

	LOGD("checking for gapless play option");

	if (player->build_audio_offload) {
		LOGE("offload path is not supportable.");
		goto ERROR;
	}

	if (player->pipeline->textbin) {
		LOGE("subtitle path is enabled. gapless play is not supported.");
		goto ERROR;
	}

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("fail to get attributes.");
		goto ERROR;
	}

	mm_attrs_multiple_get(player->attrs, NULL,
			"content_video_found", &video,
			"profile_play_count", &count,
			MM_PLAYER_GAPLESS_MODE, &gapless, NULL);

	/* gapless playback is not supported in case of video at TV profile. */
	profile_tv = __mmplayer_check_profile();
	if (profile_tv && video) {
		LOGW("not support video gapless playback");
		goto ERROR;
	}

	/* check repeat count in case of audio */
	if (!gapless &&
		(video || (count != REPEAT_COUNT_INFINITE && count < REPEAT_COUNT_MIN))) {
		LOGW("gapless is disabled");
		goto ERROR;
	}

	num_of_uri = g_list_length(player->uri_info.uri_list);

	LOGD("repeat count = %d, num_of_list = %d", count, num_of_uri);

	if (num_of_uri == ORIGINAL_URI_ONLY) {
		/* audio looping path */
		if (count >= REPEAT_COUNT_MIN) {
			/* decrease play count */
			/* we succeeded to rewind. update play count and then wait for next EOS */
			count--;
			mm_player_set_attribute((MMHandleType)player, NULL, "profile_play_count", count, NULL);
		} else if (count != REPEAT_COUNT_INFINITE) {
			LOGD("there is no next uri and no repeat");
			goto ERROR;
		}
		LOGD("looping cnt %d", count);
	} else {
		/* gapless playback path */
		if (!__mmplayer_get_next_uri(player)) {
			LOGE("failed to get next uri");
			goto ERROR;
		}
	}
	return TRUE;

ERROR:
	LOGE("unable to play gapless path. EOS will be posted soon");
	return FALSE;
}

static gboolean
__mmplayer_deactivate_selector(mmplayer_t *player, mmplayer_track_type_e type)
{
	mmplayer_selector_t *selector = &player->selector[type];
	mmplayer_gst_element_t *sinkbin = NULL;
	main_element_id_e selectorId = MMPLAYER_M_NUM;
	main_element_id_e sinkId = MMPLAYER_M_NUM;
	GstPad *srcpad = NULL;
	GstPad *sinkpad = NULL;
	gboolean send_notice = FALSE;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);

	LOGD("type %d", type);

	switch (type) {
	case MM_PLAYER_TRACK_TYPE_AUDIO:
		selectorId = MMPLAYER_M_A_INPUT_SELECTOR;
		sinkId = MMPLAYER_A_BIN;
		sinkbin = player->pipeline->audiobin;
		break;
	case MM_PLAYER_TRACK_TYPE_VIDEO:
		selectorId = MMPLAYER_M_V_INPUT_SELECTOR;
		sinkId = MMPLAYER_V_BIN;
		sinkbin = player->pipeline->videobin;
		send_notice = TRUE;
		break;
	case MM_PLAYER_TRACK_TYPE_TEXT:
		selectorId = MMPLAYER_M_T_INPUT_SELECTOR;
		sinkId = MMPLAYER_T_BIN;
		sinkbin = player->pipeline->textbin;
		break;
	default:
		LOGE("requested type is not supportable");
		return FALSE;
		break;
	}

	if (player->pipeline->mainbin[selectorId].gst) {
		gint n;

		srcpad = gst_element_get_static_pad(player->pipeline->mainbin[selectorId].gst, "src");

		if (selector->event_probe_id != 0)
			gst_pad_remove_probe(srcpad, selector->event_probe_id);
		selector->event_probe_id = 0;

		if ((sinkbin) && (sinkbin[sinkId].gst)) {
			sinkpad = gst_element_get_static_pad(sinkbin[sinkId].gst, "sink");

			if (srcpad && sinkpad) {
				/* after getting drained signal there is no data flows, so no need to do pad_block */
				LOGD("unlink %s:%s, %s:%s", GST_DEBUG_PAD_NAME(srcpad), GST_DEBUG_PAD_NAME(sinkpad));
				gst_pad_unlink(srcpad, sinkpad);

				/* send custom event to sink pad to handle it at video sink */
				if (send_notice) {
					LOGD("send custom event to sinkpad");
					GstStructure *s = gst_structure_new_empty("tizen/flush-buffer");
					GstEvent *event = gst_event_new_custom(GST_EVENT_CUSTOM_DOWNSTREAM, s);
					gst_pad_send_event(sinkpad, event);
				}
			}

			gst_object_unref(sinkpad);
			sinkpad = NULL;
		}
		gst_object_unref(srcpad);
		srcpad = NULL;

		LOGD("selector release");

		/* release and unref requests pad from the selector */
		for (n = 0; n < selector->channels->len; n++) {
			GstPad *sinkpad = g_ptr_array_index(selector->channels, n);
			gst_element_release_request_pad((player->pipeline->mainbin[selectorId].gst), sinkpad);
		}
		g_ptr_array_set_size(selector->channels, 0);

		gst_element_set_state(player->pipeline->mainbin[selectorId].gst, GST_STATE_NULL);
		gst_bin_remove(GST_BIN_CAST(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst), player->pipeline->mainbin[selectorId].gst);

		player->pipeline->mainbin[selectorId].gst = NULL;
		selector = NULL;
	}

	return TRUE;
}

static void
__mmplayer_deactivate_old_path(mmplayer_t *player)
{
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	if ((!__mmplayer_deactivate_selector(player, MM_PLAYER_TRACK_TYPE_AUDIO)) ||
		(!__mmplayer_deactivate_selector(player, MM_PLAYER_TRACK_TYPE_VIDEO)) ||
		(!__mmplayer_deactivate_selector(player, MM_PLAYER_TRACK_TYPE_TEXT))) {
		LOGE("deactivate selector error");
		goto ERROR;
	}

	_mmplayer_track_destroy(player);
	__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_AUTOPLUG);

	if (player->streamer) {
		_mm_player_streaming_initialize(player->streamer, FALSE);
		_mm_player_streaming_destroy(player->streamer);
		player->streamer = NULL;
	}

	MMPLAYER_PLAYBACK_LOCK(player);
	MMPLAYER_GAPLESS_PLAY_THREAD_SIGNAL(player);

	MMPLAYER_FLEAVE();
	return;

ERROR:

	if (!player->msg_posted) {
		MMMessageParamType msg = {0,};

		/*post error*/
		msg.code = MM_ERROR_PLAYER_INTERNAL;
		LOGE("gapless_uri_play> deactivate error");

		MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg);
		player->msg_posted = TRUE;
	}
	return;
}

int
_mmplayer_set_uri(MMHandleType hplayer, const char *uri)
{
	int result = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(uri, MM_ERROR_INVALID_ARGUMENT);

	if (mm_player_set_attribute(hplayer, NULL,
			"profile_uri", uri, strlen(uri), NULL) != MM_ERROR_NONE) {
		LOGE("failed to set attribute");
		result = MM_ERROR_PLAYER_INTERNAL;
	} else {
		if (_mmplayer_set_next_uri(hplayer, uri, TRUE) != MM_ERROR_NONE)
			LOGE("failed to add the original uri in the uri list.");
	}

	MMPLAYER_FLEAVE();
	return result;
}

int
_mmplayer_set_next_uri(MMHandleType hplayer, const char *uri, bool is_first_path)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	guint num_of_list = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(uri, MM_ERROR_INVALID_ARGUMENT);

	if (player->pipeline && player->pipeline->textbin) {
		LOGE("subtitle path is enabled.");
		return MM_ERROR_PLAYER_INVALID_STATE;
	}

	num_of_list = g_list_length(player->uri_info.uri_list);

	if (is_first_path) {
		if (num_of_list == 0) {
			player->uri_info.uri_list = g_list_append(player->uri_info.uri_list, g_strdup(uri));
			SECURE_LOGD("add original path : %s", uri);
		} else {
			player->uri_info.uri_list = g_list_delete_link(player->uri_info.uri_list, g_list_nth(player->uri_info.uri_list, 0));
			player->uri_info.uri_list = g_list_insert(player->uri_info.uri_list, g_strdup(uri), 0);

			SECURE_LOGD("change original path : %s", uri);
		}
	} else {
		MMHandleType attrs = 0;
		attrs = MMPLAYER_GET_ATTRS(player);

		if (num_of_list == 0) {
			char *original_uri = NULL;

			if (attrs) {
				mm_attrs_get_string_by_name(attrs, "profile_uri", &original_uri);

				if (!original_uri) {
					LOGE("there is no original uri.");
					return MM_ERROR_PLAYER_INVALID_STATE;
				}

				player->uri_info.uri_list = g_list_append(player->uri_info.uri_list, g_strdup(original_uri));
				player->uri_info.uri_idx = 0;

				SECURE_LOGD("add original path at first : %s", original_uri);
			}
		}

		player->uri_info.uri_list = g_list_append(player->uri_info.uri_list, g_strdup(uri));
		SECURE_LOGD("add new path : %s(total num of list = %d)", uri, g_list_length(player->uri_info.uri_list));
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_get_next_uri(MMHandleType hplayer, char **uri)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	char *next_uri = NULL;
	guint num_of_list = 0;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	num_of_list = g_list_length(player->uri_info.uri_list);

	if (num_of_list > 0) {
		gint uri_idx = player->uri_info.uri_idx;

		if (uri_idx < num_of_list-1)
			uri_idx++;
		else
			uri_idx = 0;

		next_uri = g_list_nth_data(player->uri_info.uri_list, uri_idx);
		LOGE("next uri idx : %d, uri = %s", uri_idx, next_uri);

		*uri = g_strdup(next_uri);
	}

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

static void
__mmplayer_gst_decode_unknown_type(GstElement *elem,  GstPad *pad,
	GstCaps *caps, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	const gchar *klass = NULL;
	const gchar *mime = NULL;
	gchar *caps_str = NULL;

	klass = gst_element_factory_get_metadata(gst_element_get_factory(elem), GST_ELEMENT_METADATA_KLASS);
	mime = gst_structure_get_name(gst_caps_get_structure(caps, 0));
	caps_str = gst_caps_to_string(caps);

	LOGW("unknown type of caps : %s from %s",
					caps_str, GST_ELEMENT_NAME(elem));

	MMPLAYER_FREEIF(caps_str);

	/* There is no available codec. */
	__mmplayer_check_not_supported_codec(player, klass, mime);
}

static gboolean
__mmplayer_gst_decode_autoplug_continue(GstElement *bin,  GstPad *pad,
	GstCaps *caps,  gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	const char *mime = NULL;
	gboolean ret = TRUE;

	MMPLAYER_LOG_GST_CAPS_TYPE(caps);
	mime = gst_structure_get_name(gst_caps_get_structure(caps, 0));

	if (g_str_has_prefix(mime, "audio")) {
		GstStructure *caps_structure = NULL;
		gint samplerate = 0;
		gint channels = 0;
		gchar *caps_str = NULL;

		caps_structure = gst_caps_get_structure(caps, 0);
		gst_structure_get_int(caps_structure, "rate", &samplerate);
		gst_structure_get_int(caps_structure, "channels", &channels);

		if ((channels > 0 && samplerate == 0)) {
			LOGD("exclude audio...");
			ret = FALSE;
		}

		caps_str = gst_caps_to_string(caps);
		/* set it directly because not sent by TAG */
		if (g_strrstr(caps_str, "mobile-xmf"))
			mm_player_set_attribute((MMHandleType)player, NULL,
					"content_audio_codec", "mobile-xmf", strlen("mobile-xmf"), NULL);

		MMPLAYER_FREEIF(caps_str);
	} else if (g_str_has_prefix(mime, "video") && player->videodec_linked) {
		LOGD("already video linked");
		ret = FALSE;
	} else {
		LOGD("found new stream");
	}

	return ret;
}

static gboolean
__mmplayer_is_audio_offload_device_type(mmplayer_t *player)
{
	gboolean ret = FALSE;
	GDBusConnection *conn = NULL;
	GError *err = NULL;
	GVariant *result = NULL;
	const gchar *dbus_device_type = NULL;
	const gchar *dbus_ret = NULL;
	gint idx = 0;

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
	if (!conn || err) {
		LOGE("failed g_bus_get_sync() (%s)", err ? err->message : NULL);
		g_error_free(err);
		goto DONE;
	}

	result = g_dbus_connection_call_sync(conn,
					"org.pulseaudio.Server",
					"/org/pulseaudio/StreamManager",
					"org.pulseaudio.StreamManager",
					"GetCurrentMediaRoutingPath",
					g_variant_new("(s)", "out"),
					G_VARIANT_TYPE("(ss)"),
					G_DBUS_CALL_FLAGS_NONE,
					2000,
					NULL,
					&err);
	if (!result || err) {
		LOGE("failed g_dbus_connection_call_sync() (%s)", err ? err->message : NULL);
		g_error_free(err);
		goto DONE;
	}

	/* device type is listed in stream-map.json at mmfw-sysconf */
	g_variant_get(result, "(&s&s)", &dbus_device_type, &dbus_ret);

	LOGI("g_dbus_connection_call_sync() success (%s, %s)", dbus_device_type, dbus_ret);
	if (strncmp("STREAM_MANAGER_RETURN_OK", dbus_ret, strlen(dbus_ret)))
		goto DONE;

	/* the device type is listed in ini file among audio-jack, bt-a2dp, usb-audio, builtin-speaker */
	for (idx = 0; player->ini.audio_offload_device_type[idx][0] != '\0'; idx++) {
		if (strstr(dbus_device_type, player->ini.audio_offload_device_type[idx])) {
			LOGD("audio offload is supportable");
			ret = TRUE;
			goto DONE;
		}
	}

	LOGD("audio offload is not supportable");

DONE:
	g_variant_unref(result);
	if (conn)
		g_object_unref(conn);

	return ret;
}

static void __mmplayer_rebuild_audio_pipeline(mmplayer_t *player)
{
	mmplayer_state_e current_state = MM_PLAYER_STATE_NONE;
	gint64 position = 0;

	MMPLAYER_RETURN_IF_FAIL(player && player->attrs &&
		player->pipeline && player->pipeline->mainbin);

	MMPLAYER_CMD_LOCK(player);
	current_state = MMPLAYER_CURRENT_STATE(player);

	if (!gst_element_query_position(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &position))
		LOGW("getting current position failed in paused");

	_mmplayer_unrealize((MMHandleType)player);
	_mmplayer_realize((MMHandleType)player);

	_mmplayer_set_position((MMHandleType)player, position);

	/* async not to be blocked in streaming case */
	mm_player_set_attribute((MMHandleType)player, NULL, "profile_prepare_async", TRUE, NULL);

	_mmplayer_pause((MMHandleType)player);

	if (current_state == MM_PLAYER_STATE_PLAYING)
		_mmplayer_start((MMHandleType)player);
	MMPLAYER_CMD_UNLOCK(player);

	LOGD("rebuilding audio pipeline is completed.");
}

void __mmplayer_audio_device_connected_cb(MMSoundDevice_t device_h, bool is_connected, void *user_data)
{
	mmplayer_t *player = (mmplayer_t *)user_data;
	mm_sound_device_type_e dev_type = MM_SOUND_DEVICE_TYPE_BUILTIN_SPEAKER;
	gboolean is_supportable = FALSE;

	if (mm_sound_get_device_type(device_h, &dev_type) != MM_ERROR_NONE)
		LOGW("failed to get device type");
	else
		LOGD("dev type (%d), connected (%d)", dev_type, is_connected);

	if ((dev_type != MM_SOUND_DEVICE_TYPE_BLUETOOTH_A2DP) &&
		(dev_type != MM_SOUND_DEVICE_TYPE_AUDIOJACK) &&
		(dev_type != MM_SOUND_DEVICE_TYPE_USB_AUDIO)) {
		LOGD("ignore this dev connected info");
		return;
	}

	is_supportable = __mmplayer_is_audio_offload_device_type(player);
	if (player->build_audio_offload == is_supportable) {
		LOGD("keep current pipeline without re-building");
		return;
	}

	/* rebuild pipeline */
	LOGD("re-build pipeline - offload: %d", is_supportable);
	player->build_audio_offload = FALSE;
	__mmplayer_rebuild_audio_pipeline(player);

	return;
}

static gboolean
__mmplayer_add_audio_device_connected_cb(mmplayer_t *player)
{
	unsigned int id = 0;

	if (player->audio_device_cb_id != 0) {
		LOGW("audio device connected cb was already added (%u)", player->audio_device_cb_id);
		return TRUE;
	}

	if (mm_sound_add_device_connected_callback(MM_SOUND_DEVICE_IO_DIRECTION_OUT_FLAG,
				__mmplayer_audio_device_connected_cb, player, &id) == MM_ERROR_NONE) {
		LOGD("added device connected cb (%u)", id);
		player->audio_device_cb_id = id;
	} else {
		LOGW("failed to add device connected cb");
		return FALSE;
	}

	return TRUE;
}

int _mmplayer_audio_offload_is_activated(MMHandleType hplayer, bool *activated)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(activated, MM_ERROR_INVALID_ARGUMENT);

	*activated = player->build_audio_offload;

	LOGD("offload activated : %d", (int)*activated);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

static gboolean
__mmplayer_is_offload_supported_type(mmplayer_t *player)
{
	/* NOTE :
	   this function need to be updated according to the supported media format
	   @see player->ini.audio_offload_media_format */

	if (__mmplayer_is_only_mp3_type(player->type)) {
		LOGD("offload supportable media format type");
		return TRUE;
	}

	return FALSE;
}

static gboolean
__mmplayer_can_build_audio_offload_path(mmplayer_t *player)
{
	gboolean ret = FALSE;
	GstElementFactory *factory = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->attrs, FALSE);

	LOGD("current stream : %s, sink: %s", player->type, player->ini.audio_offload_sink_element);
	if (!__mmplayer_is_offload_supported_type(player))
		goto DONE;

	if (!strcmp(player->ini.audio_offload_sink_element, "")) {
		LOGD("there is no audio offload sink");
		goto DONE;
	}

	if (player->ini.audio_offload_device_type[0][0] == '\0') {
		LOGW("there is no audio device type to support offload");
		goto DONE;
	}

	factory = gst_element_factory_find(player->ini.audio_offload_sink_element);
	if (!factory) {
		LOGW("there is no installed audio offload sink element");
		goto DONE;
	}
	gst_object_unref(factory);

	if (__mmplayer_acquire_hw_resource(player,
			MMPLAYER_RESOURCE_TYPE_AUDIO_OFFLOAD) != MM_ERROR_NONE) {
		LOGE("failed to acquire audio offload decoder resource");
		goto DONE;
	}

	if (!__mmplayer_add_audio_device_connected_cb(player))
		goto DONE;

	if (!__mmplayer_is_audio_offload_device_type(player))
		goto DONE;

	LOGD("audio offload can be built");
	ret = TRUE;

DONE:
	if (!ret)
		__mmplayer_release_hw_resource(player, MMPLAYER_RESOURCE_TYPE_AUDIO_OFFLOAD);

	MMPLAYER_FLEAVE();
	return ret;
}

static GstAutoplugSelectResult
__mmplayer_check_codec_info(mmplayer_t *player, const char *klass, GstCaps *caps, char *factory_name)
{
	GstAutoplugSelectResult ret = GST_AUTOPLUG_SELECT_TRY;
	int audio_offload = 0;

	if ((g_strrstr(klass, "Codec/Decoder/Audio"))) {
		mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_AUDIO_OFFLOAD, &audio_offload); /* user requirement */

		if (audio_offload && __mmplayer_can_build_audio_offload_path(player)) {
			LOGD("expose audio path to build offload output path");
			player->build_audio_offload = TRUE;
			/* update codec info */
			player->not_supported_codec &= MISSING_PLUGIN_VIDEO;
			player->can_support_codec |= FOUND_PLUGIN_AUDIO;
			player->audiodec_linked = 1;

			ret = GST_AUTOPLUG_SELECT_EXPOSE;
			goto DONE;
		}

		/* FIXME: If HW audio decoder is selected, related resource have to be acquired here.
				  And need to consider the multi-track audio content.
		          There is no HW audio decoder in public. */

		/* set stream information */
		if (!player->audiodec_linked)
			__mmplayer_set_audio_attrs(player, caps);

		/* update codec info */
		player->not_supported_codec &= MISSING_PLUGIN_VIDEO;
		player->can_support_codec |= FOUND_PLUGIN_AUDIO;
		player->audiodec_linked = 1;

	} else if (g_strrstr(klass, "Codec/Decoder/Video")) {

		if ((strlen(player->ini.videocodec_element_hw) > 0) &&
			(g_strrstr(factory_name, player->ini.videocodec_element_hw))) {

			/* mark video decoder for acquire */
			if (player->hw_resource[MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER] != NULL) {
				LOGW("video decoder resource is already acquired, skip it.");
				ret = GST_AUTOPLUG_SELECT_SKIP;
				goto DONE;
			}

			if (__mmplayer_acquire_hw_resource(player, MMPLAYER_RESOURCE_TYPE_VIDEO_DECODER) != MM_ERROR_NONE) {
				LOGE("failed to acquire video decoder resource");
				ret = GST_AUTOPLUG_SELECT_SKIP;
				goto DONE;
			}
			player->interrupted_by_resource = FALSE;
		}

		/* update codec info */
		player->not_supported_codec &= MISSING_PLUGIN_AUDIO;
		player->can_support_codec |= FOUND_PLUGIN_VIDEO;
		player->videodec_linked = 1;
	}

DONE:
	return ret;
}

GValueArray *
_mmplayer_gst_decode_autoplug_sort(GstElement *bin,
		GstPad *pad, GstCaps *caps, GValueArray *factories, gpointer data)
{
#define DEFAULT_IDX 0xFFFF
#define MIN_FACTORY_NUM 2
	mmplayer_t *player = (mmplayer_t *)data;
	GValueArray *new_factories = NULL;
	GValue val = { 0, };
	GstElementFactory *factory = NULL;
	const gchar *klass = NULL;
	gchar *factory_name = NULL;
	guint hw_dec_idx = DEFAULT_IDX;
	guint first_sw_dec_idx = DEFAULT_IDX;
	guint last_sw_dec_idx = DEFAULT_IDX;
	guint new_pos = DEFAULT_IDX;
	guint rm_pos = DEFAULT_IDX;
	int audio_codec_type;
	int video_codec_type;
	mmplayer_codec_type_e codec_type = MM_PLAYER_CODEC_TYPE_DEFAULT;

	if (factories->n_values < MIN_FACTORY_NUM)
		return NULL;

	mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_VIDEO_CODEC_TYPE, &video_codec_type);
	mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_AUDIO_CODEC_TYPE, &audio_codec_type);

#ifdef __DEBUG__
	LOGD("num of factory : %d, codec type %d, %d", factories->n_values, video_codec_type, audio_codec_type);
#endif
	for (int i = 0 ; i < factories->n_values ; i++) {
		gchar *hw_dec_info = NULL;
		gchar (*sw_dec_info)[PLAYER_INI_MAX_STRLEN] = {NULL, };

		factory = g_value_get_object(g_value_array_get_nth(factories, i));
		klass = gst_element_factory_get_klass(factory);
		factory_name = GST_OBJECT_NAME(factory);

#ifdef __DEBUG__
		LOGD("Klass [%s] Factory [%s]", klass, factory_name);
#endif
		if (g_strrstr(klass, "Codec/Decoder/Audio")) {
			if (!player->need_audio_dec_sorting) {
				LOGD("sorting is not required");
				return NULL;
			}
			codec_type = audio_codec_type;
			hw_dec_info = player->ini.audiocodec_element_hw;
			sw_dec_info = player->ini.audiocodec_element_sw;
		} else if (g_strrstr(klass, "Codec/Decoder/Video")) {
			if (!player->need_video_dec_sorting) {
				LOGD("sorting is not required");
				return NULL;
			}
			codec_type = video_codec_type;
			hw_dec_info = player->ini.videocodec_element_hw;
			sw_dec_info = player->ini.videocodec_element_sw;
		} else {
			continue;
		}

		if (g_strrstr(factory_name, hw_dec_info)) {
			hw_dec_idx = i;
		} else {
			for (int j = 0; sw_dec_info[j][0] != '\0'; j++) {
				if (strstr(factory_name, sw_dec_info[j])) {
					last_sw_dec_idx = i;
					if (first_sw_dec_idx == DEFAULT_IDX) {
						first_sw_dec_idx = i;
					}
				}
			}

			if (first_sw_dec_idx == DEFAULT_IDX)
				LOGW("unknown codec %s", factory_name);
		}
	}

	if (hw_dec_idx == DEFAULT_IDX || first_sw_dec_idx == DEFAULT_IDX)
		return NULL;

	if (codec_type == MM_PLAYER_CODEC_TYPE_HW) {
		if (hw_dec_idx < first_sw_dec_idx)
			return NULL;
		new_pos = first_sw_dec_idx - 1;
		rm_pos = hw_dec_idx + 1;
	} else if (codec_type == MM_PLAYER_CODEC_TYPE_SW) {
		if (last_sw_dec_idx < hw_dec_idx)
			return NULL;
		new_pos = last_sw_dec_idx + 1;
		rm_pos = hw_dec_idx;
	} else {
		return NULL;
	}

	/* change position - insert H/W decoder according to the new position */
	new_factories = g_value_array_copy(factories);
	factory = g_value_get_object(g_value_array_get_nth(factories, hw_dec_idx));
	g_value_init (&val, G_TYPE_OBJECT);
	g_value_set_object (&val, factory);
	g_value_array_insert(new_factories, new_pos, &val);
	g_value_unset (&val);
	g_value_array_remove(new_factories, rm_pos);	/* remove previous H/W element */

	for (int i = 0 ; i < new_factories->n_values ; i++) {
		factory = g_value_get_object(g_value_array_get_nth(new_factories, i));

		LOGD("[Re-arranged] Klass [%s] Factory [%s]",
			gst_element_factory_get_klass(factory), GST_OBJECT_NAME (factory));
	}

	return new_factories;
}

gint
_mmplayer_gst_decode_autoplug_select(GstElement *bin,  GstPad *pad,
	GstCaps *caps, GstElementFactory *factory, gpointer data)
{
	GstAutoplugSelectResult result = GST_AUTOPLUG_SELECT_TRY;
	mmplayer_t *player = (mmplayer_t *)data;

	gchar *factory_name = NULL;
	gchar *caps_str = NULL;
	const gchar *klass = NULL;
	gint idx = 0;

	factory_name = GST_OBJECT_NAME(factory);
	klass = gst_element_factory_get_metadata(factory, GST_ELEMENT_METADATA_KLASS);
	caps_str = gst_caps_to_string(caps);

	LOGD("[handle: %p] found new element [%s] to link", player, factory_name);

	/* store type string */
	if (player->type == NULL) {
		player->type = gst_caps_to_string(caps);
		__mmplayer_update_content_type_info(player);
	}

	/* filtering exclude keyword */
	for (idx = 0; player->ini.exclude_element_keyword[idx][0] != '\0'; idx++) {
		if (strstr(factory_name, player->ini.exclude_element_keyword[idx])) {
			LOGW("skipping [%s] by exculde keyword [%s]",
					factory_name, player->ini.exclude_element_keyword[idx]);

			result = GST_AUTOPLUG_SELECT_SKIP;
			goto DONE;
		}
	}

	for (idx = 0; player->ini.unsupported_codec_keyword[idx][0] != '\0'; idx++) {
		if (caps_str && strstr(caps_str, player->ini.unsupported_codec_keyword[idx])) {
			LOGW("skipping [%s] by unsupported codec keyword [%s]",
				factory_name, player->ini.unsupported_codec_keyword[idx]);
			result = GST_AUTOPLUG_SELECT_SKIP;
			goto DONE;
		}
	}

	/* exclude webm format */
	/* NOTE : MSL have to post MM_ERROR_PLAYER_NOT_SUPPORTED_FORMAT
	 * because webm format is not supportable.
	 * If webm is disabled in "autoplug-continue", there is no state change
	 * failure or error because the decodebin will expose the pad directly.
	 * It make MSL invoke _prepare_async_callback.
	 * So, we need to disable webm format in "autoplug-select" */
	if (caps_str && strstr(caps_str, "webm")) {
		LOGW("webm is not supported");
		result = GST_AUTOPLUG_SELECT_SKIP;
		goto DONE;
	}

	/* check factory class for filtering */
	/* NOTE : msl don't need to use image plugins.
	 * So, those plugins should be skipped for error handling.
	 */
	if (g_strrstr(klass, "Codec/Decoder/Image")) {
		LOGD("skipping [%s] by not required", factory_name);
		result = GST_AUTOPLUG_SELECT_SKIP;
		goto DONE;
	}

	if ((MMPLAYER_IS_MS_BUFF_SRC(player)) &&
		(g_strrstr(klass, "Codec/Demuxer") || (g_strrstr(klass, "Codec/Parser")))) {
		// TO CHECK : subtitle if needed, add subparse exception.
		LOGD("skipping parser/demuxer [%s] in es player by not required", factory_name);
		result = GST_AUTOPLUG_SELECT_SKIP;
		goto DONE;
	}

	if (g_strrstr(factory_name, "mpegpsdemux")) {
		LOGD("skipping PS container - not support");
		result = GST_AUTOPLUG_SELECT_SKIP;
		goto DONE;
	}

	if (g_strrstr(factory_name, "mssdemux"))
		player->smooth_streaming = TRUE;

	if ((g_strrstr(klass, "Codec/Parser/Converter/Video")) ||
		(g_strrstr(klass, "Codec/Decoder/Video"))) {
		gint stype = 0;
		gint width = 0;
		GstStructure *str = NULL;
		mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &stype);

		/* don't make video because of not required */
		if ((stype == MM_DISPLAY_SURFACE_NULL) &&
			(!player->set_mode.video_export)) {
			LOGD("no need video decoding, expose pad");
			result = GST_AUTOPLUG_SELECT_EXPOSE;
			goto DONE;
		}

		/* get w/h for omx state-tune */
		/* FIXME: deprecated? */
		str = gst_caps_get_structure(caps, 0);
		gst_structure_get_int(str, "width", &width);

		if (width != 0) {
			if (player->v_stream_caps) {
				gst_caps_unref(player->v_stream_caps);
				player->v_stream_caps = NULL;
			}

			player->v_stream_caps = gst_caps_copy(caps);
			LOGD("take caps for video state tune");
			MMPLAYER_LOG_GST_CAPS_TYPE(player->v_stream_caps);
		}
	}

	if (g_strrstr(klass, "Codec/Decoder")) {
		result = __mmplayer_check_codec_info(player, klass, caps, factory_name);
		if (result != GST_AUTOPLUG_SELECT_TRY) {
			LOGW("skip add decoder");
			goto DONE;
		}
	}

DONE:
	MMPLAYER_FREEIF(caps_str);

	return result;
}

static void
__mmplayer_gst_decode_pad_removed(GstElement *elem,  GstPad *new_pad,
	gpointer data)
{
	//mmplayer_t *player = (mmplayer_t *)data;
	GstCaps *caps = NULL;

	LOGD("[Decodebin2] pad-removed signal");

	caps = gst_pad_query_caps(new_pad, NULL);
	if (!caps) {
		LOGW("query caps is NULL");
		return;
	}

	gchar *caps_str = NULL;
	caps_str = gst_caps_to_string(caps);

	LOGD("pad removed caps : %s from %s", caps_str, GST_ELEMENT_NAME(elem));

	MMPLAYER_FREEIF(caps_str);
	gst_caps_unref(caps);
}

static void
__mmplayer_gst_decode_drained(GstElement *bin, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	GstIterator *iter = NULL;
	GValue item = { 0, };
	GstPad *pad = NULL;
	gboolean done = FALSE;
	gboolean is_all_drained = TRUE;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	LOGD("__mmplayer_gst_decode_drained");

	if (!MMPLAYER_CMD_TRYLOCK(player)) {
		LOGW("Fail to get cmd lock");
		return;
	}

	if (!player->gapless.reconfigure && /* If it is already checked, skip verify. */
		!__mmplayer_verify_gapless_play_path(player)) {
		LOGD("decoding is finished.");
		__mmplayer_reset_gapless_state(player);
		MMPLAYER_CMD_UNLOCK(player);
		return;
	}

	player->gapless.reconfigure = TRUE;

	/* check decodebin src pads whether they received EOS or not */
	iter = gst_element_iterate_src_pads(player->pipeline->mainbin[MMPLAYER_M_AUTOPLUG].gst);

	while (!done) {
		switch (gst_iterator_next(iter, &item)) {
		case GST_ITERATOR_OK:
			pad = g_value_get_object(&item);
			if (pad && !GST_PAD_IS_EOS(pad)) {
				LOGW("[%s:%s] not received EOS yet.", GST_DEBUG_PAD_NAME(pad));
				is_all_drained = FALSE;
				break;
			}
			g_value_reset(&item);
			break;
		case GST_ITERATOR_RESYNC:
			gst_iterator_resync(iter);
			break;
		case GST_ITERATOR_ERROR:
		case GST_ITERATOR_DONE:
			done = TRUE;
			break;
		}
	}
	g_value_unset(&item);
	gst_iterator_free(iter);

	if (!is_all_drained) {
		LOGD("Wait util the all pads get EOS.");
		MMPLAYER_CMD_UNLOCK(player);
		MMPLAYER_FLEAVE();
		return;
	}

	player->gapless.update_segment[MM_PLAYER_TRACK_TYPE_AUDIO] = FALSE;
	player->gapless.update_segment[MM_PLAYER_TRACK_TYPE_VIDEO] = FALSE;

	/* deactivate pipeline except sinkbins to set up the new pipeline of next uri*/
	MMPLAYER_POST_MSG(player, MM_MESSAGE_GAPLESS_CONSTRUCTION, NULL); /* post message for gapless */
	__mmplayer_deactivate_old_path(player);
	MMPLAYER_CMD_UNLOCK(player);

	MMPLAYER_FLEAVE();
}

void
_mmplayer_gst_element_added(GstElement *bin, GstElement *element, gpointer data)
{
	mmplayer_t *player = (mmplayer_t *)data;
	const gchar *klass = NULL;
	gchar *factory_name = NULL;

	klass = gst_element_factory_get_metadata(gst_element_get_factory(element), GST_ELEMENT_METADATA_KLASS);
	factory_name = GST_OBJECT_NAME(gst_element_get_factory(element));

	LOGD("new elem klass: %s, factory_name: %s, new elem name : %s", klass, factory_name, GST_ELEMENT_NAME(element));

	if (__mmplayer_add_dump_buffer_probe(player, element))
		LOGD("add buffer probe");

	if (g_strrstr(klass, "Codec/Decoder/Audio")) {
		gchar *selected = NULL;
		selected = g_strdup(GST_ELEMENT_NAME(element));
		player->audio_decoders = g_list_append(player->audio_decoders, selected);
	}

	if (g_strrstr(klass, "Demuxer/Adaptive")) {
		player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].id = MMPLAYER_M_ADAPTIVE_DEMUX;
		player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].gst = element;

		LOGD("set max variant limit: %d, %d %d", player->adaptive_info.limit.bandwidth,
						player->adaptive_info.limit.width, player->adaptive_info.limit.height);

		g_object_set(player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].gst,
						"max-bandwidth", player->adaptive_info.limit.bandwidth,
						"max-video-width", player->adaptive_info.limit.width,
						"max-video-height", player->adaptive_info.limit.height, NULL);

	} else if (g_strrstr(klass, "Demuxer")) {
#ifdef __DEBUG__
		LOGD("plugged element is demuxer. take it");
#endif
		player->pipeline->mainbin[MMPLAYER_M_DEMUX].id = MMPLAYER_M_DEMUX;
		player->pipeline->mainbin[MMPLAYER_M_DEMUX].gst = element;
	}

	if (g_strrstr(factory_name, "asfdemux") || g_strrstr(factory_name, "qtdemux") || g_strrstr(factory_name, "avidemux")) {
		int surface_type = 0;

		mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &surface_type);
	}

	// to support trust-zone only
	if (g_strrstr(factory_name, "asfdemux")) {
		LOGD("set file-location %s", player->profile.uri);
		g_object_set(G_OBJECT(element), "file-location", player->profile.uri, NULL);
	} else if (g_strrstr(factory_name, "legacyh264parse")) {
		LOGD("[%s] output-format to legacyh264parse", "mssdemux");
		g_object_set(G_OBJECT(element), "output-format", 1, NULL); /* NALU/Byte Stream format */
	} else if (g_strrstr(factory_name, "mpegaudioparse")) {
		if ((MMPLAYER_IS_HTTP_STREAMING(player)) &&
			(__mmplayer_is_only_mp3_type(player->type))) {
			LOGD("[mpegaudioparse] set streaming pull mode.");
			g_object_set(G_OBJECT(element), "http-pull-mp3dec", TRUE, NULL);
		}
	} else if (g_strrstr(factory_name, player->ini.videocodec_element_hw)) {
		player->pipeline->mainbin[MMPLAYER_M_DEC1].gst = element;
	}

	if ((player->pipeline->mainbin[MMPLAYER_M_DEMUX].gst) &&
		(g_strrstr(GST_ELEMENT_NAME(element), "multiqueue"))) {
		LOGD("plugged element is multiqueue. take it %s", GST_ELEMENT_NAME(element));

		player->pipeline->mainbin[MMPLAYER_M_DEMUXED_S_BUFFER].id = MMPLAYER_M_DEMUXED_S_BUFFER;
		player->pipeline->mainbin[MMPLAYER_M_DEMUXED_S_BUFFER].gst = element;

		if ((MMPLAYER_IS_HTTP_STREAMING(player)) ||
			(MMPLAYER_IS_HTTP_LIVE_STREAMING(player)) ||
			(MMPLAYER_IS_DASH_STREAMING(player))) {
			/* in case of multiqueue, max bytes size is defined with fixed value in mm_player_streaming.h*/
			_mm_player_streaming_set_multiqueue(player->streamer, element);
			_mm_player_streaming_sync_property(player->streamer, player->pipeline->mainbin[MMPLAYER_M_AUTOPLUG].gst);
		}

	}

	return;
}

static void
__mmplayer_release_misc(mmplayer_t *player)
{
	int i;
	bool cur_mode = player->set_mode.rich_audio;
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player);

	player->video_decoded_cb = NULL;
	player->video_decoded_cb_user_param = NULL;
	player->video_stream_prerolled = false;

	player->audio_decoded_cb = NULL;
	player->audio_decoded_cb_user_param = NULL;
	player->audio_extract_opt = MM_PLAYER_AUDIO_EXTRACT_DEFAULT;

	player->audio_stream_changed_cb = NULL;
	player->audio_stream_changed_cb_user_param = NULL;

	player->sent_bos = FALSE;
	player->playback_rate = DEFAULT_PLAYBACK_RATE;

	player->seek_state = MMPLAYER_SEEK_NONE;

	player->total_bitrate = 0;
	player->total_maximum_bitrate = 0;

	player->not_found_demuxer = 0;

	player->last_position = 0;
	player->duration = 0;
	player->http_content_size = 0;
	player->not_supported_codec = MISSING_PLUGIN_NONE;
	player->can_support_codec = FOUND_PLUGIN_NONE;
	player->pending_seek.is_pending = false;
	player->pending_seek.pos = 0;
	player->msg_posted = FALSE;
	player->has_many_types = FALSE;
	player->is_subtitle_force_drop = FALSE;
	player->play_subtitle = FALSE;
	player->adjust_subtitle_pos = 0;
	player->has_closed_caption = FALSE;
	player->set_mode.video_export = false;
	player->profile.uri_type = MM_PLAYER_URI_TYPE_NONE;
	memset(&player->set_mode, 0, sizeof(mmplayer_setting_mode_t));
	/* recover mode */
	player->set_mode.rich_audio = cur_mode;

	if (player->audio_device_cb_id > 0 &&
		mm_sound_remove_device_connected_callback(player->audio_device_cb_id) != MM_ERROR_NONE)
		LOGW("failed to remove audio device_connected_callback");
	player->audio_device_cb_id = 0;

	for (i = 0; i < MM_PLAYER_STREAM_COUNT_MAX; i++) {
		player->bitrate[i] = 0;
		player->maximum_bitrate[i] = 0;
	}

	/* free memory related to audio effect */
	MMPLAYER_FREEIF(player->audio_effect_info.custom_ext_level_for_plugin);

	if (player->adaptive_info.var_list) {
		g_list_free_full(player->adaptive_info.var_list, g_free);
		player->adaptive_info.var_list = NULL;
	}

	player->adaptive_info.limit.bandwidth = ADAPTIVE_VARIANT_DEFAULT_VALUE;
	player->adaptive_info.limit.width = ADAPTIVE_VARIANT_DEFAULT_VALUE;
	player->adaptive_info.limit.height = ADAPTIVE_VARIANT_DEFAULT_VALUE;

	/* Reset video360 settings to their defaults in case if the pipeline is to be
	 * re-created.
	 * */
	player->video360_metadata.is_spherical = -1;
	player->is_openal_plugin_used = FALSE;

	player->is_content_spherical = FALSE;
	player->is_video360_enabled = TRUE;
	player->video360_metadata.projection_type = VIDEO360_PROJECTION_TYPE_EQUIRECTANGULAR;
	player->video360_metadata.stereo_mode = VIDEO360_MODE_MONOSCOPIC;
	player->video360_yaw_radians = PLAYER_SPHERICAL_DEFAULT_YAW;
	player->video360_pitch_radians = PLAYER_SPHERICAL_DEFAULT_PITCH;
	player->video360_zoom = 1.0f;
	player->video360_horizontal_fov = PLAYER_SPHERICAL_DEFAULT_H_FOV;
	player->video360_vertical_fov = PLAYER_SPHERICAL_DEFAULT_V_FOV;

	player->sound.rg_enable = false;

	__mmplayer_initialize_video_roi(player);
	MMPLAYER_FLEAVE();
}

static void
__mmplayer_release_misc_post(mmplayer_t *player)
{
	char *original_uri = NULL;
	MMPLAYER_FENTER();

	/* player->pipeline is already released before. */

	MMPLAYER_RETURN_IF_FAIL(player);

	mm_player_set_attribute((MMHandleType)player, NULL, "content_video_found", 0, NULL);

	/* clean found audio decoders */
	if (player->audio_decoders) {
		GList *a_dec = player->audio_decoders;
		for (; a_dec; a_dec = g_list_next(a_dec)) {
			gchar *name = a_dec->data;
			MMPLAYER_FREEIF(name);
		}
		g_list_free(player->audio_decoders);
		player->audio_decoders = NULL;
	}

	/* clean the uri list except original uri */
	if (player->uri_info.uri_list) {
		original_uri = g_list_nth_data(player->uri_info.uri_list, 0);

		if (!original_uri)
			LOGW("failed to get original uri info");

		mm_player_set_attribute((MMHandleType)player, NULL, "profile_uri",
				original_uri, (original_uri) ? strlen(original_uri) : (0), NULL);

		GList *uri_list = player->uri_info.uri_list;
		for (; uri_list; uri_list = g_list_next(uri_list)) {
			gchar *uri = uri_list->data;
			MMPLAYER_FREEIF(uri);
		}
		g_list_free(player->uri_info.uri_list);
		player->uri_info.uri_list = NULL;
	}

	/* clear the audio stream buffer list */
	_mmplayer_audio_stream_clear_buffer(player, FALSE);

	/* clear the video stream bo list */
	__mmplayer_video_stream_destroy_bo_list(player);
	__mmplayer_initialize_storage_info(player, MMPLAYER_PATH_MAX);

	if (player->profile.input_mem.buf) {
		free(player->profile.input_mem.buf);
		player->profile.input_mem.buf = NULL;
	}
	player->profile.input_mem.len = 0;
	player->profile.input_mem.offset = 0;

	player->uri_info.uri_idx = 0;
	MMPLAYER_FLEAVE();
}

gboolean
__mmplayer_check_subtitle(mmplayer_t *player)
{
	MMHandleType attrs = 0;
	char *subtitle_uri = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);

	/* get subtitle attribute */
	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs)
		return FALSE;

	mm_attrs_get_string_by_name(attrs, "subtitle_uri", &subtitle_uri);
	if (!subtitle_uri || !strlen(subtitle_uri))
		return FALSE;

	SECURE_LOGD("subtitle uri is %s[%zu]", subtitle_uri, strlen(subtitle_uri));
	player->is_external_subtitle_present = TRUE;

	MMPLAYER_FLEAVE();

	return TRUE;
}

void
_mmplayer_cancel_eos_timer(mmplayer_t *player)
{
	MMPLAYER_RETURN_IF_FAIL(player);

	if (player->eos_timer) {
		LOGD("cancel eos timer");
		__mmplayer_remove_g_source_from_context(player->context.global_default, player->eos_timer);
		player->eos_timer = 0;
	}

	return;
}

static void
__mmplayer_add_sink(mmplayer_t *player, GstElement *sink)
{
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player);
	MMPLAYER_RETURN_IF_FAIL(sink);

	player->sink_elements = g_list_append(player->sink_elements, sink);

	MMPLAYER_FLEAVE();
}

static void
__mmplayer_del_sink(mmplayer_t *player, GstElement *sink)
{
	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player);
	MMPLAYER_RETURN_IF_FAIL(sink);

	player->sink_elements = g_list_remove(player->sink_elements, sink);

	MMPLAYER_FLEAVE();
}

void
_mmplayer_add_signal_connection(mmplayer_t *player, GObject *object,
	mmplayer_signal_type_e type, const gchar *signal, GCallback cb_funct, gpointer u_data)
{
	mmplayer_signal_item_t *item = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	if (type >= MM_PLAYER_SIGNAL_TYPE_MAX) {
		LOGE("invalid signal type [%d]", type);
		return;
	}

	item = (mmplayer_signal_item_t *)g_try_malloc(sizeof(mmplayer_signal_item_t));
	if (!item) {
		LOGE("cannot connect signal [%s]", signal);
		return;
	}

	item->obj = object;
	item->sig = g_signal_connect(object, signal, cb_funct, u_data);
	player->signals[type] = g_list_append(player->signals[type], item);

	MMPLAYER_FLEAVE();
	return;
}

/* NOTE : be careful with calling this api. please refer to below glib comment
 * glib comment : Note that there is a bug in GObject that makes this function much
 * less useful than it might seem otherwise. Once gobject is disposed, the callback
 * will no longer be called, but, the signal handler is not currently disconnected.
 * If the instance is itself being freed at the same time than this doesn't matter,
 * since the signal will automatically be removed, but if instance persists,
 * then the signal handler will leak. You should not remove the signal yourself
 * because in a future versions of GObject, the handler will automatically be
 * disconnected.
 *
 * It's possible to work around this problem in a way that will continue to work
 * with future versions of GObject by checking that the signal handler is still
 * connected before disconnected it:
 *
 *  if (g_signal_handler_is_connected(instance, id))
 *    g_signal_handler_disconnect(instance, id);
 */
static void
__mmplayer_release_signal_connection(mmplayer_t *player, mmplayer_signal_type_e type)
{
	GList *sig_list = NULL;
	mmplayer_signal_item_t *item = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_IF_FAIL(player);

	LOGD("release signals type : %d", type);

	if (type >= MM_PLAYER_SIGNAL_TYPE_ALL) {
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_AUTOPLUG);
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_VIDEOBIN);
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_AUDIOBIN);
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_TEXTBIN);
		__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_OTHERS);
		return;
	}

	sig_list = player->signals[type];

	for (; sig_list; sig_list = sig_list->next) {
		item = sig_list->data;

		if (item && item->obj && GST_IS_ELEMENT(item->obj)) {
			if (g_signal_handler_is_connected(item->obj, item->sig))
				g_signal_handler_disconnect(item->obj, item->sig);
		}

		MMPLAYER_FREEIF(item);
	}

	g_list_free(player->signals[type]);
	player->signals[type] = NULL;

	MMPLAYER_FLEAVE();

	return;
}

int
_mmplayer_change_videosink(MMHandleType handle, MMDisplaySurfaceType surface_type, int wl_surface_id)
{
	mmplayer_t *player = 0;
	int prev_display_surface_type = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(handle, MM_ERROR_COMMON_INVALID_ARGUMENT);

	player = MM_PLAYER_CAST(handle);

	/* check video sinkbin is created */
	if (_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_NUM)) {
		LOGW("Videosink is already created");
		return MM_ERROR_NONE;
	}

	LOGD("videosink element is not yet ready");

	if (surface_type >= MM_DISPLAY_SURFACE_NUM) {
		LOGE("Not support this surface type(%d) for changing vidoesink", surface_type);
		MMPLAYER_FLEAVE();
		return MM_ERROR_INVALID_ARGUMENT;
	}

	/* load previous attributes */
	if (player->attrs) {
		mm_attrs_get_int_by_name(player->attrs, "display_surface_type", &prev_display_surface_type);
		LOGD("[0: Video surface, 4: EVAS surface] previous surface type(%d), new surface type(%d)", prev_display_surface_type, surface_type);
		if (prev_display_surface_type == surface_type) {
			LOGD("incoming display surface type is same as previous one, do nothing..");
			MMPLAYER_FLEAVE();
			return MM_ERROR_NONE;
		}
	} else {
		LOGE("failed to load attributes");
		MMPLAYER_FLEAVE();
		return MM_ERROR_PLAYER_INTERNAL;
	}

	/* videobin is not created yet, so we just set attributes related to display surface */
	LOGD("store display attribute for given surface type(%d)", surface_type);
	mm_player_set_attribute(handle, NULL, "display_surface_type", surface_type,
			"display_overlay", wl_surface_id, NULL);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

/* Note : if silent is true, then subtitle would not be displayed. :*/
int
_mmplayer_set_subtitle_silent(MMHandleType hplayer, int silent)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	player->set_mode.subtitle_off = silent;

	LOGD("subtitle is %s.", player->set_mode.subtitle_off ? "ON" : "OFF");

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_sync_subtitle_pipeline(mmplayer_t *player)
{
	mmplayer_gst_element_t *mainbin = NULL;
	mmplayer_gst_element_t *textbin = NULL;
	GstStateChangeReturn ret = GST_STATE_CHANGE_FAILURE;
	GstState current_state = GST_STATE_VOID_PENDING;
	GstState element_state = GST_STATE_VOID_PENDING;
	GstState element_pending_state = GST_STATE_VOID_PENDING;
	gint64 time = 0;
	GstEvent *event = NULL;
	int result = MM_ERROR_NONE;

	GstClock *curr_clock = NULL;
	GstClockTime base_time, start_time, curr_time;


	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
								player->pipeline &&
								player->pipeline->mainbin &&
								player->pipeline->textbin, MM_ERROR_PLAYER_NOT_INITIALIZED);

	mainbin = player->pipeline->mainbin;
	textbin = player->pipeline->textbin;

	current_state = GST_STATE(mainbin[MMPLAYER_M_PIPE].gst);

	// sync clock with current pipeline
	curr_clock = GST_ELEMENT_CLOCK(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst);
	curr_time = gst_clock_get_time(curr_clock);

	base_time = gst_element_get_base_time(GST_ELEMENT_CAST(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst));
	start_time = gst_element_get_start_time(GST_ELEMENT_CAST(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst));

	LOGD("state: %d, base_time=%" GST_TIME_FORMAT " start_time=%" GST_TIME_FORMAT " curr_time=%" GST_TIME_FORMAT,
		current_state, GST_TIME_ARGS(base_time), GST_TIME_ARGS(start_time), GST_TIME_ARGS(curr_time));

	if (current_state > GST_STATE_READY) {
		// sync state with current pipeline
		gst_element_set_state(textbin[MMPLAYER_T_BIN].gst, GST_STATE_PAUSED);
		gst_element_set_state(mainbin[MMPLAYER_M_SUBPARSE].gst, GST_STATE_PAUSED);
		gst_element_set_state(mainbin[MMPLAYER_M_SUBSRC].gst, GST_STATE_PAUSED);

		ret = gst_element_get_state(mainbin[MMPLAYER_M_SUBSRC].gst, &element_state, &element_pending_state, 5 * GST_SECOND);
		if (GST_STATE_CHANGE_FAILURE == ret) {
			LOGE("fail to state change.");
			result = MM_ERROR_PLAYER_INTERNAL;
			goto ERROR;
		}
	}
	gst_element_set_base_time(textbin[MMPLAYER_T_BIN].gst, base_time);
	gst_element_set_start_time(textbin[MMPLAYER_T_BIN].gst, start_time);

	if (curr_clock) {
		gst_element_set_clock(textbin[MMPLAYER_T_BIN].gst, curr_clock);
		gst_object_unref(curr_clock);
	}

	// seek to current position
	if (!gst_element_query_position(mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &time)) {
		result = MM_ERROR_PLAYER_INVALID_STATE;
		LOGE("gst_element_query_position failed, invalid state");
		goto ERROR;
	}

	LOGD("seek time = %"G_GINT64_FORMAT", rate = %f", time, player->playback_rate);
	event = gst_event_new_seek(player->playback_rate, GST_FORMAT_TIME, (GstSeekFlags)(GST_SEEK_FLAG_FLUSH), GST_SEEK_TYPE_SET, time, GST_SEEK_TYPE_NONE, -1);
	if (event) {
		_mmplayer_gst_send_event_to_sink(player, event);
	} else {
		result = MM_ERROR_PLAYER_INTERNAL;
		LOGE("gst_event_new_seek failed"); /* pipeline will got error and can not be recovered */
		goto ERROR;
	}

	/* sync state with current pipeline */
	gst_element_sync_state_with_parent(textbin[MMPLAYER_T_BIN].gst);
	gst_element_sync_state_with_parent(mainbin[MMPLAYER_M_SUBPARSE].gst);
	gst_element_sync_state_with_parent(mainbin[MMPLAYER_M_SUBSRC].gst);

	return MM_ERROR_NONE;

ERROR:
	/* release text pipeline resource */
	player->textsink_linked = 0;

	/* release signal */
	__mmplayer_release_signal_connection(player, MM_PLAYER_SIGNAL_TYPE_TEXTBIN);

	/* release textbin with it's childs */
	MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->textbin, MMPLAYER_T_BIN);
	MMPLAYER_FREEIF(player->pipeline->textbin);
	player->pipeline->textbin = NULL;

	/* release subtitle elem */
	MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->mainbin, MMPLAYER_M_SUBSRC);
	MMPLAYER_RELEASE_ELEMENT(player, player->pipeline->mainbin, MMPLAYER_M_SUBPARSE);

	return result;
}

static int
__mmplayer_change_external_subtitle_language(mmplayer_t *player, const char *filepath)
{
	GstStateChangeReturn ret = GST_STATE_CHANGE_FAILURE;
	GstState current_state = GST_STATE_VOID_PENDING;

	MMHandleType attrs = 0;
	mmplayer_gst_element_t *mainbin = NULL;
	mmplayer_gst_element_t *textbin = NULL;

	gchar *subtitle_uri = NULL;
	int result = MM_ERROR_NONE;
	const gchar *charset = NULL;

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player &&
								player->pipeline &&
								player->pipeline->mainbin &&
								player->pipeline->textbin, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(filepath, MM_ERROR_COMMON_INVALID_ARGUMENT);

	mainbin = player->pipeline->mainbin;
	textbin = player->pipeline->textbin;

	current_state = GST_STATE(mainbin[MMPLAYER_M_PIPE].gst);
	if (current_state < GST_STATE_READY) {
		result = MM_ERROR_PLAYER_INVALID_STATE;
		LOGE("Pipeline is not in proper state");
		goto EXIT;
	}

	attrs = MMPLAYER_GET_ATTRS(player);
	if (!attrs) {
		LOGE("cannot get content attribute");
		result = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	mm_attrs_get_string_by_name(attrs, "subtitle_uri", &subtitle_uri);
	if (!subtitle_uri || strlen(subtitle_uri) < 1) {
		LOGE("subtitle uri is not proper filepath");
		result = MM_ERROR_PLAYER_INVALID_URI;
		goto EXIT;
	}

	if (!_mmplayer_get_storage_info(filepath, &player->storage_info[MMPLAYER_PATH_TEXT])) {
		LOGE("failed to get storage info of subtitle path");
		result = MM_ERROR_PLAYER_INVALID_URI;
		goto EXIT;
	}

	SECURE_LOGD("old subtitle file path is [%s]", subtitle_uri);
	SECURE_LOGD("new subtitle file path is [%s]", filepath);

	if (!strcmp(filepath, subtitle_uri)) {
		LOGD("subtitle path is not changed");
		goto EXIT;
	} else {
		if (mm_player_set_attribute((MMHandleType)player, NULL,
				"subtitle_uri", filepath, strlen(filepath), NULL) != MM_ERROR_NONE) {
			LOGE("failed to set attribute");
			goto EXIT;
		}
	}

	//gst_pad_set_blocked_async(src-srcpad, TRUE)
	MMPLAYER_SUBTITLE_INFO_LOCK(player);
	player->subtitle_language_list = NULL;
	MMPLAYER_SUBTITLE_INFO_UNLOCK(player);

	ret = gst_element_set_state(textbin[MMPLAYER_T_BIN].gst, GST_STATE_READY);
	if (ret != GST_STATE_CHANGE_SUCCESS) {
		LOGE("failed to change state of textbin to READY");
		result = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	ret = gst_element_set_state(mainbin[MMPLAYER_M_SUBPARSE].gst, GST_STATE_READY);
	if (ret != GST_STATE_CHANGE_SUCCESS) {
		LOGE("failed to change state of subparse to READY");
		result = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	ret = gst_element_set_state(mainbin[MMPLAYER_M_SUBSRC].gst, GST_STATE_READY);
	if (ret != GST_STATE_CHANGE_SUCCESS) {
		LOGE("failed to change state of filesrc to READY");
		result = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	__mmplayer_initialize_storage_info(player, MMPLAYER_PATH_TEXT);

	g_object_set(G_OBJECT(mainbin[MMPLAYER_M_SUBSRC].gst), "location", filepath, NULL);

	charset = _mmplayer_get_charset(filepath);
	if (charset) {
		LOGD("detected charset is %s", charset);
		g_object_set(G_OBJECT(mainbin[MMPLAYER_M_SUBPARSE].gst), "subtitle-encoding", charset, NULL);
	}

	result = _mmplayer_sync_subtitle_pipeline(player);

EXIT:
	MMPLAYER_FLEAVE();
	return result;
}

/* API to switch between external subtitles */
int
_mmplayer_set_external_subtitle_path(MMHandleType hplayer, const char *filepath)
{
	int result = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	char *path = NULL;

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	/* filepath can be null in idle state */
	if (filepath) {
		/* check file path */
		if ((path = strstr(filepath, "file://")))
			result = _mmplayer_exist_file_path(path + 7);
		else
			result = _mmplayer_exist_file_path(filepath);

		if (result != MM_ERROR_NONE) {
			LOGE("invalid subtitle path 0x%X", result);
			return result; /* file not found or permission denied */
		}
	}

	if (!player->pipeline) {
		/* IDLE state */
		if (mm_player_set_attribute(hplayer, NULL, "subtitle_uri", filepath,
				(filepath)?(strlen(filepath)):(0), NULL) != MM_ERROR_NONE) {
			LOGE("failed to set attribute");
			return MM_ERROR_PLAYER_INTERNAL;
		}
	} else {
		/* cur state <> IDLE(READY, PAUSE, PLAYING..) */
		/* check filepath */
		MMPLAYER_RETURN_VAL_IF_FAIL(filepath, MM_ERROR_COMMON_INVALID_ARGUMENT);

		if (!__mmplayer_check_subtitle(player)) {
			if (mm_player_set_attribute(hplayer, NULL, "subtitle_uri",
					filepath, strlen(filepath), NULL) != MM_ERROR_NONE) {
				LOGE("failed to set attribute");
				return MM_ERROR_PLAYER_INTERNAL;
			}

			if (__mmplayer_gst_create_text_pipeline(player) != MM_ERROR_NONE) {
				LOGE("fail to create text pipeline");
				return MM_ERROR_PLAYER_INTERNAL;
			}

			result = _mmplayer_sync_subtitle_pipeline(player);
		} else {
			result = __mmplayer_change_external_subtitle_language(player, filepath);
		}

		MMPLAYER_BUS_MSG_THREAD_SIGNAL(player);
		player->is_external_subtitle_added_now = TRUE;

		MMPLAYER_SUBTITLE_INFO_LOCK(player);
		if (!player->subtitle_language_list) {
			gint64 timeout = g_get_monotonic_time() + G_TIME_SPAN_SECOND; /* wait 1 sec */
			if (!MMPLAYER_SUBTITLE_INFO_WAIT_UNTIL(player, timeout))
				LOGW("subtitle language list is not updated yet");
		}
		MMPLAYER_SUBTITLE_INFO_UNLOCK(player);
	}

	MMPLAYER_FLEAVE();
	return result;
}

static int
__mmplayer_change_selector_pad(mmplayer_t *player, mmplayer_track_type_e type, int index)
{
	int result = MM_ERROR_NONE;
	gchar *change_pad_name = NULL;
	GstPad *sinkpad = NULL;
	mmplayer_gst_element_t *mainbin = NULL;
	main_element_id_e elem_idx = MMPLAYER_M_NUM;
	GstCaps *caps = NULL;
	gint total_track_num = 0;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->pipeline && player->pipeline->mainbin,
													MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("Change Track(%d) to %d", type, index);

	mainbin = player->pipeline->mainbin;

	if (type == MM_PLAYER_TRACK_TYPE_AUDIO) {
		elem_idx = MMPLAYER_M_A_INPUT_SELECTOR;
	} else if (type == MM_PLAYER_TRACK_TYPE_TEXT) {
		elem_idx = MMPLAYER_M_T_INPUT_SELECTOR;
	} else {
		/* Changing Video Track is not supported. */
		LOGE("Track Type Error");
		goto EXIT;
	}

	if (mainbin[elem_idx].gst == NULL) {
		result = MM_ERROR_PLAYER_NO_OP;
		LOGD("Req track doesn't exist");
		goto EXIT;
	}

	total_track_num = player->selector[type].total_track_num;
	if (total_track_num <= 0) {
		result = MM_ERROR_PLAYER_NO_OP;
		LOGD("Language list is not available");
		goto EXIT;
	}

	if ((index < 0) || (index >= total_track_num)) {
		result = MM_ERROR_INVALID_ARGUMENT;
		LOGD("Not a proper index : %d", index);
		goto EXIT;
	}

	/*To get the new pad from the selector*/
	change_pad_name = g_strdup_printf("sink_%u", index);
	if (change_pad_name == NULL) {
		result = MM_ERROR_PLAYER_INTERNAL;
		LOGD("Pad does not exists");
		goto EXIT;
	}

	LOGD("new active pad name: %s", change_pad_name);

	sinkpad = gst_element_get_static_pad(mainbin[elem_idx].gst, change_pad_name);
	if (sinkpad == NULL) {
		LOGD("sinkpad is NULL");
		result = MM_ERROR_PLAYER_INTERNAL;
		goto EXIT;
	}

	LOGD("Set Active Pad - %s:%s", GST_DEBUG_PAD_NAME(sinkpad));
	g_object_set(mainbin[elem_idx].gst, "active-pad", sinkpad, NULL);

	caps = gst_pad_get_current_caps(sinkpad);
	MMPLAYER_LOG_GST_CAPS_TYPE(caps);

	if (sinkpad)
		gst_object_unref(sinkpad);

	if (type == MM_PLAYER_TRACK_TYPE_AUDIO)
		__mmplayer_set_audio_attrs(player, caps);

EXIT:
	MMPLAYER_FREEIF(change_pad_name);
	return result;
}

int
_mmplayer_change_track_language(MMHandleType hplayer, mmplayer_track_type_e type, int index)
{
	int result = MM_ERROR_NONE;
	mmplayer_t *player = NULL;
	mmplayer_gst_element_t *mainbin = NULL;

	gint current_active_index = 0;

	GstState current_state = GST_STATE_VOID_PENDING;
	GstEvent *event = NULL;
	gint64 time = 0;

	MMPLAYER_FENTER();

	player = (mmplayer_t *)hplayer;
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	if (!player->pipeline) {
		LOGE("Track %d pre setting -> %d", type, index);

		player->selector[type].active_pad_index = index;
		goto EXIT;
	}

	mainbin = player->pipeline->mainbin;

	current_active_index = player->selector[type].active_pad_index;

	/*If index is same as running index no need to change the pad*/
	if (current_active_index == index)
		goto EXIT;

	if (!gst_element_query_position(mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &time)) {
		result = MM_ERROR_PLAYER_INVALID_STATE;
		goto EXIT;
	}

	current_state = GST_STATE(mainbin[MMPLAYER_M_PIPE].gst);
	if (current_state < GST_STATE_PAUSED) {
		result = MM_ERROR_PLAYER_INVALID_STATE;
		LOGW("Pipeline not in porper state");
		goto EXIT;
	}

	result = __mmplayer_change_selector_pad(player, type, index);
	if (result != MM_ERROR_NONE) {
		LOGE("change selector pad error");
		goto EXIT;
	}

	player->selector[type].active_pad_index = index;

	if (current_state == GST_STATE_PLAYING) {
		event = gst_event_new_seek(player->playback_rate, GST_FORMAT_TIME,
			(GstSeekFlags)(GST_SEEK_FLAG_SEGMENT | GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_SKIP),
			GST_SEEK_TYPE_SET, time, GST_SEEK_TYPE_NONE, -1);
		if (event) {
			_mmplayer_gst_send_event_to_sink(player, event);
		} else {
			result = MM_ERROR_PLAYER_INTERNAL;
			goto EXIT;
		}
	}

EXIT:
	return result;
}

int
_mmplayer_get_subtitle_silent(MMHandleType hplayer, int *silent)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	/* check player handle */
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	*silent = player->set_mode.subtitle_off;

	LOGD("subtitle is %s.", silent ? "ON" : "OFF");

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

static gboolean
__mmplayer_add_dump_buffer_probe(mmplayer_t *player, GstElement *element)
{
	MMPLAYER_RETURN_VAL_IF_FAIL(player, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(element, FALSE);

	gchar *factory_name = GST_OBJECT_NAME(gst_element_get_factory(element));
	gchar dump_file_name[PLAYER_INI_MAX_STRLEN*2];

	int idx = 0;

	for (idx = 0; player->ini.dump_element_keyword[idx][0] != '\0'; idx++) {
		if (g_strrstr(factory_name, player->ini.dump_element_keyword[idx])) {
			LOGD("dump [%s] sink pad", player->ini.dump_element_keyword[idx]);
			mmplayer_dump_t *dump_s;
			dump_s = g_try_malloc(sizeof(mmplayer_dump_t));
			if (dump_s == NULL) {
				LOGE("malloc fail");
				return FALSE;
			}

			dump_s->dump_element_file = NULL;
			dump_s->dump_pad = NULL;
			dump_s->dump_pad = gst_element_get_static_pad(element, "sink");

			if (dump_s->dump_pad) {
				memset(dump_file_name, 0x00, PLAYER_INI_MAX_STRLEN * 2);
				snprintf(dump_file_name, PLAYER_INI_MAX_STRLEN * 2, "%s/%s_sink_pad.dump", player->ini.dump_element_path, player->ini.dump_element_keyword[idx]);
				dump_s->dump_element_file = fopen(dump_file_name, "w+");
				dump_s->probe_handle_id = gst_pad_add_probe(dump_s->dump_pad, GST_PAD_PROBE_TYPE_BUFFER, __mmplayer_dump_buffer_probe_cb, dump_s->dump_element_file, NULL);
				/* add list for removed buffer probe and close FILE */
				player->dump_list = g_list_append(player->dump_list, dump_s);
				LOGD("%s sink pad added buffer probe for dump", factory_name);
				return TRUE;
			} else {
				MMPLAYER_FREEIF(dump_s);
				LOGE("failed to get %s sink pad added", factory_name);
			}
		}
	}
	return FALSE;
}

static GstPadProbeReturn
__mmplayer_dump_buffer_probe_cb(GstPad *pad,  GstPadProbeInfo *info, gpointer u_data)
{
	FILE *dump_data = (FILE *)u_data;
//	int written = 0;
	GstBuffer *buffer = gst_pad_probe_info_get_buffer(info);
	GstMapInfo probe_info = GST_MAP_INFO_INIT;

	MMPLAYER_RETURN_VAL_IF_FAIL(dump_data, GST_PAD_PROBE_PASS);

	gst_buffer_map(buffer, &probe_info, GST_MAP_READ);
#ifdef __DEBUG__
	LOGD("buffer timestamp = %" GST_TIME_FORMAT, GST_TIME_ARGS(GST_BUFFER_TIMESTAMP(buffer)));
#endif
	fwrite(probe_info.data, 1, probe_info.size , dump_data);

	gst_buffer_unmap(buffer, &probe_info);

	return GST_PAD_PROBE_OK;
}

static void
__mmplayer_release_dump_list(GList *dump_list)
{
	GList *d_list = dump_list;

	if (!d_list)
		return;

	for (; d_list; d_list = g_list_next(d_list)) {
		mmplayer_dump_t *dump_s = d_list->data;
		if (dump_s->dump_pad) {
			if (dump_s->probe_handle_id)
				gst_pad_remove_probe(dump_s->dump_pad, dump_s->probe_handle_id);
			gst_object_unref(GST_OBJECT(dump_s->dump_pad));
		}
		if (dump_s->dump_element_file) {
			fclose(dump_s->dump_element_file);
			dump_s->dump_element_file = NULL;
		}
		MMPLAYER_FREEIF(dump_s);
	}
	g_list_free(dump_list);
	dump_list = NULL;
}

int
_mmplayer_has_closed_caption(MMHandleType hplayer, bool *exist)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(exist, MM_ERROR_INVALID_ARGUMENT);

	*exist = (bool)player->has_closed_caption;

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

void
_mm_player_video_stream_internal_buffer_unref(void *buffer)
{
	MMPLAYER_FENTER();
	if (buffer) {
#ifdef __DEBUG__
		LOGD("unref internal gst buffer %p", buffer);
#endif
		gst_buffer_unref((GstBuffer *)buffer);
		buffer = NULL;
	}
	MMPLAYER_FLEAVE();
}

int
_mmplayer_get_timeout(MMHandleType hplayer, int *timeout)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(timeout, MM_ERROR_COMMON_INVALID_ARGUMENT);

	if (MMPLAYER_IS_STREAMING(player))
		*timeout = (int)player->ini.live_state_change_timeout;
	else
		*timeout = (int)player->ini.localplayback_state_change_timeout;

	LOGD("timeout = %d", *timeout);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

static void
__mmplayer_initialize_storage_info(mmplayer_t *player, mmplayer_path_type_e path_type)
{
	int i = 0;
	MMPLAYER_FENTER();
	MMPLAYER_RETURN_IF_FAIL(player);

	for (i = 0; i < MMPLAYER_PATH_MAX; i++) {

		if (path_type == MMPLAYER_PATH_MAX || path_type == i) {
			player->storage_info[i].type = STORAGE_TYPE_INTERNAL;
			player->storage_info[i].state = STORAGE_STATE_UNMOUNTABLE;
			player->storage_info[i].id = -1;
			memset(player->storage_info[i].path, 0x00, MM_MAX_URL_LEN);

			if (path_type != MMPLAYER_PATH_MAX)
				break;
		}
	}

	MMPLAYER_FLEAVE();
}

int
_mmplayer_manage_external_storage_state(MMHandleType hplayer, int id, int state)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	MMMessageParamType msg_param = {0, };

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGW("state changed storage %d:%d", id, state);

	if (state != STORAGE_STATE_UNMOUNTABLE && state != STORAGE_STATE_REMOVED)
		return MM_ERROR_NONE;

	/* FIXME: text path should be handled seperately. */
	if (((player->storage_info[MMPLAYER_PATH_VOD].type == STORAGE_TYPE_EXTERNAL)
		&& (player->storage_info[MMPLAYER_PATH_VOD].id == id)) ||
		((player->storage_info[MMPLAYER_PATH_TEXT].type == STORAGE_TYPE_EXTERNAL)
		&& (player->storage_info[MMPLAYER_PATH_TEXT].id == id))) {
		LOGW("external storage is removed");

		if (player->msg_posted == FALSE) {
			memset(&msg_param, 0, sizeof(MMMessageParamType));
			msg_param.code = MM_ERROR_PLAYER_INVALID_URI;
			MMPLAYER_POST_MSG(player, MM_MESSAGE_ERROR, &msg_param);
			player->msg_posted = TRUE;
		}

		/* unrealize the player */
		ret = _mmplayer_unrealize(hplayer);
		if (ret != MM_ERROR_NONE)
			LOGE("failed to unrealize");
	}

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_adaptive_variant_info(MMHandleType hplayer, int *num, char **var_info)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int idx = 0, total = 0;
	gchar *result = NULL, *tmp = NULL;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(num && var_info, MM_ERROR_COMMON_INVALID_ARGUMENT);

	total = *num = g_list_length(player->adaptive_info.var_list);
	if (total <= 0) {
		LOGW("There is no stream variant info.");
		return ret;
	}

	result = g_strdup("");
	for (idx = 0 ; idx < total ; idx++) {
		stream_variant_t *v_data = NULL;
		v_data = g_list_nth_data(player->adaptive_info.var_list, idx);

		if (v_data) {
			gchar data[64] = {0};
			snprintf(data, sizeof(data), "%d,%d,%d,", v_data->bandwidth, v_data->width, v_data->height);

			tmp = g_strconcat(result, data, NULL);
			g_free(result);
			result = tmp;
		} else {
			LOGW("There is no variant data in %d", idx);
			(*num)--;
		}
	}

	*var_info = (char *)result;

	LOGD("variant info %d:%s", *num, *var_info);
	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_set_max_adaptive_variant_limit(MMHandleType hplayer, int bandwidth, int width, int height)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("set limit to [b]%d, [w]%d, [h]%d", bandwidth, width, height);

	player->adaptive_info.limit.bandwidth = (bandwidth >= ADAPTIVE_VARIANT_DEFAULT_VALUE) ? (bandwidth) : (ADAPTIVE_VARIANT_DEFAULT_VALUE);
	player->adaptive_info.limit.width = (width >= ADAPTIVE_VARIANT_DEFAULT_VALUE) ? (width) : (ADAPTIVE_VARIANT_DEFAULT_VALUE);
	player->adaptive_info.limit.height = (height >= ADAPTIVE_VARIANT_DEFAULT_VALUE) ? (height) : (ADAPTIVE_VARIANT_DEFAULT_VALUE);

	if (player->pipeline && player->pipeline->mainbin && player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].gst) {
		LOGD("update max limit of %s", GST_ELEMENT_NAME(player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].gst));
		g_object_set(player->pipeline->mainbin[MMPLAYER_M_ADAPTIVE_DEMUX].gst,
						"max-bandwidth", bandwidth, "max-video-width", width, "max-video-height", height, NULL);

		/* FIXME: seek to current position for applying new variant limitation */
	}

	MMPLAYER_FLEAVE();
	return ret;

}

int
_mmplayer_get_max_adaptive_variant_limit(MMHandleType hplayer, int *bandwidth, int *width, int *height)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(bandwidth && width && height, MM_ERROR_COMMON_INVALID_ARGUMENT);

	*bandwidth = player->adaptive_info.limit.bandwidth;
	*width = player->adaptive_info.limit.width;
	*height = player->adaptive_info.limit.height;

	LOGD("get limit to [b]%d, [w]%d, [h]%d", *bandwidth, *width, *height);

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_get_streaming_buffering_time(MMHandleType hplayer, int *prebuffer_ms, int *rebuffer_ms)
{
	int ret = MM_ERROR_NONE;
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player && player->streamer, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(prebuffer_ms && rebuffer_ms, MM_ERROR_COMMON_INVALID_ARGUMENT);
	MMPLAYER_RETURN_VAL_IF_FAIL(MMPLAYER_IS_STREAMING(player), MM_ERROR_PLAYER_NO_OP);

	*prebuffer_ms = player->streamer->buffering_req.prebuffer_time;

	if (player->streamer->buffering_req.rebuffer_time > MIN_BUFFERING_TIME)
		*rebuffer_ms = player->streamer->buffering_req.rebuffer_time;
	else /* live case */
		*rebuffer_ms = DEFAULT_LIVE_REBUFFER_TIME;

	LOGD("buffering time %d ms / %d ms", *prebuffer_ms, *rebuffer_ms);

	MMPLAYER_FLEAVE();
	return ret;
}

int
_mmplayer_set_codec_type(MMHandleType hplayer, mmplayer_stream_type_e stream_type, mmplayer_codec_type_e codec_type)
{
#define IDX_FIRST_SW_CODEC 0
	mmplayer_t *player = (mmplayer_t *)hplayer;
	const char *attr_name = (stream_type == MM_PLAYER_STREAM_TYPE_AUDIO) ? (MM_PLAYER_AUDIO_CODEC_TYPE) : (MM_PLAYER_VIDEO_CODEC_TYPE);

	MMPLAYER_FENTER();
	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	LOGD("ini setting : [a][h:%s][s:%s] / [v][h:%s][s:%s]",
		player->ini.audiocodec_element_hw, player->ini.audiocodec_element_sw[IDX_FIRST_SW_CODEC],
		player->ini.videocodec_element_hw, player->ini.videocodec_element_sw[IDX_FIRST_SW_CODEC]);

	switch (stream_type) {
	case MM_PLAYER_STREAM_TYPE_AUDIO:
	/* to support audio codec selection, codec info have to be added in ini file as below.
	   audio codec element hw = xxxx
	   audio codec element sw = avdec
	   and in case of audio hw codec is supported and selected,
	   audio filter elements should be applied depending on the hw capabilities.
	 */
		if (((codec_type == MM_PLAYER_CODEC_TYPE_HW) &&
			 (!strcmp(player->ini.audiocodec_element_hw, ""))) ||
			((codec_type == MM_PLAYER_CODEC_TYPE_SW) &&
			 (!strcmp(player->ini.audiocodec_element_sw[IDX_FIRST_SW_CODEC], "")))) {
			LOGE("There is no audio codec info for codec_type %d", codec_type);
			return MM_ERROR_PLAYER_NO_OP;
		}
	break;
	case MM_PLAYER_STREAM_TYPE_VIDEO:
	/* to support video codec selection, codec info have to be added in ini file as below.
	   video codec element hw = omx
	   video codec element sw = avdec */
		if (((codec_type == MM_PLAYER_CODEC_TYPE_HW) &&
			 (!strcmp(player->ini.videocodec_element_hw, ""))) ||
			((codec_type == MM_PLAYER_CODEC_TYPE_SW) &&
			 (!strcmp(player->ini.videocodec_element_sw[IDX_FIRST_SW_CODEC], "")))) {
			LOGE("There is no video codec info for codec_type %d", codec_type);
			return MM_ERROR_PLAYER_NO_OP;
		}
	break;
	default:
		LOGE("Invalid stream type %s", MMPLAYER_STREAM_TYPE_GET_NAME(stream_type));
		return MM_ERROR_COMMON_INVALID_ARGUMENT;
	break;
	}

	LOGD("update %s codec_type to %d", attr_name, codec_type);
	mm_player_set_attribute(hplayer, NULL, attr_name, codec_type, NULL);

	MMPLAYER_FLEAVE();
	return MM_ERROR_NONE;
}

int
_mmplayer_set_replaygain_enabled(MMHandleType hplayer, bool enabled)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	GstElement *rg_vol_element = NULL;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	player->sound.rg_enable = enabled;

	/* just hold rgvolume enable value if pipeline is not ready */
	if (!player->pipeline || !player->pipeline->audiobin) {
		LOGD("pipeline is not ready. holding rgvolume enable value");
		return MM_ERROR_NONE;
	}

	rg_vol_element = player->pipeline->audiobin[MMPLAYER_A_RGVOL].gst;

	if (!rg_vol_element) {
		LOGD("rgvolume element is not created");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	if (enabled)
		g_object_set(rg_vol_element, "enable-rgvolume", TRUE, NULL);
	else
		g_object_set(rg_vol_element, "enable-rgvolume", FALSE, NULL);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_is_replaygain_enabled(MMHandleType hplayer, bool *enabled)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	GstElement *rg_vol_element = NULL;
	gboolean enable = FALSE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(enabled, MM_ERROR_INVALID_ARGUMENT);

	/* just hold enable_rg value if pipeline is not ready */
	if (!player->pipeline || !player->pipeline->audiobin) {
		LOGD("pipeline is not ready. holding rgvolume value (%d)", player->sound.rg_enable);
		*enabled = player->sound.rg_enable;
		return MM_ERROR_NONE;
	}

	rg_vol_element = player->pipeline->audiobin[MMPLAYER_A_RGVOL].gst;

	if (!rg_vol_element) {
		LOGD("rgvolume element is not created");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	g_object_get(rg_vol_element, "enable-rgvolume", &enable, NULL);
	*enabled = (bool)enable;

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_set_video_roi_area(MMHandleType hplayer, double scale_x, double scale_y, double scale_width, double scale_height)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	MMHandleType attrs = 0;
	int handle = 0;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	attrs = MMPLAYER_GET_ATTRS(player);
	MMPLAYER_RETURN_VAL_IF_FAIL(attrs, MM_ERROR_PLAYER_INTERNAL);

	mm_attrs_get_int_by_name(attrs, "display_overlay", &handle);
	if (!handle) {
		LOGE("Display handle is NULL, after setting window handle, set video roi area");
		return MM_ERROR_PLAYER_INTERNAL;
	}

	player->video_roi.scale_x = scale_x;
	player->video_roi.scale_y = scale_y;
	player->video_roi.scale_width = scale_width;
	player->video_roi.scale_height = scale_height;

	/* check video sinkbin is created */
	if (!_mmplayer_is_videosink_ready(player, MM_DISPLAY_SURFACE_NUM))
		return MM_ERROR_NONE;

	if (!gst_video_overlay_set_video_roi_area(
		GST_VIDEO_OVERLAY(player->pipeline->videobin[MMPLAYER_V_SINK].gst),
		scale_x, scale_y, scale_width, scale_height))
		ret = MM_ERROR_PLAYER_INTERNAL;
	else
		LOGD("set video param : video roi area scale value: x(%f) y(%f) width(%f) height(%f)",
			scale_x, scale_y, scale_width, scale_height);

	MMPLAYER_FLEAVE();

	return ret;
}

int
_mmplayer_get_video_roi_area(MMHandleType hplayer, double *scale_x, double *scale_y, double *scale_width, double *scale_height)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	int ret = MM_ERROR_NONE;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(scale_x && scale_y && scale_width && scale_height, MM_ERROR_INVALID_ARGUMENT);

	*scale_x = player->video_roi.scale_x;
	*scale_y = player->video_roi.scale_y;
	*scale_width = player->video_roi.scale_width;
	*scale_height = player->video_roi.scale_height;

	LOGD("get video param : video roi area scale value: x(%f) y(%f) width(%f) height(%f)",
		*scale_x, *scale_y, *scale_width, *scale_height);

	return ret;
}

int
_mmplayer_set_client_pid(MMHandleType hplayer, int pid)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);

	player->client_pid = pid;

	LOGD("client pid[%d] %p", pid, player);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

int
_mmplayer_is_audio_control_available(MMHandleType hplayer, mmplayer_audio_control_opt_e opt, bool *available)
{
	mmplayer_t *player = (mmplayer_t *)hplayer;
	mmplayer_codec_type_e codec_type = MM_PLAYER_CODEC_TYPE_DEFAULT;
	enum audio_element_id elem_id = MMPLAYER_A_NUM;

	MMPLAYER_FENTER();

	MMPLAYER_RETURN_VAL_IF_FAIL(player, MM_ERROR_PLAYER_NOT_INITIALIZED);
	MMPLAYER_RETURN_VAL_IF_FAIL(available, MM_ERROR_INVALID_ARGUMENT);

	*available = true;
	mm_attrs_get_int_by_name(player->attrs, MM_PLAYER_AUDIO_CODEC_TYPE, (int *)&codec_type);

	LOGD("current state %d, codec_type %d", MMPLAYER_CURRENT_STATE(player), codec_type);

	if (codec_type == MM_PLAYER_CODEC_TYPE_SW)
		return MM_ERROR_NONE;

	/* in case of audio codec default type is HW */
	switch(opt) {
		case MM_PLAYER_AUDIO_CONTROL_OPT_EFFECT:
			if (player->ini.support_audio_effect)
				return MM_ERROR_NONE;
			elem_id = MMPLAYER_A_FILTER;
		break;
		case MM_PLAYER_AUDIO_CONTROL_OPT_REPLAYGAIN:
			if (player->ini.support_replaygain_control)
				return MM_ERROR_NONE;
			elem_id = MMPLAYER_A_RGVOL;
		break;
		case MM_PLAYER_AUDIO_CONTROL_OPT_PITCH:
			if (player->ini.support_pitch_control)
				return MM_ERROR_NONE;
			elem_id = MMPLAYER_A_PITCH;
		break;
		case MM_PLAYER_AUDIO_CONTROL_OPT_PCM_EXPORTING:
			if (player->ini.support_audio_effect)
				return MM_ERROR_NONE;
		break;
		/* default case handling is not required */
	}

	if (MMPLAYER_CURRENT_STATE(player) < MM_PLAYER_STATE_READY) {
		LOGW("audio control option [%d] is not available", opt);
		*available = false;
	} else {
		/* setting pcm exporting option is allowed before READY state */
		if (opt == MM_PLAYER_AUDIO_CONTROL_OPT_PCM_EXPORTING)
			return MM_ERROR_PLAYER_INVALID_STATE;

		/* check whether the audio filter exist or not after READY state,
		   because the sw codec could be added during auto-plugging in some cases */
		if (!player->pipeline ||
			!player->pipeline->audiobin ||
			!player->pipeline->audiobin[elem_id].gst) {
			LOGW("there is no audio elem [%d]", elem_id);
			*available = false;
		}
	}

	LOGD("audio control opt %d, available %d", opt, *available);

	MMPLAYER_FLEAVE();

	return MM_ERROR_NONE;
}

static gboolean
__mmplayer_update_duration_value(mmplayer_t *player)
{
	gboolean ret = FALSE;
	gint64 dur_nsec = 0;
	LOGD("try to update duration");

	if (gst_element_query_duration(player->pipeline->mainbin[MMPLAYER_M_PIPE].gst, GST_FORMAT_TIME, &dur_nsec) && (dur_nsec > 0)) {
		player->duration = dur_nsec;
		LOGW("duration : %"G_GINT64_FORMAT" msec", GST_TIME_AS_MSECONDS(dur_nsec));
		ret = TRUE;
	}

	if (player->duration < 0) {
		LOGW("duration is Non-Initialized !!!");
		player->duration = 0;
	}

	/* update streaming service type */
	player->streaming_type =  _mmplayer_get_stream_service_type(player);

	/* check duration is OK */
	if (dur_nsec == 0 && !MMPLAYER_IS_LIVE_STREAMING(player))
		/* FIXIT : find another way to get duration here. */
		LOGW("finally it's failed to get duration from pipeline. progressbar will not work correctely!");

	return ret;
}

static gboolean
__mmplayer_update_audio_attrs(mmplayer_t *player, MMHandleType attrs)
{
	/* update audio params
	NOTE : We need original audio params and it can be only obtained from src pad of audio
	decoder. Below code only valid when we are not using 'resampler' just before
	'audioconverter'. */
	GstCaps *caps_a = NULL;
	GstPad *pad = NULL;
	gint samplerate = 0, channels = 0;
	GstStructure *p = NULL;
	GstElement *aconv = NULL;

	LOGD("try to update audio attrs");

	MMPLAYER_RETURN_VAL_IF_FAIL(player->pipeline->audiobin, FALSE);

	if (player->pipeline->audiobin[MMPLAYER_A_CONV].gst) {
		aconv = player->pipeline->audiobin[MMPLAYER_A_CONV].gst;
	} else if (player->pipeline->audiobin[MMPLAYER_A_EXTRACT_CONV].gst) {
		aconv = player->pipeline->audiobin[MMPLAYER_A_EXTRACT_CONV].gst;
	} else {
		LOGE("there is no audio converter");
		return FALSE;
	}

	pad = gst_element_get_static_pad(aconv, "sink");

	if (!pad) {
		LOGW("failed to get pad from audio converter");
		return FALSE;
	}

	caps_a = gst_pad_get_current_caps(pad);
	if (!caps_a) {
		LOGW("not ready to get audio caps");
		gst_object_unref(pad);
		return FALSE;
	}

	p = gst_caps_get_structure(caps_a, 0);

	mm_attrs_get_int_by_name(attrs, "content_audio_samplerate", &samplerate);

	gst_structure_get_int(p, "rate", &samplerate);
	gst_structure_get_int(p, "channels", &channels);

	mm_player_set_attribute((MMHandleType)player, NULL,
			"content_audio_samplerate", samplerate,
			"content_audio_channels", channels, NULL);

	SECURE_LOGD("samplerate : %d	channels : %d", samplerate, channels);

	gst_caps_unref(caps_a);
	gst_object_unref(pad);

	return TRUE;
}

static gboolean
__mmplayer_update_video_attrs(mmplayer_t *player, MMHandleType attrs)
{
	LOGD("try to update video attrs");

	GstCaps *caps_v = NULL;
	GstPad *pad = NULL;
	gint tmpNu, tmpDe;
	gint width, height;
	GstStructure *p = NULL;

	MMPLAYER_RETURN_VAL_IF_FAIL(player->pipeline->videobin, FALSE);
	MMPLAYER_RETURN_VAL_IF_FAIL(player->pipeline->videobin[MMPLAYER_V_SINK].gst, FALSE);

	pad = gst_element_get_static_pad(player->pipeline->videobin[MMPLAYER_V_SINK].gst, "sink");
	if (!pad) {
		LOGD("no videosink sink pad");
		return FALSE;
	}

	caps_v = gst_pad_get_current_caps(pad);
	/* Use v_stream_caps, if fail to get video_sink sink pad*/
	if (!caps_v && player->v_stream_caps) {
		caps_v = player->v_stream_caps;
		gst_caps_ref(caps_v);
	}

	if (!caps_v) {
		LOGD("no negitiated caps from videosink");
		gst_object_unref(pad);
		return FALSE;
	}

	p = gst_caps_get_structure(caps_v, 0);
	gst_structure_get_int(p, "width", &width);
	gst_structure_get_int(p, "height", &height);

	mm_player_set_attribute((MMHandleType)player, NULL,
			MM_PLAYER_VIDEO_WIDTH, width, MM_PLAYER_VIDEO_HEIGHT, height, NULL);

	gst_structure_get_fraction(p, "framerate", &tmpNu, &tmpDe);

	SECURE_LOGD("width : %d     height : %d", width, height);

	gst_caps_unref(caps_v);
	gst_object_unref(pad);

	if (tmpDe > 0) {
		mm_player_set_attribute((MMHandleType)player, NULL,
				MM_PLAYER_VIDEO_FPS, (tmpNu/tmpDe), NULL);
		SECURE_LOGD("fps : %d", tmpNu / tmpDe);
	}

	return TRUE;
}

static gboolean
__mmplayer_update_bitrate_attrs(mmplayer_t *player, MMHandleType attrs)
{
	gboolean ret = FALSE;
	guint64 data_size = 0;
	gchar *path = NULL;
	struct stat sb;

	/* FIXIT : please make it clear the dependancy with duration/codec/uritype */
	if (!player->duration)
		return FALSE;

	if (!MMPLAYER_IS_STREAMING(player) && (player->can_support_codec & FOUND_PLUGIN_VIDEO)) {
		mm_attrs_get_string_by_name(attrs, "profile_uri", &path);
		if (stat(path, &sb) == 0)
			data_size = (guint64)sb.st_size;

	} else if (MMPLAYER_IS_HTTP_STREAMING(player)) {
		data_size = player->http_content_size;
	}

	LOGD("try to update bitrate : data_size = %"G_GUINT64_FORMAT, data_size);

	if (data_size) {
		guint64 bitrate = 0;
		guint64 msec_dur = 0;

		msec_dur = GST_TIME_AS_MSECONDS(player->duration);
		if (msec_dur > 0) {
			bitrate = data_size * 8 * 1000 / msec_dur;
			SECURE_LOGD("file size : %"G_GUINT64_FORMAT
					", video bitrate = %"G_GUINT64_FORMAT, data_size, bitrate);
			mm_player_set_attribute((MMHandleType)player, NULL,
					MM_PLAYER_VIDEO_BITRATE, (int)bitrate, NULL);
			ret = TRUE;
		} else {
			LOGD("player duration is less than 0");
		}
	}

	if (MMPLAYER_IS_RTSP_STREAMING(player)) {
		if (player->total_bitrate) {
			mm_player_set_attribute((MMHandleType)player, NULL,
					MM_PLAYER_VIDEO_BITRATE, player->total_bitrate, NULL);
			ret = TRUE;
		}
	}

	return ret;
}

static void
__mmplayer_copy_uri_and_set_type(mmplayer_parse_profile_t *data, const char *uri, int uri_type)
{
	strncpy(data->uri, uri, MM_MAX_URL_LEN - 1);
	data->uri_type = uri_type;
}

static int
__mmplayer_set_mem_uri(mmplayer_parse_profile_t *data, char *path, void *param)
{
	int ret = MM_ERROR_PLAYER_INVALID_URI;
	int mem_size = 0;
	char *buffer = NULL;
	char *seperator = strchr(path, ',');
	char ext[100] = {0,}, size[100] = {0,};

	if (seperator) {
		if ((buffer = strstr(path, "ext="))) {
			buffer += strlen("ext=");

			if (strlen(buffer)) {
				strncpy(ext, buffer, 99);

				if ((seperator = strchr(ext, ','))
					|| (seperator = strchr(ext, ' '))
					|| (seperator = strchr(ext, '\0'))) {
					seperator[0] = '\0';
				}
			}
		}

		if ((buffer = strstr(path, "size="))) {
			buffer += strlen("size=");

			if (strlen(buffer) > 0) {
				strncpy(size, buffer, 99);

				if ((seperator = strchr(size, ','))
					|| (seperator = strchr(size, ' '))
					|| (seperator = strchr(size, '\0'))) {
					seperator[0] = '\0';
				}

				mem_size = atoi(size);
			}
		}
	}

	LOGD("ext: %s, mem_size: %d, mmap(param): %p", ext, mem_size, param);

	if (mem_size && param) {
		if (data->input_mem.buf)
			free(data->input_mem.buf);
		data->input_mem.buf = malloc(mem_size);

		if (data->input_mem.buf) {
			memcpy(data->input_mem.buf, param, mem_size);
			data->input_mem.len = mem_size;
			ret = MM_ERROR_NONE;
		} else {
			LOGE("failed to alloc mem %d", mem_size);
			ret = MM_ERROR_PLAYER_INTERNAL;
		}

		data->input_mem.offset = 0;
		data->uri_type = MM_PLAYER_URI_TYPE_MEM;
	}

	return ret;
}

static int
__mmplayer_set_file_uri(mmplayer_parse_profile_t *data, const char *uri)
{
	gchar *location = NULL;
	GError *err = NULL;
	char *path = NULL;
	int ret = MM_ERROR_NONE;

	if ((path = strstr(uri, "file://"))) {
		location = g_filename_from_uri(uri, NULL, &err);
		if (!location || (err != NULL)) {
			LOGE("Invalid URI '%s' for filesrc: %s", path,
				(err != NULL) ? err->message : "unknown error");
			if (err)
				g_error_free(err);

			MMPLAYER_FREEIF(location);

			data->uri_type = MM_PLAYER_URI_TYPE_NONE;
			return MM_ERROR_PLAYER_INVALID_URI;
		}
		LOGD("path from uri: %s", location);
	}

	path = (location != NULL) ? (location) : ((char *)uri);


	ret = _mmplayer_exist_file_path(path);

	/* if no protocol prefix exist. check file existence and then give file:// as it's prefix */
	if (ret == MM_ERROR_NONE) {
		g_snprintf(data->uri,  MM_MAX_URL_LEN, "file://%s", path);
		if (_mmplayer_is_sdp_file(path)) {
			LOGD("uri is actually a file but it's sdp file. giving it to rtspsrc");
			data->uri_type = MM_PLAYER_URI_TYPE_URL_RTSP;
		} else {
			data->uri_type = MM_PLAYER_URI_TYPE_FILE;
		}
	} else if (ret == MM_ERROR_PLAYER_PERMISSION_DENIED) {
		data->uri_type = MM_PLAYER_URI_TYPE_NO_PERMISSION;
	} else {
		LOGE("invalid uri, could not play..");
		data->uri_type = MM_PLAYER_URI_TYPE_NONE;
	}

	MMPLAYER_FREEIF(location);

	return ret;
}

static mmplayer_video_decoded_data_info_t *
__mmplayer_create_stream_from_pad(GstPad *pad)
{
	GstCaps *caps = NULL;
	GstStructure *structure = NULL;
	unsigned int fourcc = 0;
	const gchar *string_format = NULL;
	mmplayer_video_decoded_data_info_t *stream = NULL;
	gint width, height;
	MMPixelFormatType format;
	GstVideoInfo info;

	caps = gst_pad_get_current_caps(pad);
	if (!caps) {
		LOGE("Caps is NULL.");
		return NULL;
	}

#ifdef __DEBUG__
	MMPLAYER_LOG_GST_CAPS_TYPE(caps);
#endif
	structure = gst_caps_get_structure(caps, 0);
	gst_structure_get_int(structure, "width", &width);
	gst_structure_get_int(structure, "height", &height);
	string_format = gst_structure_get_string(structure, "format");

	if (string_format)
		fourcc = _mmplayer_convert_fourcc_string_to_value(string_format);
	format = _mmplayer_get_pixtype(fourcc);
	gst_video_info_from_caps(&info, caps);
	gst_caps_unref(caps);

	/* moved here */
	if (width == 0 || height == 0 || format == MM_PIXEL_FORMAT_INVALID) {
		LOGE("Wrong condition!!");
		return NULL;
	}

	stream = (mmplayer_video_decoded_data_info_t *)g_try_malloc0(sizeof(mmplayer_video_decoded_data_info_t));
	if (!stream) {
		LOGE("failed to alloc mem for video data");
		return NULL;
	}

	stream->width = width;
	stream->height = height;
	stream->format = format;
	stream->plane_num = GST_VIDEO_INFO_N_PLANES(&info);

	return stream;
}

static void
__mmplayer_zerocopy_set_stride_elevation_bo(mmplayer_video_decoded_data_info_t *stream, GstMemory *mem)
{
	unsigned int pitch = 0;
	unsigned int size = 0;
	int index = 0;
	tbm_surface_h surface = gst_tizen_memory_get_surface(mem);
	tbm_bo bo = NULL;

	for (index = 0; index < gst_tizen_memory_get_num_bos(mem); index++) {
		bo = gst_tizen_memory_get_bos(mem, index);
		if (bo)
			stream->bo[index] = tbm_bo_ref(bo);
		else
			LOGE("failed to get bo for index %d", index);
	}

	for (index = 0; index < stream->plane_num; index++) {
		tbm_surface_internal_get_plane_data(surface, index, &size, NULL, &pitch);
		stream->stride[index] = pitch;
		if (pitch)
			stream->elevation[index] = size / pitch;
		else
			stream->elevation[index] = stream->height;
	}
}

static gboolean
__mmplayer_swcodec_set_stride_elevation(mmplayer_video_decoded_data_info_t *stream)
{
	if (stream->format == MM_PIXEL_FORMAT_I420) {
		int ret = TBM_SURFACE_ERROR_NONE;
		tbm_surface_h surface;
		tbm_surface_info_s info;

		surface = tbm_surface_create(stream->width, stream->height, TBM_FORMAT_YUV420);

		ret = tbm_surface_get_info(surface, &info);
		if (ret != TBM_SURFACE_ERROR_NONE) {
			tbm_surface_destroy(surface);
			return FALSE;
		}

		tbm_surface_destroy(surface);
		stream->stride[0] = info.planes[0].stride;
		stream->elevation[0] = info.planes[0].size / info.planes[0].stride;
		stream->stride[1] = info.planes[1].stride;
		stream->elevation[1] = info.planes[1].size / info.planes[1].stride;
		stream->stride[2] = info.planes[2].stride;
		stream->elevation[2] = info.planes[2].size / info.planes[2].stride;
		stream->bo_size = info.planes[0].size + info.planes[1].size + info.planes[2].size;
	} else if (stream->format == MM_PIXEL_FORMAT_RGBA) {
		stream->stride[0] = stream->width * 4;
		stream->elevation[0] = stream->height;
		stream->bo_size = stream->stride[0] * stream->height;
	} else {
		LOGE("Not support format %d", stream->format);
		return FALSE;
	}

	return TRUE;
}

static gboolean
__mmplayer_swcodec_set_bo(mmplayer_t *player, mmplayer_video_decoded_data_info_t *stream, GstMemory *mem)
{
	tbm_bo_handle thandle;
	gboolean is_mapped;
	int src_stride[MM_PLAYER_IMGB_MPLANE_MAX] = { 0, };
	int src_offset[MM_PLAYER_IMGB_MPLANE_MAX] = { 0, };
	int dest_offset[MM_PLAYER_IMGB_MPLANE_MAX] = { 0, };
	int i = 0;
	int j = 0;
	int k = 0;
	unsigned char *src = NULL;
	unsigned char *dest = NULL;
	GstMapInfo mapinfo = GST_MAP_INFO_INIT;

	is_mapped = gst_memory_map(mem, &mapinfo, GST_MAP_READWRITE);
	if (!is_mapped) {
		LOGE("fail to gst_memory_map");
		return FALSE;
	}

	if (!mapinfo.data) {
		LOGE("data pointer is wrong");
		goto ERROR;
	}

	stream->bo[0] = __mmplayer_video_stream_get_bo(player, stream->bo_size);
	if (!stream->bo[0]) {
		LOGE("Fail to tbm_bo_alloc!!");
		goto ERROR;
	}

	thandle = tbm_bo_map(stream->bo[0], TBM_DEVICE_CPU, TBM_OPTION_WRITE);
	if (!thandle.ptr) {
		LOGE("thandle pointer is wrong");
		goto ERROR;
	}

	if (stream->format == MM_PIXEL_FORMAT_I420) {
		src_stride[0] = GST_ROUND_UP_4(stream->width);
		src_stride[1] = src_stride[2] = GST_ROUND_UP_4(stream->width >> 1);
		src_offset[1] = src_stride[0] * GST_ROUND_UP_2(stream->height);
		src_offset[2] = src_offset[1] + (src_stride[1] * (GST_ROUND_UP_2(stream->height) >> 1));

		dest_offset[0] = 0;
		dest_offset[1] = stream->stride[0] * stream->elevation[0];
		dest_offset[2] = dest_offset[1] + stream->stride[1] * stream->elevation[1];

		for (i = 0; i < 3; i++) {
			src = mapinfo.data + src_offset[i];
			dest = thandle.ptr + dest_offset[i];

			if (i > 0)
				k = 1;

			for (j = 0; j < stream->height >> k; j++) {
				memcpy(dest, src, stream->width>>k);
				src += src_stride[i];
				dest += stream->stride[i];
			}
		}
	} else if (stream->format == MM_PIXEL_FORMAT_RGBA) {
		memcpy(thandle.ptr, mapinfo.data, stream->bo_size);
	} else {
		LOGE("Not support format %d", stream->format);
		goto ERROR;
	}

	tbm_bo_unmap(stream->bo[0]);
	gst_memory_unmap(mem, &mapinfo);

	return TRUE;

ERROR:
	if (stream->bo[0])
		tbm_bo_unmap(stream->bo[0]);

	if (is_mapped)
		gst_memory_unmap(mem, &mapinfo);

	return FALSE;
}

static void
__mmplayer_set_pause_state(mmplayer_t *player)
{
	if (player->sent_bos)
		return;

	/* rtsp case, get content attrs by GstMessage */
	if (MMPLAYER_IS_RTSP_STREAMING(player))
		return;

	/* it's first time to update all content attrs. */
	_mmplayer_update_content_attrs(player, ATTR_ALL);
}

static void
__mmplayer_set_playing_state(mmplayer_t *player)
{
	gchar *audio_codec = NULL;

	if (player->resumed_by_rewind && player->playback_rate < 0.0) {
		/* initialize because auto resume is done well. */
		player->resumed_by_rewind = FALSE;
		player->playback_rate = 1.0;
	}

	if (player->sent_bos)
		return;

	/* try to get content metadata */

	/* NOTE : giving ATTR_MISSING_ONLY may have dependency with
	 * c-api since c-api doesn't use _start() anymore. It may not work propery with
	 * legacy mmfw-player api
	 */
	_mmplayer_update_content_attrs(player, ATTR_MISSING_ONLY);

	if ((player->cmd == MMPLAYER_COMMAND_START)
		|| (player->cmd == MMPLAYER_COMMAND_RESUME)) {
		__mmplayer_handle_missed_plugin(player);
	}

	/* check audio codec field is set or not
	 * we can get it from typefinder or codec's caps.
	 */
	mm_attrs_get_string_by_name(player->attrs, "content_audio_codec", &audio_codec);

	/* The codec format can't be sent for audio only case like amr, mid etc.
	 * Because, parser don't make related TAG.
	 * So, if it's not set yet, fill it with found data.
	 */
	if (!audio_codec) {
		if (g_strrstr(player->type, "audio/midi"))
			audio_codec = "MIDI";
		else if (g_strrstr(player->type, "audio/x-amr"))
			audio_codec = "AMR";
		else if (g_strrstr(player->type, "audio/mpeg")
				&& !g_strrstr(player->type, "mpegversion=(int)1"))
			audio_codec = "AAC";
		else
			audio_codec = "unknown";

		if (mm_player_set_attribute((MMHandleType)player, NULL,
				"content_audio_codec", audio_codec, strlen(audio_codec), NULL) != MM_ERROR_NONE)
			LOGE("failed to set attribute");

		LOGD("set audio codec type with caps");
	}

	return;
}
